/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

#include <string.h>
#include <strings.h>

#include "galaxy/galaxy_all.h"
#define SERVER_FUNCTIONS_INCLUDE "echo_server.h"
#define USE_SERVER_DATA
#include "galaxy/server_functions.h"

/* This server will mimic what the speech recognizer receives
   and what the synthesizer sends. */

/* These #defines are for compatibility with the old MIT playbox. */

#define AUDIO_SERVER_NEW_UTT "new_utt"
#define AUDIO_SERVER_SAMPLES_DONE "done"
#define AUDIO_SERVER_SAMPLES_ABORT "cancel"

#define AUDIO_CLIENT_PLAY_SAMPLES_DONE "AUDIO_CLIENT_PLAY_SAMPLES_DONE"
#define AUDIO_CLIENT_PLAY_SAMPLES_START "AUDIO_CLIENT_PLAY_START"

/* I think I want to collect all the audio and then echo it,
   rather than doing it bit by bit. */

typedef struct __ServerDataContainer {
  Gal_ObjectType dtype;
  char *data_buf;  
  int size;
  int batch;
  int playbox_compat;
  GalIO_BrokerStruct *b;
  GalSS_BrokerProxy *bp;
  int sample_rate;
  char *encoding_format;
  char *logfile;
  FILE *logfile_fp;
  int use_proxy;
} ServerDataContainer;

static GalIO_BrokerStruct *set_up_return_broker(ServerDataContainer *c,
						GalSS_Environment *env)
{
  Gal_Frame f = Gal_MakeFrame("FromEcho", GAL_CLAUSE);
  GalIO_BrokerStruct *b = (GalIO_BrokerStruct *) NULL;

  if (c->use_proxy) {
    GalSS_BrokerProxy *bp = GalSS_ProxifyObjectType(env, (Gal_ObjectType) -1,
						    0, 10);
    if (bp) {
      b = GalSS_BrokerProxyBroker(bp);
      Gal_SetProp(f, ":synth_proxy", Gal_CreateProxyObject(bp, 0));
      c->bp = bp;
    }    
  } else {
    b = GalIO_BrokerDataOutInit(GalSS_EnvComm(env), 0, 10);
    if (b && (GalIO_GetBrokerListenPort(b) > 0)) {
      GalIO_BrokerPopulateFrame(b, f, ":synth_host", ":synth_port");
    } else {
      if (b) {
	GalIO_BrokerDataDone(b);
	GalIO_DestroyBrokerStruct(b);
      }
      b = (GalIO_BrokerStruct *) NULL;
    }
  }
  if (b) {
    if (c->sample_rate) {
      Gal_SetProp(f, ":sample_rate", Gal_IntObject(c->sample_rate));
    }
    if (c->encoding_format) {
      Gal_SetProp(f, ":encoding_format", Gal_StringObject(c->encoding_format));
    }    
    GalSS_EnvWriteFrame(env, f, 0);
  } else {
    GalUtil_Warn("Couldn't set up broker!");
  }
  if (c->playbox_compat)
    GalIO_BrokerWriteString(b, AUDIO_CLIENT_PLAY_SAMPLES_START);
  c->b = b;  
  if (c->logfile) {
    /* Try to save it to a file. */
    c->logfile_fp = fopen(c->logfile, "w");
  
    if (!c->logfile_fp) {
      GalUtil_Warn("Couldn't open logfile %s for writing\n", c->logfile);
    }
  }
  return b;
}

static void return_data_to_playbox(ServerDataContainer *c,
				   GalIO_BrokerStruct *incoming_b)
{
  /* If it's batch, we'll just set up the return here
     and ship the data. Otherwise, I will have already
     written it. */
  if (c->batch) {
    GalIO_BrokerStruct *b = set_up_return_broker(c, GalSS_BrokerGetEnvironment(incoming_b));

    if (b) {
      switch (c->dtype) {
      case GAL_INT_16:
	GalIO_BrokerWriteInt16(b, c->data_buf, (c->size / 2));
	break;
      case GAL_BINARY:
	GalIO_BrokerWriteBinary(b, c->data_buf, c->size);
	break;
      default:
	break;
      }
      /* Aargh. Between platforms, this will fry for int16. I really
	 want to avoid byte-swapping, for the moment, though. */
      if (c->logfile_fp)
	fwrite(c->data_buf, sizeof(char), c->size, c->logfile_fp);
    }
  }
  if (c->playbox_compat)
    GalIO_BrokerWriteString(c->b, AUDIO_CLIENT_PLAY_SAMPLES_DONE);
  GalIO_BrokerDataOutDone(c->b);
  if (c->logfile_fp) {
    fflush(c->logfile_fp);
    fclose(c->logfile_fp);
  }
}

static void audio_in_handler(GalIO_BrokerStruct *broker_struct,
                             void *data,
                             Gal_ObjectType data_type,
                             int n_samples)
{
  ServerDataContainer *c = (ServerDataContainer *) GalIO_GetBrokerCallerData(broker_struct);
  
  switch(data_type) {
  case GAL_INT_16:
    if (c->dtype == (Gal_ObjectType) -1) {
      c->dtype = GAL_INT_16;
    } else if (c->dtype != data_type) {
      GalUtil_Warn("audio_data_handler: data type %s does not match previous\n",
		   Gal_ObjectTypeString(data_type));
      free(data);
      break;
    }
    GalUtil_PInfo2("audio_data_handler: got %s data (%d samples)\n",
		   Gal_ObjectTypeString(data_type),
		   n_samples);
    if (c->batch) {
      c->data_buf = (char *) realloc(c->data_buf, (2 * n_samples) + c->size);
      bcopy(data, c->data_buf + c->size, 2 * n_samples);
      c->size += (2 * n_samples);
    } else {
      GalIO_BrokerWriteInt16(c->b, data, n_samples);
      /* Aargh. Between platforms, this will fry for int16. I really
	 want to avoid byte-swapping, for the moment, though. */
      if (c->logfile_fp)
	fwrite(data, sizeof(char), 2 * n_samples, c->logfile_fp);
    }
    free(data);
    break;
  case GAL_BINARY:
    if (c->dtype == (Gal_ObjectType) -1) {
      c->dtype = GAL_BINARY;
    } else if (c->dtype != data_type) {
      GalUtil_Warn("audio_data_handler: data type %s does not match previous\n",
		   Gal_ObjectTypeString(data_type));
      free(data);
      break;
    }
    GalUtil_PInfo2("audio_data_handler: got %s data (%d samples)\n",
		   Gal_ObjectTypeString(data_type),
		   n_samples);
    if (c->batch) {
      c->data_buf = (char *) realloc(c->data_buf, n_samples + c->size);
      bcopy(data, c->data_buf + c->size, n_samples);
      c->size += n_samples;
    } else {
      GalIO_BrokerWriteBinary(c->b, data, n_samples);
      if (c->logfile_fp)
	fwrite(data, sizeof(char), n_samples, c->logfile_fp);
    }
    free(data);
    break;    
  case GAL_STRING:
    if (c->playbox_compat) {
      GalUtil_PInfo2("audio_data_handler: got message %s\n",
		     (char *) data);
      if (!strcmp((char *)data, AUDIO_SERVER_NEW_UTT)) {
	free(c->data_buf);
	c->data_buf = (char *) NULL;
	c->size = 0;
      } else if (!strcmp((char *)data, AUDIO_SERVER_SAMPLES_DONE)) {
	GalIO_BrokerDataDone(broker_struct);
	/* Send the data back to the playbox. */
	return_data_to_playbox(c, broker_struct);
      } else if (!strcmp((char *)data, AUDIO_SERVER_SAMPLES_ABORT)) {
	GalIO_BrokerDataDone(broker_struct);
	free(c->data_buf);
	c->data_buf = (char *) NULL;
	c->size = 0;
      } else {
	GalUtil_Warn("audio_data_handler: Msg is unknown...\n");
      }
      free(data);
      break;
    }
      /* Otherwise, fall through. */
    default:
      GalUtil_Warn("audio_data_handler: unexpected data type %s\n",
		   Gal_ObjectTypeString(data_type));
      free(data);
      break;
    }
}

void __FreeDataContainer(void *data)
{
  ServerDataContainer *d = (ServerDataContainer *) data;
  if (d->data_buf)
    free(d->data_buf);
  if (d->encoding_format)
    free(d->encoding_format);
  if (d->logfile)
    free(d->logfile);
  free(d);
}

void __report_done(GalIO_BrokerStruct *b, void *data)
{
  ServerDataContainer *c = (ServerDataContainer *) data;

  /* Send the data back to the playbox. If we're in MIT playbox
     compatibility mode, we shouldn't do this, because we'll get
     a string telling us to do it. */
  if (!c->playbox_compat)
    return_data_to_playbox(c, b);
}

Gal_Frame handle_audio_data(Gal_Frame f, void *server_data)
{
  char *host = Gal_GetString(f, ":audio_host");
  int port = Gal_GetInt(f, ":audio_port");
  GalSS_BrokerProxy *bp = Gal_GetProxy(f, ":audio_proxy");
  int batch = Gal_GetInt(f, ":batch");
  int mit_playbox_compat = Gal_GetInt(f, ":mit_playbox_protocol");
  int sample_rate = Gal_GetInt(f, ":sample_rate");
  char *encoding_format = Gal_GetString(f, ":encoding_format");
  char *logfile = Gal_GetString(f, ":logfile");
  GalIO_BrokerStruct *b;
  
  if ((host && port) || bp) {
    ServerDataContainer *c = (ServerDataContainer *) calloc(1, sizeof(ServerDataContainer));

    c->data_buf = (char *) NULL;
    c->dtype = (Gal_ObjectType) -1;
    c->size = 0;
    c->batch = batch;
    c->playbox_compat = mit_playbox_compat;
    c->sample_rate = sample_rate;
    if (encoding_format) {
      c->encoding_format = strdup(encoding_format);
    } else {
      c->encoding_format = (char *) NULL;
    }
    c->b = (GalIO_BrokerStruct *) NULL;
    if (logfile) {
      c->logfile = strdup(logfile);
    } else {
      c->logfile = (char *) NULL;
    }

    if (host && port) {
      b = GalSS_EnvBrokerDataInInit((GalSS_Environment *) server_data,
				    host, port, f,
				    audio_in_handler, -1,
				    (void *) c,
				    /* __FreeDataContainer); */
				    NULL);
    } else {
      c->use_proxy = 1;
      b = GalSS_EnvBrokerProxyInInit((GalSS_Environment *) server_data,
				     bp, 
				     audio_in_handler, -1,
				     (void *) c,
				     /* __FreeDataContainer); */
				     NULL);
    }
    if (b) {
      GalIO_SetBrokerActive(b);
      GalIO_AddBrokerCallback(b, GAL_BROKER_DATA_DONE_EVENT,
			      __report_done, (void *) c);
      GalSS_EnvStartBroker((GalSS_Environment *) server_data, b, 0);
    }
    
    if (!batch) {
      /* If batch, wait till the end to set up the broker.
	 Otherwise, set it up now. */
      set_up_return_broker(c, (GalSS_Environment *) server_data);
    }

  }
  return f;
}
