/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
 * (c) Copyright 2000-2002 The MITRE Corporation (MITRE)
 *
 * Permission is hereby granted, without payment, to copy, use,
 * modify, display and distribute this software and its documentation,
 * if any, for any purpose, provided, first, that the US Government
 * and any of its agencies will not be charged any license fee and/or
 * royalties for the use of or access to said copyright software, and
 * provided further that the above copyright notice and the following
 * three paragraphs shall appear in all copies of this software,
 * including derivatives utilizing any portion of the copyright
 * software.  Use of this software constitutes acceptance of these
 * terms and conditions.
 *
 * IN NO EVENT SHALL MITRE BE LIABLE TO ANY PARTY FOR DIRECT,
 * INDIRECT,SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF MITRE
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * MITRE SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTIES
 * INCLUDING,BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND
 * NON-INFRINGEMENT.
 *
 * THE SOFTWARE IS PROVIDED "AS IS."  MITRE HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 */

package org.mitre.audiogui;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.Cursor;
import javax.swing.*;
import java.io.IOException;
import java.util.Collection;
import java.util.EventListener;
import java.util.EventObject;
import galaxy.server.ServerArgParser;
import galaxy.server.Server;
import galaxy.server.MainServer;
import galaxy.server.DispatchError;
import galaxy.util.Logger;
import galaxy.lang.Clause;
import galaxy.lang.GFrame;
import org.mitre.audiogui.event.*;

/**
 * The <code>Controller</code> class keeps track of the operating mode and 
 * state of an audio server, processes requests for mode and state changes, and
 * dispatches events indicating those changes to listeners implementing the 
 * {@link ControllerListener} interface or extending the 
 * {@link ControllerAdapter} adapter class.
 *
 * <p>In the context of an audio server, "states" refer to inherently dynamic
 * operating conditions.  States currently supported include:
 * <ul>
 * <li>Barge-in</li>
 * <li>GUI visibility</li>
 * <li>Keypad visibility</li>
 * <li>VAD muting state</li>
 * <li>On/off hook status</li>
 * <li>Playback status</li>
 * <li>Recording status</li>
 * <li>VAD on/off status</li>
 * <li>Voice detection status</li>
 * </ul>
 *
 * <p> To track event timing in situations where latency exists
 * between "start/stop" requests dispatched to other objects and activation of
 * requests by those objects, many states are tracked by "started" and "active"
 * fields.  "started" fields are typically (ultimately universally) manipulated
 * through "start/stop" methods invoked by an object requesting a particular
 * operation, whereas "active" fields are typically manipulated through
 * "activate/deactivate" methods invoked by the object responsible for 
 * implementing the request.
 *
 * <p> By contrast, "modes" are quasi-stationary conditions related to the 
 * general operation of the audio server.  Currently supported modes include:
 * <ul>
 * <li>Barge-in capability mode</li>
 * <li>Debugging printout capability mode</li>
 * <li>Controller GUI availability mode</li>
 * <li>Keypad availability mode</li>
 * <li>Half-duplex I/O mode</li>
 * <li>MIT audio playbox brokering protocol mode</li>
 * <li>Push-to-talk mode</li>
 * <li>Streaming record mode</li>
 * </ul>
 * Modes are manipulated through "enable/disable" methods invoked by objects 
 * requesting mode changes.
 *
 * @see ControllerListener
 * @see ControllerAdapter
 */

public class Controller implements EventGenerator 
{   
    // Mode flags.
    private boolean bargeInEnabled;
    private boolean debugEnabled;
    private boolean controllerEnabled;
    private boolean halfDuplexEnabled;
    private boolean keypadEnabled;
    private boolean playboxEnabled;
    private boolean pttEnabled;
    private boolean streamingEnabled;
    
    // State flags.
    private boolean initialized = false;
    private boolean guiVisible=false;
    private boolean keypadVisible=false;
    private boolean muteActive=true;
    private boolean onHook=true;
    private boolean recordStarted=false;
        
    private volatile boolean voiceActive=false;
    private volatile boolean recordStartable=false;
    private volatile boolean audioResendable=false;

    // Maintains the list of ControllerListeners.
    private Collection listeners = new java.util.ArrayList();
    
    // GUIs.
    private ControllerGUI controllerGUI=null;
    private KeypadGUI keypadGUI=null;

    // The Galaxy Communicator server associated with the controller.
    private Server server = null;

    // The Galaxy Communicator "main server" associated with the controller.
    private MainServer mainServer = null;

    private GFrame replyFrame = null;
    private long waitTimeoutMsec = 0;
    private Logger logger = null;

    /**
     * The Controller class constructor processes command line arguments to 
     * set mode parameters and default audio format settings.  Upon completion,
     * the constructor generates an {@link UpdateEvent} handled by 
     * listener objects.
     *
     * @param mainServer the server that owns this controller
     */
    public Controller(MainServer mainServer) 
    {

	this.mainServer = mainServer;
	logger = Logger.getLogger();
	
	ServerArgParser argParser = mainServer.getArgParser();
	debugEnabled = argParser.isArg("-debug");
	controllerEnabled = !argParser.isArg("-no_gui");
	keypadEnabled = !argParser.isArg("-no_keypad");
	waitTimeoutMsec = 1000 * argParser.getIntegerArg("-timeout_sec", 0);
	if(waitTimeoutMsec < 0)
	    waitTimeoutMsec = 0;
	
	Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
	Point p=new Point(screenSize.width/50,screenSize.height/50);
	
	if (controllerEnabled) {
	    controllerGUI = new ControllerGUI(this);
	    controllerGUI.setLocation(p);
	    controllerGUI.pack();
	}
	
	if (keypadEnabled) {
	    keypadGUI = new KeypadGUI(this, !controllerEnabled);
	    if (controllerGUI == null)
		keypadGUI.setLocation(p);
	    else {
		p.translate(controllerGUI.getWidth()+ screenSize.width/50,0);
		keypadGUI.setLocation(p);
	    }
	    keypadGUI.pack();
	}
    }
    
    /**
     * Updatest the state of this controller and its GUI's based on the 
     * information in the specified frame.
     *
     * @param frame the state frame
     */
    public void updateState(GFrame frame)
    { 
	Boolean status = null;

	status = frame.getIntAsBoolean(":barge_in_enabled");
	if(status != null) {
	    bargeInEnabled = status.booleanValue();
	}

	status = frame.getIntAsBoolean(":half_duplex_enabled");
	if(status != null) {
	    halfDuplexEnabled = status.booleanValue();
	}

	status = frame.getIntAsBoolean(":playbox_enabled");
	if(status != null) {
	    playboxEnabled = status.booleanValue();
	}

	status = frame.getIntAsBoolean(":ptt_enabled");
	if(status != null) {
	    pttEnabled = status.booleanValue();
	}

	status = frame.getIntAsBoolean(":streaming_enabled");
	if(status != null) {
	    streamingEnabled = status.booleanValue();
	}

	status = frame.getIntAsBoolean(":mute_active");
	if(status != null) {
	    muteActive = status.booleanValue();
	}

	status = frame.getIntAsBoolean(":on_hook");
	if(status != null) {
	     onHook = status.booleanValue();
	}

	status = frame.getIntAsBoolean(":record_started");
	if(status != null) {
	    recordStarted = status.booleanValue();
	} 

	status = frame.getIntAsBoolean(":record_startable");
	if(status != null) {
	    recordStartable = status.booleanValue();
	} 

	status = frame.getIntAsBoolean(":voice_active");
	if(status != null) {
	    voiceActive = status.booleanValue();
	}

	status = frame.getIntAsBoolean(":audio_resendable");
	if(status != null) {
	    audioResendable = status.booleanValue();
	}

	notifySubscribers(this, new UpdateEvent(this));

	if(controllerEnabled && !controllerGUI.isVisible())
	    setGUIVisible(true);

	if(keypadEnabled && !keypadGUI.isVisible())
	    setKeypadVisible(true);
    }

    /**
     * Sets the server associated with this controller.
     *
     * @param server the server
     */
    public void setServer(Server server) 
    {
	this.server = server;
    }

    /**
     * Sets the reply frame (which this contorller should be waiting for.
     *
     * @param replyFrame the reply frame
     */
    public synchronized void setReplyFrame(GFrame replyFrame) 
    {
	this.replyFrame = replyFrame;
    }

    /**
     * Causes this controller to wait (i.e., when waiting for a response
     * from the audio server).
     */
    private synchronized void doWait() 
    {
	// If replyFrame is not null, then the response has already arrived
	// and is awaiting processing.
	if(replyFrame == null) {
	    try {
		wait(waitTimeoutMsec);
	    } catch(Exception ex) {
		logger.logErrorMessage("Error while waiting.", ex, "Controller.doWait()");
	    }
	}
    }

    /**
     * Causes this controller to stop waiting.
     */
    public synchronized void doNotifyAll() 
    {
	try {
	    notifyAll();
	} catch(Exception ex) { 
	    logger.logErrorMessage("Error while notifying controller to wake up.", ex, "Controller.doNotifyAll()");
	}
    }

    /**
     * Sets the GUI mouse cursors to be the "wait" cursor.
     */
    private void waitCursor() 
    {
	if(keypadGUI != null)
	    keypadGUI.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
	if(controllerGUI != null)
	    controllerGUI.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
    }

    /**
     * Sets the GUI mouse cursors to be the default "arrow" cursor.
     */
    private void defaultCursor() 
    {
	if(keypadGUI != null)
	    keypadGUI.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
	if(controllerGUI != null)
	    controllerGUI.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
    }
    
    /** Indicates whether barge-in is enabled. */
    public boolean isBargeInEnabled() 
    {
	return bargeInEnabled;
    }
    
    /** Indicates whether debugging output is enabled. */ 
    public boolean isDebugEnabled() 
    {
	return debugEnabled;
    }
    
    /** Indicates whether the controller GUI is enabled. */
    public boolean isControllerEnabled() 
    {
	return controllerEnabled;
    }
    
    /** Indicates whether half-duplex operation is enabled. */
    public boolean isHalfDuplexEnabled() 
    {
	return halfDuplexEnabled;
    }
    
    /** Indicates whether the telephony keypad GUI is enabled. */
    public boolean isKeypadEnabled() 
    {
	return keypadEnabled;
    }
    
    /** 
     * Indicates whether the MIT audio playbox audio brokering protocol is 
     * enabled. 
     */
    public boolean isPlayboxEnabled() 
    {
	return playboxEnabled;
    }
    
    /** Indicates whether push-to-talk operation is enabled. */
    public boolean isPTTEnabled() 
    {
	return pttEnabled;
    }
    
    /** Indicates whether streaming audio brokering is enabled. */
    public boolean isStreamingEnabled() 
    {
	return streamingEnabled;
    }
    
    // Access information about current state.

    /** Indicates whether the controller GUI is visible. */
    public boolean isGUIVisible() 
    {
	return guiVisible;
    }
    
    /** Indicates whether the keypad GUI is visible. */
    public boolean isKeypadVisible() 
    {
	return keypadVisible;
    }
    
    /** Indicates whether VAD mute is active. */
    public boolean isMuteActive() 
    {
	return muteActive;
    }
    
    /** Indicates whether the audio server is in an on-hook state. */
    public boolean isOnHook() 
    {
	return onHook;
    }
    
    /** Indicates whether recording may be started. */
    public boolean isRecordStartable() 
    {
	return recordStartable;
    }
    
    /** Indicates whether a start recording request is in force. */
    public boolean isRecordStarted() 
    {
	return recordStarted;
    }
    
    public boolean isAudioResendable() 
    {
	return audioResendable; 
    }
    
    /** Indicates whether the voice state is active. */
    public boolean isVoiceActive() 
    {
	return voiceActive;
    }
  
    /**
     * Dispatches a frame to the audio server and tests the reply. 
     *
     * @param messageName name of the frame to send
     * @param keyName name to the key to set
     * @param keyValue value of key
     * @return if the reply corresponds to the message that was sent (i.e., the
     * key name and value match what was sent), return true. Else, return 
     * false.
     */
    private synchronized boolean dispatchFrame(String messageName, String keyName, boolean keyValue) 
    {
	GFrame frame = new Clause(messageName);
	int intValue =  keyValue ? 1 : 0;
	frame.setProperty(keyName, intValue);
	
	try {
	    server.getCurrentEnvironment().writeFrame(frame);
	    waitCursor();
	    doWait();
	} catch (IOException ioex) {
	    logger.logErrorMessage("Error writing frame.", ioex, "Controller.dispatchFrame(String, String, boolean)");
	    return false;
	} finally {
	    defaultCursor();
	}

	updateState(replyFrame);

	boolean result = ((replyFrame != null) && 
			  (replyFrame.containsProperty(keyName)) &&
			  (replyFrame.getInt(keyName).intValue() == intValue));
	replyFrame = null;
	return result;  
    }

    /**
     * Utility method for sending Galaxy Communicator frames.
     *
     * @param frame the frame to send     
     * @return true if operation successful, false otherwise
     */
    private boolean writeFrame(GFrame frame) 
    {
	try {
	    server.getCurrentEnvironment().writeFrame(frame);
	} catch (IOException ioex) {
	    logger.logErrorMessage("Error writing frame.", ioex, "Controller.writeFrame(GFrame)");
	    return false;
	}
	return true;
    }

    /** 
     * Sets debugging output mode to specified value.   
     *
     * @return true if operation successful, false otherwise 
     */
    public boolean setBargeInEnabled(boolean bargeInEnabled) 
    {
	if(!initialized || this.bargeInEnabled != bargeInEnabled)   
	    dispatchFrame("audio_gui", ":barge_in_enabled", bargeInEnabled);
	return this.bargeInEnabled == bargeInEnabled;
    }
    
    /** 
     * Sets debugging output mode to specified value.  Generates a 
     * corresponding {@link ModeEvent} event with an appropriate id.      
     *
     * @return true if operation successful, false otherwise
     */
    public boolean setDebugEnabled(boolean debugEnabled) 
    {
	if (this.debugEnabled != debugEnabled) {
	    this.debugEnabled = debugEnabled;
	    notifySubscribers(this, new ModeEvent(this, debugEnabled ? ModeEvent.DEBUG_ENABLED : ModeEvent.DEBUG_DISABLED));
	}
	return this.debugEnabled == debugEnabled;
    }
    
    /** 
     * Sets half-duplex mode to specified value. 
     *
     * @return true if operation successful, false otherwise
     */
    public boolean setHalfDuplexEnabled(boolean halfDuplexEnabled) 
    {
	if (!initialized || this.halfDuplexEnabled != halfDuplexEnabled)
	    dispatchFrame("audio_gui", ":half_duplex_enabled", halfDuplexEnabled);
	return this.halfDuplexEnabled == halfDuplexEnabled;
    }
    
    /** 
     * Sets MIT audio playbox brokering mode to specified value. 
     *
     * @return true if operation successful, false otherwise
     */
    public boolean setPlayboxEnabled(boolean playboxEnabled) 
    {
	if (!initialized || this.playboxEnabled != playboxEnabled)
	    dispatchFrame("audio_gui", ":playbox_enabled", playboxEnabled);
	return this.playboxEnabled == playboxEnabled;
    }

    /** 
     * Sets push-to-talk mode to specified value. 
     *
     * @return true if operation successful, false otherwise 
     */
    public boolean setPTTEnabled(boolean pttEnabled) 
    {
	if (!initialized || this.pttEnabled != pttEnabled)
	    dispatchFrame("audio_gui", ":ptt_enabled", pttEnabled);
	return this.pttEnabled == pttEnabled;
    }
    
    /** 
     * Sets streaming mode if not currently recording. 
     *
     * @return true if operation successful, false otherwise
     */
    public boolean setStreaming(boolean streamingEnabled) 
    {
	if (!initialized || this.streamingEnabled != streamingEnabled) 
	    dispatchFrame("audio_gui", ":streaming_enabled", streamingEnabled);
	return this.streamingEnabled == streamingEnabled;
    }    

    /** 
     * Sets GUI state to specified value if possible. Generates a 
     * {@link GUIEvent} event with an appropriate id.     
     *
     * @return true if operation successful, false otherwise
     */
    public boolean setGUIVisible(boolean guiVisible) 
    {
	if (this.guiVisible != guiVisible) {
	    this.guiVisible = guiVisible;
	    notifySubscribers(this, new GUIEvent(this, guiVisible ? GUIEvent.GUI_VISIBLE : GUIEvent.GUI_INVISIBLE));
	}
	return this.guiVisible == guiVisible;
    }
    
    /** 
     * Sets keypad GUI state to specified value if possible. Generates a
     * {@link GUIEvent} event with an appropriate id.     
     *
     * @return true if operation successful, false otherwise
     */
    public boolean setKeypadVisible(boolean keypadVisible) 
    {
	if (this.keypadVisible != keypadVisible) {
	    this.keypadVisible = keypadVisible;
	    notifySubscribers(this, new GUIEvent(this, keypadVisible ? GUIEvent.KEYPAD_VISIBLE : GUIEvent.KEYPAD_INVISIBLE));
	}
	return this.keypadVisible == keypadVisible;
    }
    
    /** 
     * Set VAD mute to specified value. 
     *
     * @return true if operation successful, false otherwise
     */
    public boolean setMuteActive(boolean muteActive) 
    {
	if (!initialized || this.muteActive != muteActive)
	    dispatchFrame("audio_gui", ":mute_active", muteActive);
	return this.muteActive == muteActive;
    }
        
    /** 
     * Set hook state to specified value.
     *
     * @return true if operation successful, false otherwise
     */
    public boolean setOnHook(boolean onHook) 
    {
	if (!initialized || this.onHook != onHook) 
	    dispatchFrame("audio_gui", ":on_hook", onHook);
	return this.onHook == onHook;
	
    }
    
    /** 
     * Perform on-hook immediately followed by off-hook.     
     *
     * @return true if operation successful, false otherwise
     */
    public boolean flash() 
    {
	if (!onHook)
	    return(setOnHook(true) && setOnHook(false));
	else
	    return true;
    }
    
    /** 
     * Generate {@link KeypadEvent} event with specified id value, provided 
     * the audio server is not on-hook at the time.
     *
     * @return true if operation successful, false otherwise
     *
     * @see Controller#isOnHook() isOnHook() 
     */
    public boolean touchtone(int id) 
    {
	if (!onHook) {
	    if(writeTouchtoneFrame(id))
		notifySubscribers(this, new KeypadEvent(this, id));
	    else
		return false;
	}
	return !onHook;
    }
    
    private boolean writeTouchtoneFrame(int id)
    {
	GFrame frame = new Clause("audio_gui");
	switch (id) {
	case KeypadEvent.TOUCHTONE_1:
	    frame.setProperty(":touchtone","1");
	    break;
	case KeypadEvent.TOUCHTONE_2:
	    frame.setProperty(":touchtone","2");
	    break;
	case KeypadEvent.TOUCHTONE_3:
	    frame.setProperty(":touchtone","3");
	    break;
	case KeypadEvent.TOUCHTONE_4:
	    frame.setProperty(":touchtone","4");
	    break;
	case KeypadEvent.TOUCHTONE_5:
	    frame.setProperty(":touchtone","5");
	    break;
	case KeypadEvent.TOUCHTONE_6:
	    frame.setProperty(":touchtone","6");
	    break;
	case KeypadEvent.TOUCHTONE_7:
	    frame.setProperty(":touchtone","7");
	    break;
	case KeypadEvent.TOUCHTONE_8:
	    frame.setProperty(":touchtone","8");
	    break;
	case KeypadEvent.TOUCHTONE_9:
	    frame.setProperty(":touchtone","9");
	    break;
	case KeypadEvent.TOUCHTONE_STAR:
	    frame.setProperty(":touchtone","*");
	    break;
	case KeypadEvent.TOUCHTONE_0:
	    frame.setProperty(":touchtone","0");
	    break;
	case KeypadEvent.TOUCHTONE_HASH:
	    frame.setProperty(":touchtone","#");
	    break;
	default:
	    return false;
	}
	return writeFrame(frame);
    }
    
    /** 
     * Set record start to specified value if possible.
     *
     * @return true if operation successful, false otherwise
     */
    public boolean setRecordStarted(boolean recordStarted) 
    {
	if (this.recordStarted != recordStarted)
	    dispatchFrame("audio_gui", ":record_started", recordStarted);
	return this.recordStarted == recordStarted;
    }
        
    /** 
     * Activate resend state if not recording and there is audio data to send.
     *
     * @return true if resend successful, false otherwise
     *
     * @see Controller#isRecordStarted() isRecordStarted() 
     */
    public boolean resend() 
    {
	if (isAudioResendable()) {
	    if(dispatchFrame("audio_gui", ":resend", true)) {
		if (debugEnabled)
		    logger.logMessage("Resending utterance.");
		return true;
	    }
	}
	return false;
    }
   
    //-------------------------------------------------------------------------
    // Implementation of EventGenerator interface.
    //-------------------------------------------------------------------------

    public void addEventListener(EventListener l) 
    {
	ControllerListener listener = (ControllerListener) l;
	synchronized (listeners) {
	    if (! listeners.contains(listener)) {
		listeners.add(listener);
	    }
	}
    }
    
    public void removeEventListener(EventListener l) 
    {
	ControllerListener listener = (ControllerListener) l;
	synchronized (listeners) {
	    listeners.remove(listener);
	}
    }
    
    public void notifySubscribers(Object source, EventObject e) 
    {
	ControllerListener listener;
	ControllerEvent event=(ControllerEvent) e;
	
	// Make a local copy of the current listeners so we can be thread safe.
	Object[] snapshot;
	synchronized (listeners) {
	    snapshot = listeners.toArray();
	}
	
	// Loop through each listener and pass on the event.
	for (int i = 0; i < snapshot.length; i++) {
	    listener = (ControllerListener) snapshot[i];
	    if (event instanceof UpdateEvent) {
		listener.eventGenerated((UpdateEvent) event);
	    }
	    else if (event instanceof KeypadEvent) {
		listener.eventGenerated((KeypadEvent) event);
	    }
	    else if (event instanceof ModeEvent) {
		listener.eventGenerated((ModeEvent) event);
	    }
	    else if (event instanceof GUIEvent) {
		listener.eventGenerated((GUIEvent) event);
	    }
	}
    }
}
