/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <stdlib.h>
#include <mdtm/mdtm.h>
#include <string.h>
#include "cats_audio_format.h"

static CatsAudioFormat CatsAudioFormat_destroy(CatsAudioFormat audioFormat);
static int CatsAudioFormat_getEncoding(CatsAudioFormat this);
static char *CatsAudioFormat_getEncodingString(CatsAudioFormat this);
static float CatsAudioFormat_getSampleRate(CatsAudioFormat this);
static int CatsAudioFormat_getSampleSizeInBits(CatsAudioFormat this);
static int CatsAudioFormat_getChannels(CatsAudioFormat this);
static int CatsAudioFormat_getFrameSize(CatsAudioFormat this);
static float CatsAudioFormat_getFrameRate(CatsAudioFormat this);
static int CatsAudioFormat_isBigEndian(CatsAudioFormat this);
static int CatsAudioFormat_matches(CatsAudioFormat this,
				   CatsAudioFormat format);
static char *CatsAudioFormat_getBigEndianString(CatsAudioFormat this);

struct CatsAudioFormatPrivate_str {
  /* SAM 9/19/02: This was mdtm_text, but on Windows, if you
     malloc in a DLL you need to free in a DLL, and there's
     no free function for mdtm_text. */
  char *encodingString;
  mdtm_text *bigEndianString;
  int bigEndian;
  int channels;
  int encoding;
  float frameRate;
  int frameSize;
  float sampleRate;
  int sampleSizeInBits;
};

CatsAudioFormat CatsAudioFormat_create(int encoding, float sampleRate,
				       int sampleSizeInBits, int channels,
				       int frameSize, float frameRate,
				       int bigEndian)
{
  CatsAudioFormat format=NULL;
  
  if ((format =
       (CatsAudioFormat) calloc(1,sizeof(struct CatsAudioFormat_str)))
      != NULL) {
    CatsAudioFormatPrivate private;

    format->destroy = CatsAudioFormat_destroy;
    format->getEncoding = CatsAudioFormat_getEncoding;
    format->getEncodingString = CatsAudioFormat_getEncodingString;
    format->getSampleRate = CatsAudioFormat_getSampleRate;
    format->getSampleSizeInBits = CatsAudioFormat_getSampleSizeInBits;
    format->getChannels = CatsAudioFormat_getChannels;
    format->getFrameSize = CatsAudioFormat_getFrameSize;
    format->getFrameRate = CatsAudioFormat_getFrameRate;
    format->isBigEndian = CatsAudioFormat_isBigEndian;
    format->matches = CatsAudioFormat_matches;
    format->getBigEndianString = CatsAudioFormat_getBigEndianString;

    if ((private = format->private =
	 (CatsAudioFormatPrivate)
	 calloc(1,sizeof(struct CatsAudioFormatPrivate_str)))
	== NULL)
      return(format->destroy(format));

    /* Set private fields */

    switch (encoding) {
    case CATS_ENCODING_ALAW:
      private->encodingString = strdup("ALAW");
      break;
    case CATS_ENCODING_PCM_SIGNED:
      private->encodingString = strdup("PCM_SIGNED");
      break;
    case CATS_ENCODING_PCM_UNSIGNED:
      private->encodingString = strdup("PCM_UNSIGNED");
      break;
    case CATS_ENCODING_ULAW:
      private->encodingString = strdup("ULAW");
      break;
    }

    if (private->encodingString == NULL)
      return(format->destroy(format));

    if ((private->bigEndianString =
	 mdtm_text_cpy(NULL,bigEndian ? "true" : "false")) == NULL)
      return(format->destroy(format));

    private->encoding = encoding;
    private->sampleRate = sampleRate;
    private->sampleSizeInBits = sampleSizeInBits;
    private->channels = channels;
    private->frameSize = frameSize;
    private->frameRate = frameRate;
    private->bigEndian = bigEndian;
  }

  return format;
}

static CatsAudioFormat CatsAudioFormat_destroy(CatsAudioFormat format)
{
  if (format->private != NULL) {
    if (format->private->encodingString != NULL)
      free(format->private->encodingString);
    free(format->private);
  }

  free(format);
  return NULL;
}

static int CatsAudioFormat_getEncoding(CatsAudioFormat this)
{
  return this->private->encoding;
}

static char *CatsAudioFormat_getEncodingString(CatsAudioFormat this)
{
  return this->private->encodingString;
}

static float CatsAudioFormat_getSampleRate(CatsAudioFormat this)
{
  return this->private->sampleRate;
}

static int CatsAudioFormat_getSampleSizeInBits(CatsAudioFormat this)
{
  return this->private->sampleSizeInBits;
}

static int CatsAudioFormat_getChannels(CatsAudioFormat this)
{
  return this->private->channels;
}

static int CatsAudioFormat_getFrameSize(CatsAudioFormat this)
{
  return this->private->frameSize;
}

static float CatsAudioFormat_getFrameRate(CatsAudioFormat this)
{
  return this->private->frameRate;
}

static int CatsAudioFormat_isBigEndian(CatsAudioFormat this)
{
  return this->private->bigEndian;
}

static int CatsAudioFormat_matches(CatsAudioFormat this,
				   CatsAudioFormat format)
{
  CatsAudioFormatPrivate thisPrivate=this->private;
  CatsAudioFormatPrivate formatPrivate=format->private;

  return thisPrivate->bigEndian == formatPrivate->bigEndian &&
    thisPrivate->channels == formatPrivate->channels &&
    thisPrivate->encoding == formatPrivate->encoding &&
    thisPrivate->frameRate == formatPrivate->frameRate &&
    thisPrivate->frameSize == formatPrivate->frameSize &&
    thisPrivate->sampleRate == formatPrivate->sampleRate &&
    thisPrivate->sampleSizeInBits == formatPrivate->sampleSizeInBits;
}

static char *CatsAudioFormat_getBigEndianString(CatsAudioFormat this)
{
  return this->private->bigEndianString;
}
