/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <stdlib.h>
#include <string.h>
#include <mdtm/mdtm.h>
#include <mdtm/mdtmmacros.h>
#include "cats_buffer.h"
#include "cats_lock_object.h"

#define CATS_DEFAULT_BUFFER_SIZE 1024

static CatsBuffer
CatsBuffer_destroy(CatsBuffer this);
static int CatsBuffer_readable(CatsBuffer this);
static char *CatsBuffer_read(CatsBuffer this, int len);
static int CatsBuffer_write(CatsBuffer this, char *b, int off,
				int len);
static void CatsBuffer_reset(CatsBuffer this);
static void CatsBuffer_flush(CatsBuffer this);
static int CatsBuffer_getBufferSize(CatsBuffer this);

/* Private methods */

static void CatsBuffer_lock(CatsBuffer this);
static void CatsBuffer_unlock(CatsBuffer this);

struct CatsBufferPrivate_str {
  int bufferSize;
  char *buffer;
  int readOffset;
  int readable;
  CatsLockObject lockObject;
};

CatsBuffer CatsBuffer_create(int bufferSize)
{
  CatsBuffer this=NULL;

  if ((this =
       (CatsBuffer)
       calloc(1,sizeof(struct CatsBuffer_str))) != NULL) {
    CatsBufferPrivate private=NULL;

    this->destroy = CatsBuffer_destroy;
    this->readable = CatsBuffer_readable;
    this->read = CatsBuffer_read;
    this->write = CatsBuffer_write;
    this->reset = CatsBuffer_reset;
    this->flush = CatsBuffer_flush;
    this->getBufferSize = CatsBuffer_getBufferSize;

    if ((private = this->private =
         (CatsBufferPrivate)
	 calloc(1,sizeof(struct CatsBufferPrivate_str)))
        == NULL)
      return(this->destroy(this));

    private->bufferSize = bufferSize > 0 ? bufferSize :
      CATS_DEFAULT_BUFFER_SIZE;
    if ((private->buffer = mdtm_vrchar_malloc(private->bufferSize)) == NULL)
      return(this->destroy(this));
    private->readable = private->readOffset = 0;

    if ((private->lockObject = CatsLockObject_create()) == NULL)
      return(this->destroy(this));
  }
  return this;
}

static CatsBuffer CatsBuffer_destroy(CatsBuffer this)
{
  if (this->private != NULL) {
    CatsBufferPrivate private=this->private;
    if (private->lockObject != NULL) {
      CatsBuffer_lock(this);
      private->lockObject->destroy(private->lockObject);
    }
    if (private->buffer != NULL)
      mdtm_vrchar_free(private->buffer);
    free(this->private);
  }
  free(this);
  return NULL;
}

static int CatsBuffer_write(CatsBuffer this, char *b, int off,
				int len)
{
  CatsLockObject lockObject=this->private->lockObject;
  int bytesWritten=0;

  lockObject->lock(lockObject);

  if (len > 0) {
    CatsBufferPrivate private=this->private;
    int writeable,writeOffset;

    writeOffset = private->readOffset + private->readable;
    writeable = private->bufferSize - writeOffset;

    if (len > writeable) {
      int oldBufferSize = private->bufferSize;
      int newBufferSize =
	1<<mdtm_srint_nextpow2(oldBufferSize+len-writeable);

      if ((private->buffer=
	   mdtm_vrchar_realloc(private->buffer,oldBufferSize,
			       newBufferSize)) == NULL) {
	CatsBuffer_unlock(this);
	return -1;
      }

      private->bufferSize = newBufferSize;
    }

    memcpy(private->buffer+writeOffset,b+off,len);
    bytesWritten = len;
    private->readable += len;
  }

  lockObject->unlock(lockObject);

  return bytesWritten;
}

static void CatsBuffer_reset(CatsBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;

  lockObject->lock(lockObject);
  this->private->readable += this->private->readOffset;
  this->private->readOffset = 0;
  lockObject->unlock(lockObject);
}

static void CatsBuffer_flush(CatsBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;

  lockObject->lock(lockObject);
  this->private->readable = this->private->readOffset = 0;
  lockObject->unlock(lockObject);
}

static int CatsBuffer_getBufferSize(CatsBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  int bufferSize;

  lockObject->lock(lockObject);
  bufferSize=this->private->bufferSize;
  lockObject->unlock(lockObject);
  return bufferSize;
}

static int CatsBuffer_readable(CatsBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  int retval;

  lockObject->lock(lockObject);
  retval = this->private->readable;
  lockObject->unlock(lockObject);

  return retval;
}

static char *CatsBuffer_read(CatsBuffer this, int len)
{
  CatsBufferPrivate private=this->private;
  CatsLockObject lockObject=private->lockObject;
  char *b;

  lockObject->lock(lockObject);

  while (private->readable < len)
    lockObject->notify(lockObject);

  b = private->buffer+private->readOffset;
  private->readOffset += len;
  private->readable -= len;

  lockObject->unlock(lockObject);

  return b;
}

static void CatsBuffer_lock(CatsBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  lockObject->lock(lockObject);
}

static void CatsBuffer_unlock(CatsBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  lockObject->unlock(lockObject);
}
