/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <stdio.h>
#include <stdlib.h>
#include <mdtm/mdtm.h>
#include <mdtm/mdtmmacros.h>
#include "cats_buffer.h"
#include "cats_capture.h"
#include "cats_target_data_line.h"
#include "cats_source_data_line.h"
#include "cats_event_listener.h"
#include "cats_controller_event.h"
#include "cats_thread.h"
#include "cats_vad_buffer.h"
#include "cats_lock_object.h"
#include "cats_vector.h"

#define DEBUG 0
#define CATS_LINE_BUFFER_LENGTH_SEC .25
#define CATS_FRAME_SIZE_MSEC 10

static CatsCapture CatsCapture_destroy(CatsCapture this);
static int CatsCapture_open(CatsCapture this);
static int CatsCapture_start(CatsCapture this);
static int CatsCapture_stop(CatsCapture this);
static int CatsCapture_close(CatsCapture this);
static int CatsCapture_isOpen(CatsCapture this);
static int CatsCapture_isRunning(CatsCapture this);
static int CatsCapture_isActive(CatsCapture this);
static int CatsCapture_isRecordStartable(CatsCapture this);
static int CatsCapture_startRecord(CatsCapture this);
static void CatsCapture_stopRecord(CatsCapture this);
static int CatsCapture_isResendable(CatsCapture this);
static int CatsCapture_resend(CatsCapture this);
static int CatsCapture_isVADStartable(CatsCapture this);
static int CatsCapture_startVAD(CatsCapture this);
static void CatsCapture_stopVAD(CatsCapture this);
static int CatsCapture_addListener(CatsCapture this, CatsEventListener l);
static void CatsCapture_removeListener(CatsCapture this, CatsEventListener l);

/* Private Methods */

/* Event notifier */

static void CatsCapture_notifySubscribers(CatsCapture this, CatsEvent event);

/* Controller listener Method */

static void CatsCapture_ControllerEventGenerated(CatsEventListener listener,
						 CatsEvent event);

/* Polling method */

static void CatsCapture_run(Gal_TaskPkg *taskPkg);

struct CatsCapturePrivate_str {

  /* Mode, state flags. */

  int opened;
  int started;
  int debugEnabled;
  int recordStarted;
  int vadStarted;
  int streamingEnabled;
  int muteActive;

  /* Audio-related fields. */

  CatsAudioFormat audioFormat;
  CatsTargetDataLine adLine;
  CatsSourceDataLine outBrokerLine;
  mdtm_vrchar *captureBuffer;
  int captureBufferSize;

  /* Supporting fields. */

  CatsVector listeners;
  CatsServer server;
  CatsController controller;
  CatsEventListener controllerListener;
  mdtm_text *errStr;
  Gal_TaskPkg *taskPkg;
  CatsVADBuffer vadBuffer;
  CatsBuffer lastUtterance;
  CatsAudioFormat lastAudioFormat;
  CatsLockObject lockObject;
};

CatsCapture CatsCapture_create(CatsController controller)
{
  CatsCapture this=NULL;

  if ((this = (CatsCapture) calloc(1,sizeof(struct CatsCapture_str)))
      != NULL) {
    CatsCapturePrivate private;

    if ((this->object = CatsObject_create(this,CATS_CAPTURE)) == NULL)
      return (this->destroy(this));

    this->destroy = CatsCapture_destroy;
    this->open = CatsCapture_open;
    this->start = CatsCapture_start;
    this->stop = CatsCapture_stop;
    this->close = CatsCapture_close;
    this->isOpen = CatsCapture_isOpen;
    this->isRunning = CatsCapture_isRunning;
    this->isActive = CatsCapture_isActive;
    this->isRecordStartable = CatsCapture_isRecordStartable;
    this->startRecord = CatsCapture_startRecord;
    this->stopRecord = CatsCapture_stopRecord;
    this->isResendable = CatsCapture_isResendable;
    this->resend = CatsCapture_resend;
    this->isVADStartable = CatsCapture_isVADStartable;
    this->startVAD = CatsCapture_startVAD;
    this->stopVAD = CatsCapture_stopVAD;
    this->addListener = CatsCapture_addListener;
    this->removeListener = CatsCapture_removeListener;

    if ((private = this->private =
         (CatsCapturePrivate)
	 calloc(1,sizeof(struct CatsCapturePrivate_str)))
        == NULL)
      return(this->destroy(this));

    if ((private->listeners = CatsVector_create(0,0)) == NULL)
      return (this->destroy(this));
    if ((private->lockObject = CatsLockObject_create()) == NULL)
      return (this->destroy(this));

    /* Create utterance buffers for resend */

    if ((private->lastUtterance = CatsBuffer_create(0)) == NULL)
      return (this->destroy(this));
    private->lastAudioFormat = NULL;

    private->server = controller->getServer(controller);
    private->controller = controller;

    private->opened = private->started = MDTM_FALSE;
    private->vadStarted = controller->isVADStarted(controller);
    private->recordStarted = controller->isRecordStarted(controller);
    private->streamingEnabled = controller->isStreamingEnabled(controller);
    private->muteActive = controller->isMuteActive(controller);

    private->audioFormat = NULL;
    private->adLine = NULL;
    private->outBrokerLine =
      private->server->getOutBrokerLine(private->server);

    private->captureBuffer = NULL;
    private->captureBufferSize = 0;

    /* INITIALIZE DEPENDENT FIELDS */

    /* Create controller listener instance */

    if ((private->controllerListener = CatsEventListener_create(this))
	== NULL)
      return(this->destroy(this));
    private->controllerListener->eventGenerated =
      CatsCapture_ControllerEventGenerated;
    controller->addListener(controller,private->controllerListener);

    if ((private->vadBuffer = CatsVADBuffer_create()) == NULL)
      return(this->destroy(this));

    if (controller->isRecordStarted(controller) || private->vadStarted)
      this->start(this);

    private->taskPkg = NULL;
  }

  return this;
}

static CatsCapture CatsCapture_destroy(CatsCapture this)
{
  if (this->private != NULL) {
    CatsCapturePrivate private=this->private;
    CatsEventListener controllerListener=private->controllerListener;
    CatsVector listeners=private->listeners;
    CatsLockObject lockObject=private->lockObject;
    CatsVADBuffer vadBuffer=private->vadBuffer;
    CatsBuffer lastUtterance=private->lastUtterance;
    CatsObject object=this->object;

    if (private->started)
      this->stop(this);
    if (private->opened)
      this->close(this);
    if (controllerListener != NULL)
      controllerListener->destroy(controllerListener);
    if (listeners != NULL)
      listeners->destroy(listeners);
    if (lockObject != NULL)
      lockObject->destroy(lockObject);
    if (vadBuffer != NULL)
      vadBuffer->destroy(vadBuffer);
    if (lastUtterance != NULL)
      lastUtterance->destroy(lastUtterance);
    if (object != NULL)
      object->destroy(object);
    free(private);
  }

  free(this);
  return NULL;
}

/* Opening the capture object starts the polling loop */

static int CatsCapture_open(CatsCapture this)
{
  this->private->lockObject->lock(this->private->lockObject);

  if (!this->private->opened) {
    CatsBuffer lastUtterance=this->private->lastUtterance;

    this->private->taskPkg =
      Gal_AddTask(CatsCapture_run,(void *)this,0,GAL_TRUE,NULL);
    this->private->opened = GAL_TRUE;

    /* Reset last utterance fields */

    lastUtterance->flush(lastUtterance);
    this->private->lastAudioFormat = NULL;
  }

  this->private->lockObject->unlock(this->private->lockObject);

  return this->private->opened == GAL_TRUE;
}

/* Tells the object to begin a capture session, sampling from the A/D
   and brokering data out */

static int CatsCapture_start(CatsCapture this)
{
  CatsCapturePrivate private=this->private;
  CatsLockObject lockObject=private->lockObject;

  lockObject->lock(lockObject);

  if (!private->started) {
    CatsController controller=private->controller;
    CatsTargetDataLine adLine;
    CatsAudioFormat audioFormat;
    CatsEvent event;
    int bytesPerSecond;
    int adLineBufferSize;

    /* Open the object if necessary to start polling loop */

    if (!private->opened)
      this->open(this);

    /* The audio format can change between sessions, so update it. */

    audioFormat = private->audioFormat =
      controller->getDefaultRecordAudioFormat(controller);

    bytesPerSecond = audioFormat->getFrameRate(audioFormat) *
      audioFormat->getFrameSize(audioFormat);

    /* Create capture buffer */

    private->captureBufferSize =
      CATS_FRAME_SIZE_MSEC*bytesPerSecond/1000;
          
    if ((private->captureBuffer =
	 mdtm_vrchar_malloc(private->captureBufferSize)) == NULL)
      goto unlock_return;

    /* Start A/D line last */

    adLine = private->adLine = controller->getADLine(controller);

    adLineBufferSize = bytesPerSecond*CATS_LINE_BUFFER_LENGTH_SEC;

#if (DEBUG)
    printf("AD line buffer size is %d\n", adLineBufferSize);
    fflush(stdout);
#endif

    if (!adLine->open(adLine,audioFormat,adLineBufferSize))
      goto unlock_return;

#if (DEBUG)
    printf("Capture buffer size is %d\n", private->captureBufferSize);
    fflush(stdout);
#endif

    if ((event =
	 CatsEvent_create(this,CATS_CAPTURE_EVENT,CATS_CAPTURE_LINE_OPENED))
	!= NULL) {
      CatsCapture_notifySubscribers(this,event);
      event->destroy(event);
    }

    adLine->start(adLine);

    private->started = MDTM_TRUE;
  }

 unlock_return:
  lockObject->unlock(lockObject);
  return private->started == MDTM_TRUE;
}

/* Stops a capture session */

static int CatsCapture_stop(CatsCapture this)
{
  CatsCapturePrivate private=this->private;

  private->lockObject->lock(private->lockObject);

  if (private->started) {
    CatsTargetDataLine adLine=private->adLine;
    CatsEvent event;

    /* Close down A/D line */

    private->started = MDTM_FALSE;
    adLine->stop(adLine);
    adLine->close(adLine);

    if ((event =
	 CatsEvent_create(this,CATS_CAPTURE_EVENT,CATS_CAPTURE_LINE_CLOSED))
	!= NULL) {
      CatsCapture_notifySubscribers(this,event);
      event->destroy(event);
    }

    /* Free capture buffer */

    private->captureBuffer =
      mdtm_vrchar_free(private->captureBuffer);
    private->captureBufferSize = 0;
  }

  private->lockObject->unlock(private->lockObject);

  return private->started == MDTM_FALSE;
}

/* Closing the capture object stops the polling loop, stopping the
   object first if necessary */

static int CatsCapture_close(CatsCapture this)
{
  CatsCapturePrivate private=this->private;

  private->lockObject->lock(private->lockObject);

  if (private->opened) {
    CatsBuffer lastUtterance=private->lastUtterance;

    private->opened = MDTM_FALSE;

    if (private->started)
      this->stop(this);

    /* Remove polling function task */

    Gal_RemoveTask(private->taskPkg);
    private->taskPkg = NULL;

    /* Reset last utterance fields */

    lastUtterance->flush(lastUtterance);
    private->lastAudioFormat = NULL;
  }

  private->lockObject->unlock(private->lockObject);

  return private->opened == MDTM_FALSE;
}

static int CatsCapture_isOpen(CatsCapture this)
{
  return this->private->opened;
}

static int CatsCapture_isRunning(CatsCapture this)
{
  return this->private->started;
}

static int CatsCapture_isActive(CatsCapture this)
{
  return this->private->recordStarted;
}

static int CatsCapture_isRecordStartable(CatsCapture this)
{
  CatsCapturePrivate private=this->private;
  CatsController controller=private->controller;

  return !(private->recordStarted || !private->opened ||
	   controller->getDefaultRecordAudioFormat(controller) == NULL ||
	   controller->getADLine(controller) == NULL);
}

static int CatsCapture_startRecord(CatsCapture this)
{
  CatsCapturePrivate private=this->private;

  if (this->isRecordStartable(this))
    private->recordStarted = MDTM_TRUE;
  return private->recordStarted == MDTM_TRUE;
}

static void CatsCapture_stopRecord(CatsCapture this)
{
  this->private->recordStarted = MDTM_FALSE;
}

static int CatsCapture_isResendable(CatsCapture this)
{
  CatsCapturePrivate private=this->private;
  CatsSourceDataLine outBrokerLine=private->outBrokerLine;
  CatsBuffer lastUtterance=private->lastUtterance;

  return !(private->recordStarted || private->lastAudioFormat == NULL ||
	   outBrokerLine == NULL || outBrokerLine->isOpen(outBrokerLine) ||
	   lastUtterance->readable(lastUtterance) == 0);
}

static int CatsCapture_resend(CatsCapture this)
{
  this->private->lockObject->lock(this->private->lockObject);

  if (this->isResendable(this)) {
    CatsCapturePrivate private=this->private;
    CatsSourceDataLine outBrokerLine=private->outBrokerLine;
    CatsBuffer lastUtterance=private->lastUtterance;
    CatsAudioFormat lastAudioFormat=private->lastAudioFormat;
    int len=lastUtterance->readable(lastUtterance);
    char *b=lastUtterance->read(lastUtterance,len);

    outBrokerLine->open(outBrokerLine,lastAudioFormat,0);
    outBrokerLine->start(outBrokerLine);
    outBrokerLine->write(outBrokerLine,b,0,len);
    outBrokerLine->stop(outBrokerLine);
    outBrokerLine->close(outBrokerLine);
    lastUtterance->reset(lastUtterance);

    private->lockObject->unlock(private->lockObject);
    return MDTM_TRUE;
  }

  this->private->lockObject->unlock(this->private->lockObject);
  return MDTM_FALSE;
}

static int CatsCapture_isVADStartable(CatsCapture this)
{
  CatsCapturePrivate private=this->private;
  CatsController controller=private->controller;

  return !(private->vadStarted || !private->opened ||
	   controller->getDefaultRecordAudioFormat(controller) == NULL ||
	   controller->getADLine(controller) == NULL);
}

static int CatsCapture_startVAD(CatsCapture this)
{
  CatsCapturePrivate private=this->private;

  if (this->isVADStartable(this))
    private->vadStarted = MDTM_TRUE;
  return private->vadStarted == MDTM_TRUE;
}

static void CatsCapture_stopVAD(CatsCapture this)
{
  this->private->vadStarted = MDTM_FALSE;
}

/* EventGenerator implementation. */

static int CatsCapture_addListener(CatsCapture this, CatsEventListener l)
{
  CatsServer server=this->private->server;
  CatsVector listeners=this->private->listeners;
  int retval;

  if (!(retval = listeners->add(listeners,l))) {
    GalUtil_Error("%s\n","Failed to add CatsEventListener");
    server->shutDown(server);
  }

  return retval;
}

static void CatsCapture_notifySubscribers(CatsCapture this, CatsEvent event)
{
  CatsVector listeners=this->private->listeners;
  CatsEventListener listener;
  int i;

  for (i=0; i<listeners->size(listeners); i++) {
    listener = listeners->elementAt(listeners,i);
    listener->eventGenerated(listener,event);
  }
}

static void CatsCapture_removeListener(CatsCapture this, CatsEventListener l)
{
  CatsVector listeners=this->private->listeners;
  listeners->removeElement(listeners,l);
}

/* Controller listener method - turns on/off appropriate flags,
   nothing more. */

static void CatsCapture_ControllerEventGenerated(CatsEventListener listener,
						 CatsEvent event)
{
  CatsCapture this = (CatsCapture) listener->getSource(listener);

  switch (event->eventType) {

  case CATS_MODE_EVENT:
    switch (event->eventID) {
    case CATS_STREAMING_DISABLED:
      this->private->streamingEnabled = MDTM_FALSE;
      break;
    case CATS_STREAMING_ENABLED:
      this->private->streamingEnabled = MDTM_TRUE;
      break;
    }
    break;

  case CATS_VAD_EVENT:
    switch (event->eventID) {
    case CATS_MUTE_ACTIVE:
      this->private->muteActive = MDTM_TRUE;
      break;
    case CATS_MUTE_INACTIVE:
      this->private->muteActive = MDTM_FALSE;
      break;
    }
    break;

  }
}

/* Polling function. */

/* SAM 9/23/02: We should drain the ring buffer whenever we have the
   chance. We will do this by looping while there are bytes available
   and those bytes are greater than the buffer size. Then we'll
   read in buffer size chunks. This buffer will be small enough so that
   we won't have to worry about leaving too much audio in the
   ring buffer that the PortAudio populates. */

static void CatsCapture_run(Gal_TaskPkg *taskPkg)
{
  CatsCapture this=Gal_TaskPkgData(taskPkg);
  CatsCapturePrivate private=this->private;
  CatsEvent event;
  int restart_ms = 5;

  if (private->opened) {
    CatsController controller=private->controller;

    /* Start or stop capture object according to conditions */

    if (!private->started && (private->recordStarted || private->vadStarted))
      this->start(this);

    if (private->started) {
      CatsTargetDataLine adLine=private->adLine;
      CatsSourceDataLine outBrokerLine=private->outBrokerLine;
      CatsVADBuffer vadBuffer=private->vadBuffer;
      char *captureBuffer=private->captureBuffer;
      int captureBufferSize=private->captureBufferSize;
      int bytesPassed;
      /* We could check this every time, or check it in
	 the beginning. The latter is a little riskier, but
	 I'm afraid that if the read thread continues to
	 push data, this loop will never exit. */
      int available = adLine->available(adLine);
      int totalPassed = 0;
      int first_time = 1;

      if (private->vadStarted) {

	/* Open VAD buffer if necessary */

	if (!vadBuffer->isOpen(vadBuffer)) {
	  int vadRaiseFactor=controller->getDefaultVadRaiseFactor(controller);
	  int vadLowerFactor=controller->getDefaultVadLowerFactor(controller);

	  if (!vadBuffer->open(vadBuffer,private->audioFormat,
			       private->captureBufferSize,vadRaiseFactor,
			       vadLowerFactor)) {
	    this->close(this);
	    return;
	  }

	  if ((event = CatsEvent_create(this,CATS_VAD_EVENT,CATS_VAD_STARTED))
	      != NULL) {
	    CatsCapture_notifySubscribers(this,event);
	    event->destroy(event);
	  }
	}
      }

      /* SAM 9/23/02: In this loop, there are things
	 which need to happen whether or not we've read any data.
	 So I need to set up this loop so that it always runs
	 at least once. */

      while (first_time || (available > captureBufferSize)) {
	int data_available = (available > captureBufferSize);

	if (data_available) {
	  bytesPassed = adLine->read(adLine,captureBuffer,0,captureBufferSize);
	  available -= bytesPassed;
	  totalPassed += bytesPassed;

	  if (bytesPassed != captureBufferSize)
	    break;
	} else {
	  bytesPassed = 0;
	}

	/* Perform VAD buffering if VAD started */

	if (private->vadStarted) {
	  int voiceActive;

	  /* EBG - Need to consider failure behavior here */

	  if (data_available) {
	    if (vadBuffer->write(vadBuffer,captureBuffer,0,bytesPassed) == -1 ||
		(voiceActive =
		 vadBuffer->read(vadBuffer,captureBuffer,0,bytesPassed)) == -1) {
	      this->close(this);
	      return;
	    }

	    /* SAM 9/23/02: Don't make any decisions about VAD
	       unless there's data available. */

	    /* Activate/deactivate voice state as necessary */

	    if (voiceActive) {
	      if (controller->isVoiceActivateable(controller) &&
		  !controller->isRequestRedundant(controller,
						  controller->activateVoice,NULL,
						  this->object))
		controller->activateVoice(controller,NULL,this->object);
	    } else if (controller->isVoiceActive(controller) &&
		       !controller->isRequestRedundant(controller,
						       controller->deactivateVoice,
						       NULL,this->object)) {
	      controller->deactivateVoice(controller,NULL,this->object);
	    }
	  }
	} else if (vadBuffer->isOpen(vadBuffer)) {
	  vadBuffer->close(vadBuffer);

	  if ((event = CatsEvent_create(this,CATS_VAD_EVENT,CATS_VAD_STOPPED))
	      != NULL) {
	    CatsCapture_notifySubscribers(this,event);
	    event->destroy(event);
	  }
	} /* if (private->vadStarted) */

	/* If recording started, write data to output broker line,
	   otherwise shut down broker as necessary. */

	if (private->recordStarted) {
	  CatsBuffer lastUtterance=private->lastUtterance;

	  /* Recording doesn't actually start until broker line is started */

	  if (!outBrokerLine->isOpen(outBrokerLine)) {
	    lastUtterance->flush(lastUtterance);
	    private->lastAudioFormat = private->audioFormat;

	    /* EBG - need to consider error behavior here */

	    if (!outBrokerLine->open(outBrokerLine,private->audioFormat,0)) {
	      this->close(this);
	      return;
	    }
	    outBrokerLine->start(outBrokerLine);

	    if ((event = CatsEvent_create(this,CATS_CAPTURE_EVENT,
					  CATS_RECORD_STARTED)) != NULL) {
	      CatsCapture_notifySubscribers(this,event);
	      event->destroy(event);
	    }
	  }

	  if (data_available) {
	    outBrokerLine->write(outBrokerLine,captureBuffer,0,bytesPassed);
	    lastUtterance->write(lastUtterance,captureBuffer,0,bytesPassed);
	  }

	} else if (outBrokerLine->isOpen(outBrokerLine)) {
	  outBrokerLine->stop(outBrokerLine);
	  outBrokerLine->close(outBrokerLine);
	  if ((event = CatsEvent_create(this,CATS_CAPTURE_EVENT,
					CATS_RECORD_STOPPED)) != NULL) {
	    CatsCapture_notifySubscribers(this,event);
	    event->destroy(event);
	  }
	}
	first_time = 0;
      }
            
#if (DEBUG)
      printf("Capture passing %d bytes total\n", totalPassed);
      fflush(stdout);
#endif

    } else { /* Sleep if not started to avoid runaway condition */

      restart_ms = CATS_FRAME_SIZE_MSEC * 10;

    }

    if (private->started &&
	!(private->recordStarted || private->vadStarted))
      this->stop(this);

    /* Set task to repeat polling function */

    Gal_ReAddTask(taskPkg,(void *)this,restart_ms,GAL_TRUE,NULL);

  } /* if (private->opened) */

}
