/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <stdlib.h>
#include <galaxy/galaxy.h>
#include <galaxy/util.h>
#include <mdtm/mdtm.h>
#include "cats_controller.h"
#include "cats_controller_event.h"
#include "cats_server.h"
#include "cats_source_data_line.h"
#include "cats_target_data_line.h"
#include "cats_capture.h"
#include "cats_playback.h"
#include "cats_switchboard.h"
#include "cats_vector.h"
#include "cats_request_queue.h"
#include "cats_lock_object.h"

#define CATS_POLL_MSEC 10

static int nextId=0;

static CatsController CatsController_destroy(CatsController this);
static int CatsController_open(CatsController this);
static void CatsController_close(CatsController this);
static void CatsController_shutDown(CatsController this);
static int CatsController_isBargeInEnabled(CatsController this);
static int CatsController_isDebugEnabled(CatsController this);
static int CatsController_isHalfDuplexEnabled(CatsController this);
static int CatsController_isMuteActive(CatsController this);
static int CatsController_isOnHook(CatsController this);
static int CatsController_isRecordStartable(CatsController this);
static int CatsController_isRecordStarted(CatsController this);
static int CatsController_isResendable(CatsController this);
static int CatsController_isPlaybackStartable(CatsController this);
static int CatsController_isPlaybackStarted(CatsController this);
static int CatsController_isPlayboxEnabled(CatsController this);
static int CatsController_isPTTEnabled(CatsController this);
static int CatsController_isStreamingEnabled(CatsController this);
static int CatsController_isCaptureLineOpen(CatsController this);
static int CatsController_isPlaybackLineOpen(CatsController this);
static int CatsController_useOutboundBrokers(CatsController this);
static int CatsController_isVADStartable(CatsController this);
static int CatsController_isVADStarted(CatsController this);
static int CatsController_isVoiceActivateable(CatsController this);
static int CatsController_isVoiceActive(CatsController this);
static int
CatsController_isRequestRedundant(CatsController this, void *caller,
				  void (*requesterCallback)(CatsRequest),
				  CatsObject requester);
static CatsAudioFormat CatsController_getRecordFormat(CatsController this);
static CatsAudioFormat CatsController_getPlaybackFormat(CatsController this);
static float CatsController_getDefaultVadRaiseFactor(CatsController this);
static float CatsController_getDefaultVadLowerFactor(CatsController this);
static CatsServer CatsController_getServer(CatsController this);
static void CatsController_setADLine(CatsController this,
				     CatsTargetDataLine line);
static CatsTargetDataLine CatsController_getADLine(CatsController this);
static void CatsController_setDALine(CatsController this,
				     CatsSourceDataLine line);
static CatsSourceDataLine CatsController_getDALine(CatsController this);
static void CatsController_setPhoneLine(CatsController this,
					CatsPhoneLine line);
static CatsPhoneLine CatsController_getPhoneLine(CatsController this);
static int CatsController_ensureSwitchboard(CatsController this);

static CatsRequest
CatsController_disableBargeIn(CatsController this,
			      void (*requesterCallback)(CatsRequest),
			      CatsObject requester);
static CatsRequest
CatsController_enableBargeIn(CatsController this,
			     void (*requesterCallback)(CatsRequest),
			     CatsObject requester);
static CatsRequest
CatsController_setBargeInEnabled(CatsController this,
				 void (*requesterCallback)(CatsRequest),
				 CatsObject requester, int bargeInEnabled);
static CatsRequest
CatsController_disableHalfDuplex(CatsController this,
				 void (*requesterCallback)(CatsRequest),
				 CatsObject requester);
static CatsRequest
CatsController_enableHalfDuplex(CatsController this,
				void (*requesterCallback)(CatsRequest),
				CatsObject requester);
static CatsRequest
CatsController_setHalfDuplexEnabled(CatsController this,
				    void (*requesterCallback)(CatsRequest),
				    CatsObject requester,
				    int halfDuplexEnabled);
static CatsRequest
CatsController_disablePlaybox(CatsController this,
			      void (*requesterCallback)(CatsRequest),
			      CatsObject requester);
static CatsRequest
CatsController_enablePlaybox(CatsController this,
			     void (*requesterCallback)(CatsRequest),
			     CatsObject requester);
static CatsRequest
CatsController_setPlayboxEnabled(CatsController this,
				 void (*requesterCallback)(CatsRequest),
				 CatsObject requester,
				 int playboxEnabled);
static CatsRequest
CatsController_disablePTT(CatsController this,
			  void (*requesterCallback)(CatsRequest),
			  CatsObject requester);
static CatsRequest
CatsController_enablePTT(CatsController this,
			 void (*requesterCallback)(CatsRequest),
			 CatsObject requester);
static CatsRequest
CatsController_setPTTEnabled(CatsController this,
			     void (*requesterCallback)(CatsRequest),
			     CatsObject requester, int pttEnabled);
static CatsRequest
CatsController_disableStreaming(CatsController this,
				void (*requesterCallback)(CatsRequest),
				CatsObject requester);
static CatsRequest
CatsController_enableStreaming(CatsController this,
			       void (*requesterCallback)(CatsRequest),
			       CatsObject requester);
static CatsRequest
CatsController_setStreamingEnabled(CatsController this,
				   void (*requesterCallback)(CatsRequest),
				   CatsObject requester,
				   int streamingEnabled);
static CatsRequest
CatsController_disableDebug(CatsController this,
			    void (*requesterCallback)(CatsRequest),
			    CatsObject requester);
static CatsRequest
CatsController_enableDebug(CatsController this,
			   void (*requesterCallback)(CatsRequest),
			   CatsObject requester);
static CatsRequest
CatsController_setDebugEnabled(CatsController this,
			       void (*requesterCallback)(CatsRequest),
			       CatsObject requester,
			       int streamingEnabled);
static CatsRequest
CatsController_activateMute(CatsController this,
			    void (*requesterCallback)(CatsRequest),
			    CatsObject requester);
static CatsRequest
CatsController_deactivateMute(CatsController this,
			      void (*requesterCallback)(CatsRequest),
			      CatsObject requester);
static CatsRequest
CatsController_offHook(CatsController this,
		       void (*requesterCallback)(CatsRequest),
		       CatsObject requester);
static CatsRequest
CatsController_onHook(CatsController this,
		      void (*requesterCallback)(CatsRequest),
		      CatsObject requester);
static CatsRequest
CatsController_touchtone(CatsController this, int id,
			 void (*requesterCallback)(CatsRequest),
			 CatsObject requester);
static CatsRequest
CatsController_startPlayback(CatsController this,
			    void (*requesterCallback)(CatsRequest),
			    CatsObject requester);
static CatsRequest
CatsController_stopPlayback(CatsController this,
			    void (*requesterCallback)(CatsRequest),
			    CatsObject requester);
static CatsRequest
CatsController_bargeIn(CatsController this,
		       void (*requesterCallback)(CatsRequest),
		       CatsObject requester);
static CatsRequest
CatsController_startRecord(CatsController this,
			   void (*requesterCallback)(CatsRequest),
			   CatsObject requester);
static CatsRequest
CatsController_stopRecord(CatsController this,
			  void (*requesterCallback)(CatsRequest),
			  CatsObject requester);
static CatsRequest
CatsController_resend(CatsController this,
		      void (*requesterCallback)(CatsRequest),
		      CatsObject requester);
static CatsRequest
CatsController_startVAD(CatsController this,
			void (*requesterCallback)(CatsRequest),
			CatsObject requester);
static CatsRequest
CatsController_stopVAD(CatsController this,
		       void (*requesterCallback)(CatsRequest),
		       CatsObject requester);
static CatsRequest
CatsController_activateVoice(CatsController this,
			     void (*requesterCallback)(CatsRequest),
			     CatsObject requester);
static CatsRequest
CatsController_deactivateVoice(CatsController this,
			       void (*requesterCallback)(CatsRequest),
			       CatsObject requester);

/* EventGenerator implementation. */

static int CatsController_addListener(CatsController this,
				      CatsEventListener l);
static void CatsController_removeListener(CatsController this,
					  CatsEventListener l);

/* PRIVATE METHODS */

/* Polling method */

static void CatsController_run(Gal_TaskPkg *taskPkg);

/* Event notifier */

static void CatsController_notifySubscribers(CatsController this,
					     CatsEvent event);

/* Capture, playback event handlers */

static void CatsController_captureEventGenerated(CatsEventListener listener,
						 CatsEvent event);
static void CatsController_playbackEventGenerated(CatsEventListener listener,
						  CatsEvent event);

/* Request monitor callbacks */

static void CatsController_disableBargeInCallback(CatsRequest request);
static void CatsController_enableBargeInCallback(CatsRequest request);
static void CatsController_disableHalfDuplexCallback(CatsRequest request);
static void CatsController_enableHalfDuplexCallback(CatsRequest request);
static void CatsController_disablePlayboxCallback(CatsRequest request);
static void CatsController_enablePlayboxCallback(CatsRequest request);
static void CatsController_disablePTTCallback(CatsRequest request);
static void CatsController_enablePTTCallback(CatsRequest request);
static void CatsController_disableStreamingCallback(CatsRequest request);
static void CatsController_enableStreamingCallback(CatsRequest request);
static void CatsController_disableDebugCallback(CatsRequest request);
static void CatsController_enableDebugCallback(CatsRequest request);

static void CatsController_activateMuteCallback(CatsRequest request);
static void CatsController_deactivateMuteCallback(CatsRequest request);
static void CatsController_offHookCallback(CatsRequest request);
static void CatsController_onHookCallback(CatsRequest request);
static void CatsController_touchtoneCallback(CatsRequest request);
static void CatsController_startPlaybackCallback(CatsRequest request);
static void CatsController_stopPlaybackCallback(CatsRequest request);
static void CatsController_bargeInCallback(CatsRequest request);
static void CatsController_startRecordCallback(CatsRequest request);
static void CatsController_stopRecordCallback(CatsRequest request);
static void CatsController_resendCallback(CatsRequest request);
static void CatsController_startVADCallback(CatsRequest request);
static void CatsController_stopVADCallback(CatsRequest request);
static void CatsController_activateVoiceCallback(CatsRequest request);
static void CatsController_deactivateVoiceCallback(CatsRequest request);

struct CatsControllerPrivate_str {

  /* Serially assigned unique identifier */

  int id;

  CatsControllerParameters parameters;
  /* Mode flags. */

  int bargeInDisabled;
  int debugEnabled;
  int halfDuplexEnabled;
  int playboxEnabled;
  int pttEnabled;
  int streamingDisabled;

  /* State flags. */

  int opened;
  int muteActive;
  int onHook;
  int playbackLineOpen;
  int playbackStarted;
  int captureLineOpen;
  int recordStarted;
  int vadStarted;
  int voiceActive;
  int bargeInStarted;

  /* Operating parameters */

  float defaultVadRaiseFactor;
  float defaultVadLowerFactor;
  CatsAudioFormat defaultRecordAudioFormat;
  CatsAudioFormat defaultPlaybackAudioFormat;

  /* Before 4.0, this should always be 1. In 4.0 or after,
     it should be zero by default, and can be enabled on the
     command line. In 4.0, broker proxies are the default. */
  int useBrokers;

  /* Supporting fields */

  /* Event listener fields */

  CatsVector listeners;
  CatsEventListener captureListener;
  CatsEventListener playbackListener;

  CatsServer server;
  CatsCapture capture;
  CatsPlayback playback;
  CatsSwitchboard switchboard;
  CatsLockObject lockObject;

  /* Audio device lines */

  CatsTargetDataLine adLine;
  CatsSourceDataLine daLine;

  /* Phone device line */

  CatsPhoneLine phoneLine;

  /* Timed task loop */

  Gal_TaskPkg *taskPkg;

  /* Request queue */

  CatsRequestQueue requestQueue;
};

/* Constructor */

CatsController CatsController_create(CatsControllerParameters parameters,
				     CatsServer server)
{
  CatsController this=NULL;

  if ((this =
       (CatsController) calloc(1,sizeof(struct CatsController_str)))
      != NULL) {
    CatsControllerPrivate private;
    CatsCapture capture;
    CatsPlayback playback;
    int frameSize;
    float frameRate;

    if ((this->object = CatsObject_create(this,CATS_CONTROLLER)) == NULL)
      return(this->destroy(this));

    /* Register public methods */

    this->destroy = CatsController_destroy;
    this->open = CatsController_open;
    this->close = CatsController_close;
    this->shutDown = CatsController_shutDown;
    this->isBargeInEnabled = CatsController_isBargeInEnabled;
    this->isDebugEnabled = CatsController_isDebugEnabled;
    this->isHalfDuplexEnabled = CatsController_isHalfDuplexEnabled;
    this->isPlayboxEnabled = CatsController_isPlayboxEnabled;
    this->isPTTEnabled = CatsController_isPTTEnabled;
    this->useOutboundBrokers = CatsController_useOutboundBrokers;
    this->isStreamingEnabled = CatsController_isStreamingEnabled;
    this->isCaptureLineOpen = CatsController_isCaptureLineOpen;
    this->isPlaybackLineOpen = CatsController_isPlaybackLineOpen;
    this->isMuteActive = CatsController_isMuteActive;
    this->isOnHook = CatsController_isOnHook;
    this->isRecordStartable = CatsController_isRecordStartable;
    this->isRecordStarted = CatsController_isRecordStarted;
    this->isResendable = CatsController_isResendable;
    this->isPlaybackStartable = CatsController_isPlaybackStartable;
    this->isPlaybackStarted = CatsController_isPlaybackStarted;
    this->isVADStartable = CatsController_isVADStartable;
    this->isVADStarted = CatsController_isVADStarted;
    this->isVoiceActivateable = CatsController_isVoiceActivateable;
    this->isVoiceActive = CatsController_isVoiceActive;
    this->isRequestRedundant = CatsController_isRequestRedundant;
    this->getDefaultRecordAudioFormat = CatsController_getRecordFormat;
    this->getDefaultPlaybackAudioFormat =
      CatsController_getPlaybackFormat;
    this->getDefaultVadRaiseFactor = CatsController_getDefaultVadRaiseFactor;
    this->getDefaultVadLowerFactor = CatsController_getDefaultVadLowerFactor;
    this->getServer = CatsController_getServer;
    this->setADLine = CatsController_setADLine;
    this->getADLine = CatsController_getADLine;
    this->setDALine = CatsController_setDALine;
    this->getDALine = CatsController_getDALine;
    this->setPhoneLine = CatsController_setPhoneLine;
    this->getPhoneLine = CatsController_getPhoneLine;
    this->ensureSwitchboard = CatsController_ensureSwitchboard;
    this->disableBargeIn = CatsController_disableBargeIn;
    this->enableBargeIn = CatsController_enableBargeIn;
    this->setBargeInEnabled = CatsController_setBargeInEnabled;
    this->disableHalfDuplex = CatsController_disableHalfDuplex;
    this->enableHalfDuplex = CatsController_enableHalfDuplex;
    this->setHalfDuplexEnabled = CatsController_setHalfDuplexEnabled;
    this->disablePlaybox = CatsController_disablePlaybox;
    this->enablePlaybox = CatsController_enablePlaybox;
    this->setPlayboxEnabled = CatsController_setPlayboxEnabled;
    this->disablePTT = CatsController_disablePTT;
    this->enablePTT = CatsController_enablePTT;
    this->setPTTEnabled = CatsController_setPTTEnabled;
    this->disableStreaming = CatsController_disableStreaming;
    this->enableStreaming = CatsController_enableStreaming;
    this->setStreamingEnabled = CatsController_setStreamingEnabled;
    this->disableDebug = CatsController_disableDebug;
    this->enableDebug = CatsController_enableDebug;
    this->setDebugEnabled = CatsController_setDebugEnabled;
    this->addListener = CatsController_addListener;
    this->removeListener = CatsController_removeListener;
    this->activateMute = CatsController_activateMute;
    this->deactivateMute = CatsController_deactivateMute;
    this->offHook = CatsController_offHook;
    this->onHook = CatsController_onHook;
    this->touchtone = CatsController_touchtone;
    this->startPlayback = CatsController_startPlayback;
    this->stopPlayback = CatsController_stopPlayback;
    this->bargeIn = CatsController_bargeIn;
    this->startRecord = CatsController_startRecord;
    this->stopRecord = CatsController_stopRecord;
    this->resend = CatsController_resend;
    this->startVAD = CatsController_startVAD;
    this->stopVAD = CatsController_stopVAD;
    this->activateVoice = CatsController_activateVoice;
    this->deactivateVoice = CatsController_deactivateVoice;

    if ((private = this->private =
	 (CatsControllerPrivate)
	 calloc(1,sizeof(struct CatsControllerPrivate_str)))
	== NULL)
      return(this->destroy(this));

    if ((private->listeners = CatsVector_create(0,0)) == NULL)
      return(this->destroy(this));

    if ((private->lockObject = CatsLockObject_create()) == NULL)
      return(this->destroy(this));

    private->id = nextId++;

    /* Initialize state */

    private->captureLineOpen = private->playbackLineOpen =
      private->playbackStarted = private->recordStarted =
      private->voiceActive = private->vadStarted = GAL_FALSE;
    private->muteActive = private->onHook = GAL_TRUE;

    /* SAM 10/1/02: the parameters struct is allocated outside
       this function, and is the responsibility of the caller
       to free. */
    
    private->parameters = parameters;
    private->debugEnabled = parameters->debugEnabled;
    private->halfDuplexEnabled = parameters->halfDuplexEnabled;
    private->bargeInDisabled = parameters->bargeInDisabled;
    private->streamingDisabled = parameters->streamingDisabled;
    private->playboxEnabled = parameters->playboxEnabled;
    private->pttEnabled = parameters->pttEnabled;
    
    /*
    ** Extract default record/playback audio format information:
    **
    ** '-record_encoding' specifies the encoding type used for recording
    ** '-record_sample_rate' specifies the sample rate used for recording
    ** '-record_sample_size' specifies the # of bits to use for recording
    ** '-record_channels' specifies the number of channels to record
    ** '-record_little_endian' specifies to use little endian byte order in
    **   recording
    **
    ** Similar '-playback_*' options are available.
    */

    frameSize = ((parameters->recordSampleSizeInBits+7)>>3)*(parameters->recordChannels);
    frameRate = parameters->recordSampleRate;

    if ((private->defaultRecordAudioFormat =
	 CatsAudioFormat_create(parameters->recordEncoding,
				parameters->recordSampleRate,
				parameters->recordSampleSizeInBits,
				parameters->recordChannels,
				frameSize,frameRate,
				!parameters->recordLittleEndian)) == NULL) {
      GalUtil_Error("%s","Failed to create record audio format\n");
      return(this->destroy(this));
    }

    frameSize = ((parameters->playbackSampleSizeInBits+7)>>3)*(parameters->playbackChannels);
    frameRate = parameters->playbackSampleRate;

    if ((private->defaultPlaybackAudioFormat =
	 CatsAudioFormat_create(parameters->playbackEncoding,
				parameters->playbackSampleRate,
				parameters->playbackSampleSizeInBits,
				parameters->playbackChannels,
				frameSize,frameRate,
				!parameters->playbackLittleEndian)) == NULL) {
      GalUtil_Error("%s","Failed to create playback audio format\n");
      return(this->destroy(this));
    }

    /* Extract VAD operating parameters if specified */

    private->defaultVadRaiseFactor = parameters->defaultVadRaiseFactor;
    private->defaultVadLowerFactor = parameters->defaultVadLowerFactor;

    /* SAM 10/1/02: If we're before GC 4.0, use brokers. If we're
       doing playbox compatibility, use brokers. If we've defaulted
       to using brokers on the command line, use brokers. Otherwise,
       use broker proxies. */
#if (GC_VERSION >= 0x40000)
    private->useBrokers = (private->playboxEnabled || parameters->useBrokers);
#else
    private->useBrokers = 1;
#endif

    private->server = server;

    /* INITIALIZE DEPENDENT FIELDS */

    if ((capture = private->capture = CatsCapture_create(this)) == NULL)
      return(this->destroy(this));
    capture->open(capture);

    if ((private->captureListener = CatsEventListener_create(this)) == NULL)
      return(this->destroy(this));
    private->captureListener->eventGenerated =
      CatsController_captureEventGenerated;
    capture->addListener(capture,private->captureListener);

    if ((playback = private->playback = CatsPlayback_create(this)) == NULL)
      return(this->destroy(this));
    playback->open(playback);

    if ((private->playbackListener = CatsEventListener_create(this)) == NULL)
      return(this->destroy(this));
    private->playbackListener->eventGenerated =
      CatsController_playbackEventGenerated;
    playback->addListener(playback,private->playbackListener);

    private->taskPkg = NULL;

    if ((private->requestQueue = CatsRequestQueue_create()) == NULL)
      return(this->destroy(this));

    /* CatsController_notifySubscribers(this,
       CATS_INITIALIZATION_EVENT,CATS_INITIALIZED);*/
  }

  return this;
}

static CatsController CatsController_destroy(CatsController this)
{
  CatsControllerPrivate private=this->private;
  if (private != NULL) {
    CatsAudioFormat recordFormat=private->defaultRecordAudioFormat;
    CatsAudioFormat playbackFormat=private->defaultPlaybackAudioFormat;
    CatsCapture capture=private->capture;
    CatsPlayback playback=private->playback;
    CatsVector listeners=private->listeners;
    CatsLockObject lockObject=private->lockObject;
    CatsObject object=this->object;
    CatsPhoneLine phoneLine = private->phoneLine;
    CatsSwitchboard switchboard = private->switchboard;
    
    if (lockObject != NULL)
      lockObject->lock(lockObject);

    this->close(this);

    if (recordFormat != NULL)
      recordFormat->destroy(recordFormat);
    if (playbackFormat != NULL)
      playbackFormat->destroy(playbackFormat);
    if (capture != NULL)
      capture->destroy(capture);
    if (playback != NULL)
      playback->destroy(playback);
    if (listeners != NULL)
      listeners->destroy(listeners);
    if (phoneLine != NULL)
      phoneLine->destroy(phoneLine);
    if (switchboard != NULL)
      switchboard->destroy(switchboard);

    if (lockObject != NULL) {
      lockObject->unlock(lockObject);
      lockObject->destroy(lockObject);
    }

    if (object != NULL)
      object->destroy(object);

    free(private);
  }

  free(this);
  return NULL;
}

/* PUBLIC METHOD IMPLEMENTATIONS */

/* Opening the controller object starts the polling loop */

static int CatsController_open(CatsController this)
{
  this->private->lockObject->lock(this->private->lockObject);

  if (!this->private->opened) {
    this->private->taskPkg =
      Gal_AddTask(CatsController_run,(void *)this,0,GAL_TRUE,NULL);
    this->private->opened = GAL_TRUE;
  }

  this->private->lockObject->unlock(this->private->lockObject);

  return this->private->opened == GAL_TRUE;
}

static void CatsController_close(CatsController this)
{
  CatsControllerPrivate private=this->private;

  private->lockObject->lock(private->lockObject);

  if (private->opened) {
    private->opened = GAL_FALSE;

    /* Remove polling function task */

    Gal_RemoveTask(private->taskPkg);
    private->taskPkg = NULL;
  }

  private->lockObject->unlock(private->lockObject);
}

static void CatsController_shutDown(CatsController this)
{
  CatsServer server=this->private->server;

  server->shutDown(server);
}

/* Access information about operating modes */

static int CatsController_isBargeInEnabled(CatsController this)
{
  return !this->private->bargeInDisabled;
}

static int CatsController_isDebugEnabled(CatsController this)
{
  return this->private->debugEnabled;
}

static int CatsController_isHalfDuplexEnabled(CatsController this)
{
  return this->private->halfDuplexEnabled;
}

static int CatsController_isPlayboxEnabled(CatsController this)
{
  return this->private->playboxEnabled;
}

static int CatsController_useOutboundBrokers(CatsController this)
{
  return this->private->useBrokers;
}

static int CatsController_isPTTEnabled(CatsController this)
{
  return this->private->pttEnabled;
}

static int CatsController_isStreamingEnabled(CatsController this)
{
  return !this->private->streamingDisabled;
}

/* Access information about current state. */

static int CatsController_isCaptureLineOpen(CatsController this)
{
  return this->private->captureLineOpen;
}

static int CatsController_isPlaybackLineOpen(CatsController this)
{
  return this->private->playbackLineOpen;
}

static int CatsController_isMuteActive(CatsController this)
{
  return this->private->muteActive;
}

static int CatsController_isOnHook(CatsController this)
{
  return this->private->onHook;
}

static int CatsController_isRecordStartable(CatsController this)
{
  CatsControllerPrivate private=this->private;
  CatsCapture capture=private->capture;

  return !(private->onHook || private->recordStarted ||
	   !capture->isRecordStartable(capture) ||
	   (this->isPlaybackStarted(this) && private->bargeInDisabled));
}

static int CatsController_isRecordStarted(CatsController this)
{
  return this->private->recordStarted;
}

static int CatsController_isResendable(CatsController this)
{
  CatsCapture capture=this->private->capture;
  return !this->private->onHook && capture->isResendable(capture);
}

static int CatsController_isPlaybackStartable(CatsController this)
{
  CatsControllerPrivate private=this->private;
  CatsPlayback playback=private->playback;
  CatsServer server=private->server;

  return !(private->onHook || private->playbackStarted ||
	   !playback->isPlaybackStartable(playback) ||
	   server->dataInBrokersAvailable(server) == 0 ||
	   (private->halfDuplexEnabled && private->recordStarted));
}

static int CatsController_isPlaybackStarted(CatsController this)
{
  return this->private->playbackStarted;
}

static int CatsController_isVADStartable(CatsController this)
{
  CatsControllerPrivate private=this->private;
  CatsServer server=private->server;
  CatsCapture capture=private->capture;
  CatsSourceDataLine outBrokerLine=server->getOutBrokerLine(server);

  return !(private->onHook || private->vadStarted ||
	   !capture->isVADStartable(capture) || outBrokerLine == NULL
	   || private->pttEnabled || private->muteActive ||
	   (private->halfDuplexEnabled && this->isPlaybackStarted(this)));
}

static int CatsController_isVADStarted(CatsController this)
{
  return this->private->vadStarted;
}

static int CatsController_isVoiceActivateable(CatsController this)
{
  CatsControllerPrivate private=this->private;
  return !(private->voiceActive || private->muteActive ||
	   !private->vadStarted || !this->isRecordStartable(this));
}

static int CatsController_isVoiceActive(CatsController this)
{
  return this->private->voiceActive;
}

static int
CatsController_isRequestRedundant(CatsController this, void *caller,
				  void (*requesterCallback)(CatsRequest),
				  CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;

  return requestQueue->isRedundant(requestQueue,caller,requesterCallback,
				   requester);
}

/* Get default operating parameters. */

static CatsAudioFormat CatsController_getRecordFormat(CatsController this)
{
  return this->private->defaultRecordAudioFormat;
}

static CatsAudioFormat CatsController_getPlaybackFormat(CatsController this)
{
  return this->private->defaultPlaybackAudioFormat;
}

static float CatsController_getDefaultVadRaiseFactor(CatsController this)
{
  return this->private->defaultVadRaiseFactor;
}

static float CatsController_getDefaultVadLowerFactor(CatsController this)
{
  return this->private->defaultVadLowerFactor;
}

static CatsServer CatsController_getServer(CatsController this)
{
  return this->private->server;
}

static void CatsController_setADLine(CatsController this,
				     CatsTargetDataLine line)
{
  this->private->adLine = line;
}

static CatsTargetDataLine CatsController_getADLine(CatsController this)
{
  return this->private->adLine;
}

static void CatsController_setDALine(CatsController this,
				     CatsSourceDataLine line)
{
  this->private->daLine = line;
}

static CatsSourceDataLine CatsController_getDALine(CatsController this)
{
  return this->private->daLine;
}

static void CatsController_setPhoneLine(CatsController this,
					CatsPhoneLine line)
{
  this->private->phoneLine = line;
}

static CatsPhoneLine CatsController_getPhoneLine(CatsController this)
{
  return this->private->phoneLine;
}


static int CatsController_ensureSwitchboard(CatsController this)
{
  if (this->private->switchboard) {
    return 1;
  } else {
    CatsSwitchboard sb = CatsSwitchboard_create(this);
    if (!sb) {
      return 0;
    } else if (!sb->open(sb)) {
      sb->destroy(sb);
      return 0;
    } else {
      this->private->switchboard = sb;
      return 1;
    }
  }
}


/* Methods to set operating modes. */

static CatsRequest
CatsController_disableBargeIn(CatsController this,
			      void (*requesterCallback)(CatsRequest),
			      CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->disableBargeIn,
				  CatsController_disableBargeInCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_disableBargeInCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (!this->private->bargeInDisabled) {
      CatsEvent event;

      this->private->bargeInDisabled = GAL_TRUE;
      if ((event = CatsEvent_create(requester,CATS_MODE_EVENT,
				    CATS_BARGE_IN_DISABLED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_enableBargeIn(CatsController this,
			     void (*requesterCallback)(CatsRequest),
			     CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->enableBargeIn,
				  CatsController_enableBargeInCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_enableBargeInCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->private->bargeInDisabled) {
      CatsEvent event;

      this->private->bargeInDisabled = GAL_FALSE;
      if ((event = CatsEvent_create(requester,CATS_MODE_EVENT,
				    CATS_BARGE_IN_ENABLED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_setBargeInEnabled(CatsController this,
				 void (*requesterCallback)(CatsRequest),
				 CatsObject requester, int bargeInEnabled)
{
  CatsRequest request=bargeInEnabled ?
    this->enableBargeIn(this,requesterCallback,requester) :
    this->disableBargeIn(this,requesterCallback,requester);

  request->requestData = (void *)bargeInEnabled;
  return request;
}

static CatsRequest
CatsController_disableHalfDuplex(CatsController this,
				 void (*requesterCallback)(CatsRequest),
				 CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->disableHalfDuplex,
				  CatsController_disableHalfDuplexCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_disableHalfDuplexCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->private->halfDuplexEnabled) {
      CatsEvent event;

      this->private->halfDuplexEnabled = GAL_FALSE;
      if ((event =
	   CatsEvent_create(requester,CATS_MODE_EVENT,
			    CATS_HALF_DUPLEX_DISABLED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_enableHalfDuplex(CatsController this,
				void (*requesterCallback)(CatsRequest),
				CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->enableHalfDuplex,
				  CatsController_enableHalfDuplexCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_enableHalfDuplexCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (!this->private->halfDuplexEnabled) {
      CatsEvent event;

      this->private->halfDuplexEnabled = GAL_TRUE;
      if ((event = CatsEvent_create(requester,CATS_MODE_EVENT,
				    CATS_HALF_DUPLEX_ENABLED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_setHalfDuplexEnabled(CatsController this,
				    void (*requesterCallback)(CatsRequest),
				    CatsObject requester,
				    int halfDuplexEnabled)
{
  CatsRequest request=halfDuplexEnabled ?
    this->enableHalfDuplex(this,requesterCallback,requester) :
    this->disableHalfDuplex(this,requesterCallback,requester);

  request->requestData=(void *)halfDuplexEnabled;
  return request;
}

static CatsRequest
CatsController_disablePlaybox(CatsController this,
			      void (*requesterCallback)(CatsRequest),
			      CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->disablePlaybox,
				  CatsController_disablePlayboxCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_disablePlayboxCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->private->playboxEnabled) {
      CatsEvent event;

      this->private->playboxEnabled = GAL_FALSE;
      if ((event = CatsEvent_create(requester,CATS_MODE_EVENT,
				    CATS_PLAYBOX_DISABLED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_enablePlaybox(CatsController this,
			     void (*requesterCallback)(CatsRequest),
			     CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->enablePlaybox,
				  CatsController_enablePlayboxCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_enablePlayboxCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (!this->private->playboxEnabled) {
      CatsEvent event;

      this->private->playboxEnabled = GAL_TRUE;
      if ((event = CatsEvent_create(requester,CATS_MODE_EVENT,
				    CATS_PLAYBOX_ENABLED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_setPlayboxEnabled(CatsController this,
				 void (*requesterCallback)(CatsRequest),
				 CatsObject requester,
				 int playboxEnabled)
{
  CatsRequest request=playboxEnabled ?
    this->enablePlaybox(this,requesterCallback,requester) :
    this->disablePlaybox(this,requesterCallback,requester);

  request->requestData=(void *)playboxEnabled;
  return request;
}

static CatsRequest
CatsController_disablePTT(CatsController this,
			  void (*requesterCallback)(CatsRequest),
			  CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->disablePTT,
				  CatsController_disablePTTCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_disablePTTCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->private->pttEnabled) {
      CatsEvent event;

      this->private->pttEnabled = GAL_FALSE;
      if ((event = CatsEvent_create(requester,CATS_MODE_EVENT,
				    CATS_PTT_DISABLED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_enablePTT(CatsController this,
			 void (*requesterCallback)(CatsRequest),
			 CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->enablePTT,
				  CatsController_enablePTTCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_enablePTTCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (!this->private->pttEnabled) {
      CatsEvent event;

      this->private->pttEnabled = GAL_TRUE;
      if ((event = CatsEvent_create(requester,CATS_MODE_EVENT,
				    CATS_PTT_ENABLED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_setPTTEnabled(CatsController this,
			     void (*requesterCallback)(CatsRequest),
			     CatsObject requester, int pttEnabled)
{
  CatsRequest request=pttEnabled ?
    this->enablePTT(this,requesterCallback,requester) :
    this->disablePTT(this,requesterCallback,requester);

  request->requestData=(void *)pttEnabled;
  return request;
}

static CatsRequest
CatsController_disableStreaming(CatsController this,
				void (*requesterCallback)(CatsRequest),
				CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->disableStreaming,
				  CatsController_disableStreamingCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_disableStreamingCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (!(this->private->streamingDisabled || this->private->recordStarted)) {
      CatsEvent event;

      this->private->streamingDisabled = GAL_TRUE;
      if ((event = CatsEvent_create(requester,CATS_MODE_EVENT,
				    CATS_STREAMING_DISABLED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_enableStreaming(CatsController this,
			       void (*requesterCallback)(CatsRequest),
			       CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->enableStreaming,
				  CatsController_enableStreamingCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_enableStreamingCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->private->streamingDisabled && !this->private->recordStarted) {
      CatsEvent event;

      this->private->streamingDisabled = GAL_FALSE;
      if ((event = CatsEvent_create(requester,CATS_MODE_EVENT,
				    CATS_STREAMING_ENABLED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_setStreamingEnabled(CatsController this,
				   void (*requesterCallback)(CatsRequest),
				   CatsObject requester, int streamingEnabled)
{
  CatsRequest request=streamingEnabled ?
    this->enableStreaming(this,requesterCallback,requester) :
    this->disableStreaming(this,requesterCallback,requester);

  request->requestData=(void *)streamingEnabled;
  return request;
}

static CatsRequest
CatsController_disableDebug(CatsController this,
			    void (*requesterCallback)(CatsRequest),
			    CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->disableDebug,
				  CatsController_disableDebugCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_disableDebugCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->private->debugEnabled) {
      CatsEvent event;

      this->private->debugEnabled = GAL_FALSE;
      if ((event = CatsEvent_create(requester,CATS_MODE_EVENT,
				    CATS_DEBUG_DISABLED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_enableDebug(CatsController this,
			   void (*requesterCallback)(CatsRequest),
			   CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->enableDebug,
				  CatsController_enableDebugCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_enableDebugCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (!this->private->debugEnabled) {
      CatsEvent event;

      this->private->debugEnabled = GAL_TRUE;
      if ((event = CatsEvent_create(requester,CATS_MODE_EVENT,
				    CATS_DEBUG_ENABLED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_setDebugEnabled(CatsController this,
			       void (*requesterCallback)(CatsRequest),
			       CatsObject requester, int debugEnabled)
{
  CatsRequest request=debugEnabled ?
    this->enableDebug(this,requesterCallback,requester) :
    this->disableDebug(this,requesterCallback,requester);

  request->requestData=(void *)debugEnabled;
  return request;
}

/* Methods to invoke requests to change current state. */

/* Mute. */

/* SAM 7/9/02: Note that we're putting everything into
   the notifier queue, even when the events are apparently
   synchronous. The reason for this is that any synchronous
   event which follows an ASYNCHRONOUS event (e.g., start playback,
   stop playback) must be processed after the asynchronous event
   actually happens. So it needs to go into the queue, always,
   because we don't know the context in which it's called. */

static CatsRequest
CatsController_activateMute(CatsController this,
			    void (*requesterCallback)(CatsRequest),
			    CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->activateMute,
				  CatsController_activateMuteCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL) {

    /* EBG: Note that invoking request methods with the current
       request object as the requester will cause requests created by
       the methods to be registered as children of the current
       request, per CatsRequest_create.  This enables detection of
       recursive requests and allows for correct dequeueing of
       dependent requests. */

    this->stopVAD(this,NULL,request->object);
    requestQueue->add(requestQueue,request);
  }

  return request;
}

static void CatsController_activateMuteCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (!this->private->muteActive) {
      CatsEvent event;

      this->private->muteActive = GAL_TRUE;
      if ((event = CatsEvent_create(requester,CATS_VAD_EVENT,
				    CATS_MUTE_ACTIVE)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_deactivateMute(CatsController this,
			      void (*requesterCallback)(CatsRequest),
			      CatsObject requester)
{
  CatsRequest request;
  CatsRequestQueue requestQueue=this->private->requestQueue;

  if ((request=CatsRequest_create(this->deactivateMute,
				  CatsController_deactivateMuteCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_deactivateMuteCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->private->muteActive) {
      CatsEvent event;

      this->private->muteActive = GAL_FALSE;
      if ((event = CatsEvent_create(requester,CATS_VAD_EVENT,
				    CATS_MUTE_INACTIVE)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

/* On/off hook. */

static CatsRequest
CatsController_offHook(CatsController this,
		       void (*requesterCallback)(CatsRequest),
		       CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->offHook,
				  CatsController_offHookCallback,this,
				  requesterCallback,requester,NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_offHookCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->private->onHook) {
      CatsEvent event;

      this->private->onHook = GAL_FALSE;
      if ((event = CatsEvent_create(requester,CATS_PHONE_EVENT,
				    CATS_OFF_HOOK)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_onHook(CatsController this,
		      void (*requesterCallback)(CatsRequest),
		      CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->onHook,
				  CatsController_onHookCallback,this,
				  requesterCallback,requester,NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_onHookCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (!this->private->onHook) {
      CatsEvent event;

      this->private->onHook = GAL_TRUE;
      if ((event = CatsEvent_create(requester,CATS_PHONE_EVENT,
				    CATS_ON_HOOK)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

static CatsRequest
CatsController_touchtone(CatsController this, int id,
			 void (*requesterCallback)(CatsRequest),
			 CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request =
       CatsRequest_create(this->touchtone,
			  CatsController_touchtoneCallback,this,
			  requesterCallback,requester,(void *)id)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_touchtoneCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (!this->private->onHook) {
      CatsEvent event;
      int eventID=(int)request->requestData;

      if ((event = CatsEvent_create(requester,CATS_PHONE_EVENT,eventID))
	  != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

/* Playback */

static CatsRequest
CatsController_startPlayback(CatsController this,
			     void (*requesterCallback)(CatsRequest),
			     CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->startPlayback,
				  CatsController_startPlaybackCallback,this,
				  requesterCallback,requester,NULL)) != NULL) {
    if (this->private->debugEnabled)
      printf("Controller queueing start playback request\n");
    requestQueue->add(requestQueue,request);
  }

  return request;
}

/* EBG: This is the first monitor callback that actually does
   something interesting. In addition to firing a request to the
   playback object, it dispatches an asynchronous event when success
   has been indicated by the playback object */

static void CatsController_startPlaybackCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->isPlaybackStartable(this)) {
      CatsPlayback playback=this->private->playback;

      playback->startPlayback(playback);
      request->status = CATS_REQUEST_PENDING;
    } else {
      request->status = CATS_REQUEST_FAILURE;
      if (this->private->debugEnabled)
	printf("Start playback request failed\n");
    }
  }

  if (request->status == CATS_REQUEST_PENDING) {
    if (this->private->playbackStarted) {
      CatsEvent event;

      if ((event = CatsEvent_create(requester,CATS_PLAYBACK_EVENT,
				    CATS_PLAYBACK_STARTED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      if (this->private->debugEnabled)
	printf("Start playback request succeeded\n");
      request->status = CATS_REQUEST_SUCCESS;
    }
  }
}

static CatsRequest
CatsController_stopPlayback(CatsController this,
			    void (*requesterCallback)(CatsRequest),
			    CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->stopPlayback,
				  CatsController_stopPlaybackCallback,this,
				  requesterCallback,requester,NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_stopPlaybackCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->private->playbackStarted) {
      CatsPlayback playback=this->private->playback;

      playback->stopPlayback(playback);
      request->status = CATS_REQUEST_PENDING;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }

  if (request->status == CATS_REQUEST_PENDING) {
    if (!this->private->playbackStarted) {
      CatsEvent event;

      if ((event = CatsEvent_create(requester,CATS_PLAYBACK_EVENT,
				    CATS_PLAYBACK_STOPPED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    }
  }
}

/* Barge-in. */

static CatsRequest
CatsController_bargeIn(CatsController this,
		       void (*requesterCallback)(CatsRequest),
		       CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->bargeIn,
				  CatsController_bargeInCallback,this,
				  requesterCallback,requester,NULL)) != NULL)
    requestQueue->add(requestQueue,request);

  return request;
}

static void CatsController_bargeInCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;
  CatsPlayback playback=this->private->playback;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (!this->private->bargeInDisabled &&
	playback->isBargeInStartable(playback)) {

      playback->bargeIn(playback);
      request->status = CATS_REQUEST_PENDING;
    } else {
      request->status = CATS_REQUEST_FAILURE;
    }
  }

  if (request->status == CATS_REQUEST_PENDING) {
    if (this->private->bargeInStarted) {
      CatsEvent event;

      this->private->bargeInStarted = GAL_FALSE;
      if ((event = CatsEvent_create(requester,CATS_PLAYBACK_EVENT,
				    CATS_BARGE_IN)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    }
  }
}

/* Recording. */

static CatsRequest
CatsController_startRecord(CatsController this,
			   void (*requesterCallback)(CatsRequest),
			   CatsObject requester)
{
  CatsRequest request;
  CatsRequestQueue requestQueue=this->private->requestQueue;

  if ((request=CatsRequest_create(this->startRecord,
				  CatsController_startRecordCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL) {
    if (!request->isRecursive(request)) {
      this->bargeIn(this,NULL,request->object);
      if (this->private->debugEnabled)
	printf("Controller queueing record start request\n");
      requestQueue->add(requestQueue,request);
    } else {
      request = request->destroy(request);
    }
  }

  return request;
}

static void CatsController_startRecordCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->isRecordStartable(this)) {
      CatsCapture capture=this->private->capture;

      capture->startRecord(capture);
      request->status = CATS_REQUEST_PENDING;
    } else {
      request->status = CATS_REQUEST_FAILURE;
      if (this->private->debugEnabled)
	printf("Record start request failed\n");
    }
  }

  if (request->status == CATS_REQUEST_PENDING) {
    if (this->private->recordStarted) {
      CatsEvent event;

      if ((event = CatsEvent_create(requester,CATS_CAPTURE_EVENT,
				    CATS_RECORD_STOPPED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
      if (this->private->debugEnabled)
	printf("Record start request succeeded\n");
    }
  }
}

static CatsRequest
CatsController_stopRecord(CatsController this,
			  void (*requesterCallback)(CatsRequest),
			  CatsObject requester)
{
  CatsRequest request;
  CatsRequestQueue requestQueue=this->private->requestQueue;

  if ((request=CatsRequest_create(this->stopRecord,
				  CatsController_stopRecordCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL) {
    if (!request->isRecursive(request)) {
      if (this->private->vadStarted)
	this->deactivateVoice(this,NULL,request->object);
      if (this->private->debugEnabled)
	printf("Controller queueing record stop request\n");
      requestQueue->add(requestQueue,request);
    } else {
      request = request->destroy(request);
    }
  }

  return request;
}

static void CatsController_stopRecordCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->private->recordStarted) {
      CatsCapture capture=this->private->capture;

      capture->stopRecord(capture);
      request->status = CATS_REQUEST_PENDING;
    } else {
      request->status = CATS_REQUEST_FAILURE;
      if (this->private->debugEnabled)
	printf("Record stop request failed\n");
    }
  }

  if (request->status == CATS_REQUEST_PENDING) {
    if (!this->private->recordStarted) {
      CatsEvent event;

      if ((event = CatsEvent_create(requester,CATS_CAPTURE_EVENT,
				    CATS_RECORD_STOPPED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
      if (this->private->debugEnabled)
	printf("Record stop request succeeded\n");
    }
  }
}

/* Resending. */

static CatsRequest
CatsController_resend(CatsController this,
		      void (*requesterCallback)(CatsRequest),
		      CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->resend,
				  CatsController_resendCallback,this,
				  requesterCallback,requester,NULL)) != NULL) {
    if (this->private->debugEnabled)
      printf("Controller queueing record stop request\n");
    requestQueue->add(requestQueue,request);
  }

  return request;
}

static void CatsController_resendCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->isResendable(this)) {
      CatsCapture capture=this->private->capture;
      CatsEvent event;

      capture->resend(capture);
      if ((event = CatsEvent_create(requester,CATS_CAPTURE_EVENT,
				    CATS_RESEND)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
      if (this->private->debugEnabled)
	printf("Resend request succeeded\n");
    } else {
      request->status = CATS_REQUEST_FAILURE;
      if (this->private->debugEnabled)
	printf("Resend request failed\n");
    }
  }
}

/* VAD monitoring state. */

static CatsRequest
CatsController_startVAD(CatsController this,
			void (*requesterCallback)(CatsRequest),
			CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->startVAD,
				  CatsController_startVADCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL) {
    if (!request->isRecursive(request)) {
      this->stopRecord(this,NULL,request->object);
      if (this->private->debugEnabled)
	printf("Controller queueing VAD start request\n");
      requestQueue->add(requestQueue,request);
    } else {
      request = request->destroy(request);
    }
  }

  return request;
}

static void CatsController_startVADCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->isVADStartable(this)) {
      CatsCapture capture=this->private->capture;

      capture->startVAD(capture);
      request->status = CATS_REQUEST_PENDING;
    } else {
      request->status = CATS_REQUEST_FAILURE;
      if (this->private->debugEnabled)
	printf("VAD start request failed\n");
    }
  }

  if (request->status == CATS_REQUEST_PENDING) {
    if (this->private->vadStarted) {
      CatsEvent event;

      if ((event = CatsEvent_create(requester,CATS_CAPTURE_EVENT,
				    CATS_VAD_STARTED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
      if (this->private->debugEnabled)
	printf("VAD start request succeeded\n");
    }
  }
}

static CatsRequest
CatsController_stopVAD(CatsController this,
		       void (*requesterCallback)(CatsRequest),
		       CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->stopVAD,
				  CatsController_stopVADCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL) {
    if (!request->isRecursive(request)) {
      this->deactivateVoice(this,NULL,request->object);
      if (this->private->debugEnabled)
	printf("Controller queueing VAD stop request\n");
      requestQueue->add(requestQueue,request);
    } else {
      request = request->destroy(request);
    }
  }

  return request;
}

static void CatsController_stopVADCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->isVADStarted(this)) {
      CatsCapture capture=this->private->capture;

      capture->stopVAD(capture);
      request->status = CATS_REQUEST_PENDING;
    } else {
      request->status = CATS_REQUEST_FAILURE;
      if (this->private->debugEnabled)
	printf("VAD stop request failed\n");
    }
  }

  if (request->status == CATS_REQUEST_PENDING) {
    if (!this->private->vadStarted) {
      CatsEvent event;

      if ((event = CatsEvent_create(requester,CATS_VAD_EVENT,
				    CATS_VAD_STOPPED)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
      if (this->private->debugEnabled)
	printf("VAD stop request succeeded\n");
    }
  }
}

/* Voice activity. */

static CatsRequest
CatsController_activateVoice(CatsController this,
			     void (*requesterCallback)(CatsRequest),
			     CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->activateVoice,
				  CatsController_activateVoiceCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL) {
    if (!request->isRecursive(request)) {
      this->startRecord(this,NULL,request->object);
      requestQueue->add(requestQueue,request);
      if (this->private->debugEnabled)
	printf("Controller queueing activate voice request\n");
    } else {
      request = request->destroy(request);
    }
  }

  return request;
}

static void CatsController_activateVoiceCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    CatsEvent event;

    this->private->voiceActive = GAL_TRUE;

    if ((event = CatsEvent_create(requester,CATS_VAD_EVENT,
				  CATS_VOICE_ACTIVE)) != NULL) {
      CatsController_notifySubscribers(this,event);
      event->destroy(event);
    }

    request->status = CATS_REQUEST_SUCCESS;
    if (this->private->debugEnabled)
      printf("Activate voice request succeeded\n");
  }
}

static CatsRequest
CatsController_deactivateVoice(CatsController this,
			       void (*requesterCallback)(CatsRequest),
			       CatsObject requester)
{
  CatsRequestQueue requestQueue=this->private->requestQueue;
  CatsRequest request;

  if ((request=CatsRequest_create(this->deactivateVoice,
				  CatsController_deactivateVoiceCallback,
				  this,requesterCallback,requester,
				  NULL)) != NULL) {
    if (!request->isRecursive(request)) {
      this->stopRecord(this,NULL,request->object);
      if (this->private->debugEnabled)
	printf("Controller queueing deactivate voice request\n");
      requestQueue->add(requestQueue,request);
      if (this->private->halfDuplexEnabled)
	this->stopVAD(this,NULL,request->object);
    } else {
      request = request->destroy(request);
    }
  }

  return request;
}

static void CatsController_deactivateVoiceCallback(CatsRequest request)
{
  CatsController this=(CatsController)request->monitor;
  void *requester=request->requester;

  if (request->status == CATS_REQUEST_INACTIVE) {
    if (this->private->voiceActive) {
      CatsEvent event;

      this->private->voiceActive = GAL_FALSE;
      if ((event = CatsEvent_create(requester,CATS_VAD_EVENT,
				    CATS_VOICE_INACTIVE)) != NULL) {
	CatsController_notifySubscribers(this,event);
	event->destroy(event);
      }
      request->status = CATS_REQUEST_SUCCESS;
    } else {
      if (this->private->debugEnabled)
	printf("Deactivate voice request failed\n");
      request->status = CATS_REQUEST_FAILURE;
    }
  }
}

/* EventGenerator implementation. */

static int CatsController_addListener(CatsController this,
				      CatsEventListener l)
{
  CatsServer server=this->private->server;
  CatsVector listeners=this->private->listeners;
  int retval;

  if (!(retval = listeners->add(listeners,l))) {
    GalUtil_Error("%s\n","Failed to add CatsEventListener");
    server->shutDown(server);
  }

  return retval;
}

static void CatsController_notifySubscribers(CatsController this,
					     CatsEvent event)
{
  CatsVector listeners=this->private->listeners;
  CatsEventListener listener;
  int i;

  for (i=0; i<listeners->size(listeners); i++) {
    listener = listeners->elementAt(listeners,i);
    listener->eventGenerated(listener,event);
  }
}

static void CatsController_removeListener(CatsController this,
					  CatsEventListener l)
{
  CatsVector listeners=this->private->listeners;
  listeners->removeElement(listeners,l);
}

static void CatsController_captureEventGenerated(CatsEventListener listener,
						 CatsEvent event)
{
  CatsController this=listener->getSource(listener);

  switch (event->eventType) {
  case CATS_CAPTURE_EVENT:
    switch(event->eventID) {
    case CATS_RECORD_STARTED:
      this->private->recordStarted = GAL_TRUE;
      break;
    case CATS_RECORD_STOPPED:
      this->private->recordStarted = GAL_FALSE;
      break;
    case CATS_CAPTURE_LINE_OPENED:
      this->private->captureLineOpen = GAL_TRUE;
      break;
    case CATS_CAPTURE_LINE_CLOSED:
      this->private->captureLineOpen = GAL_FALSE;
      break;
    }
    break;
  case CATS_VAD_EVENT:
    switch(event->eventID) {
    case CATS_VAD_STARTED:
      this->private->vadStarted = GAL_TRUE;
      break;
    case CATS_VAD_STOPPED:
      this->private->vadStarted = GAL_FALSE;
      break;
    }
    break;
  }
}

static void CatsController_playbackEventGenerated(CatsEventListener listener,
						  CatsEvent event)
{
  CatsController this=listener->getSource(listener);

  switch (event->eventType) {
  case CATS_PLAYBACK_EVENT:
    switch(event->eventID) {
    case CATS_PLAYBACK_STARTED:
      this->private->playbackStarted = GAL_TRUE;
      break;
    case CATS_PLAYBACK_STOPPED:
      this->private->playbackStarted = GAL_FALSE;
      break;
    case CATS_PLAYBACK_LINE_OPENED:
      this->private->playbackLineOpen = GAL_TRUE;
      break;
    case CATS_PLAYBACK_LINE_CLOSED:
      this->private->playbackLineOpen = GAL_FALSE;
      break;
    case CATS_BARGE_IN:
      this->private->bargeInStarted = GAL_TRUE;
      break;
    }
    break;
  }
}

/* Polling method */

static void CatsController_run(Gal_TaskPkg *taskPkg)
{
  CatsController this=Gal_TaskPkgData(taskPkg);
  CatsControllerPrivate private=this->private;

  if (private->opened) {
    CatsRequestQueue requestQueue=private->requestQueue;

    /* Always try to start VAD */

    if (this->isVADStartable(this) &&
	!this->isRequestRedundant(this,this->startVAD,NULL,this->object))
      this->startVAD(this,NULL,this->object);

    /* Service request queue */

    requestQueue->service(requestQueue);

    /* Refire task */

    Gal_ReAddTask(taskPkg,(void *)this,CATS_POLL_MSEC,GAL_FALSE,NULL);
  }
}
