/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <stdio.h>
#include <stdlib.h>
#include <galaxy/galaxy_io.h>
#include <galaxy/generic-server.h>
#include <mdtm/mdtm.h>
#include "cats_data_in_broker.h"
#include "cats_ring_buffer.h"
#include "cats_target_data_line.h"

#define DEBUG 0

static CatsDataInBroker CatsDataInBroker_destroy(CatsDataInBroker this);
static CatsAudioFormat CatsDataInBroker_getAudioFormat(CatsDataInBroker this);
static CatsTargetDataLine CatsDataInBroker_getLine(CatsDataInBroker this);
static int CatsDataInBroker_isFinished(CatsDataInBroker this);

/* CatsTargetDataLine implementation */

static int CatsDataInBroker_open(CatsTargetDataLine line,
				 CatsAudioFormat audioFormat, int bufferSize);
static int CatsDataInBroker_start(CatsTargetDataLine line);
static int CatsDataInBroker_read(CatsTargetDataLine line, char *b,int off,
				 int len);
static int CatsDataInBroker_stop(CatsTargetDataLine line);
static int CatsDataInBroker_close(CatsTargetDataLine line);
static int CatsDataInBroker_available(CatsTargetDataLine line);
static int CatsDataInBroker_isOpen(CatsTargetDataLine line);
static int CatsDataInBroker_isRunning(CatsTargetDataLine line);
static int CatsDataInBroker_isActive(CatsTargetDataLine line);
static int CatsDataInBroker_getBufferSize(CatsTargetDataLine line);
static CatsAudioFormat CatsDataInBroker_getFormat(CatsTargetDataLine line);
static int CatsDataInBroker_getFramePosition(CatsTargetDataLine line);

/* PRIVATE METHODS */

/* BrokerStruct callbacks */

static void CatsDataInBroker_run(GalIO_BrokerStruct *broker_struct, 
				 void *data, Gal_ObjectType data_type, 
				 int n_samples);
static void CatsDataInBroker_abort(GalIO_BrokerStruct *brokerStruct,
				   void *data);
static void CatsDataInBroker_done(GalIO_BrokerStruct *brokerStruct,
				  void *data);
static void CatsDataInBroker_cleanup(GalIO_BrokerStruct *brokerStruct,
				     void *data);

static CatsAudioFormat
CatsDataInBroker_frameToFormat(Gal_Frame frame,
			       CatsAudioFormat defaultFormat);

struct CatsDataInBrokerPrivate_str {
  CatsServer server;
  int opened;
  int started;
  int activated;
  CatsTargetDataLine line;
  CatsRingBuffer buffer;
  GalIO_BrokerStruct *brokerStruct;
  CatsAudioFormat audioFormat;
  int frameSize;
  int framePosition;
};

CatsDataInBroker CatsDataInBroker_create(CatsServer server,
					 Gal_Frame brokerFrame)
{
  CatsDataInBroker this=NULL;

  if ((this =
       (CatsDataInBroker)
       calloc(1,sizeof(struct CatsDataInBroker_str))) != NULL) {
    CatsDataInBrokerPrivate private;
    CatsTargetDataLine line;
    GalSS_Environment *env=server->getEnvironment(server);
    CatsController controller=server->getController(server);
    CatsAudioFormat defaultFormat=
      controller->getDefaultPlaybackAudioFormat(controller);
    CatsAudioFormat audioFormat;
    GalIO_BrokerStruct *brokerStruct;
    char *host=Gal_GetString(brokerFrame,":binary_host");
    unsigned short port=Gal_GetInt(brokerFrame,":binary_port");
#if (GC_VERSION >= 0x40000)
    GalSS_BrokerProxy *bp = Gal_GetProxy(brokerFrame, ":binary_proxy");
    int initial_poll_ms = -1;
#else
    void *bp = (void *) NULL;
    int initial_poll_ms = 0;
#endif

    this->destroy = CatsDataInBroker_destroy;
    this->getAudioFormat = CatsDataInBroker_getAudioFormat;
    this->getLine = CatsDataInBroker_getLine;
    this->isFinished = CatsDataInBroker_isFinished;

    if ((private = this->private =
         (CatsDataInBrokerPrivate)
	 calloc(1,sizeof(struct CatsDataInBrokerPrivate_str)))
        == NULL)
      return(this->destroy(this));

    /* Create buffer */

    if ((private->buffer = CatsRingBuffer_create(0)) == NULL)
      return(this->destroy(this));

    /* Create and activate broker struct */

    if (!((host && port) || bp))
      return(this->destroy(this));

    if (host && port) {
      if ((private->brokerStruct = brokerStruct = 
	   GalSS_EnvBrokerDataInInit(env,host,port,brokerFrame,
				     CatsDataInBroker_run,
				     initial_poll_ms,this,NULL)) == NULL) {
	return(this->destroy(this));
      }
    } 
#if (GC_VERSION >= 0x40000)
    else if ((private->brokerStruct = brokerStruct = 
		GalSS_EnvBrokerProxyInInit(env,bp,
					  CatsDataInBroker_run,-1,this,NULL)) == NULL) {
      return(this->destroy(this));
    }
#endif
    
    GalIO_AddBrokerCallback(brokerStruct,GAL_BROKER_ABORT_EVENT, 
			    CatsDataInBroker_abort,this);
    GalIO_AddBrokerCallback(brokerStruct,GAL_BROKER_DATA_DONE_EVENT, 
			    CatsDataInBroker_done,this);
    GalIO_AddBrokerCallback(brokerStruct, GAL_BROKER_DESTRUCTION_EVENT,
			    CatsDataInBroker_cleanup,this);
    GalIO_SetBrokerActive(private->brokerStruct);
#if (GC_VERSION >= 0x40000)
    GalSS_EnvStartBroker(env, private->brokerStruct, 0);
#endif
    private->activated = GAL_TRUE;

    /* Generate audio format from broker frame */

    audioFormat = private->audioFormat =
      CatsDataInBroker_frameToFormat(brokerFrame,defaultFormat);
    private->frameSize = audioFormat->getFrameSize(audioFormat);
    private->framePosition = 0;

    if ((line = private->line = CatsTargetDataLine_create(this)) == NULL)
      return(this->destroy(this));

    line->open = CatsDataInBroker_open;
    line->start = CatsDataInBroker_start;
    line->read = CatsDataInBroker_read;
    line->stop = CatsDataInBroker_stop;
    line->close = CatsDataInBroker_close;
    line->available = CatsDataInBroker_available;
    line->isOpen = CatsDataInBroker_isOpen;
    line->isRunning = CatsDataInBroker_isRunning;
    line->isActive = CatsDataInBroker_isActive;
    line->getBufferSize = CatsDataInBroker_getBufferSize;
    line->getFormat = CatsDataInBroker_getFormat;
    line->getFramePosition = CatsDataInBroker_getFramePosition;

    private->server = server;
    private->opened = private->started = GAL_FALSE;

  }

  return this;
}

static CatsDataInBroker CatsDataInBroker_destroy(CatsDataInBroker this)
{
  if (this->private != NULL) {
    CatsRingBuffer buffer=this->private->buffer;
    CatsTargetDataLine line=this->private->line;

    if (line != NULL) {
      if (line->isRunning(line))
	line->stop(line);
      if (line->isOpen(line))
	line->close(line);
      line->destroy(line);
    }

    if (buffer != NULL)
      buffer->destroy(buffer);
    free(this->private);
  }
  free(this);
  return NULL;
}

static CatsAudioFormat CatsDataInBroker_getAudioFormat(CatsDataInBroker this)
{
  return this->private->audioFormat;
}

static CatsTargetDataLine CatsDataInBroker_getLine(CatsDataInBroker this)
{
  return this->private->line;
}

/* CatsTargetDataLine implementation */

static int CatsDataInBroker_open(CatsTargetDataLine line,
				 CatsAudioFormat audioFormat, int bufferSize)
{
  CatsDataInBroker this=line->getSource(line);
  CatsDataInBrokerPrivate private=this->private;

  if (!private->opened &&
      audioFormat->matches(audioFormat,private->audioFormat))
    private->opened = GAL_TRUE;

  return private->opened == GAL_TRUE;
}

static int CatsDataInBroker_start(CatsTargetDataLine line)
{
  CatsDataInBroker this=line->getSource(line);
  CatsDataInBrokerPrivate private=this->private;

  if (private->opened && !private->started)
    private->started = GAL_TRUE;

  return private->started == GAL_TRUE;
}

static int CatsDataInBroker_read(CatsTargetDataLine line, char *b,int off,
				 int len)
{
  CatsDataInBroker this=line->getSource(line);
  CatsRingBuffer buffer=this->private->buffer;
  return (buffer->read(buffer,b,off,len));
}

/* Stopping the broker line terminates the broker */

static int CatsDataInBroker_stop(CatsTargetDataLine line)
{
  CatsDataInBroker this=line->getSource(line);
  if (this->private->started) {
    GalIO_BrokerStruct *brokerStruct=this->private->brokerStruct;
    this->private->started = GAL_FALSE;
    if (brokerStruct != NULL) {
      GalIO_BrokerDataDone(brokerStruct);
      GalIO_DestroyBrokerStruct(brokerStruct);
    }
  }
  return this->private->started == GAL_FALSE;
}

/* EBG: Closing the broker destroys it.  For obvious reasons,
   reopening it is not recommended. */

static int CatsDataInBroker_close(CatsTargetDataLine line)
{
  CatsDataInBroker this=line->getSource(line);

  if (this->private->opened) {
    this->private->opened = GAL_FALSE;
    if (line->isRunning(line))
      line->stop(line);
    this->destroy(this);
    return GAL_TRUE;
  }

  return GAL_FALSE;
}

static int CatsDataInBroker_available(CatsTargetDataLine line)
{
  CatsDataInBroker this=line->getSource(line);
  CatsRingBuffer buffer=this->private->buffer;
  return buffer->readable(buffer);
}

static int CatsDataInBroker_isOpen(CatsTargetDataLine line)
{
  CatsDataInBroker this=line->getSource(line);
  return this->private->opened;
}

static int CatsDataInBroker_isRunning(CatsTargetDataLine line)
{
  CatsDataInBroker this=line->getSource(line);
  return this->private->started;
}

static int CatsDataInBroker_isActive(CatsTargetDataLine line)
{
  CatsDataInBroker this=line->getSource(line);
  return this->private->activated;
}

static int CatsDataInBroker_getBufferSize(CatsTargetDataLine line)
{
  CatsDataInBroker this=line->getSource(line);
  CatsRingBuffer buffer=this->private->buffer;
  return buffer->getBufferSize(buffer);
}

static CatsAudioFormat CatsDataInBroker_getFormat(CatsTargetDataLine line)
{
  CatsDataInBroker this=line->getSource(line);
  return this->private->audioFormat;
}

static int CatsDataInBroker_getFramePosition(CatsTargetDataLine line)
{
  CatsDataInBroker this=line->getSource(line);
  return this->private->framePosition;
}

static int CatsDataInBroker_isFinished(CatsDataInBroker this)
{
  CatsTargetDataLine line=this->private->line;
  return line->available(line) == 0 && !this->private->started;
}

static void CatsDataInBroker_run(GalIO_BrokerStruct *broker_struct, 
				 void *data, Gal_ObjectType data_type, 
				 int n_samples)
{
  CatsDataInBroker this=GalIO_GetBrokerData(broker_struct);
  CatsRingBuffer buffer=this->private->buffer;
  int len=0;
  int found_data = 0;

  switch(data_type) {
  case GAL_BINARY:
#if (DEBUG)
    printf("Input broker receiving %d bytes\n",n_samples);
    fflush(stdout);
#endif
    len = n_samples;
    found_data = 1;
    break;
  case GAL_INT_16:
    len = n_samples << 1;
    found_data = 1;
    break;
  default:
	break;
  }

  if (found_data) {
    buffer->write(buffer,(char *)data,0,len);

    this->private->framePosition += len / this->private->frameSize;
  }
  /* SAM 10/1/02: The really clean way of doing this is to
     make sure that you free according to type. This function should
     be in the core, but it isn't. */
  
  switch (data_type) {
  case GAL_FRAME:
    Gal_FreeFrame((Gal_Frame) data);
    break;
  case GAL_PROXY:
    GalSS_FreeBrokerProxy((GalSS_BrokerProxy *) data);
    break;
  case GAL_LIST:
    if (data) {
      int i;

      for (i = 0; i < n_samples; i++) {
	Gal_FreeObject(((Gal_Object *) data)[i]);
      }
      free(data);
    }
    break;
  case GAL_STRING:
  case GAL_TOKEN:
  case GAL_FLOAT:
  case GAL_BINARY:
  case GAL_INT_16:
  case GAL_INT_32:
  case GAL_INT_64:
  case GAL_FLOAT_32:
  case GAL_FLOAT_64:
    free(data);
    break;
  default:
    /* Don't do anything with GAL_SYMBOL, GAL_KEYWORD,
       GAL_TAG, GAL_INT. */
    break;
  }
}

static void CatsDataInBroker_abort(GalIO_BrokerStruct *brokerStruct,
				   void *data)
{
  CatsDataInBroker this=(CatsDataInBroker)data;
  CatsServer server=this->private->server;
  GalSS_Environment *env=server->getEnvironment(server);
  Gal_Frame frame;

  this->private->activated = GAL_FALSE;
  frame = Gal_MakeFrame("notify", GAL_CLAUSE); 
  Gal_SetProp(frame, ":notification", Gal_StringObject("Audio aborted.")); 
  GalSS_EnvWriteFrame(env, frame, 0); 
  Gal_FreeFrame(frame);
}

static void CatsDataInBroker_done(GalIO_BrokerStruct *brokerStruct,
				  void *data)
{
  CatsDataInBroker this=(CatsDataInBroker)data;
  this->private->activated = GAL_FALSE;
}

static void CatsDataInBroker_cleanup(GalIO_BrokerStruct *brokerStruct,
				     void *data)
{
  CatsDataInBroker this=(CatsDataInBroker)data;
  this->private->brokerStruct = NULL;
}

/* Helper method to generate audio format from frame */

static CatsAudioFormat
CatsDataInBroker_frameToFormat(Gal_Frame frame,
			       CatsAudioFormat defaultFormat)
{
  Gal_Object galObject;
  int encoding,sampleSizeInBits,channels,frameSize,bigEndian;
  float sampleRate,frameRate;

  encoding=defaultFormat->getEncoding(defaultFormat);
  if ((galObject=Gal_GetObject(frame,":encoding")) != NULL) {
    char *encodingString=Gal_StringValue(galObject);
    if (mdtm_text_cmp(encodingString,"ALAW"))
      encoding=CATS_ENCODING_ALAW;
    else if (mdtm_text_cmp(encodingString,"PCM_SIGNED"))
      encoding=CATS_ENCODING_PCM_SIGNED;
    else if (mdtm_text_cmp(encodingString,"PCM_UNSIGNED"))
      encoding=CATS_ENCODING_PCM_UNSIGNED;
    else if (mdtm_text_cmp(encodingString,"ULAW"))
      encoding=CATS_ENCODING_PCM_UNSIGNED;
  }

  sampleRate=defaultFormat->getSampleRate(defaultFormat);
  if ((galObject=Gal_GetObject(frame,":sample_rate")) != NULL)
    sampleRate = (float) Gal_IntValue(galObject);

  sampleSizeInBits=defaultFormat->getSampleSizeInBits(defaultFormat);
  if ((galObject=Gal_GetObject(frame,":sample_size_in_bits")) != NULL)
    sampleSizeInBits = Gal_IntValue(galObject);

  channels=defaultFormat->getChannels(defaultFormat);
  if ((galObject=Gal_GetObject(frame,":channels")) != NULL)
    channels = Gal_IntValue(galObject);

  frameSize=defaultFormat->getFrameSize(defaultFormat);
  if ((galObject=Gal_GetObject(frame,":frame_size")) != NULL)
    frameSize = Gal_IntValue(galObject);

  frameRate=defaultFormat->getFrameRate(defaultFormat);
  if ((galObject=Gal_GetObject(frame,":frame_rate")) != NULL)
    frameRate = Gal_IntValue(galObject);

  bigEndian=defaultFormat->isBigEndian(defaultFormat);
  if ((galObject=Gal_GetObject(frame,":big_endian")) != NULL)
    bigEndian = mdtm_text_cmp(Gal_StringValue(galObject),"true");

  return CatsAudioFormat_create(encoding,sampleRate,sampleSizeInBits,channels,
				frameSize,frameRate,bigEndian);
}
