/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <limits.h>
#include <stdlib.h>
#include <galaxy/galaxy_io.h>
#include <galaxy/generic-server.h>
#include "cats_data_out_broker.h"
#include "cats_source_data_line.h"
#include "cats_buffer.h"

static CatsDataOutBroker CatsDataOutBroker_destroy(CatsDataOutBroker this);
static CatsSourceDataLine CatsDataOutBroker_getLine(CatsDataOutBroker this);

/* CatsSourceDataLine implementation */

static int CatsDataOutBroker_open(CatsSourceDataLine line,
				  CatsAudioFormat audioFormat, int bufferSize);
static int CatsDataOutBroker_start(CatsSourceDataLine line);
static int CatsDataOutBroker_write(CatsSourceDataLine line, char *b, int off,
				   int len);
static int CatsDataOutBroker_stop(CatsSourceDataLine line);
static int CatsDataOutBroker_close(CatsSourceDataLine line);
static int CatsDataOutBroker_available(CatsSourceDataLine line);
static int CatsDataOutBroker_isOpen(CatsSourceDataLine line);
static int CatsDataOutBroker_isRunning(CatsSourceDataLine line);
static int CatsDataOutBroker_isActive(CatsSourceDataLine line);
static int CatsDataOutBroker_getBufferSize(CatsSourceDataLine line);
static CatsAudioFormat CatsDataOutBroker_getFormat(CatsSourceDataLine line);
static int CatsDataOutBroker_getFramePosition(CatsSourceDataLine line);

/* PRIVATE METHODS */

static void CatsDataOutBroker_startBroker(CatsDataOutBroker this,
					  CatsAudioFormat audioFormat);
static void CatsDataOutBroker_writeBroker(CatsDataOutBroker this, char *b,
					  int off, int len);
static void CatsDataOutBroker_stopBroker(CatsDataOutBroker this);

struct CatsDataOutBrokerPrivate_str {
  CatsServer server;
  int opened;
  int started;
  int activated;
  int streaming;
  CatsSourceDataLine line;
  GalIO_BrokerStruct *brokerStruct;
#if (GC_VERSION >= 0x40000)
  GalSS_BrokerProxy *brokerProxy;
#endif
  CatsAudioFormat audioFormat;
  CatsBuffer outputCache;
  int frameSize;
  int framePosition;
  CatsEventListener *listeners;
  int nListeners;
};

CatsDataOutBroker CatsDataOutBroker_create(CatsServer server)
{
  CatsDataOutBroker this=NULL;

  if ((this =
       (CatsDataOutBroker)
       calloc(1,sizeof(struct CatsDataOutBroker_str))) != NULL) {
    CatsDataOutBrokerPrivate private;
    CatsSourceDataLine line;

    this->destroy = CatsDataOutBroker_destroy;
    this->getLine = CatsDataOutBroker_getLine;

    if ((private = this->private =
         (CatsDataOutBrokerPrivate)
	 calloc(1,sizeof(struct CatsDataOutBrokerPrivate_str)))
        == NULL)
      return(this->destroy(this));

    if ((line = private->line = CatsSourceDataLine_create(this)) == NULL)
      return(this->destroy(this));

    line->open = CatsDataOutBroker_open;
    line->start = CatsDataOutBroker_start;
    line->write = CatsDataOutBroker_write;
    line->stop = CatsDataOutBroker_stop;
    line->close = CatsDataOutBroker_close;
    line->available = CatsDataOutBroker_available;
    line->isOpen = CatsDataOutBroker_isOpen;
    line->isRunning = CatsDataOutBroker_isRunning;
    line->isActive = CatsDataOutBroker_isActive;
    line->getBufferSize = CatsDataOutBroker_getBufferSize;
    line->getFormat = CatsDataOutBroker_getFormat;
    line->getFramePosition = CatsDataOutBroker_getFramePosition;

    if ((private->outputCache = CatsBuffer_create(0)) == NULL)
      return(this->destroy(this));

    private->server = server;
    private->opened = private->started = private->activated = GAL_FALSE;
    private->audioFormat = NULL;
    private->brokerStruct=NULL;

  }

  return this;
}

static CatsDataOutBroker CatsDataOutBroker_destroy(CatsDataOutBroker this)
{
  if (this->private != NULL) {
    CatsSourceDataLine line=this->private->line;
    CatsBuffer outputCache=this->private->outputCache;

    if (line != NULL)
      line->destroy(line);
    if (outputCache != NULL)
      outputCache->destroy(outputCache);
    free(this->private);
  }
  free(this);
  return NULL;
}

static CatsSourceDataLine CatsDataOutBroker_getLine(CatsDataOutBroker this)
{
  return this->private->line;
}

/* CatsSourceDataLine implementation */

static int CatsDataOutBroker_open(CatsSourceDataLine line,
				  CatsAudioFormat audioFormat, int bufferSize)
{
  CatsDataOutBroker this=line->getSource(line);

  if (!this->private->opened) {
    CatsDataOutBrokerPrivate private=this->private;
    CatsServer server=private->server;
    CatsController controller=server->getController(server);

    private->audioFormat = audioFormat;
    private->frameSize = audioFormat->getFrameSize(audioFormat);
    private->framePosition = 0;
    private->streaming = controller->isStreamingEnabled(controller);
    private->opened = GAL_TRUE;
  }
  return this->private->opened == GAL_TRUE;
}

static int CatsDataOutBroker_start(CatsSourceDataLine line)
{
  CatsDataOutBroker this=line->getSource(line);

  if (this->private->opened && !this->private->started) {
    this->private->started = GAL_TRUE;
    if (this->private->streaming)
      CatsDataOutBroker_startBroker(this,this->private->audioFormat);
  }

  return this->private->started == GAL_TRUE;
}

static int CatsDataOutBroker_write(CatsSourceDataLine line, char *b, int off,
				   int len)
{
  CatsDataOutBroker this=line->getSource(line);
  CatsDataOutBrokerPrivate private=this->private;
  CatsBuffer outputCache=private->outputCache;

  if (!private->started || len % private->frameSize != 0)
    return -1;

  if (private->streaming)
    CatsDataOutBroker_writeBroker(this,b,off,len);
  else
    outputCache->write(outputCache,b,off,len);

  private->framePosition += len / private->frameSize;
  private->activated = GAL_TRUE;
  return len;
}

static int CatsDataOutBroker_stop(CatsSourceDataLine line)
{
  CatsDataOutBroker this=line->getSource(line);
  CatsDataOutBrokerPrivate private=this->private;

  if (private->started) {
    private->started = GAL_FALSE;

    if (!private->streaming) {
      CatsBuffer outputCache=private->outputCache;
      int size=outputCache->readable(outputCache);
      void *data;

      data=outputCache->read(outputCache,size);
      CatsDataOutBroker_startBroker(this,this->private->audioFormat);
      CatsDataOutBroker_writeBroker(this,data,0,size);
      outputCache->flush(outputCache);
    }

    CatsDataOutBroker_stopBroker(this);
  }

  return this->private->started == GAL_FALSE;
}

static int CatsDataOutBroker_close(CatsSourceDataLine line)
{
  CatsDataOutBroker this=line->getSource(line);
  CatsDataOutBrokerPrivate private=this->private;

  if (private->opened && !private->started) {
    private->opened = GAL_FALSE;
  }

  return private->opened == GAL_FALSE;
}

static int CatsDataOutBroker_available(CatsSourceDataLine line)
{
  return INT_MAX;
}

static int CatsDataOutBroker_isOpen(CatsSourceDataLine line)
{
  CatsDataOutBroker this=line->getSource(line);
  return this->private->opened;
}

static int CatsDataOutBroker_isRunning(CatsSourceDataLine line)
{
  CatsDataOutBroker this=line->getSource(line);
  return this->private->started;
}

static int CatsDataOutBroker_isActive(CatsSourceDataLine line)
{
  CatsDataOutBroker this=line->getSource(line);
  return this->private->activated;
}

static int CatsDataOutBroker_getBufferSize(CatsSourceDataLine line)
{
  return INT_MAX;
}

static CatsAudioFormat CatsDataOutBroker_getFormat(CatsSourceDataLine line)
{
  CatsDataOutBroker this=line->getSource(line);
  return this->private->audioFormat;
}

static int CatsDataOutBroker_getFramePosition(CatsSourceDataLine line)
{
  CatsDataOutBroker this=line->getSource(line);
  return this->private->framePosition;
}

/* PRIVATE METHODS */

static void CatsDataOutBroker_startBroker(CatsDataOutBroker this,
					  CatsAudioFormat audioFormat)
{
  GalIO_BrokerStruct *brokerStruct=this->private->brokerStruct;
#if (GC_VERSION >= 0x40000)
  GalSS_BrokerProxy *bp = this->private->brokerProxy;
#else
  void *bp = (void *) NULL;
#endif
  CatsServer server=this->private->server;
  CatsController controller = server->getController(server);
  /* This will always be true pre-4.0. Otherwise, proxies
     are the default. */
  int useOutboundBrokers = controller->useOutboundBrokers(controller);
  
  if ((useOutboundBrokers && (brokerStruct == NULL)) ||
      ((!useOutboundBrokers) && (bp == NULL))) {
    GalSS_Environment *env=server->getEnvironment(server);
    GalIO_CommStruct *gcomm=server->getCommStruct(server);
    Gal_Frame brokerFrame = (Gal_Frame) NULL;
    
    if (useOutboundBrokers) {
      this->private->brokerStruct = brokerStruct = 
	GalIO_BrokerDataOutInit(gcomm,0,10);
      if (brokerStruct && GalIO_GetBrokerListenPort(brokerStruct) > 0) {
	brokerFrame = Gal_MakeFrame(server->getServerID(server),
				    GAL_CLAUSE);

	GalIO_BrokerPopulateFrame(brokerStruct,brokerFrame,":binary_host",
				  ":binary_port");
	if (controller->isPlayboxEnabled(controller)) {
	  GalIO_BrokerWriteString(brokerStruct,"new_utt");
	}
      }
    } 
#if (GC_VERSION >= 0x40000) 
    else {      
      this->private->brokerProxy = bp =
	GalSS_ProxifyObjectType(env, GAL_BINARY, 0, 10);
      if (bp) {
	brokerFrame = Gal_MakeFrame(server->getServerID(server),
				    GAL_CLAUSE);
	Gal_SetProp(brokerFrame, ":binary_proxy",
		    Gal_CreateProxyObject(bp, 0));
      }
    }
#endif
    if (brokerFrame) {
      Gal_SetProp(brokerFrame,":encoding",
		  Gal_StringObject(audioFormat->getEncodingString(audioFormat)));
      Gal_SetProp(brokerFrame,":sample_rate",
		  Gal_IntObject((int)audioFormat->getSampleRate(audioFormat)));
      Gal_SetProp(brokerFrame,":sample_size_in_bits",
		  Gal_IntObject(audioFormat->getSampleSizeInBits(audioFormat)));
      Gal_SetProp(brokerFrame,":channels",
		  Gal_IntObject(audioFormat->getChannels(audioFormat)));
      Gal_SetProp(brokerFrame,":frame_size",
		  Gal_IntObject(audioFormat->getFrameSize(audioFormat)));
      Gal_SetProp(brokerFrame,":frame_rate",
		  Gal_IntObject((int)audioFormat->getFrameRate(audioFormat)));
      Gal_SetProp(brokerFrame,":big_endian",
		  Gal_StringObject(audioFormat->
				   getBigEndianString(audioFormat)));

      GalSS_EnvWriteFrame(env,brokerFrame,GAL_FALSE);
      Gal_FreeFrame(brokerFrame);
    } else {
      server->shutDown(server);
    }
  }
}

static void CatsDataOutBroker_writeBroker(CatsDataOutBroker this, char *b,
					  int off, int len)
{
  CatsServer server=this->private->server;
  CatsController controller=server->getController(server);
  GalIO_BrokerStruct *brokerStruct=this->private->brokerStruct;

  if (controller->isPlayboxEnabled(controller)) {
    GalIO_BrokerWriteInt16(brokerStruct,b+off,len>>1);
  } else if (controller->useOutboundBrokers(controller)) {
    GalIO_BrokerWriteBinary(brokerStruct,b+off,len);
  }
#if (GC_VERSION >= 0x40000)
  else {
    GalSS_ProxyArrayAdd(this->private->brokerProxy, b+off, len);
  }
#endif
}

static void CatsDataOutBroker_stopBroker(CatsDataOutBroker this)
{
  GalIO_BrokerStruct *brokerStruct=this->private->brokerStruct;
#if (GC_VERSION >= 0x40000)
  GalSS_BrokerProxy *bp = this->private->brokerProxy;
#else
  void *bp = (void *) NULL;
#endif
  CatsServer server=this->private->server;
  CatsController controller = server->getController(server);
  /* This will always be true pre-4.0. Otherwise, proxies
     are the default. */
  int useOutboundBrokers = controller->useOutboundBrokers(controller);
  
  if ((useOutboundBrokers && (brokerStruct != NULL)) ||
      ((!useOutboundBrokers) && (bp != NULL))) {

    if (controller->isPlayboxEnabled(controller))
      GalIO_BrokerWriteString(brokerStruct,"done");
    if (controller->useOutboundBrokers(controller)) {
      GalIO_BrokerDataOutDone(brokerStruct);
      /* SAM: We had a call to GalIO_DestroyBrokerStruct here,
	 but it isn't necessary, because we're using the normal
	 timeout method for out brokers, and in fact it will
	 cause problems if we call it before anyone contacts
	 the broker. */
      this->private->brokerStruct = NULL;
    }
#if (GC_VERSION >= 0x40000)
    else {
      GalSS_ProxyDone(this->private->brokerProxy);
      GalSS_FreeBrokerProxy(this->private->brokerProxy);
      this->private->brokerProxy = NULL;
    }
#endif
  }
}
