/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <galaxy/galaxy_all.h>
#define SERVER_FUNCTIONS_INCLUDE "cats_galaxy.h"
#define USE_SERVER_DATA
#include <galaxy/server_functions.h>

#include <string.h>
#include "cats_server.h"
#include "cats_controller_event.h"
#include "pa_mixer.h"
#ifdef ENABLE_CFONE
#include "cf_phone_line.h"
#include "cats_switchboard.h"
#endif

static char *oas[] = {
  "-debug","Enable debugging output",
  "-half_duplex","Enable half duplex operation",
  "-nobarge","Disable server-driven barge-in",
  "-nostreaming","Disable streaming mode",
  "-playbox","MIT playbox compatibility mode",
  "-ptt","Allow for push-to-talk operation",
  "-record_encoding encStr","Encoding used for recording",NULL,
  "-record_sample_rate rate","Sample rate used for recording",NULL,
  "-record_sample_size size","Sample size in bits/sample used for recording",NULL,
  "-record_channels channels","Number of channels to record",NULL,
  "-record_little_endian","Record using little endian byte ordering rather than big endian",
  "-playback_encoding encStr","Encoding used for playback",NULL,
  "-playback_sample_rate rate","Sample rate used for playback",NULL,
  "-playback_sample_size size","Sample size in bits/sample used for playback",NULL,
  "-playback_channels channels","Number of channels to playback",NULL,
  "-playback_little_endian","Playback using little endian byte ordering rather than big endian",
  "-vad_raise factor","VAD threshold raise adaptation factor",NULL,
  "-vad_lower factor","VAD threshold raise adaptation factor",NULL,
#ifdef ENABLE_CFONE
  "-use_cfone tty","Set up Computerfone support on the specified tty device", NULL,
#endif
#if (GC_VERSION >= 0x40000)
  "-outbound_brokers", "Force CATS to write to brokers instead of broker proxies",
#endif
  NULL
};

/* Static functions to get commstruct, server objects from dispatch
   function data */

static GalIO_CommStruct *__get_gcomm(void *server_data);
static CatsServer __get_server(void *server_data);
static void __connection_shutdown_callback(GalIO_CommStruct *gcomm,
					   void *callback_data);

/* Galaxy server initializer */

typedef struct CatsGalaxyServerData_str {
  CatsControllerParameters parameters;
#ifdef ENABLE_CFONE
  char *cfoneDevice;
#endif
} *CatsGalaxyServerData;

typedef struct CatsGalaxyGCommData_str {
  CatsServer server;
#ifdef ENABLE_CFONE
  CfPhoneLine phoneLine;
#endif
} *CatsGalaxyGCommData;

void *_GalSS_init_server(Gal_Server *s, int argc, char **argv) 
{

  CatsControllerParameters parameters = (CatsControllerParameters) calloc(1,sizeof(struct CatsControllerParameters_str));
  char *encodingString = (char *) NULL;
  CatsGalaxyServerData data = (CatsGalaxyServerData) calloc(1, sizeof(struct CatsGalaxyServerData_str));
  
  if (!parameters || !data) {
    GalUtil_Warn("Could not create structures for argument parameters");
    exit(1);
  }

  if (!GalUtil_OACheckUsage(argc, argv, oas, NULL)) 
    exit(1);

  /* Parse mode flags */

  GalUtil_OAExtract(argc, argv, oas, "-debug", GAL_OA_INT,
		    &parameters->debugEnabled);
  GalUtil_OAExtract(argc, argv, oas, "-half_duplex", GAL_OA_INT,
		    &parameters->halfDuplexEnabled);
  GalUtil_OAExtract(argc, argv, oas, "-nobarge", GAL_OA_INT,
		    &parameters->bargeInDisabled);
  GalUtil_OAExtract(argc, argv, oas, "-nostreaming", GAL_OA_INT,
		    &parameters->streamingDisabled);
  GalUtil_OAExtract(argc, argv, oas, "-playbox", GAL_OA_INT,
		    &parameters->playboxEnabled);
  GalUtil_OAExtract(argc, argv, oas, "-ptt", GAL_OA_INT,
		    &parameters->pttEnabled);

  /*
  ** Extract default record/playback audio format information:
  **
  ** '-record_encoding' specifies the encoding type used for recording
  ** '-record_sample_rate' specifies the sample rate used for recording
  ** '-record_sample_size' specifies the # of bits to use for recording
  ** '-record_channels' specifies the number of channels to record
  ** '-record_little_endian' specifies to use little endian byte order in
  **   recording
  **
  ** Similar '-playback_*' options are available.
  */

  GalUtil_OAExtract(argc, argv, oas, "-record_encoding", GAL_OA_STRING,
		    &encodingString);
  if (!encodingString) {
    parameters->recordEncoding = CATS_DEFAULT_ENCODING;
  } else if (Gal_StringEq(encodingString,"PCM_SIGNED")) {
    parameters->recordEncoding = CATS_ENCODING_PCM_SIGNED;
  } else if (Gal_StringEq(encodingString,"PCM_UNSIGNED")) {
    parameters->recordEncoding = CATS_ENCODING_PCM_UNSIGNED;
  } else if (Gal_StringEq(encodingString,"ULAW")) {
    parameters->recordEncoding = CATS_ENCODING_ULAW;
  } else if (Gal_StringEq(encodingString,"ALAW")) {
    parameters->recordEncoding = CATS_ENCODING_ALAW;
  } else {
    GalUtil_Error("Unsupported audio format %s\n", encodingString);
    exit(1);
  }

  if (!GalUtil_OAExtract(argc, argv, oas, "-record_sample_rate", GAL_OA_FLOAT,
			 &parameters->recordSampleRate)) {
    parameters->recordSampleRate = CATS_DEFAULT_SAMPLE_RATE;
  }
  if (!GalUtil_OAExtract(argc, argv, oas, "-record_sample_size", GAL_OA_INT,
			 &parameters->recordSampleSizeInBits)) {
    parameters->recordSampleSizeInBits = CATS_DEFAULT_SAMPLE_SIZE_IN_BITS;
  }
  if (!GalUtil_OAExtract(argc, argv, oas, "-record_channels", GAL_OA_INT,
			 &parameters->recordChannels)) {
    parameters->recordChannels = CATS_DEFAULT_CHANNELS;
  }
  GalUtil_OAExtract(argc, argv, oas, "-record_little_endian", GAL_OA_INT,
		    &parameters->recordLittleEndian);

  encodingString = (char *) NULL;
  GalUtil_OAExtract(argc, argv, oas, "-playback_encoding", GAL_OA_STRING,
		    &encodingString);

  if (!encodingString) {
    parameters->playbackEncoding = CATS_DEFAULT_ENCODING;
  } else if (Gal_StringEq(encodingString,"PCM_SIGNED")) {
    parameters->playbackEncoding = CATS_ENCODING_PCM_SIGNED;
  } else if (Gal_StringEq(encodingString,"PCM_UNSIGNED")) {
    parameters->playbackEncoding = CATS_ENCODING_PCM_UNSIGNED;
  } else if (Gal_StringEq(encodingString,"ULAW")) {
    parameters->playbackEncoding = CATS_ENCODING_ULAW;
  } else if (Gal_StringEq(encodingString,"ALAW")) {
    parameters->playbackEncoding = CATS_ENCODING_ALAW;
  } else {
    GalUtil_Error("Unsupported audio format %s\n", encodingString);
    exit(1);
  }

  if (!GalUtil_OAExtract(argc, argv, oas, "-playback_sample_rate",
			 GAL_OA_FLOAT, &parameters->playbackSampleRate)) {
    parameters->playbackSampleRate = CATS_DEFAULT_SAMPLE_RATE;
  }
  if (!GalUtil_OAExtract(argc, argv, oas, "-playback_sample_size", GAL_OA_INT,
			 &parameters->playbackSampleSizeInBits)) {
    parameters->playbackSampleSizeInBits = CATS_DEFAULT_SAMPLE_SIZE_IN_BITS;
  }
  if (!GalUtil_OAExtract(argc, argv, oas, "-playback_channels", GAL_OA_INT,
			 &parameters->playbackChannels)) {
    parameters->playbackChannels = CATS_DEFAULT_CHANNELS;
  }
  GalUtil_OAExtract(argc, argv, oas, "-playback_little_endian", GAL_OA_INT,
		    &parameters->playbackLittleEndian);

  /* Extract VAD operating parameters if specified */

  if (!GalUtil_OAExtract(argc, argv, oas, "-vad_raise", GAL_OA_FLOAT,
			 &parameters->defaultVadRaiseFactor)) {
    parameters->defaultVadRaiseFactor = CATS_DEFAULT_T_RAISE_FACTOR;
  }
  if (!GalUtil_OAExtract(argc, argv, oas, "-vad_lower", GAL_OA_FLOAT,
			 &parameters->defaultVadLowerFactor)) {
    parameters->defaultVadLowerFactor = CATS_DEFAULT_T_LOWER_FACTOR;
  }

#ifdef ENABLE_CFONE
  data->cfoneDevice = (char *) NULL;
  GalUtil_OAExtract(argc, argv, oas, "-use_cfone", GAL_OA_STRING,
		    &data->cfoneDevice);
#endif
#if (GC_VERSION >= 0x40000)
  /* 4.0 and after, proxies are the default. */
  parameters->useBrokers = GalUtil_OAExtract(argc, argv, oas, "-outbound_brokers", GAL_OA_INT, NULL);
#endif

  data->parameters = parameters;
  return (void *) data;
}

/* DISPATCH FUNCTIONS */

Gal_Frame reinitialize(Gal_Frame f, void *server_data)
{
  CatsServer server;
  CatsController controller;
  GalIO_CommStruct *gcomm=__get_gcomm(server_data);
  PaMixer paMixer;
  CatsGalaxyServerData data = (CatsGalaxyServerData) GalIO_GetCommServerData(gcomm);
  char *initial_session_id=Gal_GetString(f,":session_id");
  CatsGalaxyGCommData commData = (CatsGalaxyGCommData) calloc(1, sizeof(struct CatsGalaxyGCommData_str));

  if (!commData) {
    GalUtil_Fatal("Failed to create connection data!");
  }
  
  if ((server = CatsServer_create(data->parameters,server_data,
				  initial_session_id)) == NULL)
    GalUtil_Fatal("Failed to create server object!");

  commData->server = server;
  controller = server->getController(server);

  if ((paMixer = PaMixer_create(controller->isHalfDuplexEnabled(controller)))
      == NULL)
    GalUtil_Fatal("Failed to create PortAudio Mixer!");

  controller->setADLine(controller,paMixer->getTargetDataLine(paMixer));
  controller->setDALine(controller,paMixer->getSourceDataLine(paMixer));

#ifdef ENABLE_CFONE
  if (data->cfoneDevice) {
    CfPhoneLine cfPhoneLine;
    if ((cfPhoneLine = CfPhoneLine_create(data->cfoneDevice)) == NULL)
      GalUtil_Fatal("Failed to create Computerfone line!\n");
    commData->phoneLine = cfPhoneLine;
    controller->setPhoneLine(controller,cfPhoneLine->getPhoneLine(cfPhoneLine));
    /* Now that the controller has the phone line, we'd better make
       sure there's a switchboard. Not sure where this should happen. */
    if (!controller->ensureSwitchboard(controller))
      GalUtil_Fatal("Failed to create switchboard!\n");
  }
#endif

  GalIO_AddConnectionCallback(gcomm,GAL_CONNECTION_SHUTDOWN_EVENT,
			      __connection_shutdown_callback,
			      server);
  GalIO_SetCommData(gcomm, commData, NULL);

  return f;
}

Gal_Frame receive_audio(Gal_Frame f, void *server_data)
{
  CatsServer server = __get_server(server_data);

  server->receiveAudio(server,f);

  return f;
}

Gal_Frame off_hook(Gal_Frame f, void *server_data)
{
  CatsServer server = __get_server(server_data);

  server->offHook(server);

  return f;
}

Gal_Frame on_hook(Gal_Frame f, void *server_data)
{
  CatsServer server = __get_server(server_data);
  server->onHook(server);

  return f;
}

Gal_Frame touchtone(Gal_Frame f, void *server_data)
{
  CatsServer server = __get_server(server_data);
  char *idString;

  if ((idString = Gal_GetString(f,":touchtone")) != NULL)
    server->touchtone(server,idString);

  return f;
}

Gal_Frame mute(Gal_Frame f, void *server_data)
{
  CatsServer server = __get_server(server_data);

  server->mute(server);

  return f;
}

Gal_Frame unmute(Gal_Frame f, void *server_data)
{
  CatsServer server = __get_server(server_data);

  server->unmute(server);

  return f;
}

Gal_Frame start_recording(Gal_Frame f, void *server_data)
{
  CatsServer server = __get_server(server_data);

  server->startRecording(server);

  return f;
}

Gal_Frame stop_recording(Gal_Frame f, void *server_data)
{
  CatsServer server = __get_server(server_data);

  server->stopRecording(server);

  return f;
}

Gal_Frame resend_audio(Gal_Frame f, void *server_data)
{
  CatsServer server=__get_server(server_data);

  server->resendAudio(server);

  return f;
}

Gal_Frame query_status(Gal_Frame f, void *server_data)
{
  CatsServer server = __get_server(server_data);

  server->queryStatus(server,GAL_TRUE);

  return f;
}

Gal_Frame set_mode(Gal_Frame f, void *server_data)
{
  CatsServer server=__get_server(server_data);

  server->setMode(server,f);

  return f;
}

static GalIO_CommStruct *__get_gcomm(void *server_data)
{
  return GalSS_EnvComm((GalSS_Environment *) server_data);
}

static CatsServer __get_server(void *server_data)
{
  GalIO_CommStruct *gcomm=__get_gcomm(server_data);
  CatsGalaxyGCommData data = (CatsGalaxyGCommData) GalIO_GetCommData(gcomm);

  if (data) {
    return data->server;
  } else {
    return NULL;
  }
}

static void __connection_shutdown_callback(GalIO_CommStruct *gcomm,
					   void *callback_data)
{
  CatsServer server=(CatsServer)callback_data;
  CatsController controller=server->getController(server);
  CatsTargetDataLine adLine=controller->getADLine(controller);
  PaMixer paMixer=(PaMixer)adLine->getSource(adLine);
  CatsGalaxyGCommData data = (CatsGalaxyGCommData) GalIO_GetCommData(gcomm);
  
#ifdef ENABLE_CFONE
  /* The controller has the phone line, and it should probably
     destroy it. The problem is that the CfPhoneLine ALSO has the
     line, and will also try to destroy it. So first we need to
     unset the phone line from the controller, since that's the
     one that allows us to set it, and then the controller will
     skip freeing the phone line. It will, however, free the
     switchboard, which will NOT try to free the phone line. */
  if (data->phoneLine) {    
    controller->setPhoneLine(controller, (CatsPhoneLine) NULL);
    data->phoneLine->destroy(data->phoneLine);
  }
#endif
  free(data);
  server->destroy(server);
  paMixer->destroy(paMixer);
}
