/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <galaxy/gthread.h>
#include <mdtm/mdtm.h>
#include <mdtm/mdtmmacros.h>
#include "cats_lock_object.h"
#include "cats_thread.h"
#include "cats_array_list.h"

static CatsLockObject CatsLockObject_destroy(CatsLockObject this);
static void CatsLockObject_lock(CatsLockObject this);
static void CatsLockObject_unlock(CatsLockObject this);
static void CatsLockObject_release(CatsLockObject this);
static void CatsLockObject_relock(CatsLockObject this);
static int CatsLockObject_isLockable(CatsLockObject this);
static int CatsLockObject_isLocked(CatsLockObject this);
static int CatsLockObject_isReleased(CatsLockObject this);
static void CatsLockObject_notify(CatsLockObject this);

/* Private data */

struct CatsLockObjectPrivate_str {
  GalUtil_LocalMutex *mutex;
  int numLocks;
  GalUtil_ThreadID lockedBy;
  CatsArrayList releases;
};

typedef struct CatsLockObjectRelease_str {
  int releaseNumLocks;
  GalUtil_ThreadID releasedBy;
} *CatsLockObjectRelease;

/* Private methods */

static CatsLockObjectRelease
CatsLockObjectRelease_create(int releaseNumLocks,GalUtil_ThreadID releasedBy);
static int CatsLockObject_getReleaseIndex(CatsLockObject this);

CatsLockObject CatsLockObject_create(void)
{
  CatsLockObject this=NULL;

  if ((this =
       (CatsLockObject) calloc(1,sizeof(struct CatsLockObject_str)))
      != NULL) {
    CatsLockObjectPrivate private=NULL;

    this->destroy = CatsLockObject_destroy;
    this->lock = CatsLockObject_lock;
    this->unlock = CatsLockObject_unlock;
    this->release = CatsLockObject_release;
    this->relock = CatsLockObject_relock;
    this->isLockable = CatsLockObject_isLockable;
    this->isLocked = CatsLockObject_isLocked;
    this->isReleased = CatsLockObject_isReleased;
    this->notify = CatsLockObject_notify;

    if ((private = this->private =
         (CatsLockObjectPrivate)
	 calloc(1,sizeof(struct CatsLockObjectPrivate_str)))
        == NULL)
      return(this->destroy(this));

    if ((private->mutex = GalUtil_CreateLocalMutex()) == NULL)
      return(this->destroy(this));

    private->numLocks = 0;
    private->lockedBy = 0;

    if ((private->releases = CatsArrayList_create(0,0)) == NULL)
      return (this->destroy(this));

  }
  return this;
}

static CatsLockObject CatsLockObject_destroy(CatsLockObject this)
{
  if (this->private != NULL) {
    CatsArrayList releases=this->private->releases;

    if (this->private->mutex != NULL) {
      this->lock(this);
      GalUtil_DestroyLocalMutex(this->private->mutex);
    }

    if (releases != NULL)
      releases->destroy(releases);

    free(this->private);
  }
  free(this);
  return NULL;
}

static void CatsLockObject_lock(CatsLockObject this)
{
  CatsLockObjectPrivate private=this->private;
  if (this->isLocked(this)) {
    private->numLocks++;
  } else {
    GalUtil_LockLocalMutex(private->mutex);
    private->numLocks = 1;
    private->lockedBy = GalUtil_CurrentThreadID();
  }
}

static void CatsLockObject_unlock(CatsLockObject this)
{
  if (this->isLocked(this)) {
    CatsLockObjectPrivate private=this->private;
    private->numLocks--;
    if (private->numLocks == 0)
      GalUtil_UnlockLocalMutex(private->mutex);
  }
}

static void CatsLockObject_release(CatsLockObject this)
{
  if (this->isLocked(this) && !this->isReleased(this)) {
    CatsLockObjectPrivate private=this->private;
    CatsArrayList releases=private->releases;
    CatsLockObjectRelease release;

    if ((release = CatsLockObjectRelease_create(private->numLocks,
						private->lockedBy)) == NULL)
      return;

    releases->add(releases,release);
    private->numLocks = 0;
    GalUtil_UnlockLocalMutex(private->mutex);
  }
}

static void CatsLockObject_relock(CatsLockObject this)
{
  int releaseIndex;

  GalUtil_LockLocalMutex(this->private->mutex);

  releaseIndex=CatsLockObject_getReleaseIndex(this);

  if (releaseIndex != -1) {
    CatsArrayList releases;
    CatsLockObjectRelease release;

    releases=this->private->releases;
    release=releases->remove(releases,releaseIndex);

    this->private->numLocks = release->releaseNumLocks;
    this->private->lockedBy = release->releasedBy;

    free(release);
    return;
  }

  GalUtil_LockLocalMutex(this->private->mutex);
  return;
}

static int CatsLockObject_isLockable(CatsLockObject this)
{
  return this->private->numLocks == 0 ||
    GalUtil_EqualThreadID(GalUtil_CurrentThreadID(),this->private->lockedBy);
}

static int CatsLockObject_isLocked(CatsLockObject this)
{
  return this->private->numLocks > 0 &&
    GalUtil_EqualThreadID(GalUtil_CurrentThreadID(),this->private->lockedBy);
}

static int CatsLockObject_isReleased(CatsLockObject this)
{
  return CatsLockObject_getReleaseIndex(this) != -1;
}

static void CatsLockObject_notify(CatsLockObject this)
{
  this->release(this);
  CatsThread_sleep(0,1);
  this->relock(this);
}

static CatsLockObjectRelease
CatsLockObjectRelease_create(int releaseNumLocks,GalUtil_ThreadID releasedBy)
{
  CatsLockObjectRelease this=NULL;

  if ((this = calloc(1,sizeof(struct CatsLockObjectRelease_str)))
      != NULL) {
    this->releaseNumLocks = releaseNumLocks;
    this->releasedBy = releasedBy;
  }

  return this;
}

static int CatsLockObject_getReleaseIndex(CatsLockObject this)
{
  CatsArrayList releases=this->private->releases;
  CatsLockObjectRelease release;
  int nReleases=releases->size(releases);
  int i;

  for (i=0; i<nReleases; i++) {
    release = (CatsLockObjectRelease) releases->elementAt(releases,i);
    if (GalUtil_EqualThreadID(release->releasedBy,GalUtil_CurrentThreadID()))
      return i;
  }

  return -1;
}
