/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <stdio.h>
#include <stdlib.h>
#include <mdtm/mdtm.h>
#include <mdtm/mdtmmacros.h>
#include "cats_playback.h"
#include "cats_target_data_line.h"
#include "cats_source_data_line.h"
#include "cats_event_listener.h"
#include "cats_controller_event.h"
#include "cats_thread.h"
#include "cats_vector.h"

#define DEBUG 0
#define CATS_LINE_BUFFER_LENGTH_SEC .25
#define CATS_FRAME_SIZE_MSEC 10

static CatsPlayback CatsPlayback_destroy(CatsPlayback this);
static int CatsPlayback_open(CatsPlayback this);
static int CatsPlayback_start(CatsPlayback this);
static int CatsPlayback_stop(CatsPlayback this);
static int CatsPlayback_close(CatsPlayback this);
static int CatsPlayback_isPlaybackStartable(CatsPlayback this);
static int CatsPlayback_startPlayback(CatsPlayback this);
static void CatsPlayback_stopPlayback(CatsPlayback this);
static int CatsPlayback_isBargeInStartable(CatsPlayback this);
static int CatsPlayback_bargeIn(CatsPlayback this);
static int CatsPlayback_addListener(CatsPlayback this, CatsEventListener l);
static void CatsPlayback_removeListener(CatsPlayback this,
					CatsEventListener l);

/* Controller listener method */

static void
CatsPlayback_ControllerEventGenerated(CatsEventListener listener,
				      CatsEvent event);

/* Private methods */

/* Event notifier */

static void CatsPlayback_notifySubscribers(CatsPlayback this,
					   CatsEvent event);
static void CatsPlayback_run(Gal_TaskPkg *taskPkg);

struct CatsPlaybackPrivate_str {

  /* Mode, state flags. */

  int opened;
  int started;
  int debugEnabled;
  int bargeInEnabled;
  int playbackStarted;
  int bargeInStarted;

  /* Audio-related fields. */

  CatsAudioFormat audioFormat;
  CatsTargetDataLine inBrokerLine;
  CatsSourceDataLine daLine;
  mdtm_vrchar *playbackBuffer;
  int playbackBufferSize;

  /* Supporting fields. */

  CatsVector listeners;
  CatsServer server;
  CatsController controller;
  CatsEventListener controllerListener;
  mdtm_text *errStr;
  int frameSize;
  Gal_TaskPkg *taskPkg;
};

CatsPlayback CatsPlayback_create(CatsController controller)
{
  CatsPlayback this=NULL;

  if ((this = (CatsPlayback) calloc(1,sizeof(struct CatsPlayback_str)))
      != NULL) {
    CatsPlaybackPrivate private;

    if ((this->object = CatsObject_create(this,CATS_PLAYBACK)) == NULL)
      return this->destroy(this);

    this->destroy = CatsPlayback_destroy;
    this->open = CatsPlayback_open;
    this->start = CatsPlayback_start;
    this->stop = CatsPlayback_stop;
    this->close = CatsPlayback_close;
    this->addListener = CatsPlayback_addListener;
    this->removeListener = CatsPlayback_removeListener;
    this->isPlaybackStartable = CatsPlayback_isPlaybackStartable;
    this->startPlayback = CatsPlayback_startPlayback;
    this->stopPlayback = CatsPlayback_stopPlayback;
    this->isBargeInStartable = CatsPlayback_isBargeInStartable;
    this->bargeIn = CatsPlayback_bargeIn;

    if ((private = this->private =
         (CatsPlaybackPrivate)
	 calloc(1,sizeof(struct CatsPlaybackPrivate_str)))
        == NULL)
      return(this->destroy(this));

    private->bargeInEnabled = controller->isBargeInEnabled(controller);

    /* Listener vector */

    if ((private->listeners = CatsVector_create(0,0)) == NULL)
      return(this->destroy(this));

    /* Get inBrokerLine, Server instances */

    private->server = controller->getServer(controller);
    private->inBrokerLine = NULL;
    private->daLine = NULL;

    /* Controller, controller listener instances */

    private->controller = controller;
    if ((private->controllerListener =
	 CatsEventListener_create(this)) == NULL)
      return(this->destroy(this));
    private->controllerListener->eventGenerated =
      CatsPlayback_ControllerEventGenerated;
    controller->addListener(controller,private->controllerListener);

    private->taskPkg = NULL;
  }

  return this;
}

static CatsPlayback CatsPlayback_destroy(CatsPlayback this)
{
  if (this->private != NULL) {
    CatsObject object=this->object;
    CatsVector listeners=this->private->listeners;
    CatsEventListener
      controllerListener=this->private->controllerListener;

    if (object != NULL)
      object->destroy(object);
    if (listeners != NULL)
      listeners->destroy(listeners);
    if (controllerListener != NULL)
      controllerListener->destroy(controllerListener);
    if (this->private->started) {
      this->bargeIn(this);
      this->stop(this);
    }
    if (this->private->opened)
      this->close(this);
    free(this->private);
  }

  free(this);
  return NULL;
}

static int CatsPlayback_open(CatsPlayback this)
{
  if (!this->private->opened) {
    this->private->taskPkg = 
      Gal_AddTask(CatsPlayback_run,(void *)this,0,GAL_TRUE,NULL);
    this->private->opened = GAL_TRUE;
  }
  return this->private->opened == GAL_TRUE;
}

static int CatsPlayback_start(CatsPlayback this)
{
  CatsPlaybackPrivate private=this->private;

  if (private->opened && !private->started) {
    CatsServer server=private->server;
    CatsController controller=server->getController(server);
    CatsTargetDataLine inBrokerLine;
    CatsSourceDataLine daLine;
    CatsAudioFormat audioFormat;
    CatsEvent event;
    int bytesPerSecond;
    int daLineBufferSize;

    /* EBG: Order is important here.  We have to get the audio format
       first, since it is derived from the broker line that is taken
       out of the queue by the getInBrokerLine method */

    daLine = private->daLine = controller->getDALine(controller);
    audioFormat = private->audioFormat =
      server->getBrokerAudioFormat(server);
    inBrokerLine = private->inBrokerLine = server->getInBrokerLine(server);

    bytesPerSecond = audioFormat->getFrameRate(audioFormat) *
      audioFormat->getFrameSize(audioFormat);

    daLineBufferSize = bytesPerSecond*CATS_LINE_BUFFER_LENGTH_SEC;

    if (!(daLine->open(daLine,audioFormat,daLineBufferSize) &&
	  inBrokerLine->open(inBrokerLine,audioFormat,0)))
      return MDTM_FALSE;

    if ((event = CatsEvent_create(this,CATS_PLAYBACK_EVENT,
				  CATS_PLAYBACK_LINE_OPENED)) != NULL) {
      CatsPlayback_notifySubscribers(this,event);
      event->destroy(event);
    }

    inBrokerLine->start(inBrokerLine);

    private->playbackBufferSize = daLine->getBufferSize(daLine);

    if ((private->playbackBuffer =
	 mdtm_vrchar_malloc(private->playbackBufferSize)) == NULL)
      return MDTM_FALSE;

    private->started = GAL_TRUE;

    if ((event = CatsEvent_create(this,CATS_PLAYBACK_EVENT,
				  CATS_PLAYBACK_STARTED)) != NULL) {
      CatsPlayback_notifySubscribers(this,event);
      event->destroy(event);
    }
  }

  return private->started == GAL_TRUE;
}

static int CatsPlayback_stop(CatsPlayback this)
{
  if (this->private->started) {
    CatsTargetDataLine inBrokerLine=this->private->inBrokerLine;
    CatsSourceDataLine daLine=this->private->daLine;
    CatsEvent event;

    this->private->started = GAL_FALSE;

    if (!this->private->bargeInStarted)
      daLine->drain(daLine);
    else {
      CatsServer server=this->private->server;

      daLine->flush(daLine);
      server->flushInBrokers(server);
      if ((event = CatsEvent_create(this,CATS_PLAYBACK_EVENT,
				    CATS_BARGE_IN)) != NULL) {
	CatsPlayback_notifySubscribers(this,event);
	event->destroy(event);
      }
      this->private->bargeInStarted = MDTM_FALSE;
    }

    daLine->close(daLine);
    if ((event = CatsEvent_create(this,CATS_PLAYBACK_EVENT,
				  CATS_PLAYBACK_LINE_CLOSED)) != NULL) {
      CatsPlayback_notifySubscribers(this,event);
      event->destroy(event);
    }

    inBrokerLine->stop(inBrokerLine);
    inBrokerLine->close(inBrokerLine);
    if ((event = CatsEvent_create(this,CATS_PLAYBACK_EVENT,
				  CATS_PLAYBACK_STOPPED)) != NULL) {
      CatsPlayback_notifySubscribers(this,event);
      event->destroy(event);
    }

    this->private->playbackBuffer =
      mdtm_vrchar_free(this->private->playbackBuffer);
    this->private->playbackBufferSize = 0;
  }

  return this->private->started == GAL_FALSE;
}

static int CatsPlayback_close(CatsPlayback this)
{
  if (this->private->opened) {
    this->private->opened = GAL_FALSE;

    if (this->private->started) {
      this->bargeIn(this);
      this->stop(this);
    }

    /* Remove polling function task */

    Gal_RemoveTask(this->private->taskPkg);
    this->private->taskPkg = NULL;
  }

  return this->private->opened == GAL_FALSE;
}

static int CatsPlayback_isPlaybackStartable(CatsPlayback this)
{
  CatsPlaybackPrivate private=this->private;
  CatsController controller=private->controller;

  return !(private->playbackStarted || !private->opened ||
	   controller->getDALine(controller) == NULL);
}

static int CatsPlayback_startPlayback(CatsPlayback this)
{
  CatsPlaybackPrivate private=this->private;

  if (this->isPlaybackStartable(this))
    private->playbackStarted = MDTM_TRUE;
  return private->playbackStarted == MDTM_TRUE;
}

static void CatsPlayback_stopPlayback(CatsPlayback this)
{
  this->private->playbackStarted = MDTM_FALSE;
}

static int CatsPlayback_isBargeInStartable(CatsPlayback this)
{
  return this->private->bargeInEnabled && this->private->playbackStarted;
}

static int CatsPlayback_bargeIn(CatsPlayback this)
{
  CatsPlaybackPrivate private=this->private;
  if (private->bargeInEnabled) {
    private->playbackStarted = MDTM_FALSE;
    private->bargeInStarted = MDTM_TRUE;
  }
  return private->bargeInStarted == MDTM_TRUE;
}

static int CatsPlayback_addListener(CatsPlayback this, CatsEventListener l)
{
  CatsServer server=this->private->server;
  CatsVector listeners=this->private->listeners;
  int retval;

  if (!(retval = listeners->add(listeners,l))) {
    GalUtil_Error("%s\n","Failed to add CatsEventListener");
    server->shutDown(server);
  }

  return retval;
}

static void CatsPlayback_notifySubscribers(CatsPlayback this,
					   CatsEvent event)
{
  CatsVector listeners=this->private->listeners;
  CatsEventListener listener;
  int i;

  for (i=0; i<listeners->size(listeners); i++) {
    listener = listeners->elementAt(listeners,i);
    listener->eventGenerated(listener,event);
  }
}

static void CatsPlayback_removeListener(CatsPlayback this,
					CatsEventListener l)
{
  CatsVector listeners=this->private->listeners;
  listeners->removeElement(listeners,l);
}

/* PRIVATE METHODS */

/* Controller listener method */

static void
CatsPlayback_ControllerEventGenerated(CatsEventListener listener,
				      CatsEvent event)
{
  CatsPlayback this = (CatsPlayback) listener->getSource(listener);

  switch (event->eventType) {

  case CATS_MODE_EVENT:
    switch (event->eventID) {
    case CATS_BARGE_IN_DISABLED:
      this->private->bargeInEnabled = MDTM_FALSE;
      break;
    case CATS_BARGE_IN_ENABLED:
      this->private->bargeInEnabled = MDTM_TRUE;
      break;
    }
    break;

  }

}

static void CatsPlayback_run(Gal_TaskPkg *taskPkg)
{
  CatsPlayback this=Gal_TaskPkgData(taskPkg);
  CatsPlaybackPrivate private=this->private;
  int restart_ms = 5;
  
  if (private->opened) {
    CatsController controller=this->private->controller;

    if (controller->isPlaybackStartable(controller) &&
	!controller->isRequestRedundant(controller,controller->startPlayback,
					NULL,this->object)) {
      controller->startPlayback(controller,NULL,this->object);
    }

    if (!private->started && private->playbackStarted)
      this->start(this);

    if (private->started) {
      CatsTargetDataLine inBrokerLine=private->inBrokerLine;
      CatsSourceDataLine daLine=private->daLine;
      char *playbackBuffer=private->playbackBuffer;
      int playbackBufferSize=private->playbackBufferSize;
      int bytesToPass;
      int bytesPassed;

      bytesToPass = MDTM_MIN(playbackBufferSize,
			     inBrokerLine->available(inBrokerLine));

      if (bytesToPass > 0) {
	bytesToPass = MDTM_MIN(bytesToPass,daLine->available(daLine));

	if (bytesToPass > 0) {
	  bytesPassed = inBrokerLine->read(inBrokerLine,playbackBuffer,0,
					   bytesToPass);
#if (DEBUG)
	  printf("Playback passing %d bytes\n",bytesPassed);
	  fflush(stdout);
#endif
	  daLine->write(daLine,playbackBuffer,0,bytesPassed);
	} else if (!daLine->isRunning(daLine)) {
	  daLine->start(daLine);
	}
      } else if (!inBrokerLine->isActive(inBrokerLine) &&
		 !controller->isRequestRedundant(controller,
						 controller->startPlayback,
						 NULL,this->object)) {
	controller->stopPlayback(controller,NULL,this->object);
      } else {
	restart_ms = CATS_FRAME_SIZE_MSEC;
      }
    }

    if (private->started && !private->playbackStarted)
      this->stop(this);

    /* Set task to repeat polling function */

    Gal_ReAddTask(taskPkg,(void *)this,restart_ms,GAL_TRUE,NULL);

  } /* if (private->opened) */

}
