/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <stdlib.h>
#include <mdtm/mdtmmacros.h>
#include "cats_request.h"

static CatsRequest CatsRequest_destroy(CatsRequest this);
static int CatsRequest_addChild(CatsRequest this, CatsRequest request);
static void CatsRequest_removeChild(CatsRequest this, CatsRequest request);
static int CatsRequest_equals(CatsRequest this, CatsRequest request);
static int CatsRequest_isRecursive(CatsRequest this);

CatsRequest
CatsRequest_create(void *caller, void (*monitorCallback)(CatsRequest),
		   void *monitor, void (*requesterCallback)(CatsRequest),
		   CatsObject requester, void *requestData)
{
  CatsRequest this=NULL;

  /* Everything else is optional, but we must have a valid monitor
     callback function */

  if (monitorCallback == NULL)
    return NULL;

  if ((this =
       (CatsRequest)
       calloc(1,sizeof(struct CatsRequest_str))) != NULL) {

    this->destroy = CatsRequest_destroy;
    this->addChild = CatsRequest_addChild;
    this->removeChild = CatsRequest_removeChild;
    this->equals = CatsRequest_equals;
    this->isRecursive = CatsRequest_isRecursive;

    if ((this->object = CatsObject_create(this,CATS_REQUEST)) == NULL)
      return this->destroy(this);
    if ((this->children = CatsVector_create(0,0)) == NULL)
      return this->destroy(this);

    /* Initialize fields passed as arguments */

    this->caller = caller;
    this->monitorCallback = monitorCallback;
    this->monitor = monitor;
    this->requesterCallback = requesterCallback;
    this->requester = requester;
    this->requestData = requestData;

    /* Initialize independent fields */

    this->requestQueue = NULL;
    this->status = CATS_REQUEST_INACTIVE;

    /* If requester is itself a request, make this request a child of it */

    if (requester->objectType == CATS_REQUEST) {
      CatsRequest parent=requester->handle;
      parent->addChild(parent,this);
    }
  }

  return this;
}

static CatsRequest CatsRequest_destroy(CatsRequest this)
{
  CatsObject object=this->object;
  CatsVector children=this->children;
  CatsRequest parent=this->parent;

  if (object != NULL)
    object->destroy(object);

  if (children != NULL) {
    CatsRequestQueue requestQueue=this->requestQueue;

    while (children->size(children) > 0) {
      CatsRequest child=children->remove(children,0);

      /* EBG: Destroying a child request while it's still on the
         request queue is a Bad Idea */

      /* SAM 9/23/02: But first, we should always remove
	 the parent, so that we don't get bad backpointers. */
      
      child->parent = NULL;
      if (requestQueue == NULL || !requestQueue->contains(requestQueue,child))
	child->destroy(child);
    }

    children->destroy(children);
  }

  if (parent != NULL)
    parent->removeChild(parent,this);

  free(this);
  return NULL;
}

static int CatsRequest_addChild(CatsRequest this, CatsRequest request)
{
  CatsVector children=this->children;
  int retval=children->add(children,request);

  if (retval)
    request->parent = this;

  return retval;
}

static void CatsRequest_removeChild(CatsRequest this, CatsRequest request)
{
  CatsVector children=this->children;
  children->removeElement(children,request);
}

static int CatsRequest_equals(CatsRequest this, CatsRequest request)
{
  return this->monitor == request->monitor &&
    this->monitorCallback == request->monitorCallback &&
    this->requestData == request->requestData;
}

static int CatsRequest_isRecursive(CatsRequest this)
{
  CatsRequest parent=this->parent;

  while (parent != NULL) {
    if (this->equals(this,parent))
      return MDTM_TRUE;
    parent=parent->parent;
  }

  return MDTM_FALSE;
}
