/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include "cats_request_queue.h"
#include "cats_vector.h"
#include <mdtm/mdtmmacros.h>

static CatsRequestQueue CatsRequestQueue_destroy(CatsRequestQueue this);
static int CatsRequestQueue_add(CatsRequestQueue this, CatsRequest request);
static int CatsRequestQueue_contains(CatsRequestQueue this,
				     CatsRequest request);
static int CatsRequestQueue_removeElement(CatsRequestQueue this,
					  CatsRequest request);
static int
CatsRequestQueue_isRedundant(CatsRequestQueue this, void *caller,
			     void (*requesterCallback)(CatsRequest),
			     CatsObject requester);
static void CatsRequestQueue_service(CatsRequestQueue this);

struct CatsRequestQueuePrivate_str {
  CatsVector requests;
};

CatsRequestQueue CatsRequestQueue_create(void)
{
  CatsRequestQueue this=NULL;

  if ((this =
       (CatsRequestQueue)
       calloc(1,sizeof(struct CatsRequestQueue_str))) != NULL) {
    CatsRequestQueuePrivate private;

    this->destroy = CatsRequestQueue_destroy;
    this->add = CatsRequestQueue_add;
    this->contains = CatsRequestQueue_contains;
    this->removeElement = CatsRequestQueue_removeElement;
    this->isRedundant = CatsRequestQueue_isRedundant;
    this->service = CatsRequestQueue_service;

    if ((private = this->private =
         (CatsRequestQueuePrivate)
	 calloc(1,sizeof(struct CatsRequestQueuePrivate_str)))
        == NULL)
      return(this->destroy(this));

    if ((private->requests = CatsVector_create(0,0)) == NULL)
      return(this->destroy(this));
  }

  return this;
}

static CatsRequestQueue CatsRequestQueue_destroy(CatsRequestQueue this)
{
  if (this->private != NULL) {
    CatsVector requests=this->private->requests;

    if (requests != NULL) {
      CatsRequest request;
      while (requests->size(requests) > 0) {
	request = requests->remove(requests,0);
	request->destroy(request);
      }
      requests->destroy(requests);
    }

    free(this->private);
  }
  free(this);
  return NULL;
}

static int CatsRequestQueue_add(CatsRequestQueue this, CatsRequest request)
{
  CatsVector requests=this->private->requests;

  if (requests->add(requests,request) == -1)
    return MDTM_FALSE;

  request->requestQueue=this;

  return MDTM_TRUE;
}

static int CatsRequestQueue_contains(CatsRequestQueue this,
				     CatsRequest request)
{
  CatsVector requests=this->private->requests;
  return requests->contains(requests,request);
}

static int CatsRequestQueue_removeElement(CatsRequestQueue this,
					  CatsRequest request)
{
  CatsVector requests=this->private->requests;
  int index;

  if ((index = requests->removeElement(requests,request)) != -1) {
    CatsVector children=request->children;
    CatsRequest child;

    request->status = CATS_REQUEST_ABORTED;
    if (request->requesterCallback != NULL)
      request->requesterCallback(request);

    while (children->size(children) > 0) {
      child = children->remove(children,0);
      this->removeElement(this,child);
    }
  }

  return index;
}

static int
CatsRequestQueue_isRedundant(CatsRequestQueue this, void *caller,
			     void (*requesterCallback)(CatsRequest),
			     CatsObject requester)
{
  CatsVector requests=this->private->requests;
  CatsRequest request;
  int i;

  for (i=0; i<requests->size(requests); i++) {
    request = requests->elementAt(requests,i);
    if (request->caller == caller &&
	request->requesterCallback == requesterCallback &&
	request->requester == requester)
      return MDTM_TRUE;
  }

  return MDTM_FALSE;
}

static void CatsRequestQueue_service(CatsRequestQueue this)
{
  CatsVector requests=this->private->requests;

  requests->lock(requests);

  while (requests->size(requests) > 0) {
    CatsRequest request=requests->firstElement(requests);

    request->monitorCallback(request);

    /* On any non-pending request, respond via callback, and pop and
       destroy the request.  Break at first pending request. */

    if (request->status != CATS_REQUEST_PENDING) {
      if (request->requesterCallback != NULL)
	request->requesterCallback(request);
      requests->remove(requests,0);
      request->destroy(request);
    } else {
      break;
    }
  }

  requests->unlock(requests);
}
