/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <stdlib.h>
#include <string.h>
#include <mdtm/mdtm.h>
#include <mdtm/mdtmmacros.h>
#include "cats_ring_buffer.h"
#include "cats_lock_object.h"

#define CATS_DEFAULT_BUFFER_SIZE 1024

static CatsRingBuffer
CatsRingBuffer_destroy(CatsRingBuffer this);
static int CatsRingBuffer_readable(CatsRingBuffer this);
static int CatsRingBuffer_read(CatsRingBuffer this, char *b, int off,
			       int len);
static int CatsRingBuffer_writeable(CatsRingBuffer this);
static int CatsRingBuffer_write(CatsRingBuffer this, char *b, int off,
				int len);
static void CatsRingBuffer_flush(CatsRingBuffer this);
static int CatsRingBuffer_getBufferSize(CatsRingBuffer this);
static void CatsRingBuffer_lock(CatsRingBuffer this);
static void CatsRingBuffer_unlock(CatsRingBuffer this);
static void CatsRingBuffer_release(CatsRingBuffer this);
static void CatsRingBuffer_relock(CatsRingBuffer this);
static int CatsRingBuffer_isLocked(CatsRingBuffer this);
static int CatsRingBuffer_isReleased(CatsRingBuffer this);
static void CatsRingBuffer_notify(CatsRingBuffer this);

struct CatsRingBufferPrivate_str {
  int bufferSize;
  char *buffer;
  int writeOffset;
  int writeable;
  int blocking;
  CatsLockObject lockObject;
};

CatsRingBuffer CatsRingBuffer_create(int bufferSize)
{
  CatsRingBuffer this=NULL;

  if ((this =
       (CatsRingBuffer)
       calloc(1,sizeof(struct CatsRingBuffer_str))) != NULL) {
    CatsRingBufferPrivate private=NULL;

    this->destroy = CatsRingBuffer_destroy;
    this->readable = CatsRingBuffer_readable;
    this->read = CatsRingBuffer_read;
    this->writeable = CatsRingBuffer_writeable;
    this->write = CatsRingBuffer_write;
    this->flush = CatsRingBuffer_flush;
    this->getBufferSize = CatsRingBuffer_getBufferSize;
    this->lock = CatsRingBuffer_lock;
    this->unlock = CatsRingBuffer_unlock;
    this->release = CatsRingBuffer_release;
    this->relock = CatsRingBuffer_relock;
    this->isLocked = CatsRingBuffer_isLocked;
    this->isReleased = CatsRingBuffer_isReleased;
    this->notify = CatsRingBuffer_notify;

    if ((private = this->private =
         (CatsRingBufferPrivate)
	 calloc(1,sizeof(struct CatsRingBufferPrivate_str)))
        == NULL)
      return(this->destroy(this));

    private->blocking = bufferSize != 0;
    private->bufferSize = private->blocking ? bufferSize :
      CATS_DEFAULT_BUFFER_SIZE;
    if ((private->buffer = mdtm_vrchar_malloc(private->bufferSize)) == NULL)
      return(this->destroy(this));
    private->writeable = private->bufferSize;
    private->writeOffset = 0;

    if ((private->lockObject = CatsLockObject_create()) == NULL)
      return(this->destroy(this));
  }
  return this;
}

static CatsRingBuffer CatsRingBuffer_destroy(CatsRingBuffer this)
{
  if (this->private != NULL) {
    CatsRingBufferPrivate private=this->private;
    if (private->lockObject != NULL) {
      CatsRingBuffer_lock(this);
      private->lockObject->destroy(private->lockObject);
    }
    if (private->buffer != NULL)
      mdtm_vrchar_free(private->buffer);
    free(this->private);
  }
  free(this);
  return NULL;
}

static int CatsRingBuffer_write(CatsRingBuffer this, char *b, int off,
				int len)
{
  CatsLockObject lockObject=this->private->lockObject;
  int bytesWritten=0;

  lockObject->lock(lockObject);

  while (len > 0) {
    CatsRingBufferPrivate private=this->private;
    int writeable,writeOffset,bytesToWrite,bytesToWriteBegin,
      bytesToWriteEnd;

    writeable = private->writeable;
    writeOffset = private->writeOffset;

    if (!private->blocking && len > writeable) {
      int oldBufferSize = private->bufferSize;
      int newBufferSize =
	1<<mdtm_srint_nextpow2(oldBufferSize+len-writeable);
      int readOffset = writeOffset + writeable;

      if ((private->buffer=
	   mdtm_vrchar_realloc(private->buffer,oldBufferSize,
			       newBufferSize)) == NULL) {
	CatsRingBuffer_unlock(this);
	return -1;
      }

      if (readOffset < oldBufferSize) {
	char *cptr=private->buffer+readOffset;
	memmove(cptr+newBufferSize-oldBufferSize,cptr,
		oldBufferSize-readOffset);
      }

      private->bufferSize = newBufferSize;
      writeable = private->writeable +=
	(newBufferSize-oldBufferSize);
    }

    bytesToWrite = MDTM_MIN(len,writeable);
    bytesToWriteEnd = MDTM_MIN(bytesToWrite,
			       private->bufferSize-writeOffset);

    memcpy(private->buffer+writeOffset,b+off,bytesToWriteEnd);
    off += bytesToWriteEnd;
    writeOffset = (writeOffset + bytesToWriteEnd) % private->bufferSize;
    bytesToWriteBegin = bytesToWrite - bytesToWriteEnd;

    if (bytesToWriteBegin > 0) {
      memcpy(private->buffer,b+off,bytesToWriteBegin);
      off += bytesToWriteBegin;
      writeOffset = bytesToWriteBegin;
    }

    private->writeable -= bytesToWrite;
    len -= bytesToWrite;
    private->writeOffset = writeOffset;
    bytesWritten += bytesToWrite;

    if (len > 0)
      lockObject->notify(lockObject);
  }

  lockObject->unlock(lockObject);

  return bytesWritten;
}

static void CatsRingBuffer_flush(CatsRingBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;

  lockObject->lock(lockObject);
  this->private->writeable = this->private->bufferSize;
  this->private->writeOffset = 0;
  lockObject->unlock(lockObject);
}

static int CatsRingBuffer_getBufferSize(CatsRingBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  int bufferSize;

  lockObject->lock(lockObject);
  bufferSize=this->private->bufferSize;
  lockObject->unlock(lockObject);
  return bufferSize;
}

static int CatsRingBuffer_readable(CatsRingBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  int retval;

  lockObject->lock(lockObject);
  retval = this->private->bufferSize-this->private->writeable;
  lockObject->unlock(lockObject);

  return retval;
}

static int CatsRingBuffer_read(CatsRingBuffer this, char *b, int off, int len)
{
  CatsLockObject lockObject=this->private->lockObject;
  int bytesRead=0;

  lockObject->lock(lockObject);

  while (len > 0) {
    CatsRingBufferPrivate private=this->private;
    int available,readOffset,bytesToRead,bytesToReadBegin,bytesToReadEnd;

    available = private->bufferSize-private->writeable;
    readOffset = (private->writeOffset + private->writeable)
      % private->bufferSize;

    bytesToRead = MDTM_MIN(len,available);
    bytesToReadEnd = MDTM_MIN(bytesToRead,private->bufferSize-readOffset);

    memcpy(b+off,private->buffer+readOffset,bytesToReadEnd);
    off += bytesToReadEnd;
    bytesToReadBegin = bytesToRead - bytesToReadEnd;

    if (bytesToReadBegin > 0) {
      memcpy(b+off,private->buffer,bytesToReadBegin);
      off += bytesToReadBegin;
    }

    private->writeable += bytesToRead;
    len -= bytesToRead;
    bytesRead += bytesToRead;

    if (len > 0)
      lockObject->notify(lockObject);

  }

  lockObject->unlock(lockObject);

  return bytesRead;
}

static int CatsRingBuffer_writeable(CatsRingBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  int retval;

  lockObject->lock(lockObject);
  retval = this->private->writeable;
  lockObject->unlock(lockObject);
  return retval;
}

static void CatsRingBuffer_lock(CatsRingBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  lockObject->lock(lockObject);
}

static void CatsRingBuffer_unlock(CatsRingBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  lockObject->unlock(lockObject);
}

static void CatsRingBuffer_release(CatsRingBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  lockObject->release(lockObject);
}

static void CatsRingBuffer_relock(CatsRingBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  lockObject->relock(lockObject);
}

static int CatsRingBuffer_isLocked(CatsRingBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  return lockObject->isLocked(lockObject);
}

static int CatsRingBuffer_isReleased(CatsRingBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  return lockObject->isReleased(lockObject);
}

static void CatsRingBuffer_notify(CatsRingBuffer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  lockObject->notify(lockObject);
}
