/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <stdlib.h>
#include <galaxy/galaxy_all.h>
#include <mdtm/mdtm.h>
#include <string.h>

#include "cats_server.h"
#include "cats_server_event.h"
#include "cats_phone_event.h"
#include "cats_data_in_broker.h"
#include "cats_data_out_broker.h"
#include "cats_target_data_line.h"
#include "cats_source_data_line.h"
#include "cats_controller_event.h"
#include "cats_vector.h"
#include "cats_lock_object.h"

#define SERVER_ID "cats"

/* Key strings */

#define SERVER_RESPONSE_KEY ":server_response"
#define ON_HOOK_KEY ":on_hook"
#define MUTE_ACTIVE_KEY ":mute_active"
#define RECORD_STARTABLE_KEY ":record_startable"
#define RECORD_STARTED_KEY ":record_started"
#define AUDIO_RESENDABLE_KEY ":audio_resendable"
#define RESEND_KEY ":resend"
#define BARGE_IN_KEY ":barge_in"
#define BARGE_IN_ENABLED_KEY ":barge_in_enabled"
#define DEBUG_ENABLED_KEY ":debug_enabled"
#define HALF_DUPLEX_ENABLED_KEY ":half_duplex_enabled"
#define PTT_ENABLED_KEY ":ptt_enabled"
#define STREAMING_ENABLED_KEY ":streaming_enabled"
#define PLAYBOX_ENABLED_KEY ":playbox_enabled"
#define VOICE_ACTIVE_KEY ":voice_active"
#define VAD_STARTED_KEY ":vad_started"
#define VAD_ACTIVE_KEY ":vad_active"
#define PLAYBACK_STARTED_KEY ":playback_started"
#define TOUCHTONE_KEY ":touchtone"
#define AUDIO_READY_KEY ":audio_ready"

/* Galaxy shorthand */

#define GAL_INT_0 Gal_IntObject(0)
#define GAL_INT_1 Gal_IntObject(1)

/* PUBLIC METHOD PROTOTYPES */

static CatsServer CatsServer_destroy(CatsServer this);
static void CatsServer_receiveAudio(CatsServer this, Gal_Frame f);
static void CatsServer_offHook(CatsServer this);
static void CatsServer_onHook(CatsServer this);
static void CatsServer_touchtone(CatsServer this, char *idString);
static void CatsServer_mute(CatsServer this);
static void CatsServer_unmute(CatsServer this);
static void CatsServer_startRecording(CatsServer this);
static void CatsServer_stopRecording(CatsServer this);
static int CatsServer_dataInBrokersAvailable(CatsServer this);
static CatsAudioFormat CatsServer_getBrokerAudioFormat(CatsServer this);
static CatsSourceDataLine CatsServer_getOutBrokerLine(CatsServer this);
static CatsTargetDataLine CatsServer_getInBrokerLine(CatsServer this);
static void CatsServer_flushInBrokers(CatsServer this);
static GalIO_CommStruct *CatsServer_getCommStruct(CatsServer this);
static CatsController CatsServer_getController(CatsServer this);
static GalSS_Environment *CatsServer_getEnvironment(CatsServer this);
static void CatsServer_resendAudio(CatsServer this);
static void CatsServer_queryStatus(CatsServer this, int server_response);
static void CatsServer_setMode(CatsServer this, Gal_Frame f);
static char *CatsServer_getServerID(CatsServer this);
static void CatsServer_shutDown(CatsServer this);

/* EventGenerator methods */

static int CatsServer_addListener(CatsServer this, CatsEventListener l);
static void CatsServer_removeListener(CatsServer this, CatsEventListener l);

/* PRIVATE METHOD PROTOTYPES */

/* Event notifier */

static void CatsServer_notifySubscribers(CatsServer this, CatsEvent event);

/* EventListener implementation */

static void CatsServer_controllerEventGenerated(CatsEventListener listener,
						CatsEvent event);

/* Callback response methods */

static void CatsServer_offHookCallback(CatsRequest request);
static void CatsServer_onHookCallback(CatsRequest request);
static void CatsServer_touchtoneCallback(CatsRequest request);
static void CatsServer_muteCallback(CatsRequest request);
static void CatsServer_unmuteCallback(CatsRequest request);
static void CatsServer_startRecordingCallback(CatsRequest request);
static void CatsServer_stopRecordingCallback(CatsRequest request);
static void CatsServer_resendAudioCallback(CatsRequest request);
static void CatsServer_setBargeInEnabledCallback(CatsRequest request);
static void CatsServer_setHalfDuplexEnabledCallback(CatsRequest request);
static void CatsServer_setPlayboxEnabledCallback(CatsRequest request);
static void CatsServer_setPTTEnabledCallback(CatsRequest request);
static void CatsServer_setStreamingEnabledCallback(CatsRequest request);
static void CatsServer_setDebugEnabledCallback(CatsRequest request);

/* Helper methods */

/* Response frame generators */

static Gal_Frame CatsServer_captureFrame(CatsServer this, int serverResponse,
					 int eventID);
static Gal_Frame CatsServer_initializationFrame(CatsServer this);
static Gal_Frame CatsServer_phoneFrame(CatsServer this, int serverResponse,
				       int eventID);
static Gal_Frame CatsServer_modeFrame(CatsServer this, int serverResponse,
				      int eventID);
static Gal_Frame CatsServer_playbackFrame(CatsServer this, int serverResponse,
					  int eventID);
static Gal_Frame CatsServer_vadFrame(CatsServer this, int serverResponse,
				     int eventID);

/* Private data */

struct CatsServerPrivate_str {

  /* Server identifier string */

  char *serverID;

  /* Galaxy environment */

  GalSS_Environment *env;

  /* Controller */

  CatsController controller;

  /* Input/output brokers - note that there can be many input but only
     one output */

  CatsVector dataInBrokers;
  CatsDataOutBroker dataOutBroker;

  /* Listener objects */

  /* Vector of those listening to server events */

  CatsVector listeners;

  /* Controller listener */

  CatsEventListener controllerListener;

  /* Lock object */

  CatsLockObject lockObject;

  /* Supporting fields */

  int debugEnabled;
};

/* Constructor */

CatsServer CatsServer_create(CatsControllerParameters parameters,
			     void *server_data,
			     char *initial_session_id)
{
  CatsServer this=NULL;

  if ((this =
       (CatsServer) calloc(1,sizeof(struct CatsServer_str))) != NULL) {
    CatsServerPrivate private;
    CatsController controller;
    GalSS_Environment *env;
    GalIO_CommStruct *gcomm;

    if ((this->object = CatsObject_create(this,CATS_SERVER)) == NULL)
      return(this->destroy(this));

    this->destroy = CatsServer_destroy;
    this->receiveAudio = CatsServer_receiveAudio;
    this->offHook = CatsServer_offHook;
    this->onHook = CatsServer_onHook;
    this->touchtone = CatsServer_touchtone;
    this->mute = CatsServer_mute;
    this->unmute = CatsServer_unmute;
    this->startRecording = CatsServer_startRecording;
    this->stopRecording = CatsServer_stopRecording;
    this->dataInBrokersAvailable = CatsServer_dataInBrokersAvailable;
    this->getBrokerAudioFormat = CatsServer_getBrokerAudioFormat;
    this->getOutBrokerLine = CatsServer_getOutBrokerLine;
    this->getInBrokerLine = CatsServer_getInBrokerLine;
    this->flushInBrokers = CatsServer_flushInBrokers;
    this->getCommStruct = CatsServer_getCommStruct;
    this->getController = CatsServer_getController;
    this->getEnvironment = CatsServer_getEnvironment;
    this->resendAudio = CatsServer_resendAudio;
    this->queryStatus = CatsServer_queryStatus;
    this->setMode = CatsServer_setMode;
    this->getServerID = CatsServer_getServerID;
    this->shutDown = CatsServer_shutDown;
    this->addListener = CatsServer_addListener;
    this->removeListener = CatsServer_removeListener;

    /* Create private data object */

    if ((private = this->private =
	 (CatsServerPrivate) calloc(1,sizeof(struct CatsServerPrivate_str)))
	== NULL)
      return(this->destroy(this));

    /* INITIALIZE INDEPENDENT FIELDS */

    if ((private->dataInBrokers = CatsVector_create(0,0)) == NULL)
      return(this->destroy(this));
    /* SAM 9/19/02: If you malloc in a DLL on Windows, you
       need to free in the DLL as well, or the debugger will
       break. This used to be mdtm_text, but there's no
       free function for mdtm_text. */
    if ((private->serverID = strdup(SERVER_ID)) == NULL)
      return(this->destroy(this));
    if ((private->dataOutBroker=CatsDataOutBroker_create(this)) == NULL)
      return(this->destroy(this));
    if ((private->lockObject = CatsLockObject_create()) == NULL)
      return(this->destroy(this));
    if ((private->listeners = CatsVector_create(0,0)) == NULL)
      return(this->destroy(this));

    /* INITIALIZE DEPENDENT FIELDS */

    /* Set controller, create EventListener instance to handle
       controller events */

    if ((controller = private->controller =
	 CatsController_create(parameters,this))	== NULL)
      return(this->destroy(this));
    controller->open(controller);
    private->debugEnabled=controller->isDebugEnabled(controller);

    if ((private->controllerListener = CatsEventListener_create(this))
	== NULL)
      return(this->destroy(this));
    private->controllerListener->eventGenerated=
      CatsServer_controllerEventGenerated;
    controller->addListener(controller,private->controllerListener);

    /* Initialize Galaxy environment */

    env = (GalSS_Environment *) server_data;
    gcomm = GalSS_EnvComm(env);
    GalSS_EnvMaintainInLocation(gcomm,initial_session_id,&this->private->env);

    /* Throw ready frame back at the hub */

    this->queryStatus(this,GAL_FALSE);
  }

  return this;
}

/* Destructor implementation */

static CatsServer CatsServer_destroy(CatsServer this)
{
  CatsServerPrivate private=this->private;
  CatsObject object=this->object;

  if (object != NULL)
    object->destroy(object);

  if (private != NULL) {
    CatsController controller=private->controller;
    CatsEventListener controllerListener=private->controllerListener;
    CatsVector dataInBrokers=private->dataInBrokers;
    CatsDataOutBroker dataOutBroker=private->dataOutBroker;
    CatsLockObject lockObject=private->lockObject;
    CatsDataInBroker dataInBroker;

    if (dataInBrokers != NULL) {
      while (dataInBrokers->size(dataInBrokers) > 0) {
	dataInBroker = dataInBrokers->remove(dataInBrokers,0);
	dataInBroker->destroy(dataInBroker);
      }
      dataInBrokers->destroy(dataInBrokers);
    }

    if (private->serverID != NULL)
      free(private->serverID);

    if (controller != NULL)
      controller->destroy(controller);
    if (controllerListener != NULL)
      controllerListener->destroy(controllerListener);
    if (dataOutBroker != NULL)
      dataOutBroker->destroy(dataOutBroker);
    if (lockObject != NULL)
      lockObject->destroy(lockObject);

    free(this->private);
  }

  free(this);
  return NULL;
}

static void CatsServer_receiveAudio(CatsServer this, Gal_Frame f)
{
  CatsVector dataInBrokers=this->private->dataInBrokers;
  CatsDataInBroker dataInBroker;
  CatsEvent event;

  if ((dataInBroker=CatsDataInBroker_create(this,f)) == NULL) {
    this->shutDown(this);
    return;
  } else if (!dataInBrokers->add(dataInBrokers,dataInBroker)) {
    GalUtil_Error("%s\n","Failed to allocate frame buffer!");
    this->shutDown(this);
    return;
  }

  if ((event = CatsEvent_create(this,CATS_AUDIO_BROKER_EVENT,
				CATS_BROKER_IN_AVAILABLE)) != NULL)
    CatsServer_notifySubscribers(this,event);

}

static void CatsServer_offHook(CatsServer this)
{
  CatsServerPrivate private=this->private;
  CatsLockObject lockObject=private->lockObject;
  CatsController controller=private->controller;

  lockObject->lock(lockObject);

  controller->offHook(controller,CatsServer_offHookCallback,this->object);

  lockObject->unlock(lockObject);
}

static void CatsServer_offHookCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  Gal_Frame responseFrame;

  if ((responseFrame = CatsServer_phoneFrame(this,GAL_TRUE,CATS_OFF_HOOK))
      != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static void CatsServer_onHook(CatsServer this)
{
  CatsServerPrivate private=this->private;
  CatsLockObject lockObject=private->lockObject;
  CatsController controller=private->controller;

  lockObject->lock(lockObject);

  controller->onHook(controller,CatsServer_onHookCallback,this->object);

  lockObject->unlock(lockObject);
}

static void CatsServer_onHookCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  Gal_Frame responseFrame;

  if ((responseFrame = CatsServer_phoneFrame(this,GAL_TRUE,CATS_ON_HOOK))
      != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static void CatsServer_touchtone(CatsServer this, char *idString)
{
  CatsLockObject lockObject=this->private->lockObject;
  CatsController controller=this->private->controller;
  int eventID = -1;

  lockObject->lock(lockObject);

  if (mdtm_text_cmp(idString,"1"))
    eventID = CATS_TOUCHTONE_1;
  else if (mdtm_text_cmp(idString,"2"))
    eventID = CATS_TOUCHTONE_2;
  else if (mdtm_text_cmp(idString,"3"))
    eventID = CATS_TOUCHTONE_3;
  else if (mdtm_text_cmp(idString,"4"))
    eventID = CATS_TOUCHTONE_4;
  else if (mdtm_text_cmp(idString,"5"))
    eventID = CATS_TOUCHTONE_5;
  else if (mdtm_text_cmp(idString,"6"))
    eventID = CATS_TOUCHTONE_6;
  else if (mdtm_text_cmp(idString,"7"))
    eventID = CATS_TOUCHTONE_7;
  else if (mdtm_text_cmp(idString,"8"))
    eventID = CATS_TOUCHTONE_8;
  else if (mdtm_text_cmp(idString,"9"))
    eventID = CATS_TOUCHTONE_9;
  else if (mdtm_text_cmp(idString,"*"))
    eventID = CATS_TOUCHTONE_STAR;
  else if (mdtm_text_cmp(idString,"0"))
    eventID = CATS_TOUCHTONE_0;
  else if (mdtm_text_cmp(idString,"#"))
    eventID = CATS_TOUCHTONE_POUND;

  controller->touchtone(controller,eventID,CatsServer_touchtoneCallback,
			this->object);

  lockObject->unlock(lockObject);
}

static void CatsServer_touchtoneCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  int eventID=(int)request->requestData;
  Gal_Frame responseFrame;

  if (request->status == CATS_REQUEST_SUCCESS &&
      (responseFrame = CatsServer_phoneFrame(this,GAL_TRUE,eventID))
      != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static void CatsServer_mute(CatsServer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  CatsController controller=this->private->controller;

  lockObject->lock(lockObject);

  controller->activateMute(controller,CatsServer_muteCallback,this->object);

  lockObject->unlock(lockObject);
}

static void CatsServer_muteCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  Gal_Frame responseFrame;

  if ((responseFrame = CatsServer_vadFrame(this,GAL_TRUE,CATS_MUTE_ACTIVE))
      != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static void CatsServer_unmute(CatsServer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  CatsController controller=this->private->controller;

  lockObject->lock(lockObject);

  controller->deactivateMute(controller,CatsServer_unmuteCallback,
			     this->object);

  lockObject->unlock(lockObject);
}

static void CatsServer_unmuteCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  Gal_Frame responseFrame;

  if ((responseFrame = CatsServer_vadFrame(this,GAL_TRUE,CATS_MUTE_INACTIVE))
      != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static void CatsServer_startRecording(CatsServer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  CatsController controller=this->private->controller;

  lockObject->lock(lockObject);

  controller->startRecord(controller,CatsServer_startRecordingCallback,
			  this->object);

  lockObject->unlock(lockObject);
}

static void CatsServer_startRecordingCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  Gal_Frame responseFrame;

  if ((responseFrame =
       CatsServer_captureFrame(this,GAL_TRUE,CATS_RECORD_STARTED)) != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static void CatsServer_stopRecording(CatsServer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  CatsController controller=this->private->controller;

  lockObject->lock(lockObject);

  controller->stopRecord(controller,CatsServer_stopRecordingCallback,
			 this->object);

  lockObject->unlock(lockObject);
}

static void CatsServer_stopRecordingCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  Gal_Frame responseFrame;

  if ((responseFrame =
       CatsServer_captureFrame(this,GAL_TRUE,CATS_RECORD_STOPPED)) != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static void CatsServer_resendAudio(CatsServer this)
{
  CatsLockObject lockObject=this->private->lockObject;
  CatsController controller=this->private->controller;

  lockObject->lock(lockObject);

  controller->resend(controller,CatsServer_resendAudioCallback,this->object);

  lockObject->unlock(lockObject);
}

static void CatsServer_resendAudioCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  Gal_Frame responseFrame;

  if (request->status == CATS_REQUEST_SUCCESS)
    responseFrame = CatsServer_captureFrame(this,GAL_TRUE,CATS_RESEND);
  else {
    responseFrame = Gal_MakeFrame(this->private->serverID,GAL_CLAUSE);
    Gal_SetProp(responseFrame,RESEND_KEY,GAL_INT_0);
  }

  if (responseFrame != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static void CatsServer_queryStatus(CatsServer this, int server_response)
{
  CatsServerPrivate private=this->private;
  CatsLockObject lockObject=private->lockObject;
  Gal_Frame responseFrame=NULL;

  lockObject->lock(lockObject);

  if ((responseFrame = Gal_MakeFrame(private->serverID,GAL_CLAUSE))
      != NULL) {
    CatsController controller=private->controller;

    Gal_SetProp(responseFrame,SERVER_RESPONSE_KEY,
		Gal_IntObject(server_response));
    Gal_SetProp(responseFrame,":query_status",GAL_INT_1);

    Gal_SetProp(responseFrame,BARGE_IN_ENABLED_KEY,
		Gal_IntObject(controller->isBargeInEnabled(controller)));
    Gal_SetProp(responseFrame,HALF_DUPLEX_ENABLED_KEY,
		Gal_IntObject(controller->isHalfDuplexEnabled(controller)));
    Gal_SetProp(responseFrame,PTT_ENABLED_KEY,
		Gal_IntObject(controller->isPTTEnabled(controller)));
    Gal_SetProp(responseFrame,STREAMING_ENABLED_KEY,
		Gal_IntObject(controller->isStreamingEnabled(controller)));
    Gal_SetProp(responseFrame,PLAYBOX_ENABLED_KEY,
		Gal_IntObject(controller->isPlayboxEnabled(controller)));
    Gal_SetProp(responseFrame,MUTE_ACTIVE_KEY,
		Gal_IntObject(controller->isMuteActive(controller)));
    Gal_SetProp(responseFrame,ON_HOOK_KEY,
		Gal_IntObject(controller->isOnHook(controller)));
    Gal_SetProp(responseFrame,RECORD_STARTED_KEY,
		Gal_IntObject(controller->isRecordStarted(controller)));
    Gal_SetProp(responseFrame,RECORD_STARTABLE_KEY,
		Gal_IntObject(controller->isRecordStartable(controller)));
    Gal_SetProp(responseFrame,VOICE_ACTIVE_KEY,
		Gal_IntObject(controller->isVoiceActive(controller)));
    Gal_SetProp(responseFrame,AUDIO_RESENDABLE_KEY,
		Gal_IntObject(controller->isResendable(controller)));

    GalSS_EnvWriteFrame(private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }

  lockObject->unlock(lockObject);
}

static void CatsServer_setMode(CatsServer this, Gal_Frame f)
{
  CatsServerPrivate private=this->private;
  CatsLockObject lockObject=private->lockObject;
  CatsController controller=private->controller;
  Gal_Frame responseFrame=NULL;

  lockObject->lock(lockObject);

  if ((responseFrame = Gal_MakeFrame(private->serverID,GAL_CLAUSE))
      != NULL) {
    Gal_Object mode_object;

    if ((mode_object = Gal_GetObject(f,BARGE_IN_ENABLED_KEY)) != NULL)
      controller->setBargeInEnabled(controller,
				    CatsServer_setBargeInEnabledCallback,
				    this->object,Gal_IntValue(mode_object));
    else if ((mode_object = Gal_GetObject(f,HALF_DUPLEX_ENABLED_KEY)) != NULL)
      controller->setHalfDuplexEnabled(controller,
				       CatsServer_setHalfDuplexEnabledCallback,
				       this->object,
				       Gal_IntValue(mode_object));
    else if ((mode_object = Gal_GetObject(f,PLAYBOX_ENABLED_KEY)) != NULL)
      controller->setPlayboxEnabled(controller,
				    CatsServer_setPlayboxEnabledCallback,
				    this->object,
				    Gal_IntValue(mode_object));
    else if ((mode_object = Gal_GetObject(f,PTT_ENABLED_KEY)) != NULL)
      controller->setPTTEnabled(controller,
				CatsServer_setPTTEnabledCallback,
				this->object,Gal_IntValue(mode_object));
    else if ((mode_object = Gal_GetObject(f,STREAMING_ENABLED_KEY)) != NULL)
      controller->setStreamingEnabled(controller,
				      CatsServer_setStreamingEnabledCallback,
				      this->object,Gal_IntValue(mode_object));
    else if ((mode_object = Gal_GetObject(f,DEBUG_ENABLED_KEY)) != NULL)
      controller->setDebugEnabled(controller,
				  CatsServer_setDebugEnabledCallback,
				  this->object,Gal_IntValue(mode_object));
  }

  lockObject->unlock(lockObject);
}

static void CatsServer_setBargeInEnabledCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  int requestData=(int)request->requestData;
  int eventID=requestData ? CATS_BARGE_IN_ENABLED : CATS_BARGE_IN_DISABLED;
  Gal_Frame responseFrame;

  if ((responseFrame = CatsServer_modeFrame(this,GAL_TRUE,eventID)) != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static void CatsServer_setHalfDuplexEnabledCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  int requestData=(int)request->requestData;
  int eventID=requestData ? CATS_HALF_DUPLEX_ENABLED :
    CATS_HALF_DUPLEX_DISABLED;
  Gal_Frame responseFrame;

  if ((responseFrame = CatsServer_modeFrame(this,GAL_TRUE,eventID)) != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static void CatsServer_setPlayboxEnabledCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  int requestData=(int)request->requestData;
  int eventID=requestData ? CATS_PLAYBOX_ENABLED : CATS_PLAYBOX_DISABLED;
  Gal_Frame responseFrame;

  if ((responseFrame = CatsServer_modeFrame(this,GAL_TRUE,eventID)) != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static void CatsServer_setPTTEnabledCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  int requestData=(int)request->requestData;
  int eventID=requestData ? CATS_PTT_ENABLED : CATS_PTT_DISABLED;
  Gal_Frame responseFrame;

  if ((responseFrame = CatsServer_modeFrame(this,GAL_TRUE,eventID)) != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static void CatsServer_setStreamingEnabledCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  int requestData=(int)request->requestData;
  int eventID=requestData ? CATS_STREAMING_ENABLED : CATS_STREAMING_DISABLED;
  Gal_Frame responseFrame;

  if ((responseFrame = CatsServer_modeFrame(this,GAL_TRUE,eventID)) != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static void CatsServer_setDebugEnabledCallback(CatsRequest request)
{
  CatsObject object=request->requester;
  CatsServer this=(CatsServer)object->handle;
  int requestData=(int)request->requestData;
  int eventID=requestData ? CATS_DEBUG_ENABLED : CATS_DEBUG_DISABLED;
  Gal_Frame responseFrame;

  if ((responseFrame = CatsServer_modeFrame(this,GAL_TRUE,eventID)) != NULL) {
    GalSS_EnvWriteFrame(this->private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }
}

static int CatsServer_dataInBrokersAvailable(CatsServer this)
{
  CatsVector dataInBrokers=this->private->dataInBrokers;
  return dataInBrokers->size(dataInBrokers);
}

static CatsAudioFormat CatsServer_getBrokerAudioFormat(CatsServer this)
{
  CatsVector dataInBrokers=this->private->dataInBrokers;
  CatsDataInBroker dataInBroker=dataInBrokers->firstElement(dataInBrokers);
  return dataInBroker->getAudioFormat(dataInBroker);
}

static CatsSourceDataLine CatsServer_getOutBrokerLine(CatsServer this)
{
  CatsDataOutBroker broker=this->private->dataOutBroker;
  return broker->getLine(broker);
}

/* EBG: It is imperative for the data input broker line to destroy the
   input broker object when the line closes, to avoid a massive memory
   leak. */

static CatsTargetDataLine CatsServer_getInBrokerLine(CatsServer this)
{
  CatsVector dataInBrokers=this->private->dataInBrokers;
  CatsDataInBroker dataInBroker=dataInBrokers->remove(dataInBrokers,0);

  if (this->private->debugEnabled)
    printf("Server removing input broker, returning broker line\n");

  return dataInBroker == NULL ? NULL : dataInBroker->getLine(dataInBroker);
}

static void CatsServer_flushInBrokers(CatsServer this)
{
  CatsVector dataInBrokers=this->private->dataInBrokers;
  CatsDataInBroker dataInBroker;

  while (dataInBrokers->size(dataInBrokers) > 0) {
    dataInBroker = dataInBrokers->remove(dataInBrokers,0);
    dataInBroker->destroy(dataInBroker);
  }
}

static GalIO_CommStruct *CatsServer_getCommStruct(CatsServer this)
{
  return GalSS_EnvComm(this->private->env);
}

static CatsController CatsServer_getController(CatsServer this)
{
  return this->private->controller;
}

static GalSS_Environment *CatsServer_getEnvironment(CatsServer this)
{
  return this->private->env;
}

static char *CatsServer_getServerID(CatsServer this)
{
  return this->private->serverID;
}

static void CatsServer_shutDown(CatsServer this)
{
  GalIO_SetCommDone(this->getCommStruct(this));
}

/* EventGenerator implementation. */

static int CatsServer_addListener(CatsServer this, CatsEventListener l)
{
  CatsVector listeners=this->private->listeners;
  int retval;

  if (!(retval = listeners->add(listeners,l)))
    GalUtil_Error("%s\n","Failed to add CatsEventListener");

  return retval;
}

static void CatsServer_notifySubscribers(CatsServer this, CatsEvent event)
{
  CatsVector listeners=this->private->listeners;
  CatsEventListener listener;
  int i, nListeners=listeners->size(listeners);

  for (i=0; i<nListeners; i++) {
    listener = listeners->elementAt(listeners,i);
    listener->eventGenerated(listener,event);
  }
}

static void CatsServer_removeListener(CatsServer this, CatsEventListener l)
{
  CatsVector listeners=this->private->listeners;
  listeners->removeElement(listeners,l);
}

/* PRIVATE METHODS */

/* EventListener implementations */

/* Controller event listener */

static void CatsServer_controllerEventGenerated(CatsEventListener listener,
						CatsEvent event)
{
  CatsServer this=(CatsServer) listener->getSource(listener);
  CatsObject object=this->object;
  CatsServerPrivate private=this->private;
  CatsLockObject lockObject=private->lockObject;
  int eventType=event->eventType;
  int eventID=event->eventID;
  void *eventSource=event->source;
  
  /* BRYAN: I just added the initialization here. I'm pretty sure
     this will keep bad things from happening if you've
     forgotten to set it otherwise. */

  Gal_Frame responseFrame=NULL;

  lockObject->lock(lockObject);

  switch (eventType) {

  case CATS_CAPTURE_EVENT:
    if (eventSource != object)
      responseFrame = CatsServer_captureFrame(this,GAL_FALSE,eventID);
    break;

  case CATS_INITIALIZATION_EVENT:
    responseFrame = CatsServer_initializationFrame(this);
    break;

  case CATS_PHONE_EVENT:
    if (eventSource != object)
      responseFrame = CatsServer_phoneFrame(this,GAL_FALSE,eventID);
    break;

  case CATS_MODE_EVENT:
    if (eventSource != object)
      responseFrame = CatsServer_modeFrame(this,GAL_FALSE,eventID);
    break;

  case CATS_PLAYBACK_EVENT:
    if (eventSource != object)
      responseFrame = CatsServer_playbackFrame(this,GAL_FALSE,eventID);
    break;

  case CATS_VAD_EVENT:
    if (eventSource != object)
      responseFrame = CatsServer_vadFrame(this,GAL_FALSE,eventID);
    break;
  }

  if (responseFrame) {
    GalSS_EnvWriteFrame(private->env,responseFrame,GAL_FALSE);
    Gal_FreeFrame(responseFrame);
  }

  lockObject->unlock(lockObject);
}

/* PRIVATE METHODS */

/* Frame generators */

static Gal_Frame CatsServer_captureFrame(CatsServer this, int serverResponse,
					 int eventID)
{
  Gal_Frame responseFrame=NULL;
  CatsController controller=this->private->controller;

  if ((responseFrame = Gal_MakeFrame(this->private->serverID,GAL_CLAUSE))
      == NULL) {
    this->shutDown(this);
    return NULL;
  }

  Gal_SetProp(responseFrame,SERVER_RESPONSE_KEY,
	      Gal_IntObject(serverResponse));

  switch (eventID) {
  case CATS_RECORD_STARTED: case CATS_RECORD_STOPPED:
    Gal_SetProp(responseFrame,RECORD_STARTED_KEY,
		Gal_IntObject(controller->isRecordStarted(controller)));
    Gal_SetProp(responseFrame,AUDIO_RESENDABLE_KEY,
		Gal_IntObject(controller->isResendable(controller)));
    break;
  case CATS_RESEND:
    Gal_SetProp(responseFrame,RESEND_KEY,GAL_INT_1);
    break;
  }

  return responseFrame;
}

static Gal_Frame CatsServer_initializationFrame(CatsServer this)
{
  Gal_Frame responseFrame=NULL;

  if ((responseFrame = Gal_MakeFrame(this->private->serverID,GAL_CLAUSE))
      == NULL) {
    this->shutDown(this);
    return NULL;
  }

  Gal_SetProp(responseFrame,SERVER_RESPONSE_KEY,GAL_INT_0);
  Gal_SetProp(responseFrame,AUDIO_READY_KEY,GAL_INT_1);

  return responseFrame;
}

static Gal_Frame CatsServer_phoneFrame(CatsServer this, int serverResponse,
				       int eventID)
{
  Gal_Frame responseFrame=NULL;
  CatsController controller=this->private->controller;

  if ((responseFrame = Gal_MakeFrame(this->private->serverID,GAL_CLAUSE))
      == NULL) {
    this->shutDown(this);
    return NULL;
  }

  Gal_SetProp(responseFrame,SERVER_RESPONSE_KEY,
	      Gal_IntObject(serverResponse));

  switch (eventID) {
  case CATS_OFF_HOOK: case CATS_ON_HOOK:
    Gal_SetProp(responseFrame,ON_HOOK_KEY,
		Gal_IntObject(controller->isOnHook(controller)));
    Gal_SetProp(responseFrame,RECORD_STARTABLE_KEY,
		Gal_IntObject(controller->isRecordStartable(controller)));
    Gal_SetProp(responseFrame,AUDIO_RESENDABLE_KEY,
		Gal_IntObject(controller->isResendable(controller)));
    break;
  case CATS_TOUCHTONE_1:
    Gal_SetProp(responseFrame,TOUCHTONE_KEY,Gal_StringObject("1"));
    break;
  case CATS_TOUCHTONE_2:
    Gal_SetProp(responseFrame,TOUCHTONE_KEY,Gal_StringObject("2"));
    break;
  case CATS_TOUCHTONE_3:
    Gal_SetProp(responseFrame,TOUCHTONE_KEY,Gal_StringObject("3"));
    break;
  case CATS_TOUCHTONE_4:
    Gal_SetProp(responseFrame,TOUCHTONE_KEY,Gal_StringObject("4"));
    break;
  case CATS_TOUCHTONE_5:
    Gal_SetProp(responseFrame,TOUCHTONE_KEY,Gal_StringObject("5"));
    break;
  case CATS_TOUCHTONE_6:
    Gal_SetProp(responseFrame,TOUCHTONE_KEY,Gal_StringObject("6"));
    break;
  case CATS_TOUCHTONE_7:
    Gal_SetProp(responseFrame,TOUCHTONE_KEY,Gal_StringObject("7"));
    break;
  case CATS_TOUCHTONE_8:
    Gal_SetProp(responseFrame,TOUCHTONE_KEY,Gal_StringObject("8"));
    break;
  case CATS_TOUCHTONE_9:
    Gal_SetProp(responseFrame,TOUCHTONE_KEY,Gal_StringObject("9"));
    break;
  case CATS_TOUCHTONE_STAR:
    Gal_SetProp(responseFrame,TOUCHTONE_KEY,Gal_StringObject("*"));
    break;
  case CATS_TOUCHTONE_0:
    Gal_SetProp(responseFrame,TOUCHTONE_KEY,Gal_StringObject("0"));
    break;
  case CATS_TOUCHTONE_POUND:
    Gal_SetProp(responseFrame,TOUCHTONE_KEY,Gal_StringObject("#"));
    break;
  }

  return responseFrame;
}

static Gal_Frame CatsServer_modeFrame(CatsServer this, int serverResponse,
				      int eventID)
{
  Gal_Frame responseFrame=NULL;
  CatsController controller=this->private->controller;

  if ((responseFrame = Gal_MakeFrame(this->private->serverID,GAL_CLAUSE))
      == NULL) {
    this->shutDown(this);
    return NULL;
  }

  Gal_SetProp(responseFrame,SERVER_RESPONSE_KEY,
	      Gal_IntObject(serverResponse));

  switch (eventID) {
  case CATS_BARGE_IN_DISABLED: case CATS_BARGE_IN_ENABLED:
    Gal_SetProp(responseFrame,BARGE_IN_ENABLED_KEY,
		Gal_IntObject(controller->isBargeInEnabled(controller)));
    Gal_SetProp(responseFrame,RECORD_STARTABLE_KEY,
		Gal_IntObject(controller->isRecordStartable(controller)));
    break;
  case CATS_DEBUG_DISABLED:
    this->private->debugEnabled = GAL_FALSE;
    Gal_SetProp(responseFrame,DEBUG_ENABLED_KEY,GAL_INT_0);
    break;
  case CATS_DEBUG_ENABLED:
    this->private->debugEnabled = GAL_TRUE;
    Gal_SetProp(responseFrame,DEBUG_ENABLED_KEY,GAL_INT_1);
    break;
  case CATS_HALF_DUPLEX_DISABLED: case CATS_HALF_DUPLEX_ENABLED:
    Gal_SetProp(responseFrame,HALF_DUPLEX_ENABLED_KEY,
		Gal_IntObject(controller->isHalfDuplexEnabled(controller)));
    Gal_SetProp(responseFrame,RECORD_STARTABLE_KEY,
		Gal_IntObject(controller->isRecordStartable(controller)));
    break;
  case CATS_PTT_DISABLED: case CATS_PTT_ENABLED:
    Gal_SetProp(responseFrame,PTT_ENABLED_KEY,
		Gal_IntObject(controller->isPTTEnabled(controller)));
    break;
  case CATS_STREAMING_DISABLED: case CATS_STREAMING_ENABLED:
    Gal_SetProp(responseFrame,STREAMING_ENABLED_KEY,
		Gal_IntObject(controller->isStreamingEnabled(controller)));
    break;
  case CATS_PLAYBOX_DISABLED: case CATS_PLAYBOX_ENABLED:
    Gal_SetProp(responseFrame,PLAYBOX_ENABLED_KEY,
		Gal_IntObject(controller->isPlayboxEnabled(controller)));
    break;
  }

  return responseFrame;
}

static Gal_Frame CatsServer_playbackFrame(CatsServer this, int serverResponse,
					  int eventID)
{
  Gal_Frame responseFrame=NULL;
  CatsController controller=this->private->controller;

  if ((responseFrame = Gal_MakeFrame(this->private->serverID,GAL_CLAUSE))
      == NULL) {
    this->shutDown(this);
    return NULL;
  }

  Gal_SetProp(responseFrame,SERVER_RESPONSE_KEY,
	      Gal_IntObject(serverResponse));

  switch (eventID) {
  case CATS_PLAYBACK_STARTED:
    Gal_SetProp(responseFrame,PLAYBACK_STARTED_KEY,GAL_INT_1);
    Gal_SetProp(responseFrame,RECORD_STARTABLE_KEY,
		Gal_IntObject(controller->isRecordStartable(controller)));
    break;
  case CATS_PLAYBACK_STOPPED:
    Gal_SetProp(responseFrame,PLAYBACK_STARTED_KEY,GAL_INT_0);
    Gal_SetProp(responseFrame,RECORD_STARTABLE_KEY,
		Gal_IntObject(controller->isRecordStartable(controller)));
    break;
  case CATS_BARGE_IN:
    Gal_SetProp(responseFrame,BARGE_IN_KEY,GAL_INT_1);
    break;
  }

  return responseFrame;
}

static Gal_Frame CatsServer_vadFrame(CatsServer this, int serverResponse,
				     int eventID)
{
  Gal_Frame responseFrame=NULL;
  CatsController controller=this->private->controller;

  if ((responseFrame = Gal_MakeFrame(this->private->serverID,GAL_CLAUSE))
      == NULL) {
    this->shutDown(this);
    return NULL;
  }

  Gal_SetProp(responseFrame,SERVER_RESPONSE_KEY,
	      Gal_IntObject(serverResponse));

  switch (eventID) {
  case CATS_VAD_INACTIVE:
    Gal_SetProp(responseFrame,VAD_ACTIVE_KEY,GAL_INT_0);
    break;
  case CATS_VAD_ACTIVE:
    Gal_SetProp(responseFrame,VAD_ACTIVE_KEY,GAL_INT_1);
    break;
  case CATS_VOICE_INACTIVE:
    Gal_SetProp(responseFrame,VOICE_ACTIVE_KEY,GAL_INT_0);
    break;
  case CATS_VOICE_ACTIVE:
    Gal_SetProp(responseFrame,VOICE_ACTIVE_KEY,GAL_INT_1);
    break;
  case CATS_MUTE_INACTIVE: case CATS_MUTE_ACTIVE:
    Gal_SetProp(responseFrame,MUTE_ACTIVE_KEY,
		Gal_IntObject(controller->isMuteActive(controller)));
    break;
  case CATS_VAD_STOPPED:
    Gal_SetProp(responseFrame,VAD_STARTED_KEY,GAL_INT_0);
    break;
  case CATS_VAD_STARTED:
    Gal_SetProp(responseFrame,VAD_STARTED_KEY,GAL_INT_1);
    break;
  }

  return responseFrame;
}
