/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <stdio.h>
#include <stdlib.h>
#include <mdtm/mdtm.h>
#include <mdtm/mdtmmacros.h>
#include "cats_switchboard.h"
#include "cats_event_listener.h"
#include "cats_controller_event.h"
#include "cats_lock_object.h"
#include "cats_phone_line.h"
#include "cats_vector.h"

#define DEBUG 0
#define CATS_LINE_BUFFER_LENGTH_SEC .25
#define CATS_FRAME_SIZE_MSEC 10

static CatsSwitchboard CatsSwitchboard_destroy(CatsSwitchboard this);
static int CatsSwitchboard_open(CatsSwitchboard this);
static int CatsSwitchboard_close(CatsSwitchboard this);
static int CatsSwitchboard_isOpen(CatsSwitchboard this);
static int CatsSwitchboard_addListener(CatsSwitchboard this,
				       CatsEventListener l);
static void CatsSwitchboard_removeListener(CatsSwitchboard this,
					   CatsEventListener l);

/* Private Methods */

/* Event notifier */

static void CatsSwitchboard_notifySubscribers(CatsSwitchboard this,
					      CatsEvent event);

/* Controller listener Method */

static void
CatsSwitchboard_ControllerEventGenerated(CatsEventListener listener,
					 CatsEvent event);

/* Polling method */

static void CatsSwitchboard_run(Gal_TaskPkg *taskPkg);

struct CatsSwitchboardPrivate_str {

  /* Mode, state flags. */

  int opened;
  int debugEnabled;

  /* Phone line */

  CatsPhoneLine phoneLine;

  /* Supporting fields. */

  CatsVector listeners;
  CatsController controller;
  CatsEventListener controllerListener;
  Gal_TaskPkg *taskPkg;
  CatsLockObject lockObject;
};

CatsSwitchboard CatsSwitchboard_create(CatsController controller)
{
  CatsSwitchboard this=NULL;

  if ((this = (CatsSwitchboard) calloc(1,sizeof(struct CatsSwitchboard_str)))
      != NULL) {
    CatsSwitchboardPrivate private;

    if ((this->object = CatsObject_create(this,CATS_CAPTURE)) == NULL)
      return (this->destroy(this));

    this->destroy = CatsSwitchboard_destroy;
    this->open = CatsSwitchboard_open;
    this->close = CatsSwitchboard_close;
    this->isOpen = CatsSwitchboard_isOpen;
    this->addListener = CatsSwitchboard_addListener;
    this->removeListener = CatsSwitchboard_removeListener;

    if ((private = this->private =
         (CatsSwitchboardPrivate)
	 calloc(1,sizeof(struct CatsSwitchboardPrivate_str)))
        == NULL)
      return(this->destroy(this));

    if ((private->listeners = CatsVector_create(0,0)) == NULL)
      return (this->destroy(this));
    if ((private->lockObject = CatsLockObject_create()) == NULL)
      return (this->destroy(this));

    /* Create utterance buffers for resend */

    private->controller = controller;

    private->opened = MDTM_FALSE;

    private->phoneLine = controller->getPhoneLine(controller);

    if (!private->phoneLine->isOpen(private->phoneLine)) {
      if (!private->phoneLine->open(private->phoneLine)) {
	return (this->destroy(this));
      }
    }

    /* INITIALIZE DEPENDENT FIELDS */

    /* Create controller listener instance */

    if ((private->controllerListener = CatsEventListener_create(this))
	== NULL)
      return(this->destroy(this));
    private->controllerListener->eventGenerated =
      CatsSwitchboard_ControllerEventGenerated;
    controller->addListener(controller,private->controllerListener);

    private->taskPkg = NULL;
  }

  return this;
}

static CatsSwitchboard CatsSwitchboard_destroy(CatsSwitchboard this)
{
  if (this->private != NULL) {
    CatsSwitchboardPrivate private=this->private;
    CatsEventListener controllerListener=private->controllerListener;
    CatsVector listeners=private->listeners;
    CatsLockObject lockObject=private->lockObject;
    CatsObject object=this->object;

    if (private->opened)
      this->close(this);
    if (controllerListener != NULL)
      controllerListener->destroy(controllerListener);
    if (listeners != NULL)
      listeners->destroy(listeners);
    if (lockObject != NULL)
      lockObject->destroy(lockObject);
    if (object != NULL)
      object->destroy(object);
    free(private);
  }

  free(this);
  return NULL;
}

/* Opening the capture object starts the polling loop */

static int CatsSwitchboard_open(CatsSwitchboard this)
{
  this->private->lockObject->lock(this->private->lockObject);

  if (!this->private->opened) {
    this->private->taskPkg =
      Gal_AddTask(CatsSwitchboard_run,(void *)this,0,GAL_TRUE,NULL);
    this->private->opened = GAL_TRUE;
  }

  this->private->lockObject->unlock(this->private->lockObject);

  return this->private->opened == GAL_TRUE;
}

/* Closing the capture object stops the polling loop, stopping the
   object first if necessary */

static int CatsSwitchboard_close(CatsSwitchboard this)
{
  CatsSwitchboardPrivate private=this->private;

  private->lockObject->lock(private->lockObject);

  if (private->opened) {
    private->opened = MDTM_FALSE;

    /* Remove polling function task */

    Gal_RemoveTask(private->taskPkg);
    private->taskPkg = NULL;
  }

  private->lockObject->unlock(private->lockObject);

  return private->opened == MDTM_FALSE;
}

static int CatsSwitchboard_isOpen(CatsSwitchboard this)
{
  return this->private->opened;
}

/* EventGenerator implementation. */

static int CatsSwitchboard_addListener(CatsSwitchboard this,
				       CatsEventListener l)
{
  CatsController controller=this->private->controller;
  CatsVector listeners=this->private->listeners;
  int retval;

  if (!(retval = listeners->add(listeners,l))) {
    GalUtil_Error("%s\n","Failed to add CatsEventListener");
    controller->shutDown(controller);
  }

  return retval;
}

static void CatsSwitchboard_removeListener(CatsSwitchboard this,
					   CatsEventListener l)
{
  CatsVector listeners=this->private->listeners;
  listeners->removeElement(listeners,l);
}

/* PRIVATE METHODS */

static void CatsSwitchboard_notifySubscribers(CatsSwitchboard this,
					      CatsEvent event)
{
  CatsVector listeners=this->private->listeners;
  CatsEventListener listener;
  int i;

  for (i=0; i<listeners->size(listeners); i++) {
    listener = listeners->elementAt(listeners,i);
    listener->eventGenerated(listener,event);
  }
}

/* Controller listener method - turns on/off appropriate flags,
   nothing more. */

static void
CatsSwitchboard_ControllerEventGenerated(CatsEventListener listener,
					 CatsEvent event)
{
  CatsSwitchboard this = (CatsSwitchboard) listener->getSource(listener);

  switch (event->eventType) {

  case CATS_MODE_EVENT:
    switch (event->eventID) {
    case CATS_DEBUG_DISABLED:
      this->private->debugEnabled = MDTM_FALSE;
      break;
    case CATS_DEBUG_ENABLED:
      this->private->debugEnabled = MDTM_TRUE;
      break;
    }
    break;
  }
}

/* Polling function. */

static void CatsSwitchboard_run(Gal_TaskPkg *taskPkg)
{
  CatsSwitchboard this=Gal_TaskPkgData(taskPkg);
  CatsSwitchboardPrivate private=this->private;
  CatsPhoneLine phoneLine=private->phoneLine;
  CatsEvent event;

  if (private->opened) {
    CatsController controller=private->controller;
    int result=phoneLine->poll(phoneLine);

    switch(result) {
    case CATS_PHONE_DTMF:
#if (DEBUG)
      GalUtil_PInfo1("CATS: DTMF\n");
#endif
      controller->touchtone(controller,phoneLine->getDTMF(phoneLine),NULL,
			    this->object);
      break;
    case CATS_PHONE_FAILURE:
#if (DEBUG)
      GalUtil_PInfo1("CATS: Phone failed\n");
#endif
      break;
    case CATS_PHONE_HANGUP_RECEIVED:
#if (DEBUG)
      GalUtil_PInfo1("CATS: Phone hangup\n");
#endif
      controller->onHook(controller,NULL,this->object);
      break;
    case CATS_PHONE_ANSWERED_CALL:      
#if (DEBUG)
      GalUtil_PInfo1("CATS: Answered call\n");
#endif
      controller->offHook(controller,NULL,this->object);
      break;
    }

    /* Set task to repeat polling function */

    Gal_ReAddTask(taskPkg,(void *)this,phoneLine->getPollMsec(phoneLine),
		  GAL_TRUE,NULL);

  } /* if (private->opened) */

}
