/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <math.h>
#include <stdlib.h>
#include <mdtm/mdtm.h>
#include <mdtm/mdtmmath.h>
#include <mdtm/mdtmmacros.h>

#include "cats_vad_buffer.h"
#include "cats_delay_line.h"

#define VAD_SIGNAL_DEADBAND (.06)
#define VAD_SILENCE_DEADBAND (.4+VAD_SIGNAL_DEADBAND)
#define VAD_ADAPTIVE_PERIOD (.6)

static CatsVADBuffer CatsVADBuffer_destroy(CatsVADBuffer this);
static int CatsVADBuffer_open(CatsVADBuffer this,
			      CatsAudioFormat sourceAudioFormat,
			      int frameSize, float thresholdRaiseFactor,
			      float thresholdLowerFactor);
static int CatsVADBuffer_write(CatsVADBuffer this, char *b, int off, int len);
static int CatsVADBuffer_read(CatsVADBuffer this, char *b, int off, int len);
static int CatsVADBuffer_close(CatsVADBuffer this);
static int CatsVADBuffer_isOpen(CatsVADBuffer this);

static mdtm_srint32 CatsVADBuffer_absstd(mdtm_vrint16 *frameBuffer,
					 int frameSamples);

struct CatsVADBufferPrivate_str {
  CatsDelayLine buffer;
  int sampleSize;
  int frameSize;
  int frameSamples;
  int bufferSize;
  mdtm_vrint16 *frameBuffer;
  int byteSwapFlag;

  /* Classification fields */

  int voiceActive;
  int consecutiveChangeFrames;
  double levelThreshold;
  int signalDeadBandFrames;
  int silenceDeadBandFrames;

  /* Adaptation fields */

  int adaptivePeriodFrames;
  double signalMin;
  double silenceMax;
  int signalFramesReceived;
  int silenceFramesReceived;
  float tRaiseFactor;
  float tLowerFactor;

  int opened;
};

CatsVADBuffer CatsVADBuffer_create(void)
{
  CatsVADBuffer this=NULL;

  if ((this = (CatsVADBuffer) calloc(1,sizeof(struct CatsVADBuffer_str)))
      != NULL) {
    CatsVADBufferPrivate private;

    this->destroy = CatsVADBuffer_destroy;
    this->open = CatsVADBuffer_open;
    this->write = CatsVADBuffer_write;
    this->read = CatsVADBuffer_read;
    this->close = CatsVADBuffer_close;
    this->isOpen = CatsVADBuffer_isOpen;

    if ((private = this->private =
         (CatsVADBufferPrivate)
	 calloc(1,sizeof(struct CatsVADBufferPrivate_str)))
        == NULL)
      return(this->destroy(this));

    private->opened = MDTM_FALSE;
  }

  return this;
}

static CatsVADBuffer CatsVADBuffer_destroy(CatsVADBuffer this)
{
  if (this != NULL) {
    CatsVADBufferPrivate private=this->private;

    if (private != NULL)
      free(private);
    free(this);
  }
  return NULL;
}

static int CatsVADBuffer_open(CatsVADBuffer this,
			      CatsAudioFormat sourceAudioFormat,
			      int frameSize, float thresholdRaiseFactor,
			      float thresholdLowerFactor)
{
  CatsVADBufferPrivate private=this->private;

  if (!private->opened) {
    float sampleRate;
    int sampleSize;
    int bytesPerSecond;

    /* We really need mono PCM right now */

    if (sourceAudioFormat->getEncoding(sourceAudioFormat)
	!= CATS_ENCODING_PCM_SIGNED ||
	sourceAudioFormat->getChannels(sourceAudioFormat) != 1)
      return MDTM_FALSE;

    /* We really need integer number of samples in a frame */

    if (frameSize %
	(sampleSize=sourceAudioFormat->getFrameSize(sourceAudioFormat)) != 0)
      return MDTM_FALSE;

    sampleRate = sourceAudioFormat->getSampleRate(sourceAudioFormat);
    bytesPerSecond = sampleRate*sampleSize;

    private->sampleSize = sampleSize;
    private->frameSize = frameSize;
    private->frameSamples = frameSize/sampleSize;
    private->signalDeadBandFrames =
      ceil((VAD_SIGNAL_DEADBAND*sampleRate)/private->frameSamples);
    private->silenceDeadBandFrames =
      ceil((VAD_SILENCE_DEADBAND*sampleRate)/private->frameSamples);
    private->adaptivePeriodFrames =
      ceil((VAD_ADAPTIVE_PERIOD*sampleRate)/private->frameSamples);

    private->bufferSize =
      private->signalDeadBandFrames*private->frameSamples*sampleSize;
    if ((private->buffer = CatsDelayLine_create(private->bufferSize)) == NULL)
      return MDTM_FALSE;

    private->levelThreshold = -1.;
    private->signalMin = mdtm_srdouble_inf();
    private->silenceMax = 0.;
    private->signalFramesReceived = private->silenceFramesReceived = 0;
    private->consecutiveChangeFrames = 0;

    if ((private->frameBuffer = mdtm_vrint16_malloc(private->frameSamples))
	== NULL)
      return MDTM_FALSE;

    private->byteSwapFlag = mdtm_isbigendian() !=
      sourceAudioFormat->isBigEndian(sourceAudioFormat);

    private->tRaiseFactor = thresholdRaiseFactor;
    private->tLowerFactor = thresholdLowerFactor;
    private->opened = MDTM_TRUE;
  }

  return private->opened == MDTM_TRUE;
}

static int CatsVADBuffer_write(CatsVADBuffer this, char *b, int off, int len)
{
  CatsVADBufferPrivate private=this->private;
  CatsDelayLine buffer=private->buffer;
  int frameSamples=private->frameSamples;
  int adaptivePeriodFrames=private->adaptivePeriodFrames;

  /* Local copies of private fields we'll change */

  double levelThreshold=private->levelThreshold;
  int voiceActive=private->voiceActive;
  int consecutiveChangeFrames=private->consecutiveChangeFrames;
  double signalMin=private->signalMin;
  double silenceMax=private->silenceMax;
  int signalFramesReceived=private->signalFramesReceived;
  int silenceFramesReceived=private->silenceFramesReceived;

  /* Local variables */

  mdtm_srint32 level;
  int haveSignal;

  /* Will only allow single frame write */

  if (!private->opened || len != private->frameSize ||
      buffer->write(buffer,b,off,len))
    return -1;

  mdtm_vrchar_equ((mdtm_vrchar *)private->frameBuffer,b,len);
  if (private->byteSwapFlag)
    mdtm_byteswap(private->frameBuffer,private->sampleSize,frameSamples);

  /* Perform voice activity classification */

  level = CatsVADBuffer_absstd(private->frameBuffer,private->frameSamples);

  if (levelThreshold == -1.)
    levelThreshold = 2*level;

  haveSignal = level > levelThreshold;

  if (voiceActive == haveSignal)
    consecutiveChangeFrames = 0;
  else
    consecutiveChangeFrames++;

  if ((voiceActive &&
       consecutiveChangeFrames >= private->silenceDeadBandFrames) ||
      (!voiceActive &&
       consecutiveChangeFrames >= private->signalDeadBandFrames)) {
    voiceActive = !voiceActive;
    printf(voiceActive ? "voice active\n":"voice inactive\n");
    signalMin = mdtm_srdouble_inf();
    silenceMax = 0;
    signalFramesReceived = 0;
    silenceFramesReceived = 0;
  }

  if (haveSignal) {
    signalMin = MDTM_MIN(signalMin,level);
    signalFramesReceived++;
  } else {
    silenceMax = MDTM_MAX(silenceMax,level);
    silenceFramesReceived++;
  }

  /* Perform adaptation if the frame period is reached */

  if ((signalFramesReceived + silenceFramesReceived)
      == adaptivePeriodFrames) {

    if (signalFramesReceived == adaptivePeriodFrames)
      levelThreshold += private->tRaiseFactor*(signalMin-levelThreshold);
    else if (silenceFramesReceived == adaptivePeriodFrames)
      levelThreshold -= private->tLowerFactor*(levelThreshold-silenceMax);
    else if (signalFramesReceived > silenceFramesReceived)
      levelThreshold += .5*private->tRaiseFactor*(signalMin-levelThreshold);

    signalMin = mdtm_srdouble_inf();
    silenceMax = 0;
    signalFramesReceived = silenceFramesReceived = 0;
  }

  /* Store back new values in private data */

  private->levelThreshold = levelThreshold;
  private->voiceActive = voiceActive;
  private->consecutiveChangeFrames=consecutiveChangeFrames;
  private->signalMin=signalMin;
  private->silenceMax=silenceMax;
  private->signalFramesReceived=signalFramesReceived;
  private->silenceFramesReceived=silenceFramesReceived;

  return voiceActive;
}

static int CatsVADBuffer_read(CatsVADBuffer this, char *b, int off, int len)
{
  CatsVADBufferPrivate private=this->private;
  CatsDelayLine buffer=private->buffer;

  /* Will only allow single frame write */

  if (!private->opened || len != private->frameSize)
    return -1;

  buffer->read(buffer,b,off,len);

  return private->voiceActive;
}

static int CatsVADBuffer_close(CatsVADBuffer this)
{
  CatsVADBufferPrivate private=this->private;

  if (private->opened) {
    private->buffer = private->buffer->destroy(private->buffer);
    private->frameBuffer = mdtm_vrint16_free(private->frameBuffer);
    private->opened = MDTM_FALSE;
  }

  return private->opened == MDTM_FALSE;
}

static int CatsVADBuffer_isOpen(CatsVADBuffer this)
{
  return this->private->opened;
}

/* Private methods */

static mdtm_srint32 CatsVADBuffer_absstd(mdtm_vrint16 *frameBuffer,
					 int frameSamples)
{
  mdtm_srint32 absstd=0;
  mdtm_srint32 mean=0;
  int i;

  for (i=0; i<frameSamples; i++)
    mean += frameBuffer[i];
  mean /= frameSamples;

  mdtm_vrint16_subs(frameBuffer,frameSamples,mean);
  mdtm_vrint16_abs(frameBuffer,frameSamples);

  for (i=0; i<frameSamples; i++)
    absstd += frameBuffer[i];
  absstd /= frameSamples;

  return absstd;
}
