/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/*
  This file (c) Copyright 2002 The MITRE Corporation
  
  This file is part of the Communicator Audio Telephony Server
  (CATS). It is licensed under the conditions described in the file
  COPYING in the root directory of the CATS distribution.
*/

#include <stdlib.h>
#include <mdtm/mdtm.h>
#include <mdtm/mdtmmacros.h>
#include "cats_vector.h"
#include "cats_lock_object.h"

static CatsVector CatsVector_destroy(CatsVector this);
static int CatsVector_add(CatsVector this, void *data);
static int CatsVector_contains(CatsVector this, void *elem);
static void *CatsVector_elementAt(CatsVector this, int index);
static void *CatsVector_firstElement(CatsVector this);
static void *CatsVector_remove(CatsVector this, int index);
static int CatsVector_removeElement(CatsVector this, void *obj);
static void *CatsVector_removeLast(CatsVector this);
static int CatsVector_size(CatsVector this);
static void CatsVector_lock(CatsVector this);
static void CatsVector_unlock(CatsVector this);

/* PRIVATE METHODS */

static int CatsVector_ensureCapacity(CatsVector this, int minCapacity);

struct CatsVectorPrivate_str {
  void **elementData;
  int capacity;
  int elementCount;
  int capacityIncrement;
  CatsLockObject lockObject;
};

CatsVector CatsVector_create(int initialCapacity, int capacityIncrement)
{
  CatsVector this=NULL;

  if (initialCapacity < 0)
    return NULL;

  if ((this =
       (CatsVector)
       calloc(1,sizeof(struct CatsVector_str))) != NULL) {
    CatsVectorPrivate private;

    this->destroy = CatsVector_destroy;
    this->add = CatsVector_add;
    this->contains = CatsVector_contains;
    this->elementAt = CatsVector_elementAt;
    this->firstElement = CatsVector_firstElement;
    this->remove = CatsVector_remove;
    this->removeElement = CatsVector_removeElement;
    this->removeLast = CatsVector_removeLast;
    this->size = CatsVector_size;
    this->lock = CatsVector_lock;
    this->unlock = CatsVector_unlock;

    if ((private = this->private =
         (CatsVectorPrivate)
	 calloc(1,sizeof(struct CatsVectorPrivate_str)))
        == NULL)
      return(this->destroy(this));

    private->capacity = initialCapacity == 0 ? 10 : initialCapacity;
    if ((private->elementData = calloc(private->capacity, sizeof(void *)))
	== NULL)
      return(this->destroy(this));

    private->elementCount = 0;
    private->capacityIncrement = capacityIncrement;

    if ((private->lockObject = CatsLockObject_create()) == NULL)
      return(this->destroy(this));
  }

  return this;
}

static CatsVector CatsVector_destroy(CatsVector this)
{
  if (this->private != NULL) {
    CatsLockObject lockObject=this->private->lockObject;
    if (lockObject != NULL) {
      CatsVector_lock(this);
      lockObject->destroy(lockObject);
    }
    if (this->private->elementData)
      free(this->private->elementData);
    free(this->private);
  }
  free(this);
  return NULL;
}

static int CatsVector_add(CatsVector this, void *data)
{
  int retval=MDTM_TRUE;

  CatsVector_lock(this);
  if (CatsVector_ensureCapacity(this,this->private->elementCount + 1))
    this->private->elementData[this->private->elementCount++] = data;
  else
    retval = MDTM_FALSE;
  CatsVector_unlock(this);
  return retval;
}

static int CatsVector_contains(CatsVector this, void *elem)
{
  int size;
  int i;

  CatsVector_lock(this);
  size = this->size(this);

  for (i=0; i<size; i++) {
    if (elem == this->elementAt(this,i)) {
      CatsVector_unlock(this);
      return MDTM_TRUE;
    }
  }

  CatsVector_unlock(this);

  return MDTM_FALSE;
}

static void *CatsVector_elementAt(CatsVector this, int index)
{
  void *element;
  CatsVector_lock(this);
  element = index >= this->private->elementCount ? NULL :
    this->private->elementData[index];
  CatsVector_unlock(this);
  return element;
}

static void *CatsVector_firstElement(CatsVector this)
{
  void *firstElement;

  CatsVector_lock(this);
  firstElement = this->private->elementCount == 0 ? NULL :
    this->private->elementData[0];
  CatsVector_unlock(this);
  return firstElement;
}

static void *CatsVector_remove(CatsVector this, int index)
{
  void *oldValue=NULL;

  CatsVector_lock(this);

  if (index >= 0 && index < this->private->elementCount) {
    int numMoved;

    oldValue = this->private->elementData[index];
    numMoved = --this->private->elementCount-index;

    if (numMoved > 0) {
      char *cptr = (mdtm_vrchar *)this->private->elementData
	+ index*sizeof(void *);
      mdtm_vrchar_equ(cptr,cptr+sizeof(void *),numMoved*sizeof(void *));
    }
  }

  CatsVector_unlock(this);

  return oldValue;
}

static int CatsVector_removeElement(CatsVector this, void *obj)
{
  int index;
  int size=this->size(this);
  int i;

  CatsVector_lock(this);
  index = -1;
  size = this->size(this);

  for (i=0; i<size; i++) {
    if (obj == this->elementAt(this,i)) {
      index = i;
      break;
    }
  }

  if (index != -1)
    this->remove(this,index);

  CatsVector_unlock(this);

  return index != -1;
}

static void *CatsVector_removeLast(CatsVector this)
{
  return this->remove(this,this->size(this)-1);
}

static int CatsVector_size(CatsVector this)
{
  int retval;
  CatsVector_lock(this);
  retval = this->private->elementCount;
  CatsVector_unlock(this);
  return retval;
}

/* PRIVATE METHODS */

static int CatsVector_ensureCapacity(CatsVector this, int minCapacity)
{
  int oldCapacity = this->private->capacity;
  if (minCapacity > oldCapacity) {
	int idx;
	void **tempElementData;
    /*void **elementData;*/
    int newCapacity = (this->private->capacityIncrement > 0) ?
      (oldCapacity + this->private->capacityIncrement) : (oldCapacity << 1);
    if (newCapacity < minCapacity)
      newCapacity = minCapacity;

	/* We do not fully trust using mdtm_vrchar_realloc here, so do reallocation by hand. */
	tempElementData = (void **) malloc(newCapacity*sizeof(void*));
	if (tempElementData == NULL) {
      return MDTM_FALSE;
	}

	for(idx=0; idx<oldCapacity; ++idx) {
		tempElementData[idx] = (this->private->elementData)[idx];
	}
	free(this->private->elementData);
	this->private->elementData = tempElementData;

	/*
    elementData =
      (void **) mdtm_vrchar_realloc((mdtm_vrchar *)this->private->elementData,
				    oldCapacity*sizeof(void *),
				    newCapacity*sizeof(void *));
    if (elementData == NULL)
      return MDTM_FALSE;
    this->private->elementData = elementData;
	*/

    this->private->capacity = newCapacity;
  }
  return MDTM_TRUE;
}

static void CatsVector_lock(CatsVector this)
{
  CatsLockObject lockObject=this->private->lockObject;
  lockObject->lock(lockObject);
}

static void CatsVector_unlock(CatsVector this)
{
  CatsLockObject lockObject=this->private->lockObject;
  lockObject->unlock(lockObject);
}
