/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

#include <fcntl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <mdtm/mdtm.h>
#include <mdtm/mdtmmacros.h>
#include <errno.h>
#include <galaxy/galaxy.h>
#include "cf_phone_line.h"
#include "cats_phone_line.h"
#include "cats_ring_buffer.h"
#include "cats_lock_object.h"
#include "cats_thread.h"

#define DEBUG 1

/* Operating constants */

#define CF_MAX_DATA_LENGTH 1024
#define CF_POLL_SLOW_MSEC 100
#define CF_POLL_FAST_MSEC 10

/* Computerfone command strings */

#define CF_FULL_RESET_CMD ">XX\r"
#define CF_AUTOMATIC_ANSWER_DISABLE_CMD ">A0\r"
#define CF_OFF_HOOK_CMD ">O\r"
#define CF_ON_HOOK_CMD ">H\r"
#define CF_RESET_DTMF_CMD ">XI\r"
#define CF_ENABLE_DTMF_PASS_THROUGH_CMD ">IP\r"

/* Computerfone acknowledgement strings */

#define CF_FULL_RESET_ACK_STR "CFONE 4.62\r\n"
#define CF_AUTOMATIC_ANSWER_DISABLE_ACK_STR "<A\r\n"
#define CF_INCOMING_CALL_ACK_STR "<C\r\n"
#define CF_OFF_HOOK_ACK_STR "<O\r\n"
#define CF_ON_HOOK_ACK_STR "<H\r\n"
#define CF_RESET_DTMF_ACK_STR "<X\r\n"
#define CF_ENABLE_DTMF_PASS_THROUGH_ACK_STR "<I\r\n"

enum computerfoneAcks {CF_FULL_RESET_ACK, CF_AUTOMATIC_ANSWER_DISABLE_ACK,
		       CF_INCOMING_CALL_ACK, CF_OFF_HOOK_ACK,
		       CF_ON_HOOK_ACK, CF_RESET_DTMF_ACK,
		       CF_ENABLE_DTMF_PASS_THROUGH_ACK,
		       CF_UNKNOWN_ACK};

static int CfPhoneLine_strToAck(char *str)
{
  if (!strcmp(str, CF_FULL_RESET_ACK_STR))
    return CF_FULL_RESET_ACK;
  else if (!strcmp(str, CF_AUTOMATIC_ANSWER_DISABLE_ACK_STR))
    return CF_AUTOMATIC_ANSWER_DISABLE_ACK;
  else if (!strcmp(str, CF_INCOMING_CALL_ACK_STR))
    return CF_INCOMING_CALL_ACK;
  else if (!strcmp(str, CF_OFF_HOOK_ACK_STR))
    return CF_OFF_HOOK_ACK;
  else if (!strcmp(str, CF_ON_HOOK_ACK_STR))
    return CF_ON_HOOK_ACK;
  else if (!strcmp(str, CF_RESET_DTMF_ACK_STR))
    return CF_RESET_DTMF_ACK;
  else if (!strcmp(str, CF_ENABLE_DTMF_PASS_THROUGH_ACK_STR))
    return CF_ENABLE_DTMF_PASS_THROUGH_ACK;
  else
    return CF_UNKNOWN_ACK;
}

static char *CfPhoneLine_ackToStr(int ack)
{
  switch (ack) {
  case CF_FULL_RESET_ACK:
    return "full reset acknowledgment";
  case CF_AUTOMATIC_ANSWER_DISABLE_ACK:
    return "automatic answer disable acknowledgment";
  case CF_INCOMING_CALL_ACK:
    return "incoming call acknowledgment";
  case CF_OFF_HOOK_ACK:
    return "off hook cod";
  case CF_ON_HOOK_ACK:
    return "on hook acknowledgment";
  case CF_RESET_DTMF_ACK:
    return "reset DTMF acknowledgment";
  case CF_ENABLE_DTMF_PASS_THROUGH_ACK:
    return "enable DTMF pass-through acknowledgment";
  default:
    return "unknown acknowledgment";
  }
}

enum computerfoneCodes {CF_SUCCESS,CF_PENDING,CF_FAILURE,CF_CMD_NOT_SENT,
			CF_CMD_ACK_NOT_RECEIVED,
			CF_UNEXPECTED_HANG_UP_RECEIVED,
			CF_UNEXPECTED_INCOMING_CALL_RECEIVED,
			CF_CLOSED,CF_ERROR,
			CF_INCOMING_CALL,CF_DTMF,CF_ACK,CF_ANSWERED_CALL};

static CfPhoneLine CfPhoneLine_destroy(CfPhoneLine this);
static CatsPhoneLine CfPhoneLine_getPhoneLine(CfPhoneLine this);

/* CatsPhoneLine interface implementation */

static int CfPhoneLine_open(CatsPhoneLine line);
static int CfPhoneLine_poll(CatsPhoneLine line);
static int CfPhoneLine_getDTMF(CatsPhoneLine line);
static int CfPhoneLine_getPollMsec(CatsPhoneLine line);
static int CfPhoneLine_disconnect(CatsPhoneLine line);
static int CfPhoneLine_close(CatsPhoneLine line);
static int CfPhoneLine_isOpen(CatsPhoneLine line);
static int CfPhoneLine_isConnected(CatsPhoneLine line);

/* PRIVATE METHODS */

static int CfPhoneLine_sendCommand(CfPhoneLine this, char *cmd, char *ack);
static int CfPhoneLine_write(CfPhoneLine this, char *cmd);
static int CfPhoneLine_readData(CfPhoneLine this);
static int CfPhoneLine_enableRawMode(CfPhoneLine this);
static int CfPhoneLine_disableRawMode(CfPhoneLine this);
static int CfPhoneLine_closeOnError(CfPhoneLine line);
static int CfPhoneLine_closeLine(CfPhoneLine line, int on_error);

struct CfPhoneLinePrivate_str {

  /* COMMON FIELDS */

  int opened;
  char *portName;

  /* Serial port file descriptor */

  int fd;

  /* The original, "non-raw" port settings. */

  struct termios original_termios; 

  int dtmf;
  int pollMsec;

  /* Status fields */

  int connected;
  int rawModeEnabled;
  char data[CF_MAX_DATA_LENGTH];
  int dataIndex;
  char **fsm;
  int fsmIndex;

  /* Phone line interface */

  CatsPhoneLine line;

  /* Lock object */

  CatsLockObject lockObject;

};

CfPhoneLine CfPhoneLine_create(char *portName)
{
  CfPhoneLine this=NULL;

  if ((this = (CfPhoneLine) calloc(1,sizeof(struct CfPhoneLine_str)))
      != NULL) {
    CfPhoneLinePrivate private;
    CatsPhoneLine line;

    this->destroy = CfPhoneLine_destroy;
    this->getPhoneLine = CfPhoneLine_getPhoneLine;

    if ((private = this->private =
         (CfPhoneLinePrivate) calloc(1,sizeof(struct CfPhoneLinePrivate_str)))
        == NULL)
      return(this->destroy(this));

    if ((line = private->line = CatsPhoneLine_create(this)) == NULL)
      return(this->destroy(this));

    line->open = CfPhoneLine_open;
    line->poll = CfPhoneLine_poll;
    line->getDTMF = CfPhoneLine_getDTMF;
    line->disconnect = CfPhoneLine_disconnect;
    line->close = CfPhoneLine_close;
    line->isOpen = CfPhoneLine_isOpen;
    line->isConnected = CfPhoneLine_isConnected;
    line->getPollMsec = CfPhoneLine_getPollMsec;

    if ((private->lockObject = CatsLockObject_create()) == NULL)
      return(this->destroy(this));

    private->opened = MDTM_FALSE;
    private->portName = portName;
    private->dtmf = -1;
    private->pollMsec = CF_POLL_SLOW_MSEC;
    private->fd = -1;
  }

  return this;
}

static CfPhoneLine CfPhoneLine_destroy(CfPhoneLine this)
{
  if (this->private != NULL) {
    CatsLockObject lockObject=this->private->lockObject;
    CatsPhoneLine line=this->private->line;

    if (lockObject != NULL) {
      lockObject->lock(lockObject);
      if (line && line->isOpen(line))
	line->close(line);
      lockObject->destroy(lockObject);
    }

    if (line != NULL)
      line->destroy(line);

    free(this->private);
  }

  free(this);
  return NULL;
}

static CatsPhoneLine CfPhoneLine_getPhoneLine(CfPhoneLine this)
{
  return this->private->line;
}

/* CatsPhoneLine interface implementation */

/* SAM 10/1/02: This code has "benefited" from multiple hands:
   Sam Bayer, Steve Wohlever, Bryan George. The basic insight
   seems to be that at any point, the ComputerFone can either acknowledge
   your request (say, to reset) or issue an asynchronous
   event notification. There are two sets of event notifications:
   when the phone is off hook, you can get DTMF events or "I got a hangup",
   and when the phone is on hook, you can get "I got a call".
   When you interact with the ComputerFone,
   you have to be conscious of this issue. In the process of
   searching for an acknowledgment, you need to make sure
   that you deal with those event notifications.

   In some cases (e.g., opening the line), you want to
   ensure you get the correct acknowledgment immediately. In
   other cases, you need to deal with the events as they come.

   It appears that we really should write this up as a state
   machine, I think.
*/

enum {CF_DISCONNECTED, CF_IN_DISCONNECT_FSM,
      CF_CONNECTED, CF_IN_CONNECT_FSM};

static int CfPhoneLine_open(CatsPhoneLine line)
{
  CfPhoneLine this=line->getSource(line);
  CfPhoneLinePrivate private=this->private;
  CatsLockObject lockObject=private->lockObject;
  int retval;

  lockObject->lock(lockObject);

  if (!private->opened) {
    private->connected = CF_DISCONNECTED;
    private->rawModeEnabled = MDTM_FALSE;
    (private->data)[0] = MDTM_NUL;
    if ((private->fd = open(private->portName, O_RDWR | O_NONBLOCK)) == -1)
      return MDTM_FALSE;
    /* Mark it temporarily opened, so I can do the
       read/write, and enable raw mode. */
    private->opened = MDTM_TRUE;
    CfPhoneLine_enableRawMode(this);
    if (CfPhoneLine_sendCommand(this,CF_FULL_RESET_CMD,CF_FULL_RESET_ACK_STR)
	== CF_SUCCESS && 
	CfPhoneLine_sendCommand(this,CF_AUTOMATIC_ANSWER_DISABLE_CMD,
				CF_AUTOMATIC_ANSWER_DISABLE_ACK_STR)
	== CF_SUCCESS) {
#if (DEBUG)
      GalUtil_PInfo1("Opened ComputerFone on device %s\n", private->portName);
#endif
    } else {
      /* Oops, failed. Set it back. */
      CfPhoneLine_disableRawMode(this);
      close(private->fd);
      private->fd = -1;
      private->opened = MDTM_FALSE;
    }	    
  }

  retval = private->opened == MDTM_TRUE;

  lockObject->unlock(lockObject);

  return retval;
}


/* These state machines are the machines we work our
   way through when we want to hang up (or, alternatively, when
   we get a hangup, we start partway through), or when we
   want to answer a call. In each case, if we get interrupt-
   driven state changes, we want to shift focus immediately
   to those. */

static char *HangUpFSM[] = {
  CF_ON_HOOK_CMD, CF_ON_HOOK_ACK_STR,
  CF_FULL_RESET_CMD, CF_FULL_RESET_ACK_STR,
  CF_AUTOMATIC_ANSWER_DISABLE_CMD, CF_AUTOMATIC_ANSWER_DISABLE_ACK_STR,
  NULL};

static char *CallAnswerFSM[] = {
  CF_OFF_HOOK_CMD, CF_OFF_HOOK_ACK_STR,
  CF_RESET_DTMF_CMD, CF_RESET_DTMF_ACK_STR,
  CF_ENABLE_DTMF_PASS_THROUGH_CMD, CF_ENABLE_DTMF_PASS_THROUGH_ACK_STR,
  NULL};

static void CfPhoneLine_nextFSMStep(CfPhoneLine line)
{
  CfPhoneLinePrivate private=line->private;
	
  CfPhoneLine_write(line, private->fsm[private->fsmIndex++]);
}

static void CfPhoneLine_initializeFSM(CfPhoneLine this, char **fsm,
				      int initialIndex)
{
  CfPhoneLinePrivate private=this->private;

  private->fsm = fsm;
  private->fsmIndex = initialIndex;
  CfPhoneLine_nextFSMStep(this);
}

/* Returns CF_SUCCESS if the FSM complets, CF_PENDING if it
   matches, CF_FAILURE if it doesn't. */

static int CfPhoneLine_advanceFSMState(CfPhoneLine line)
{
  CfPhoneLinePrivate private=line->private;

  if (!strcmp(private->data, private->fsm[private->fsmIndex])) {
    /* Matches the current state. */
    private->fsmIndex++;
    if (!private->fsm[private->fsmIndex]) {
      /* End of the line. */
      private->fsm = (char **) NULL;
      private->fsmIndex = 0;
      return CF_SUCCESS;
    } else {
      CfPhoneLine_nextFSMStep(line);
      return CF_PENDING;
    }
  } else {
    return CF_FAILURE;
  }
}

static int CfPhoneLine_poll(CatsPhoneLine line)
{
  CfPhoneLine this=line->getSource(line);
  CfPhoneLinePrivate private=this->private;
  int result;
  
  if (!private->opened)
    return CATS_PHONE_FAILURE;

  result = CfPhoneLine_readData(this);

  /* CF_CLOSED: the read failed because the line was already closed.
     CF_ERROR: the read failed because of an error and was closed.
     CF_PENDING: nothing to report.
     CF_DTMF: found a DTMF.
     CF_ACK: found a non-DTMF input. */

  /* The socket will be closed. */
  if ((result == CF_CLOSED) || (result == CF_ERROR)) {
    return CATS_PHONE_FAILURE;
  }

  switch (private->connected) {
  case CF_DISCONNECTED:
    /* Well, you can't get a DTMF, and the only command
       you can get that makes any sense at all is a call answer. */
    switch (result) {
#if (DEBUG)
    case CF_DTMF:
      GalUtil_Warn("Discarding DTMF found in disconnect: %s\n",
		   private->data);
      break;
#endif
    case CF_ACK:
      switch (CfPhoneLine_strToAck(private->data)) {
      case CF_INCOMING_CALL_ACK:
	CfPhoneLine_initializeFSM(this, CallAnswerFSM, 0);
	private->connected = CF_IN_CONNECT_FSM;
	break;
      default:
#if (DEBUG)
	GalUtil_Warn("Ignoring unexpected %s (%s)\n",
		     CfPhoneLine_ackToStr(CfPhoneLine_strToAck(private->data)),
		     private->data);
#endif
	break;
      }
    }
    break;
  case CF_IN_DISCONNECT_FSM:
    switch (result) {
#if (DEBUG)
    case CF_DTMF:
      GalUtil_Warn("Discarding DTMF found in pending disconnect: %s\n",
		   private->data);
      break;
#endif
    case CF_ACK:
      switch (CfPhoneLine_advanceFSMState(this)) {
      case CF_SUCCESS:
	/* FSM has been cleared. */
	private->connected = CF_DISCONNECTED;
	private->pollMsec = CF_POLL_SLOW_MSEC;
	return CATS_PHONE_HANGUP_RECEIVED;
      case CF_PENDING:
	break;
      case CF_FAILURE:
	/* If we're not on the path to the disconnect, figure out
	   what to do. Basically, the only thing we want to pay
	   attention to here is an incoming call. */
	switch (CfPhoneLine_strToAck(private->data)) {
	case CF_INCOMING_CALL_ACK:
#if (DEBUG)
	  GalUtil_Warn("Incoming call interrupted the disconnect");
#endif
	  CfPhoneLine_initializeFSM(this, CallAnswerFSM, 0);
	  private->connected = CF_IN_CONNECT_FSM;
	  break;
	default:
#if (DEBUG)
	  GalUtil_Warn("Ignoring unexpected %s (%s)\n",
		       CfPhoneLine_ackToStr(CfPhoneLine_strToAck(private->data)),
		       private->data);
#endif  
	  break;
	}
	break;
      }
      break;
    }
    break;
  case CF_IN_CONNECT_FSM:
    switch (result) {
#if (DEBUG)
    case CF_DTMF:
      GalUtil_Warn("Discarding DTMF found in pending connect: %s\n",
		   private->data);
      break;
#endif
    case CF_ACK:
      switch (CfPhoneLine_advanceFSMState(this)) {
      case CF_SUCCESS:
	/* FSM has been cleared. */
	private->connected = CF_CONNECTED;
	private->pollMsec = CF_POLL_FAST_MSEC;
	return CATS_PHONE_ANSWERED_CALL;
      case CF_PENDING:
	break;
      case CF_FAILURE:
	/* If we're not on the path to the connect, figure out
	   what to do. Basically, the only thing we want to pay
	   attention to here is a hangup. Remember, you want to
	   advance the FSM to the second pair. 
	*/
	switch (CfPhoneLine_strToAck(private->data)) {
	case CF_ON_HOOK_ACK:
#if (DEBUG)
	  GalUtil_Warn("Hangup interrupted the disconnect");
#endif
	  CfPhoneLine_initializeFSM(this, HangUpFSM, 2);
	  private->connected = CF_IN_DISCONNECT_FSM;
	  break;
	default:
#if (DEBUG)
	  GalUtil_Warn("Ignoring unexpected %s (%s)\n",
		       CfPhoneLine_ackToStr(CfPhoneLine_strToAck(private->data)),
		       private->data);
#endif  
	  break;
	}
	break;
      }
      break;
    }
    break;
  case CF_CONNECTED:
    switch (result) {
    case CF_DTMF:
      switch(private->data[0]) {
      case '1':
	private->dtmf = CATS_TOUCHTONE_1;
	break;
      case '2':
	private->dtmf = CATS_TOUCHTONE_2;
	break;
      case '3':
	private->dtmf = CATS_TOUCHTONE_3;
	break;
      case '4':
	private->dtmf = CATS_TOUCHTONE_4;
	break;
      case '5':
	private->dtmf = CATS_TOUCHTONE_5;
	break;
      case '6':
	private->dtmf = CATS_TOUCHTONE_6;
	break;
      case '7':
	private->dtmf = CATS_TOUCHTONE_7;
	break;
      case '8':
	private->dtmf = CATS_TOUCHTONE_8;
	break;
      case '9':
	private->dtmf = CATS_TOUCHTONE_9;
	break;
      case '*':
	private->dtmf = CATS_TOUCHTONE_STAR;
	break;
      case '0':
	private->dtmf = CATS_TOUCHTONE_0;
	break;
      case '#':
	private->dtmf = CATS_TOUCHTONE_POUND;
	break;
      }
      return CATS_PHONE_DTMF;
    case CF_ACK:
      /* The only thing that you can care about here is
	 getting a hangup. */
      switch (CfPhoneLine_strToAck(private->data)) {
      case CF_ON_HOOK_ACK:
#if (DEBUG)
	GalUtil_Warn("Hangup received");
#endif
	CfPhoneLine_initializeFSM(this, HangUpFSM, 2);
	private->connected = CF_IN_DISCONNECT_FSM;
	break;
      default:
#if (DEBUG)
	GalUtil_Warn("Ignoring unexpected %s (%s)\n",
		     CfPhoneLine_ackToStr(CfPhoneLine_strToAck(private->data)),
		     private->data);
#endif
	break;
      }
      break;
    }
    break;
  }
  return CATS_PHONE_NO_INPUT;
}

static int CfPhoneLine_getDTMF(CatsPhoneLine line)
{
  CfPhoneLine this=line->getSource(line);
  return this->private->dtmf;
}

static int CfPhoneLine_getPollMsec(CatsPhoneLine line)
{
  CfPhoneLine this=line->getSource(line);
  return this->private->pollMsec;
}

static int CfPhoneLine_disconnectLine(CfPhoneLine this)
{
  CfPhoneLinePrivate private=this->private;
  CatsLockObject lockObject=private->lockObject;
  int retval;

  lockObject->lock(lockObject);

  if (private->connected != CF_DISCONNECTED) {
    /* Force it on hook, ignore the acknowledgment. */
    CfPhoneLine_write(this, CF_ON_HOOK_CMD);
    private->connected = CF_DISCONNECTED;
  }

  retval = private->opened == MDTM_FALSE;

  lockObject->unlock(lockObject);

  return retval;
}

static int CfPhoneLine_disconnect(CatsPhoneLine line)
{
  return CfPhoneLine_disconnectLine(line->getSource(line));
}

static int CfPhoneLine_closeOnError(CfPhoneLine line)
{
  return CfPhoneLine_closeLine(line, 1);
}

static int CfPhoneLine_close(CatsPhoneLine line)
{
  return CfPhoneLine_closeLine(line->getSource(line), 0);
}

static int CfPhoneLine_closeLine(CfPhoneLine this, int on_error)
{
  CfPhoneLinePrivate private=this->private;
  CatsLockObject lockObject=private->lockObject;
  int retval;

  lockObject->lock(lockObject);

  if (private->opened) {
    if (on_error) {
      private->connected = CF_DISCONNECTED;
    } else {
      if (private->connected != CF_DISCONNECTED)
	CfPhoneLine_disconnectLine(this);
      CfPhoneLine_disableRawMode(this);
    } 
    close(private->fd);
    private->fd = -1;
    private->opened = MDTM_FALSE;
  }

  retval = private->opened == MDTM_FALSE;

  lockObject->unlock(lockObject);

  return retval;
}

static int CfPhoneLine_isOpen(CatsPhoneLine line)
{
  CfPhoneLine this=line->getSource(line);
  return this->private->opened;
}

static int CfPhoneLine_isConnected(CatsPhoneLine line)
{
  CfPhoneLine this=line->getSource(line);
  return (this->private->connected == CF_CONNECTED);
}

/* PRIVATE METHODS */

static int CfPhoneLine_sendCommand(CfPhoneLine this, char *cmd, char *ack)
{
  if(CfPhoneLine_write(this,cmd) == CF_CMD_NOT_SENT)
    return CF_CMD_NOT_SENT;
  else {
    while (1) {
      switch(CfPhoneLine_readData(this)) {
      case CF_CLOSED:
      case CF_ERROR:
	return CF_FAILURE;
      case CF_PENDING:
	break;
      case CF_DTMF:
	return CF_FAILURE;
      case CF_ACK:
	if(strcmp(this->private->data, ack) == 0) {
	  return CF_SUCCESS;
	} else {
	  return CF_FAILURE;
	}
      default:
	return CF_FAILURE;
      }
    }
  }
}

static int CfPhoneLine_write(CfPhoneLine this, char *cmd)
{
  CfPhoneLinePrivate private=this->private;
  int len = strlen(cmd);
  int bytes_written;

  if(!private->opened)
    return CF_CLOSED;

  /* Write the command. */

  bytes_written = write(private->fd, cmd, len);

  return bytes_written == len ? CF_SUCCESS : CF_CMD_NOT_SENT;
}

/* SAM 10/1/02: Non-blocking read. We can return one of several
   values from this function:
   CF_CLOSED: the read failed because the line was already closed.
   CF_ERROR: the read failed because of an error and was closed.
   CF_PENDING: nothing to report.
   CF_DTMF: found a DTMF.
   CF_ACK: found a non-DTMF input.
*/

static int CfPhoneLine_readData(CfPhoneLine this)
{
  CfPhoneLinePrivate private=this->private;
  char *data=private->data;
  int num_bytes_read;
  char byte;

  if (!private->opened)
    return CF_CLOSED;

  /* Read one byte off the port. */
  num_bytes_read = read(private->fd, (void*) &byte, 1);
    
  /* If there is nothing to read, sleep for a little while. */
  if (num_bytes_read == -1) {
    if (errno == EAGAIN || errno == EINTR) {
      return CF_PENDING;
    } else {
      /* Oops, we'd better close and barf. */
      CfPhoneLine_closeOnError(this);
      return CF_ERROR;
    }
  } else if (!num_bytes_read) {
    /* If we read nothing, return. */
    return CF_PENDING;
  }

  /* If we're just starting a line, then see if it's DTMF. */
  if ((private->dataIndex == 0) && strchr("0123456789*\r", byte)) {
    /* Computerfone returns # key presses as \r (carriage return). */
    if(byte == '\r')
      data[0] = '#';
    else
      data[0] = byte;
    data[1] = '\0';
    /* Reset it to 0, for the next read. */
    private->dataIndex = 0;
#if (DEBUG)
    GalUtil_PInfo1("Received DTMF %s\n", private->data);
#endif
    return CF_DTMF;
  } else {
    /* Otherwise, we keep reading until we find a line. */

    /* 
       If we are not reading a DTMF event, this is an ack to a
       previous command,
       a user hang up, or an incoming call.
    */

    data[private->dataIndex++] = byte;
#if (DEBUG)
    if(byte >= 32 && byte <=126)
      GalUtil_PInfo1("Read %c (character %d) into position %d\n", byte, byte,
		     private->dataIndex - 1); 
    else
      GalUtil_PInfo1("Read character %d into position %d\n", byte,
		     private->dataIndex - 1);
#endif

    /* Steve Wohlever says: Non-DTMF data sent by Computerfone seems to
       end with two "end
       of line" characters (carriage return and new line), though this
       is not documented in the manual as far as I can see.  */
    
    if ((private->dataIndex > 1) &&
	strchr("\n\r", data[private->dataIndex - 1]) &&
	strchr("\n\r", data[private->dataIndex - 2])) {
      data[private->dataIndex++] = '\0';
#if (DEBUG)
      GalUtil_PInfo1("Received acknowledgment %s\n", private->data);
#endif
      /* Next one starts over. */
      private->dataIndex = 0;
      return CF_ACK;
    } else {
      return CF_PENDING;
    }
  }
}

static int CfPhoneLine_enableRawMode(CfPhoneLine this)
{
  CfPhoneLinePrivate private=this->private;
  int fd=private->fd;
  struct termios new_termios;

  if(!private->opened)
    return CF_CLOSED;

  /* Get the current settings. */
  ioctl(fd, TCGETS, &new_termios);

  /* Save a copy of the current settings for later use. */
  ioctl(fd, TCGETS, &(private->original_termios));

  /* Minimum number of characters to read before read is considered 
     satisfied. Default value is not necessarily correct (e.g., I found it
     was set to 4, requiring 4 DTMF events to be read before they would be
     reported by the read.
  */
  new_termios.c_cc[VMIN] = 1; 
  
  /* Modify the settings. See man page for termio(7I). */
  new_termios.c_iflag &= IXOFF; /* Needed to report \r as \r and not \n. */
  new_termios.c_lflag &= ~ICANON; /* Needed to get DTMF events. */
  ioctl(fd, TCSETS, &new_termios);  
  private->rawModeEnabled = MDTM_TRUE;

  return CF_SUCCESS;
}

static int CfPhoneLine_disableRawMode(CfPhoneLine this)
{
  CfPhoneLinePrivate private=this->private;

  if (!private->opened)
    return CF_CLOSED;

  ioctl(private->fd, TCSETS, &(private->original_termios));
  private->rawModeEnabled = MDTM_FALSE;

  return CF_SUCCESS;
}
