/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <galaxy/util.h>
#include <mdtm/mdtm.h>
#include <mdtm/mdtmmacros.h>
#include <portaudio.h>
#include "pa_mixer.h"
#include "cats_ring_buffer.h"
#include "cats_lock_object.h"
#include "cats_thread.h"

#define DEBUG 1
#define PA_FRAMES_PER_BUFFER (256)

static PaMixer PaMixer_destroy(PaMixer this);
static CatsTargetDataLine PaMixer_getTargetDataLine(PaMixer this);
static CatsSourceDataLine PaMixer_getSourceDataLine(PaMixer this);

/* CatsTargetDataLine interface implementation */

static int PaMixerTargetDataLine_open(CatsTargetDataLine line,
				      CatsAudioFormat format,
				      int bufferSize);
static int PaMixerTargetDataLine_start(CatsTargetDataLine line);
static int PaMixerTargetDataLine_read(CatsTargetDataLine line, char *b,
				      int off, int len);
static int PaMixerTargetDataLine_stop(CatsTargetDataLine line);
static int PaMixerTargetDataLine_close(CatsTargetDataLine line);
static void PaMixerTargetDataLine_flush(CatsTargetDataLine line);
static int PaMixerTargetDataLine_available(CatsTargetDataLine line);
static int PaMixerTargetDataLine_isOpen(CatsTargetDataLine line);
static int PaMixerTargetDataLine_isRunning(CatsTargetDataLine line);
static int PaMixerTargetDataLine_isActive(CatsTargetDataLine line);
static int PaMixerTargetDataLine_getBufferSize(CatsTargetDataLine line);
static CatsAudioFormat
PaMixerTargetDataLine_getFormat(CatsTargetDataLine line);
static int PaMixerTargetDataLine_getFramePosition(CatsTargetDataLine line);

/* CatsSourceDataLine interface implementation */

static int PaMixerSourceDataLine_open(CatsSourceDataLine line,
				      CatsAudioFormat format,
				      int bufferSize);
static int PaMixerSourceDataLine_start(CatsSourceDataLine line);
static int PaMixerSourceDataLine_write(CatsSourceDataLine line, char *b,
				       int off, int len);
static int PaMixerSourceDataLine_stop(CatsSourceDataLine line);
static int PaMixerSourceDataLine_close(CatsSourceDataLine line);
static void PaMixerSourceDataLine_drain(CatsSourceDataLine line);
static void PaMixerSourceDataLine_flush(CatsSourceDataLine line);
static int PaMixerSourceDataLine_available(CatsSourceDataLine line);
static int PaMixerSourceDataLine_isOpen(CatsSourceDataLine line);
static int PaMixerSourceDataLine_isRunning(CatsSourceDataLine line);
static int PaMixerSourceDataLine_isActive(CatsSourceDataLine line);
static int PaMixerSourceDataLine_getBufferSize(CatsSourceDataLine line);
static CatsAudioFormat
PaMixerSourceDataLine_getFormat(CatsSourceDataLine line);
static int PaMixerSourceDataLine_getFramePosition(CatsSourceDataLine line);

/* PRIVATE METHODS */

static void PaMixer_open(PaMixer this, CatsAudioFormat format, int bufferSize,
			 int tdlOpen, int sdlOpen);
static void PaMixer_start(PaMixer this, int tdlStart, int sdlStart);
static void PaMixer_stop(PaMixer this, int tdlStop, int sdlStop);
static void PaMixer_close(PaMixer this, int tdlClose, int sdlClose);

/* PortAudio callback */

static int PaMixer_callback(void *inputBuffer, void *outputBuffer,
			    unsigned long framesPerBuffer,
			    PaTimestamp outTime, void *userData);

struct PaMixerPrivate_str {

  /* COMMON FIELDS */

  int debugEnabled;
  int halfDuplexEnabled;
  int opened;
  int started;

  /* Audio related fields. */

  CatsAudioFormat format;
  float frameRate;
  int frameSize;
  int bufferSize;
  int byteSwap;
  float microsecondPositionFactor;
  PortAudioStream *paStream;
  PaTimestamp lastOutTime;

  /* Lock object */

  CatsLockObject lockObject;

  /* TARGET DATA LINE FIELDS */

  /* CatsTargetDataLine interface object */

  CatsTargetDataLine tdl;

  /* Mode, state flags. */

  int tdlOpened;
  int tdlStarted;
  int tdlActivated;
  int tdlFramePosition;
  int tdlMicrosecondPosition;

  CatsRingBuffer tdlBuffer;

  /* SOURCE DATA LINE FIELDS */

  /* CatsSourceDataLine interface object */

  CatsSourceDataLine sdl;

  /* Mode, state flags. */

  int sdlOpened;
  int sdlStarted;
  int sdlActivated;
  int sdlFramePosition;
  int sdlMicrosecondPosition;

  CatsRingBuffer sdlBuffer;

};

PaMixer PaMixer_create(int halfDuplexEnabled)
{
  PaMixer this=NULL;

  if ((this =
       (PaMixer) calloc(1,sizeof(struct PaMixer_str)))
      != NULL) {
    PaMixerPrivate private;
    CatsTargetDataLine tdl;
    CatsSourceDataLine sdl;
    PaError paError;

    this->destroy = PaMixer_destroy;
    this->getTargetDataLine = PaMixer_getTargetDataLine;
    this->getSourceDataLine = PaMixer_getSourceDataLine;

    if ((private = this->private =
         (PaMixerPrivate)
	 calloc(1,sizeof(struct PaMixerPrivate_str)))
        == NULL)
      return(this->destroy(this));

    if ((tdl = private->tdl =
	 CatsTargetDataLine_create(this)) == NULL)
      return(this->destroy(this));

    tdl->open = PaMixerTargetDataLine_open;
    tdl->start = PaMixerTargetDataLine_start;
    tdl->read = PaMixerTargetDataLine_read;
    tdl->stop = PaMixerTargetDataLine_stop;
    tdl->close = PaMixerTargetDataLine_close;
    tdl->flush = PaMixerTargetDataLine_flush;
    tdl->available = PaMixerTargetDataLine_available;
    tdl->isOpen = PaMixerTargetDataLine_isOpen;
    tdl->isRunning = PaMixerTargetDataLine_isRunning;
    tdl->isActive = PaMixerTargetDataLine_isActive;
    tdl->getBufferSize = PaMixerTargetDataLine_getBufferSize;
    tdl->getFormat = PaMixerTargetDataLine_getFormat;
    tdl->getFramePosition = PaMixerTargetDataLine_getFramePosition;

    if ((sdl = private->sdl =
	 CatsSourceDataLine_create(this)) == NULL)
      return(this->destroy(this));

    sdl->open = PaMixerSourceDataLine_open;
    sdl->start = PaMixerSourceDataLine_start;
    sdl->write = PaMixerSourceDataLine_write;
    sdl->stop = PaMixerSourceDataLine_stop;
    sdl->close = PaMixerSourceDataLine_close;
    sdl->drain = PaMixerSourceDataLine_drain;
    sdl->flush = PaMixerSourceDataLine_flush;
    sdl->available = PaMixerSourceDataLine_available;
    sdl->isOpen = PaMixerSourceDataLine_isOpen;
    sdl->isRunning = PaMixerSourceDataLine_isRunning;
    sdl->isActive = PaMixerSourceDataLine_isActive;
    sdl->getBufferSize = PaMixerSourceDataLine_getBufferSize;
    sdl->getFormat = PaMixerSourceDataLine_getFormat;
    sdl->getFramePosition = PaMixerSourceDataLine_getFramePosition;

    private->halfDuplexEnabled = halfDuplexEnabled;
    private->debugEnabled = private->opened = private->opened = 
      private->tdlOpened = private->tdlStarted = private->tdlActivated =
      private->sdlOpened = private->sdlStarted = private->sdlActivated =
      MDTM_FALSE;

    if ((paError = Pa_Initialize()) != paNoError) {
      if (private->debugEnabled)
	fprintf(stderr,"PortAudio error: %s\n",Pa_GetErrorText(paError));
      return(this->destroy(this));
    }

    if ((private->lockObject = CatsLockObject_create()) == NULL)
      return(this->destroy(this));

    private->paStream = NULL;
    private->lastOutTime = 0;
  }

  return this;
}

static PaMixer PaMixer_destroy(PaMixer this)
{
  if (this->private != NULL) {
    CatsTargetDataLine tdl=this->private->tdl;
    CatsSourceDataLine sdl=this->private->sdl;
    CatsLockObject lockObject=this->private->lockObject;

    if (lockObject != NULL) {
      lockObject->lock(lockObject);
      if (tdl && tdl->isOpen(tdl))
	tdl->close(tdl);
      if (sdl && sdl->isOpen(sdl))
	sdl->close(sdl);
      lockObject->destroy(lockObject);
    }

    if (tdl != NULL)
      tdl->destroy(tdl);

    if (sdl != NULL)
      sdl->destroy(sdl);

    free(this->private);
  }

  Pa_Terminate();
  free(this);
  return NULL;
}

static CatsTargetDataLine PaMixer_getTargetDataLine(PaMixer this)
{
  return this->private->tdl;
}

static CatsSourceDataLine PaMixer_getSourceDataLine(PaMixer this)
{
  return this->private->sdl;
}

/* CatsTargetDataLine interface implementation */

static int PaMixerTargetDataLine_open(CatsTargetDataLine line,
				      CatsAudioFormat format,
				      int bufferSize)
{
  PaMixer this=line->getSource(line);
  CatsLockObject lockObject=this->private->lockObject;
  int retval;

  lockObject->lock(lockObject);

  if (!this->private->tdlOpened) {
    PaMixer_open(this,format,bufferSize,MDTM_TRUE,MDTM_FALSE);
  }

  retval = this->private->tdlOpened == MDTM_TRUE;

  lockObject->unlock(lockObject);

  return retval;
}

static int PaMixerTargetDataLine_start(CatsTargetDataLine line)
{
  PaMixer this=line->getSource(line);
  CatsLockObject lockObject=this->private->lockObject;
  int retval;

  lockObject->lock(lockObject);

  if (this->private->tdlOpened && !this->private->tdlStarted) {
    PaMixer_start(this,MDTM_TRUE,MDTM_FALSE);
  }

  retval = this->private->tdlStarted == MDTM_TRUE;

  lockObject->unlock(lockObject);

  return retval;
}

static int PaMixerTargetDataLine_read(CatsTargetDataLine line, char *b,
				      int off, int len)
{
  PaMixer this=line->getSource(line);
  CatsLockObject lockObject=this->private->lockObject;
  CatsRingBuffer tdlBuffer;
  int frameSize;
  int totalBytesRead=0;

  lockObject->lock(lockObject);

  frameSize=this->private->frameSize;

  if (len % frameSize != 0) {
    lockObject->unlock(lockObject);
    return -1;
  }

  tdlBuffer=this->private->tdlBuffer;

  /* This is overly complicated, but necessary to avoid an even more
     complicated deadlock */

  while (len > 0) {
    int readable=tdlBuffer->readable(tdlBuffer);
    int bytesRead=0;

    if (readable > 0) {
      if ((bytesRead =
	   tdlBuffer->read(tdlBuffer,b,off,MDTM_MIN(len,readable))) < 0) {
	lockObject->unlock(lockObject);
	return -1;
      }

      if (this->private->byteSwap)
	mdtm_byteswap(b+off,frameSize,bytesRead/frameSize);

      off += bytesRead;
      len -= bytesRead;
      totalBytesRead += bytesRead;

    }

    if (len > 0)
      lockObject->notify(lockObject);

  }

  lockObject->unlock(lockObject);

  return totalBytesRead;
}

static int PaMixerTargetDataLine_stop(CatsTargetDataLine line)
{
  PaMixer this=line->getSource(line);
  CatsLockObject lockObject=this->private->lockObject;
  int retval;

  lockObject->lock(lockObject);

  if (this->private->tdlStarted) {
    PaMixer_stop(this,MDTM_TRUE,MDTM_FALSE);
  } 

  retval = this->private->tdlStarted == MDTM_FALSE;

  lockObject->unlock(lockObject);

  return retval;
}

static int PaMixerTargetDataLine_close(CatsTargetDataLine line)
{
  PaMixer this=line->getSource(line);
  PaMixerPrivate private=this->private;
  CatsLockObject lockObject=private->lockObject;
  int retval;

  lockObject->lock(lockObject);

  if (private->tdlOpened) {
    PaMixer_close(this,MDTM_TRUE,MDTM_FALSE);
  }

  retval = private->tdlOpened == MDTM_FALSE;

  lockObject->unlock(lockObject);

  return retval;
}

static void PaMixerTargetDataLine_flush(CatsTargetDataLine line)
{
  PaMixer this=line->getSource(line);
  CatsLockObject lockObject=this->private->lockObject;
  CatsRingBuffer tdlBuffer;

  lockObject->lock(lockObject);

  tdlBuffer=this->private->tdlBuffer;
  tdlBuffer->flush(tdlBuffer);

  lockObject->unlock(lockObject);
}

static int PaMixerTargetDataLine_available(CatsTargetDataLine line)
{
  PaMixer this=line->getSource(line);
  CatsLockObject lockObject=this->private->lockObject;
  CatsRingBuffer tdlBuffer;
  int available;

  lockObject->lock(lockObject);

  tdlBuffer = this->private->tdlBuffer;

  available = tdlBuffer->readable(tdlBuffer);

  lockObject->unlock(lockObject);

  return available;
}

static int PaMixerTargetDataLine_isOpen(CatsTargetDataLine line)
{
  PaMixer this=line->getSource(line);
  return this->private->tdlOpened;
}

static int PaMixerTargetDataLine_isRunning(CatsTargetDataLine line)
{
  PaMixer this=line->getSource(line);
  return this->private->tdlStarted;
}

static int PaMixerTargetDataLine_isActive(CatsTargetDataLine line)
{
  PaMixer this=line->getSource(line);
  return this->private->tdlActivated;
}

static int PaMixerTargetDataLine_getBufferSize(CatsTargetDataLine line)
{
  PaMixer this=line->getSource(line);
  return this->private->bufferSize;
}

static CatsAudioFormat PaMixerTargetDataLine_getFormat(CatsTargetDataLine line)
{
  PaMixer this=line->getSource(line);
  return this->private->format;
}

static int PaMixerTargetDataLine_getFramePosition(CatsTargetDataLine line)
{
  PaMixer this=line->getSource(line);
  return this->private->tdlFramePosition;
}

/* CatsSourceDataLine interface implementation */

static int PaMixerSourceDataLine_open(CatsSourceDataLine line,
				      CatsAudioFormat format,
				      int bufferSize)
{
  PaMixer this=line->getSource(line);
  CatsLockObject lockObject=this->private->lockObject;
  int retval;

  lockObject->lock(lockObject);

  if (!this->private->sdlOpened) {
    PaMixer_open(this,format,bufferSize,MDTM_FALSE,MDTM_TRUE);
  }

  retval = this->private->sdlOpened == MDTM_TRUE;

  lockObject->unlock(lockObject);

  return retval;
}

static int PaMixerSourceDataLine_start(CatsSourceDataLine line)
{
  PaMixer this=line->getSource(line);
  CatsLockObject lockObject=this->private->lockObject;
  int retval;

  lockObject->lock(lockObject);

  if (this->private->sdlOpened && !this->private->sdlStarted) {
    PaMixer_start(this,MDTM_FALSE,MDTM_TRUE);
  }

  retval = this->private->sdlStarted == MDTM_TRUE;

  lockObject->unlock(lockObject);

  return retval;
}

static int PaMixerSourceDataLine_write(CatsSourceDataLine line, char *b,
				       int off, int len)
{
  PaMixer this=line->getSource(line);
  CatsLockObject lockObject=this->private->lockObject;
  CatsRingBuffer sdlBuffer;
  int frameSize;
  int totalBytesWritten=0;

  lockObject->lock(lockObject);

  frameSize = this->private->frameSize;

  if (len % this->private->frameSize != 0) {
    lockObject->unlock(lockObject);
    return -1;
  }

  sdlBuffer=this->private->sdlBuffer;

  /* This is overly complicated, but necessary to avoid an even more
     complicated deadlock */

  while (len > 0) {
    int writeable=sdlBuffer->writeable(sdlBuffer);
    int bytesWritten=0;

    if (writeable > 0) {
      if ((bytesWritten =
	   sdlBuffer->write(sdlBuffer,b,off,MDTM_MIN(len,writeable))) < 0) {
	lockObject->unlock(lockObject);
	return -1;
      }

      off += bytesWritten;
      len -= bytesWritten;
      totalBytesWritten += bytesWritten;
    }

    if (len > 0)
      lockObject->notify(lockObject);
  }

  lockObject->unlock(lockObject);

  return totalBytesWritten;
}

static int PaMixerSourceDataLine_stop(CatsSourceDataLine line)
{
  PaMixer this=line->getSource(line);
  CatsLockObject lockObject=this->private->lockObject;
  int retval;

  lockObject->lock(lockObject);

  if (this->private->sdlStarted) {
    PaMixer_stop(this,MDTM_FALSE,MDTM_TRUE);
  } 

  retval = this->private->sdlStarted == MDTM_FALSE;

  lockObject->unlock(lockObject);

  return retval;
}

static int PaMixerSourceDataLine_close(CatsSourceDataLine line)
{
  PaMixer this=line->getSource(line);
  PaMixerPrivate private=this->private;
  CatsLockObject lockObject=private->lockObject;
  int retval;

  lockObject->lock(lockObject);

  if (private->sdlOpened) {
    PaMixer_close(this,MDTM_FALSE,MDTM_TRUE);
  }

  retval = private->sdlOpened == MDTM_FALSE;

  lockObject->unlock(lockObject);

  return retval;
}

static void PaMixerSourceDataLine_drain(CatsSourceDataLine line)
{
  PaMixer this=line->getSource(line);
  CatsLockObject lockObject=this->private->lockObject;
  int bufferSize;

  lockObject->lock(lockObject);

  bufferSize=this->private->bufferSize;

  if (!line->isRunning(line))
    line->start(line);

  while (line->available(line) < bufferSize)
    lockObject->notify(lockObject);

  lockObject->unlock(lockObject);
}

static void PaMixerSourceDataLine_flush(CatsSourceDataLine line)
{
  PaMixer this=line->getSource(line);
  CatsLockObject lockObject=this->private->lockObject;
  CatsRingBuffer sdlBuffer;

  lockObject->lock(lockObject);

  sdlBuffer=this->private->sdlBuffer;
  sdlBuffer->flush(sdlBuffer);

  lockObject->unlock(lockObject);
}

static int PaMixerSourceDataLine_available(CatsSourceDataLine line)
{
  PaMixer this=line->getSource(line);
  CatsLockObject lockObject=this->private->lockObject;
  CatsRingBuffer sdlBuffer;
  int available;

  lockObject->lock(lockObject);

  sdlBuffer = this->private->sdlBuffer;

  available = sdlBuffer->writeable(sdlBuffer);

  lockObject->unlock(lockObject);

  return available;
}

static int PaMixerSourceDataLine_isOpen(CatsSourceDataLine line)
{
  PaMixer this=line->getSource(line);
  return this->private->sdlOpened;
}

static int PaMixerSourceDataLine_isRunning(CatsSourceDataLine line)
{
  PaMixer this=line->getSource(line);
  return this->private->sdlStarted;
}

static int PaMixerSourceDataLine_isActive(CatsSourceDataLine line)
{
  PaMixer this=line->getSource(line);
  return this->private->sdlActivated;
}

static int PaMixerSourceDataLine_getBufferSize(CatsSourceDataLine line)
{
  PaMixer this=line->getSource(line);
  return this->private->bufferSize;
}

static CatsAudioFormat PaMixerSourceDataLine_getFormat(CatsSourceDataLine line)
{
  PaMixer this=line->getSource(line);
  return this->private->format;
}

static int PaMixerSourceDataLine_getFramePosition(CatsSourceDataLine line)
{
  PaMixer this=line->getSource(line);
  return this->private->sdlFramePosition;
}

/* PRIVATE METHODS */

static void PaMixer_open(PaMixer this, CatsAudioFormat format, int bufferSize,
			 int tdlOpen, int sdlOpen)
{
  PaMixerPrivate private=this->private;

  /* EBG - attempting to open full duplex when only half enabled is a
     Bad Thing. */

  if (tdlOpen && sdlOpen && private->halfDuplexEnabled)
    return;

  /* EBG - leave line status alone if input flags false, otherwise
     gate status requests against half duplex operation. */

  if (tdlOpen && !(private->halfDuplexEnabled && private->sdlOpened))
    private->tdlOpened = MDTM_TRUE;

  if (sdlOpen && !(private->halfDuplexEnabled && private->tdlOpened))
    private->sdlOpened = MDTM_TRUE;

  if (!private->opened && (private->tdlOpened || private->sdlOpened)) {
    PaError paError;
    int numChannels;
    int sampleSizeInBits;
    PaSampleFormat sampleFormat;
    double sampleRate;
    int paNumBuffers;

    PaDeviceID inputDevice;
    int numInputChannels;

    PaDeviceID outputDevice;
    int numOutputChannels;

    private->format = format;
    sampleSizeInBits = format->getSampleSizeInBits(format);

    switch(format->getEncoding(format)) {
    case CATS_ENCODING_PCM_SIGNED:
      switch(sampleSizeInBits) {
      case 8:
	sampleFormat = paInt8;
	break;
      case 16:
	sampleFormat = paInt16;
	break;
      case 24:
	sampleFormat = paInt24;
	break;
      default:
	if (private->debugEnabled)
	  fprintf(stderr,"PortAudio error: Unsupported format\n");
	return;
      }
      break;
    case CATS_ENCODING_PCM_UNSIGNED:
      switch(sampleSizeInBits) {
      case 8:
	sampleFormat = paUInt8;
	break;
      default:
	if (private->debugEnabled)
	  fprintf(stderr,"PortAudio error: Unsupported format\n");
	return;
      }
      break;
    default:
      if (private->debugEnabled)
	fprintf(stderr,"PortAudio error: Unsupported format\n");
      return;
    }

    numChannels = format->getChannels(format);
    sampleRate = format->getSampleRate(format);
    private->frameRate = format->getFrameRate(format);
    private->frameSize = format->getFrameSize(format);
    private->byteSwap = format->isBigEndian(format) ^ mdtm_isbigendian();

    paNumBuffers =
      ceil(((double)bufferSize)/(PA_FRAMES_PER_BUFFER*private->frameSize));
    paNumBuffers = MDTM_MAX(paNumBuffers,
			    2*Pa_GetMinNumBuffers(PA_FRAMES_PER_BUFFER,
						  private->frameRate));
#if (DEBUG)
    printf("Setting buffer size PA_FRAMES_PER_BUFFER %d, framesize %d, numbuffers %d\n", PA_FRAMES_PER_BUFFER, private->frameSize,paNumBuffers); fflush(stdout);
#endif

    private->bufferSize = bufferSize =
      1 << mdtm_srint_nextpow2(PA_FRAMES_PER_BUFFER*private->frameSize*
			       paNumBuffers);

#if (DEBUG)
    printf("Buffer size is %d\n", bufferSize); fflush(stdout);
#endif
    private->microsecondPositionFactor = 1000000./private->frameRate;

    if (tdlOpen || !private->halfDuplexEnabled) {
      inputDevice = Pa_GetDefaultInputDeviceID();
      numInputChannels = numChannels;
    } else {
      inputDevice = paNoDevice;
      numInputChannels = 0;
    }

    if (sdlOpen || !private->halfDuplexEnabled) {
      outputDevice = Pa_GetDefaultOutputDeviceID();
      numOutputChannels = numChannels;
    } else {
      outputDevice = paNoDevice;
      numOutputChannels = 0;
    }

    if ((private->tdlBuffer = CatsRingBuffer_create(bufferSize)) == NULL)
      return;

    if ((private->sdlBuffer = CatsRingBuffer_create(bufferSize)) == NULL)
      return;

    /* There should be no constants defined a la 'paNoFlag' - they
       should be of the form 'PA_NOFLAG', so it's clear they're
       defined constants and that they're picked up by '#include
       "portaudio.h"' */

    if ((paError = Pa_OpenStream(&private->paStream,
				 inputDevice,
				 numInputChannels,
				 sampleFormat,
				 NULL,
				 outputDevice,
				 numOutputChannels,
				 sampleFormat,
				 NULL,
				 sampleRate,
				 PA_FRAMES_PER_BUFFER,
				 paNumBuffers,
				 paNoFlag,
				 PaMixer_callback,
				 this)) != paNoError) {
      if (private->debugEnabled)
	fprintf(stderr,"PortAudio error: %s\n",Pa_GetErrorText(paError));
      return;
    }

    this->private->opened = MDTM_TRUE;
  }

  return;
}

static void PaMixer_start(PaMixer this, int tdlStart, int sdlStart)
{
  PaMixerPrivate private=this->private;

  /* Can't start the thing if it isn't open, can we? */

  if (private->opened) {
    if (tdlStart && private->tdlOpened)
      private->tdlStarted = MDTM_TRUE;
    if (sdlStart && private->sdlOpened)
      private->sdlStarted = MDTM_TRUE;

    if (!private->started && (private->tdlStarted || private->sdlStarted)) {
      PaError paError;

      if ((paError = Pa_StartStream(private->paStream))
	  != paNoError) {
	if (private->debugEnabled)
	  fprintf(stderr,"PortAudio error: %s\n",Pa_GetErrorText(paError));
	return;
      }

      private->started = MDTM_TRUE;
    }
  }
}

static void PaMixer_stop(PaMixer this, int tdlStop, int sdlStop)
{
  PaMixerPrivate private=this->private;

  if (tdlStop) {
    private->tdlActivated = MDTM_FALSE;
    private->tdlStarted = MDTM_FALSE;
  }

  if (sdlStop) {
    private->sdlActivated = MDTM_FALSE;
    private->sdlStarted = MDTM_FALSE;
  }

  if (private->started && !(private->tdlStarted || private->sdlStarted)) {
    PaError paError;

    paError = Pa_StopStream(private->paStream);

    if (paError != paNoError) {
      if (private->debugEnabled)
	fprintf(stderr,"PortAudio error: %s\n",Pa_GetErrorText(paError));
      return;
    }

    private->started = MDTM_FALSE;
  }
}

static void PaMixer_close(PaMixer this, int tdlClose, int sdlClose)
{
  PaMixerPrivate private=this->private;
  CatsTargetDataLine tdl=this->getTargetDataLine(this);
  CatsSourceDataLine sdl=this->getSourceDataLine(this);

  if (tdlClose) {
    if (tdl->isRunning(tdl))
      tdl->stop(tdl);
    private->tdlOpened = MDTM_FALSE;
  }

  if (sdlClose) {
    if (sdl->isRunning(sdl))
      sdl->stop(sdl);
    private->sdlOpened = MDTM_FALSE;
  }

  if (private->opened && !(private->tdlOpened || private->sdlOpened)) {
    PaError paError;
    CatsRingBuffer tdlBuffer=private->tdlBuffer;

    private->tdlBuffer = tdlBuffer->destroy(tdlBuffer);
    private->bufferSize = 0;
    private->format = NULL;

    paError = Pa_CloseStream(private->paStream);
    private->paStream = NULL;

    if (paError != paNoError) {
      if (private->debugEnabled)
	fprintf(stderr,"PortAudio error: %s\n",Pa_GetErrorText(paError));
      return;
    }

    private->opened = MDTM_FALSE;
  }
}

static int PaMixer_callback(void *inputBuffer, void *outputBuffer,
			    unsigned long framesPerBuffer,
			    PaTimestamp outTime, void *userData)
{
  PaMixer this=(PaMixer)userData;
  PaMixerPrivate private=this->private;
  int frameSize;
  int len;

  frameSize=private->frameSize;
  len=framesPerBuffer*frameSize;

  /* Handle input data */

  if (inputBuffer != NULL) {
    CatsRingBuffer tdlBuffer=private->tdlBuffer;
    int bytesToWrite;

    /* Set time stamp information */

    private->tdlFramePosition = outTime;
    private->tdlMicrosecondPosition =
      outTime*private->microsecondPositionFactor;

    /* Write as much data as available, drop the rest on the floor if
       necessary */

    bytesToWrite =
      MDTM_MIN(len,tdlBuffer->writeable(tdlBuffer));

    if (private->tdlStarted && bytesToWrite > 0) {

#if (DEBUG)
      printf("Writing %d bytes (writeable %d)\n", bytesToWrite, tdlBuffer->writeable(tdlBuffer)); fflush(stdout);
#endif
      
      tdlBuffer->write(tdlBuffer,(char *)inputBuffer,0,bytesToWrite);
      private->tdlActivated = MDTM_TRUE;
    } else {
      private->tdlActivated = MDTM_FALSE;
    }

#ifdef __solaris__
    CatsThread_sleep(0,1);
#endif

  }

  /* Handle output data */

  if (outputBuffer != NULL) {
    CatsRingBuffer sdlBuffer=private->sdlBuffer;
    int bytesToRead;

    /* Set time stamp information */

    private->sdlFramePosition = outTime;
    private->sdlMicrosecondPosition =
      outTime*private->microsecondPositionFactor;

    /* Read as much data as available, zero-pad the rest of the buffer
       if necessary, byte swap the whole mess if necessary */

    bytesToRead = MDTM_MIN(len,sdlBuffer->readable(sdlBuffer));

    if (private->sdlStarted && bytesToRead > 0) {
      sdlBuffer->read(sdlBuffer,(char *)outputBuffer,0,bytesToRead);
      if (bytesToRead < len)
	mdtm_vrchar_zeros((char *)outputBuffer+bytesToRead,len-bytesToRead);
      if (private->byteSwap)
	mdtm_byteswap(outputBuffer,frameSize,framesPerBuffer);
      private->sdlActivated = MDTM_TRUE;
    } else {
      mdtm_vrchar_zeros(outputBuffer,len);
      private->sdlActivated = MDTM_FALSE;
    }

#if __solaris__
    CatsThread_sleep(0,1);
#endif

  }

  return 0;
}
