/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/* Original toplevel file sox.c from sox-12.17.1 mangled
   into a Communicator-compliant server by Sam Bayer.

   Portions (c) Copyright 2002 The MITRE Corporation

   This file is part of the Galaxy Communicator system. It is licensed
   under the conditions described in the file LICENSE in the root 
   directory of the Galaxy Communicator system.
*/

/*
 * Sox - The Swiss Army Knife of Audio Manipulation.
 *
 * This is the main function for the command line sox program.
 *
 * July 5, 1991
 * Copyright 1991 Lance Norskog And Sundry Contributors
 * This source code is freely redistributable and may be used for
 * any purpose.  This copyright notice must be maintained.
 * Lance Norskog And Sundry Contributors are not responsible for
 * the consequences of using this software.
 *
 * Change History:
 *
 * June 1, 1998 - Chris Bagwell (cbagwell@sprynet.com)
 *   Added patch to get volume working again.  Based on patch sent from
 *   Matija Nalis <mnalis@public.srce.hr>.
 *   Added command line switches to force format to ADPCM or GSM.
 *
 * September 12, 1998 - Chris Bagwell (cbagwell@sprynet.com)
 *   Reworked code that handled effects.  Wasn't correctly draining
 *   stereo effects and a few other problems.
 *   Made command usage (-h) show supported effects and file formats.
 *   (this is partially from a patch by Leigh Smith
 *    leigh@psychokiller.dialix.oz.au).
 *
 */

#include <stdio.h>
#include "st_i.h"
#include <string.h>
#include <stdlib.h>             /* for malloc() */
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#include <sys/types.h>          /* for fstat() */
#include <sys/stat.h>           /* for fstat() */
#ifdef HAVE_UNISTD_H
#include <unistd.h>             /* for unlink() */
#endif

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#ifndef HAVE_GETOPT
int getopt(int,char **,char *);
extern char *optarg;
extern int optind;
#endif
#endif

#include "galaxy/galaxy_all.h"
#define SERVER_FUNCTIONS_INCLUDE "sox_server.h"
#include "galaxy/server_functions.h"

static char *oas[] = {
  "-cats_compatibility", "recognize only audio formats recognized by CATS",
  "-outbound_brokers", "force SoX to write to brokers instead of proxies",
  NULL
};

static int CatsFormatsOnly = 0;
static int UseOutboundProxies = 1;

/* Original header from sox.c: */

/*
 * SOX main program.
 *
 * Rewrite for new nicer option syntax.  July 13, 1991.
 * Rewrite for separate effects library.  Sep. 15, 1991.
 * Incorporate Jimen Ching's fixes for real library operation: Aug 3, 1994.
 * Rewrite for multiple effects: Aug 24, 1994.
 */

/* 10/2/02: Completely rewritten for Communicator compliance.
   Added pipes and threads for broker in and broker out, frame
   deconstruction and construction, encapsulated all global
   variables, etc. */

/* local forward declarations */
static int copy_input(ft_t, char **);
static int open_input(ft_t, char **);
static int copy_output(ft_t, char **);
static int filetype(int);
static st_sample_t volumechange(st_sample_t *buf, st_ssize_t ct, double vol);

/* We parse effects into a temporary effects table and then place into
 * the real effects table.  This makes it easier to reorder some effects
 * as needed.  For instance, we can run a resampling effect before
 * converting a mono file to stereo.  This allows the resample to work
 * on half the data.
 *
 * Real effects table only needs to be 2 entries bigger then the user
 * specified table.  This is because at most we will need to add
 * a resample effect and an channel averaging effect.
 */

#define MAX_EFF 16
#define MAX_USER_EFF 14

/*
 * In efftab's, location 0 is always the input stream.
 *
 * If one was to support effects for quad-channel files, there would
 * need to be an effect tabel for each channel.
 */

/* SAM 9/24/02: We need to encapsulate each sox request,
   so all the global variables need to go. */ 

/*
 * Begin request encapsulation
 */

/* SAM 9/24/02: I'm going to want a smaller buffer size than
   8192, which is the default size for sox. It APPEARS that
   I should be able to use a smaller buffer size if I want,
   since only sox.c and raw.c use the ST_BUFSIZ constant. */

/* SAM 9/27/02: Listen very carefully. If I have an input broker,
   I need to set up a pipe and fdopen the read side,
   and set up a thread to do the main processing while the
   incoming broker callback feeds the other side of the pipe.
   Similarly, if I have an outgoing broker, I need to set
   up a pipe and fdopen the write side, and set up a thread
   to do the broker output while the main processing feeds the
   other side of the pipe. This request structure handles
   the whole thing. I free it using free_sox_request, and
   once everybody who has a claim to it is done, it's really
   freed. */

/* Bit flags for the freeing. */

#define SR_INPUT_THREAD 1
#define SR_INPUT_BROKER 2
#define SR_OUTPUT_THREAD 4

#define SR_ALL ( SR_INPUT_THREAD | SR_INPUT_BROKER | SR_OUTPUT_THREAD )

/* For the purposes of this tool, I'm going to make ST_BUFSIZ
   smaller. */

#undef ST_BUFSIZ
#define ST_BUFSIZ 1024

typedef struct __sox_request {
  ft_t informat;
  int in_filetype_copied;
  ft_t outformat;
  int out_filetype_copied;
  int verbose;
  int do_volume;
  int clipped;
  float volume;
  int swap_input;
  /* Effects tables. */
  struct st_effect efftab[MAX_EFF]; /* left/mono channel effects */
  struct st_effect efftabR[MAX_EFF];/* right channel effects */
  int neffects;                     /* # of effects to run on data */
  struct st_effect user_efftab[MAX_USER_EFF];
  int nuser_effects;
  st_sample_t ibufl[ST_BUFSIZ/2];    /* Left/right interleave buffers */
  st_sample_t ibufr[ST_BUFSIZ/2];
  st_sample_t obufl[ST_BUFSIZ/2];
  st_sample_t obufr[ST_BUFSIZ/2];
  GalSS_Environment *env;
  /* If we're getting broker input. */
  char *broker_host;
  int broker_port;
  /* I'm going to remove this once I've
     started up the outgoing broker, so I need to save
     some indication that there was one, so I can
     issue a proxy if I found a proxy when converting.
     The default behavior, actually, should be to
     issue a proxy, but at the moment, CATS doesn't
     accept proxies. I'll have to fix that this week. */
  GalSS_BrokerProxy *bp;
  int use_out_proxy;
  /* These flags tell us if we're doing brokering in
     either direction. */
  int in_broker;
  int out_broker;
  /* This mutex is for general memory management and
     checking the aborted flag. */
  GalUtil_LocalMutex mutex;
  int aborted;
  int done;
  /* This is for the input thread, if there is one. */
  int input_pipe[2];
  GalUtil_ThreadID input_tid;
  int input_thread_started;
  int input_pipe_opened;
  /* This is for the output thread, if there is one. */
  int output_pipe[2];
  GalUtil_ThreadID output_tid;
  int output_thread_started;
  int output_pipe_opened;
} sox_request;

static int process(sox_request *r, char **errstr);
static int checkeffect(sox_request *r, char **errstr);
static int flow_effect_out(sox_request *);
static int flow_effect(int, sox_request *);
static int drain_effect(int, sox_request *);
static int open_output(sox_request *, char **);

static void free_sox_request(sox_request *r, int by_whom);

static sox_request *new_sox_request(GalSS_Environment *env, Gal_Frame f, char **errstr)
{
  sox_request *r = (sox_request *) calloc(1, sizeof(sox_request));
  Gal_Object temp;

  *errstr = (char *) NULL;
  r->informat = (ft_t) malloc(sizeof(struct st_soundstream));
  r->outformat = (ft_t) malloc(sizeof(struct st_soundstream));
  st_initformat(r->informat);
  st_initformat(r->outformat);
  r->broker_host = (char *) NULL;
  r->bp = (GalSS_BrokerProxy *) NULL;
  r->env = GalSS_EnvCopy(env);
  r->done = SR_ALL;
  GalUtil_InitLocalMutex(&(r->mutex));
  
  temp = Gal_GetObject(f, ":verbose");
  if (temp) {
    r->verbose = 1;
    /* Because the sox library isn't properly encapsulated yet,
       we need to turn verbose on and off. This won't work
       well when we're handling multiple requests in multiple threads.
       Sigh. */
  }
  
  r->do_volume = 0;
  r->volume = 0.0;
  r->clipped = 0;
  temp = Gal_GetObject(f, ":volume");
  if (temp) {
    if (!Gal_Floatp(temp)) {
      free_sox_request(r, 0);
      *errstr = ":volume must be float";
      return (sox_request *) NULL;
    } else {
      r->do_volume = 1;
      r->volume = Gal_FloatValue(temp);
    }
  }
  
  if (Gal_GetObject(f, ":swap")) {
    r->swap_input = 1;
  }
  
  /* And now, we also do the effects here.
     The effects should be a list of lists, where
     the first element of each list is an effect name,
     and the subsequent elements are the arguments
     that one would pass to the effects on the command line. */

  r->nuser_effects = 0;
  temp = Gal_GetObject(f, ":effects");
  if (temp) {
    if (!Gal_Listp(temp)) {
      free_sox_request(r, 0);
      *errstr = ":effects must be a list of effects descriptions";
      return (sox_request *) NULL;
    } else {
      int i, len = 0;
      Gal_Object *elts = Gal_ListValue(temp, &len);
      for (i = 0; i < len; i++) {
	int j, sub_n = 0;
	Gal_Object *sub_elts;
	char **argv;
	int argc_effect;
	char *eff_name[1];
	eff_t effp;
	
	if (!Gal_Listp(elts[i])) {
	  free_sox_request(r, 0);
	  *errstr = "each effects description must be a list of strings";
	  return (sox_request *) NULL;
	}

	sub_elts = Gal_ListValue(elts[i], &sub_n);
	
	if (sub_n < 1) {
	  free_sox_request(r, 0);
	  *errstr = "each effects description must be a non-empty list of strings";
	  return (sox_request *) NULL;
	}

	if (r->nuser_effects >= MAX_USER_EFF) {
	  free_sox_request(r, 0);
	  *errstr = "too many effects specified";
	  return (sox_request *) NULL;
	}

	if (!Gal_Stringp(sub_elts[0])) {
	  free_sox_request(r, 0);
	  *errstr = "each effects description must be a non-empty list of strings";
	  return (sox_request *) NULL;
	}

	effp = &(r->user_efftab[r->nuser_effects]);
	  
	/* Find the effect. */
	eff_name[0] = Gal_StringValue(sub_elts[0]);
	argc_effect = st_geteffect_opt(effp, 1, eff_name);

	if (argc_effect == ST_EOF) {
	  free_sox_request(r, 0);
	  *errstr = "unknown effect";
	  return (sox_request *) NULL;
	}

	argv = (char **) calloc(sub_n - 1, sizeof(char *));
	
	/* Now, collect all the strings. */
	for (j = 1; j < sub_n; j++) {
	  if (!Gal_Stringp(sub_elts[j])) {
	    free_sox_request(r, 0);
	    *errstr = "each effects description must be a non-empty list of strings";
	    return (sox_request *) NULL;
	  }
	  argv[j - 1] = Gal_StringValue(sub_elts[j]);
	}

	(*effp->h->getopts)(effp, sub_n - 1, argv);
	free(argv);
	r->nuser_effects++;
      }
    }
  }
	
  return r;
}

static void free_ft(ft_t ft, int filetype_copied)
{
  if (ft->filename)
    free(ft->filename);
  if (ft->filetype && filetype_copied)
    free(ft->filetype);
  free(ft);
}

/* Once we set up a thread or a broker, we need
   to tell the done slot, so that it knows that
   the component has to be done. */

static void update_sox_request(sox_request *r, int flag)
{
  GalUtil_LockLocalMutex(&(r->mutex));
  r->done = r->done & ~flag;
  GalUtil_UnlockLocalMutex(&(r->mutex));
}

static void free_sox_request(sox_request *r, int by_whom)
{
  int free_it = 0;
  
  GalUtil_LockLocalMutex(&(r->mutex));
  r->done = r->done | by_whom;
  if ((r->done & SR_ALL) == SR_ALL) {
    free_it = 1;
  }
  GalUtil_UnlockLocalMutex(&(r->mutex));
  
  if (free_it) {
    if (r->input_pipe_opened) {
      close(r->input_pipe[0]);
      close(r->input_pipe[1]);
    }
    if (r->output_pipe_opened) {
      close(r->output_pipe[0]);
      close(r->output_pipe[1]);
    }	    
    free_ft(r->informat, r->in_filetype_copied);
    free_ft(r->outformat, r->out_filetype_copied);
    if (r->broker_host)
      free(r->broker_host);
    GalSS_EnvUnlock(r->env);
    free(r);
  }
}

/*
 * End request encapsulation
 */

/* For the
   first pass, we have three dispatch functions: one which
   goes from file to broker (play), one which goes from broker
   to file (record), and one which goes from broker to broker (convert).
   In each case, we may have a list of effects. */

/* Sox makes a pretty thorough assumption that it's writing
   to and reading from file pointers. Folding brokers in
   here is going to be interesting. I think what I'll do
   is set up a pipe for each broker, and then I'll call
   fdopen() on the appropriate end. The catch will be in making
   sure the reads from that type of file pointer happen
   at the right time, and that the writes really result
   in stuff being written to an outbound broker.

   I could do this with threads, or I could do this
   with timed tasks. The problem with doing this with
   timed tasks is that the startread method might read,
   if the data has a header, and we don't know how long
   those blocks are going to be. So the best thing to do is
   probably set up a thread. */

static int populate_ft_info(sox_request *r, Gal_Frame f,
			    int is_input, int filename_expected, char **errstr)
{
  ft_t ft = is_input ? r->informat : r->outformat;
  char *sr_key = is_input ? ":input_sample_rate" : ":output_sample_rate";
  char *de_key = is_input ? ":input_data_encoding" : ":output_data_encoding";
  char *c_key = is_input ? ":input_channels" : ":output_channels";
  char *ft_key = is_input ? ":input_filetype" : ":output_filetype";
  char *ds_key = is_input ? ":input_data_size" : ":output_data_size";
  char *f_key = is_input ? ":input_filename" : ":output_filename";
  Gal_Object temp;
  int is_broker = 0;

  *errstr = (char *) NULL;
  
  if (is_input && r->swap_input)
    ft->swap = 1;

  temp = Gal_GetObject(f, f_key);
  if (temp) {
    if (!filename_expected) {
      *errstr = "unexpected filename found";
      return -1;
    } else if (!Gal_Stringp(temp)) {
      *errstr = "file name must be string";
      return -1;
    } else {
      ft->filename = strdup(Gal_StringValue(temp));
      if (is_input) {
	r->in_broker = 0;
      } else {
	r->out_broker = 0;
      }
    }
  } else if (filename_expected) {
    *errstr = "expected filename not found";
    return -1;
  } else {
    ft->filename = strdup("<broker>");
    ft->filetype = "raw";
    is_broker = 1;
    if (is_input) {
      r->in_broker = 1;
    } else {
      r->out_broker = 1;
    }
  }

  /* Next, we check the broker args, for is_input.
     They should only be present if filename_expected is 0. */

  if (is_input) {

    if (is_broker && r->swap_input) {
      /* This should never happen. */
      *errstr = "Shouldn't swap input with brokers";
      return -1;
    }
  
    temp = Gal_GetObject(f, ":host");
    if (temp) {
      if (!Gal_Stringp(temp)) {
	*errstr = ":host must be string";
	return -1;
      } else if (filename_expected) {
	*errstr = "unexpected :host found";
	return -1;
      } else {
	r->broker_host = strdup(Gal_StringValue(temp));
      }
    }

    temp = Gal_GetObject(f, ":port");
    if (temp) {
      if (!Gal_Intp(temp)) {
	*errstr = ":port must be integer";
	return -1;
      } else if (filename_expected) {
	*errstr = "unexpected :port found";
	return -1;
      } else {
	r->broker_port = Gal_IntValue(temp);
      }
    }

    temp = Gal_GetObject(f, ":proxy");
    if (temp) {
      if (!Gal_Proxyp(temp)) {
	*errstr = ":proxy must be proxy object";
	return -1;
      } else if (filename_expected) {
	*errstr = "unexpected :proxy found";
	return -1;
      } else {
	r->bp = (GalSS_BrokerProxy *) Gal_ProxyValue(temp);
	r->use_out_proxy = 1;
      }
    }

    /* Finally, if you have :host or :port, you must have
       both, and if you have :proxy, you can't have :host or :port. */

    if (r->broker_host || r->broker_port) {
      if (!(r->broker_host && r->broker_port)) {
	*errstr = "both :host and :port required";
	return -1;
      }
      if (r->bp) {
	*errstr = ":proxy not permitted with :host and :port";
	return -1;
      }
    }

    if ((!filename_expected) &&
	!(r->broker_host || r->broker_port || r->bp)) {
      /* If you don't have any of them and it should be a broker. barf. */
      *errstr = "no broker input found";
      return -1;
    }
  }

  
  /* Legal types are legal file extensions in sox. Brokers must
     be raw data. */
  temp = Gal_GetObject(f, ft_key);
  if (temp) {
    if (!Gal_Stringp(temp)) {
      *errstr = "file type must be string";
      return -1;
    } else if (is_broker) {
      /* No file type for brokering. */
      *errstr = "file type not permitted for brokering";
      return -1;
    } else {
      ft->filetype = Gal_StringValue(temp);
      if (ft->filetype) {
	if (ft->filetype[0] == '.') {
	  ft->filetype = strdup(ft->filetype + 1);
	} else {
	  ft->filetype = strdup(ft->filetype);
	}
	if (is_input)
	  r->in_filetype_copied = 1;
	else
	  r->out_filetype_copied = 1;
      }
    }
  }

  temp = Gal_GetObject(f, sr_key);  
  if (temp) {
    if (!Gal_Intp(temp)) {      
      *errstr = "sample rate must be integer";
      return -1;
    } else {
      ft->info.rate = Gal_IntValue(temp);
      if (ft->info.rate <= 0) {
	*errstr = "sample rate must be positive integer";
	return -1;
      }
    }
  }

  temp = Gal_GetObject(f, c_key);
  if (temp) {
    if (!Gal_Intp(temp)) {      
      *errstr = "channels must be integer";
      return -1;
    } else {
      ft->info.channels = Gal_IntValue(temp);
      if ((ft->info.channels != 1) &&
	  (ft->info.channels != 2) &&
	  (ft->info.channels != 4)) {
	*errstr = "channels must be 1, 2 or 4";
	return -1;
      }
    }
  }

  temp = Gal_GetObject(f, ds_key);
  if (temp) {
    if (!Gal_Intp(temp)) {
      *errstr = "data size must be integer";
      return -1;
    } else {
      switch (Gal_IntValue(temp)) {
      case 8:
	ft->info.size = ST_SIZE_BYTE;
	break;
      case 16:
	ft->info.size = ST_SIZE_WORD;
	break;
      case 32:
	ft->info.size = ST_SIZE_DWORD;
	break;
      default:
	*errstr = "data size must be 8, 16, or 32";
	return -1;
      }
    }
  }

  /* Encodings. Legal values are PCM_SIGNED, PCM_UNSIGNED,
     ULAW, ALAW, ADPCM, IMA_ADPCM, GSM, FLOATING_POINT. Values
     chosen for consistency with CATS. */

  temp = Gal_GetObject(f, de_key);
  if (temp) {
    if (!Gal_Stringp(temp)) {
      *errstr = "data encoding must be string";
      return -1;
    } else {
      char *encoding = Gal_StringValue(temp);

      if (!strcmp(encoding, "PCM_SIGNED")) {
	ft->info.encoding = ST_ENCODING_SIGN2;
      } else if (!strcmp(encoding, "PCM_UNSIGNED")) {
	ft->info.encoding = ST_ENCODING_UNSIGNED;
      } else if (!strcmp(encoding, "ULAW")) {
	ft->info.encoding = ST_ENCODING_ULAW;
      } else if (!strcmp(encoding, "ALAW")) {
	ft->info.encoding = ST_ENCODING_ALAW;
      } else if (CatsFormatsOnly) {
	*errstr = "unknown data encoding";
	return -1;
      } else if (!strcmp(encoding, "FLOATING_POINT")) {
	ft->info.encoding = ST_ENCODING_FLOAT;
      } else if (!strcmp(encoding, "ADPCM")) {
	ft->info.encoding = ST_ENCODING_ADPCM;
      } else if (!strcmp(encoding, "IMA_ADPCM")) {
	ft->info.encoding = ST_ENCODING_IMA_ADPCM;
      } else if (!strcmp(encoding, "GSM")) {
	ft->info.encoding = ST_ENCODING_GSM;
      } else {
	*errstr = "unknown data encoding";
	return -1;
      }
    }
  }

  if (is_input) {
    if (copy_input(ft, errstr) < 0) {
      return -1;
    }
  } else if (filename_expected) {
    /* copy_output only makes sense with the filename. */
    if (copy_output(ft, errstr) < 0) {
      return -1;
    }
  } else if ( st_gettype(ft) ) {
    *errstr = "unknown input file format";
    return -1;
  }
  
  return 0;
}

/* The incoming broker should set up the thread. */

static int set_informat_size(sox_request *r, int size, char **errstr)
{
  if ((r->informat->info.size != -1) &&
      (r->informat->info.size != size)) {
    if (errstr)
      *errstr = "requested data size doesn't match broker";
    return -1;
  } 
  r->informat->info.size = size;
  return 0;
}

static int sox_request_abort(sox_request *r)
{
  int val = 0;
  
  GalUtil_LockLocalMutex(&(r->mutex));
  val = r->aborted;
  r->aborted = 1;
  GalUtil_UnlockLocalMutex(&(r->mutex));
  return val;
}

static int sox_request_aborted(sox_request *r)
{
  int val = 0;
  
  GalUtil_LockLocalMutex(&(r->mutex));
  val = r->aborted;
  GalUtil_UnlockLocalMutex(&(r->mutex));
  return val;
}

static void add_output_format_info(Gal_Frame f, ft_t outformat)
{
  int data_size = -1;
  char *encoding = (char *) NULL;
  
  Gal_SetProp(f, ":output_sample_rate",
	      Gal_IntObject(outformat->info.rate));
  switch (outformat->info.size) {
  case ST_SIZE_BYTE:
    data_size = 8;
    break;
  case ST_SIZE_WORD:
    data_size = 16;
    break;
  case ST_SIZE_DWORD:
    data_size = 32;
    break;
  }
  if (data_size != -1) {
    Gal_SetProp(f, ":output_data_size",
		Gal_IntObject(data_size));
  }
  Gal_SetProp(f, ":output_channels",
	      Gal_IntObject(outformat->info.channels));
  switch (outformat->info.encoding) {
  case ST_ENCODING_SIGN2:
    encoding = "PCM_SIGNED";
    break;
  case ST_ENCODING_UNSIGNED:
    encoding = "PCM_UNSIGNED";
    break;
  case ST_ENCODING_ULAW:
    encoding = "ULAW";
    break;
  case ST_ENCODING_ALAW:
    encoding = "ALAW";
    break;
  case ST_ENCODING_FLOAT:
    encoding = "FLOATING_POINT";
    break;
  case ST_ENCODING_ADPCM:
    encoding = "ADPCM";
    break;
  case ST_ENCODING_IMA_ADPCM:
    encoding = "IMA_ADPCM";
    break;
  case ST_ENCODING_GSM:
    encoding = "GSM";
    break;
  }
  if (encoding) {
    Gal_SetProp(f, ":output_data_encoding",
		Gal_StringObject(encoding));
  }
}

/* process() must make sure that it NEVER calls
   GalSS_EnvError by itself. Note that I call
   free_sox_request() here but not after process()
   in sox_play(), because sox_play will have its
   request freed by either the output thread or
   the output broker. If I had a file-to-file
   dispatch function, THAT would have to free the
   request. */

static void *thread_process(void *data)
{
  char *errstr = (char *) NULL;
  sox_request *r = (sox_request *) data;

  GalUtil_Debug1("=== Entering thread_process\n");

  /* If process returns 0, it's closed the
     file pointers. */
  if (process(r, &errstr) < 0) {
    GalSS_EnvError(r->env, errstr);
    fclose(r->informat->fp);
    fclose(r->outformat->fp);
  } else if (!r->out_broker) {
    /* If we don't have an out broker, then we'd
       better send the reply here. */
    Gal_Frame f = Gal_MakeClauseFrame("");
    Gal_SetProp(f, ":output_filename",
		Gal_StringObject(r->outformat->filename));
    add_output_format_info(f, r->outformat);
    GalSS_EnvReply(r->env, f);
  }
  
  free_sox_request(r, SR_INPUT_THREAD);
  GalUtil_Debug1("=== Exiting thread_process\n");
  return (void *) NULL;
}

/* This function should be in the GC core. */

static void free_broker_data(void *data, Gal_ObjectType data_type, int n_samples)
{
  switch (data_type) {
  case GAL_FRAME:
    Gal_FreeFrame((Gal_Frame) data);
    break;
  case GAL_PROXY:
    GalSS_FreeBrokerProxy((GalSS_BrokerProxy *) data);
    break;
  case GAL_LIST:
    if (data) {
      int i;

      for (i = 0; i < n_samples; i++) {
	Gal_FreeObject(((Gal_Object *) data)[i]);
      }
      free(data);
    }
    break;
  case GAL_STRING:
  case GAL_TOKEN:
  case GAL_FLOAT:
  case GAL_BINARY:
  case GAL_INT_16:
  case GAL_INT_32:
  case GAL_INT_64:
  case GAL_FLOAT_32:
  case GAL_FLOAT_64:
    free(data);
    break;
  default:
    /* Don't do anything with GAL_SYMBOL, GAL_KEYWORD,
       GAL_TAG, GAL_INT. */
    break;
  }
}

static void sox_in_broker_handler(GalIO_BrokerStruct *broker_struct,
				  void *data,
				  Gal_ObjectType data_type,
				  int n_samples)
{
  GalSS_Environment *env = GalSS_BrokerGetEnvironment(broker_struct);
  sox_request *r = (sox_request *) GalIO_GetBrokerData(broker_struct);
  int dsize = -1;

  /* There seems to be no option to shut down the broker from here,
     so we just have to assume it will continue to be drained. */
  
  if (sox_request_aborted(r)) {
    free_broker_data(data, data_type, n_samples);
    return;
  }
  
  switch (data_type) {
  case GAL_BINARY:
    dsize = ST_SIZE_BYTE;
    break;
  case GAL_INT_16:
    n_samples = n_samples * 2;
    dsize = ST_SIZE_WORD;
    break;
  case GAL_INT_32:
    dsize = ST_SIZE_DWORD;
    n_samples = n_samples * 4;
    break;
  default:
    break;
  }

  GalUtil_PInfo2("Received data type %s\n",
		 Gal_ObjectTypeString(data_type));
  
  if (dsize == -1) {
    /* Abort the whole thing. */
    int previously_aborted = sox_request_abort(r);
    if (!previously_aborted) {
      /* Close the pipe if it was opened, after setting aborted
	 (so we don't need mutexes). The thread will exit by itself. */
      if (r->input_thread_started) {
	close(r->input_pipe[1]);
      }
      /* Send the error. */
      GalSS_EnvError(env, "received data other than GAL_BINARY, GAL_INT_16, or GAL_INT_32");
      free_broker_data(data, data_type, n_samples);
      return;
    }
  }

  /* Bryan says that everything should be bytes, because that's
     the way all audio folks think. So we're ignoring the size
     of the input. */
  
  if (r->informat->info.size == -1) {
    r->informat->info.size = dsize;
  }
    
  if (!r->input_thread_started) {
    /* Set up the pipes and thread. */
    if (pipe(r->input_pipe) < -1) {
      GalSS_EnvError(env, "can't create input pipe");
      sox_request_abort(r);
      free_broker_data(data, data_type, n_samples);
      return;
    } else {
      r->input_pipe_opened = 1;
    }

    /* Stolen from open_input() */
    if ((r->informat->fp = fdopen(r->input_pipe[0], "rb")) == NULL) {
      GalSS_EnvError(env, "can't open input pipe");
      sox_request_abort(r);
      free_broker_data(data, data_type, n_samples);
      return;
    }
  
    /* See if this file is seekable or not */
#if     defined(DUMB_FILESYSTEM)
    r->informat->seekable = 0;
#else
    r->informat->seekable = (filetype(fileno(r->informat->fp)) == S_IFREG);
#endif
    /* End stolen from open_input() */

    update_sox_request(r, SR_INPUT_THREAD);
    /* At this point, we set up the thread to run the processor. */
    GalUtil_ThreadCreate(&(r->input_tid), thread_process, (void *) r);

    /* And mark that we've set it up. */
    r->input_thread_started = 1;
  }
  /* Write the material through the pipe. */
  write(r->input_pipe[1], data, n_samples);
  free_broker_data(data, data_type, n_samples);
}

static void sox_in_broker_done(GalIO_BrokerStruct *b, void *data)
{
  sox_request *r = (sox_request *) GalIO_GetBrokerData(b);

  /* Close the pipe. */
  if (r->input_thread_started) {
    close(r->input_pipe[1]);
  }
}

static void sox_in_broker_abort(GalIO_BrokerStruct *b, void *data)
{
  GalSS_Environment *env = GalSS_BrokerGetEnvironment(b);
  sox_request *r = (sox_request *) GalIO_GetBrokerData(b);

  if (!sox_request_abort(r)) {
    GalSS_EnvError(env, "incoming broker aborted");
  }
  if (r->input_thread_started) {
    close(r->input_pipe[1]);
  }
}

static void free_broker_sr(void *data)
{
  sox_request *r = (sox_request *) data;
  
  free_sox_request(r, SR_INPUT_BROKER);
}

static int set_up_incoming_broker(GalSS_Environment *env, Gal_Frame f,
				  sox_request *r, char **errstr)
{
  GalIO_BrokerStruct *b = (GalIO_BrokerStruct *) NULL;
  
  /* The idea is that the data type should be raw, and the broker should
     be a homogeneous one, with the appropriate data type. */
  
  if (r->bp) {
    /* First, try to figure out the type information. If the broker
       is typed, it had better be GAL_BINARY, GAL_INT_16, or GAL_INT_32.
       Oops. Bryan says that's an idiotic thing to do. */
    int new_size = -1;
    
    switch(GalSS_BrokerProxyObjectType(r->bp)) {      
    case GAL_BINARY:
      new_size = ST_SIZE_BYTE;
      break;
    case GAL_INT_16:
      new_size = ST_SIZE_WORD;
      break;
    case GAL_INT_32:
      new_size = ST_SIZE_DWORD;
      break;
    case (Gal_ObjectType) -1:
      /* Still don't know. We'll figure it out when the first one comes. */
      break;
    default:
      *errstr = "Bad type for broker proxy";
      return -1;
    }
    if (new_size != -1 && (r->informat->info.size == -1)) {
      if (set_informat_size(r, new_size, errstr) < 0) {
	return -1;
      }
    }
    /* Now, we ignore the type information in the broker callback. */
    b = GalSS_EnvBrokerProxyInInit(env, r->bp,
				   sox_in_broker_handler, -1,
				   (void *) r,
				   free_broker_sr);
    /* At this point, this will be dead data soon.
       We've also set the use_out_proxy slot, so we can
       still tell that the outbound needs a proxy. */
    r->bp = (GalSS_BrokerProxy *) NULL;
  } else {
    b = GalSS_EnvBrokerDataInInit(env, r->broker_host, r->broker_port, f,
				  sox_in_broker_handler, -1,
				  (void *) r,
				  free_broker_sr);
  }
  if (b) {
    /* For cleanliness, I'll call GalSS_EnvStartBroker, even though
       I don't anticipate running sox with tasks as threads. */
    GalIO_SetBrokerActive(b);
    update_sox_request(r, SR_INPUT_BROKER);
    GalIO_AddBrokerCallback(b, GAL_BROKER_DATA_DONE_EVENT,
			    sox_in_broker_done, (void *) NULL);
    GalIO_AddBrokerCallback(b, GAL_BROKER_ABORT_EVENT,
			    sox_in_broker_abort, (void *) NULL);
    GalSS_EnvStartBroker(env, b, 0);
    return 0;
  } else {
    *errstr = "Couldn't start up incoming broker";
    return -1;
  }
}

/* Once we've set up the input processing, under what
   circumstances should we send a response? The response
   should ALWAYS be sent by the element which instituted
   the abort: broker, auxiliary broker thread, or main
   thread. If the output is a file, we want to send an error if the
   output processing fails at any point, otherwise,
   success. If the output is a broker, on the other
   hand, we respond immediately if the out broker can be
   set up, or an error if it can't be. Once the broker is
   set up, the best we can do is abort the output.
   we can't yet send error status through a
   broker, but we'll have to fix that later. */   
  
/* In these toplevels, make sure you free the sox request
   after you send the error. Just in case. */

Gal_Frame sox_convert(Gal_Frame f, void *server_data)
{
  GalSS_Environment *env = (GalSS_Environment *) server_data;
  char *errstr;
  sox_request *r = new_sox_request(env, f, &errstr);
  
  if (!r) {
    GalSS_EnvError(env, errstr);
    return (Gal_Frame) NULL;
  }
  
  if (populate_ft_info(r, f, 1, 0, &errstr) < 0) {
    GalSS_EnvError(env, errstr);
    free_sox_request(r, 0);
    return (Gal_Frame) NULL;
  }

  if (populate_ft_info(r, f, 0, 0, &errstr) < 0) {
    GalSS_EnvError(env, errstr);
    free_sox_request(r, 0);
    return (Gal_Frame) NULL;
  }
  
  /* Now, set up a pipe from the broker, and set up
     a thread to read from the pipe, and use that as
     the input. */

  if (set_up_incoming_broker(env, f, r, &errstr) < 0) {
    GalSS_EnvError(env, errstr);
    free_sox_request(r, 0);
  } else {
    GalSS_EnvPostponeReply(env);
  }
  return (Gal_Frame) NULL;
}

Gal_Frame sox_play(Gal_Frame f, void *server_data)
{
  GalSS_Environment *env = (GalSS_Environment *) server_data;
  char *errstr;
  sox_request *r = new_sox_request(env, f, &errstr);

  if (!r) {
    GalSS_EnvError(env, errstr);
    return (Gal_Frame) NULL;
  }
  /* Always use out proxy. */
  r->use_out_proxy = UseOutboundProxies;
  
  if (populate_ft_info(r, f, 1, 1, &errstr) < 0) {
    GalSS_EnvError(env, errstr);
    free_sox_request(r, 0);
    return (Gal_Frame) NULL;
  }

  if (populate_ft_info(r, f, 0, 0, &errstr) < 0) {
    GalSS_EnvError(env, errstr);
    free_sox_request(r, 0);
    return (Gal_Frame) NULL;
  }

  /* We have a file; open it. */
  if (open_input(r->informat, &errstr) < 0) {
    GalSS_EnvError(env, errstr);
    free_sox_request(r, 0);
    return (Gal_Frame) NULL;
  }

  /* If process returns 0, it's closed the
     file pointers. */
  if (process(r, &errstr) < 0) {
    GalSS_EnvError(env, errstr);
    fclose(r->informat->fp);
    fclose(r->outformat->fp);
  } else {
    GalSS_EnvPostponeReply(env);
  }
  
  return (Gal_Frame) NULL;
}

Gal_Frame sox_record(Gal_Frame f, void *server_data)
{
  GalSS_Environment *env = (GalSS_Environment *) server_data;
  char *errstr;
  sox_request *r = new_sox_request(env, f, &errstr);

  if (!r) {
    GalSS_EnvError(env, errstr);
    return (Gal_Frame) NULL;
  }

  if (populate_ft_info(r, f, 1, 0, &errstr) < 0) {
    GalSS_EnvError(env, errstr);
    free_sox_request(r, 0);
    return (Gal_Frame) NULL;
  }

  if (populate_ft_info(r, f, 0, 1, &errstr) < 0) {
    GalSS_EnvError(env, errstr);
    free_sox_request(r, 0);
    return (Gal_Frame) NULL;
  }

  /* Now, set up a pipe from the broker, and set up
     a thread to read from the pipe, and use that as
     the input. */

  if (set_up_incoming_broker(env, f, r, &errstr) < 0) {
    GalSS_EnvError(env, errstr);
    free_sox_request(r, 0);
  } else {
    GalSS_EnvPostponeReply(env);
  }
  return (Gal_Frame) NULL;

}

static int copy_input(ft_t ft, char **errstr)
{
  *errstr = (char *) NULL;
  
  /* Let auto effect do the work if user is not overriding. */
  if (!ft->filetype)
    ft->filetype = "auto";

  if ( st_gettype(ft) ) {
    *errstr = "unknown input file format";
    return -1;
  }

  /* Default the input comment to the filename if not set from
   * command line.
   * FIXME: Should be a memory copy, not a pointer asignment.
   */
  if (!ft->comment)
    ft->comment = ft->filename;
  return 0;
}

static int open_input(ft_t ft, char **errstr)
{
  *errstr = (char *) NULL;
  
  /* Open file handler based on input name. Disabling
     stdin for Communicator-compliant server. */
  if ((ft->fp = fopen(ft->filename, "rb")) == NULL) {
    *errstr = "can't open input file";
    return -1;
  }

  /* See if this file is seekable or not */
#if     defined(DUMB_FILESYSTEM)
  ft->seekable = 0;
#else
  ft->seekable = (filetype(fileno(ft->fp)) == S_IFREG);
#endif
  return 0;
}

#if defined(DOS) || defined(WIN32)
#define LASTCHAR '\\'
#else
#define LASTCHAR '/'
#endif

static int copy_output(ft_t ft, char **errstr)
{
  *errstr = (char *) NULL;
  
  if (!ft->filetype && ft->filename) {
    /* Use filename extension to determine audio type. */
    
    /* First, chop off any path portions of filename.  This
     * prevents the next search from considering that part. */
    if ((ft->filetype = strrchr(ft->filename, LASTCHAR)) == NULL)      
      ft->filetype = ft->filename;

    /* Now look for an filename extension */
    if ((ft->filetype = strrchr(ft->filetype, '.')) != NULL)
      ft->filetype++;
    else
      ft->filetype = NULL;
  }

  if (ft->filename) {
    if ( st_gettype(ft) ) {
      *errstr = "unknown output file format";
      return -1;
    }
  }
  return 0;    
}

/* We're going to either open an outbound broker
   and set up a child thread, or open a file. */

typedef struct __outbound_broker {
  sox_request *r;
  GalSS_BrokerProxy *bp;
  GalIO_BrokerStruct *b;
} __outbound_broker;

static void *thread_feed_outbound_broker(void *data)
{
  __outbound_broker *obb = (__outbound_broker *) data;
  /* Loop, reading data in blocks. */
  char data_buf[ST_BUFSIZ];
  int n;
  
  GalUtil_Debug1("=== Entering thread_feed_outbound_broker\n");  
  while (1) {
    n = read(obb->r->output_pipe[0], data_buf, ST_BUFSIZ);
    if ((n == -1) && (errno == EINTR)) {
      continue;
    } else if (n == -1) {
      /* Oops, something bad happened. */
      break;
    } else if (n == 0) {
      /* We're done. */
      break;
    } else {
      /* Write the damn data. */
      if (obb->bp) {
	switch (obb->r->outformat->info.size) {
	case ST_SIZE_BYTE:
	  GalSS_ProxyArrayAdd(obb->bp, data_buf, n);
	  break;
	case ST_SIZE_WORD:
	  GalSS_ProxyArrayAdd(obb->bp, data_buf, n / 2);
	  break;
	case ST_SIZE_DWORD:
	  GalSS_ProxyArrayAdd(obb->bp, data_buf, n / 4);
	  break;
	}
      } else if (obb->b) {
	switch (obb->r->outformat->info.size) {
	case ST_SIZE_BYTE:
	  GalIO_BrokerWriteBinary(obb->b, data_buf, n);
	  break;
	case ST_SIZE_WORD:
	  GalIO_BrokerWriteInt16(obb->b, data_buf, n / 2);
	  break;
	case ST_SIZE_DWORD:
	  GalIO_BrokerWriteInt32(obb->b, data_buf, n / 4);
	  break;
	}
      }

      switch (obb->r->outformat->info.size) {
      case ST_SIZE_BYTE:
	GalUtil_PInfo2("Wrote data type GAL_BINARY\n");
	break;
      case ST_SIZE_WORD:
	GalUtil_PInfo2("Wrote data type GAL_INT_16\n");
	break;
      case ST_SIZE_DWORD:
	GalUtil_PInfo2("Wrote data type GAL_INT_32\n");
	break;
      }

      /* We may not have gotten everything, either because
	 of an abort or because of end of file. */
      if (n < ST_BUFSIZ) {
	break;
      }
      if (sox_request_aborted(obb->r)) {
	break;
      }
    }
  }
  /* Now that we're out of the loop, the question is
     whether something aborted or not. */
  if (sox_request_aborted(obb->r)) {
    /* Just done, not out done. */
    if (obb->bp) {
      GalIO_BrokerDataDone(GalSS_BrokerProxyBroker(obb->bp));
    } else {
      GalIO_BrokerDataDone(obb->b);
    }
  } else if (obb->bp) {
    GalSS_ProxyDone(obb->bp);
  } else {
    GalIO_BrokerDataOutDone(obb->b);
  }
  free_sox_request(obb->r, SR_OUTPUT_THREAD);
  /* The brokers themselves will be handled by the
     appropriate timed task. */
  if (obb->bp)
    GalSS_FreeBrokerProxy(obb->bp);  
  free(obb);
  GalUtil_Debug1("=== Exiting thread_feed_outbound_broker\n");  
  return (void *) NULL;
}

void *_GalSS_init_server(GalIO_ServerStruct *server, int argc, char **argv)
{
  if (GalUtil_OACheckUsage(argc, argv, oas, NULL) == 0)	
    exit(1);
  GalUtil_OAExtract(argc, argv, oas, "-cats_compatibility",
		    GAL_OA_INT, &CatsFormatsOnly);
  if (GalUtil_OAExtract(argc, argv, oas, "-outbound_brokers",
			GAL_OA_INT, NULL)) {
    UseOutboundProxies = 0;
  }
  /* Better set myname, since the verbose mode in sox needs it. */
  myname = "sox_server";
  return (void *) NULL;
}

static int open_output(sox_request *r, char **errstr)
{
  ft_t ft = r->outformat;
  *errstr = (char *) NULL;

  if (r->out_broker) {
    GalSS_BrokerProxy *bp = (GalSS_BrokerProxy *) NULL;
    GalIO_BrokerStruct *b = (GalIO_BrokerStruct *) NULL;
    __outbound_broker *obb;
    
    /* We're setting up an outbound broker. It can either
       be a proxy or not. Start by setting up the pipe. */
    if (pipe(r->output_pipe) < -1) {
      *errstr = "can't create output pipe";
      return -1;
    } else {
      r->output_pipe_opened = 1;
    }

    /* Stolen from open_output() */
    if ((ft->fp = fdopen(r->output_pipe[1], "wb")) == NULL) {
      *errstr = "can't open output pipe";
      return -1;
    }
    /* Now, set up the broker. */
    if (r->use_out_proxy) {
      /* If we're supposed to use an outbound proxy, figure
	 out what type to use and build it. */
      Gal_ObjectType t;
      
      switch (ft->info.size) {
      case ST_SIZE_BYTE:
	t = GAL_BINARY;
	break;
      case ST_SIZE_WORD:
	t = GAL_INT_16;
	break;
      case ST_SIZE_DWORD:
	t = GAL_INT_32;
	break;
      default:
	*errstr = "unknown output size";
	return -1;
      }
      
      bp = GalSS_ProxifyObjectType(r->env, t, -1, 10);
      if (!bp) {
	*errstr = "couldn't start outbound broker proxy";
	return -1;
      } else {
	/* Doesn't matter what the name is, because it
	   will get changed in the reply. */
	Gal_Frame f = Gal_MakeClauseFrame("");
	
	GalSS_EnvStartBroker(r->env, GalSS_BrokerProxyBroker(bp), 0);
	/* Send the result. */
	Gal_SetProp(f, ":proxy", Gal_CreateProxyObject(bp, 0));
	add_output_format_info(f, r->outformat);
	GalSS_EnvReply(r->env, f);
      }
    } else {
      b = GalIO_BrokerDataOutInit(GalSS_EnvComm(r->env), -1, 10);
      if (!b) {
	*errstr = "couldn't start outbound broker";
	return -1;
      } else {
	Gal_Frame f = Gal_MakeClauseFrame("");
	GalSS_EnvStartBroker(r->env, b, 0);
	GalIO_BrokerPopulateFrame(b, f, ":host", ":port");
	add_output_format_info(f, r->outformat);
	GalSS_EnvReply(r->env, f);
      }
    }
    /* Now, we set up the thread to feed the outbound broker. */
    obb = (__outbound_broker *) calloc(1, sizeof(__outbound_broker));
    obb->b = b;
    obb->bp = bp;
    obb->r = r;
    update_sox_request(r, SR_OUTPUT_THREAD);
    GalUtil_ThreadCreate(&(r->output_tid),
			 thread_feed_outbound_broker, (void *) obb);

    /* And mark that we've set it up. */
    r->output_thread_started = 1;
  } else {
    /* Disabled stdout. */

    ft->fp = fopen(ft->filename, "wb");

    if (ft->fp == NULL) {
      *errstr = "Can't open output file";
      return -1;
    }
  }

  /* stdout tends to be line-buffered.  Override this */
  /* to be Full Buffering. */
  if (setvbuf (ft->fp,NULL,_IOFBF,sizeof(char)*ST_BUFSIZ))
    {
      *errstr = "Can't set write buffer";
      return -1;
    }

#if     defined(DUMB_FILESYSTEM)
  ft->seekable = 0;
#else
  ft->seekable  = (filetype(fileno(ft->fp)) == S_IFREG);
#endif
  return 0;
}

/*
 * Process input file -> effect table -> output file
 *      one buffer at a time
 */

static int process(sox_request *r, char **errstr)
{
  int e, f, flowstatus;
  ft_t informat = r->informat;
  ft_t outformat = r->outformat;
  struct st_effect *efftab = r->efftab;
  struct st_effect *efftabR = r->efftabR;
  /* Don't pick this up from r until checkeffects() is called. */
  int neffects;

  /* First, set global verbosity. */

  *errstr = (char *) NULL;
  verbose = r->verbose;

  if (r->do_volume)
    st_report("Volume factor: %f\n", (double) r->volume);

  if (r->do_volume && (r->volume < 0.0))
    st_report("Volume adjustment is negative.  This will result in a phase change\n");
  
  /* Read and write starters can change their formats. */
  if ((* informat->h->startread)(informat) != ST_SUCCESS) {
    *errstr = "input read start failed";
    /* I'm not computing an error buffer, so I'm ignoring st_errstr. */
    GalUtil_Warn("Failed reading %s: %s",informat->filename,
		 informat->st_errstr);
    return -1;
  }

  /* Go a head and assume 1 channel audio if nothing is detected.
   * This is because libst usually doesn't set this for mono file
   * formats (for historical reasons).
   */
  if (informat->info.channels == -1)
    informat->info.channels = 1;

  if ( st_checkformat(informat) ) {
    *errstr = "bad input format";
    GalUtil_Warn("bad input format for file %s: %s",
		 informat->filename,informat->st_errstr);
    return -1;
  }

  st_report("Input file %s: using sample rate %lu\n\tsize %s, encoding %s, %d %s",
	    informat->filename, informat->info.rate,
	    st_sizes_str[(unsigned char)informat->info.size],
	    st_encodings_str[(unsigned char)informat->info.encoding],
	    informat->info.channels,
	    (informat->info.channels > 1) ? "channels" : "channel");

  if (informat->comment)
    st_report("Input file %s: comment \"%s\"\n",
	      informat->filename, informat->comment);


  /* Copy the format first, in case we're opening a broker. */
  st_copyformat(informat, outformat);

  /* Check the output format limitations. */
  if (CatsFormatsOnly) {
    if ((outformat->info.encoding ==  ST_ENCODING_FLOAT) ||
	(outformat->info.encoding ==  ST_ENCODING_ADPCM) ||
	(outformat->info.encoding ==  ST_ENCODING_IMA_ADPCM) ||
	(outformat->info.encoding ==  ST_ENCODING_GSM)) {
      *errstr = "Unrecognized output format";
      return -1;
    }
  }

  if (open_output(r, errstr) < 0) {
    GalUtil_Warn(*errstr);
    return -1;
  }

  if ((*outformat->h->startwrite)(outformat) == ST_EOF) {
    *errstr = outformat->st_errstr;
    GalUtil_Warn(outformat->st_errstr);
    return -1;
  }

  if (st_checkformat(outformat) == ST_EOF) {
    *errstr = "bad output format";
    GalUtil_Warn("bad output format: %s", outformat->st_errstr);
    return -1;
  }

  st_report("Output file %s: using sample rate %lu\n\tsize %s, encoding %s, %d %s",
	    outformat->filename, outformat->info.rate,
	    st_sizes_str[(unsigned char)outformat->info.size],
	    st_encodings_str[(unsigned char)outformat->info.encoding],
	    outformat->info.channels,
	    (outformat->info.channels > 1) ? "channels" : "channel");

  if (outformat->comment)
    st_report("Output file: comment \"%s\"\n", outformat->comment);

  /* build efftab */
  if (checkeffect(r, errstr) < 0) {
    GalUtil_Warn(*errstr);
    return -1;
  }

  neffects = r->neffects;

  /* Start all effects */
  for(e = 1; e < neffects; e++) {
    (* efftab[e].h->start)(&efftab[e]);
    if (efftabR[e].name)
      (* efftabR[e].h->start)(&efftabR[e]);
  }

  /* Reserve an output buffer for all effects */
  for(e = 0; e < neffects; e++)
    {
      efftab[e].obuf = (st_sample_t *) malloc(ST_BUFSIZ * 
						 sizeof(st_sample_t));
      if (efftab[e].obuf == NULL) {
	*errstr = "could not allocate memory";
	return -1;
      }
      if (efftabR[e].name)
        {
	  efftabR[e].obuf = (st_sample_t *) malloc(ST_BUFSIZ * 
						      sizeof(st_sample_t));
	  if (efftabR[e].obuf == NULL) {
	    *errstr = "could not allocate memory";
	    return -1;
	  }
        }
    }

  /*
   * Just like errno, we must set st_errno to known values before
   * calling I/O operations.
   */
  informat->st_errno = 0;
  outformat->st_errno = 0;

  /* Run input data through effects and get more until olen == 0 */
  do {

    efftab[0].olen = (*informat->h->read)(informat,
					     efftab[0].obuf, 
					     (st_ssize_t)ST_BUFSIZ);
    efftab[0].odone = 0;

    /* Done. */
    if (efftab[0].olen == 0)
      break;

    /* Oops, someone may have aborted it. */
    if (sox_request_aborted(r)) {
      /* Somebody else aborted, so just return. */
      fclose(informat->fp);
      fclose(outformat->fp);
      return 0;
    }

    /* Change the volume of this initial input data if needed. */
    if (r->do_volume)
      r->clipped += volumechange(efftab[0].obuf, efftab[0].olen,
				 (double) r->volume);

    /* mark chain as empty */
    for(e = 1; e < neffects; e++)
      efftab[e].odone = efftab[e].olen = 0;

    flowstatus = flow_effect_out(r);

    /* Negative flowstatus says no more output will ever be generated. */
    if (flowstatus < 0 || outformat->st_errno)
      break;

  } while (1); /* break; efftab[0].olen == 0 */

  /* If st_errno is set, probably want to barf. */
  if (outformat->st_errno) {
    GalUtil_Warn(outformat->st_errstr);
    *errstr = outformat->st_errstr;
    return -1;
  }

  /* Drain the effects out first to last,
   * pushing residue through subsequent effects */
  /* oh, what a tangled web we weave */
  
  for(f = 1; f < neffects; f++)
    {
      while (1) {

	/* Done. */
	if (drain_effect(f, r) == 0)
	  break;          /* out of while (1) */

	if (efftab[neffects-1].olen > 0)
	  (* outformat->h->write)(outformat, efftab[neffects-1].obuf,
				  (st_ssize_t) efftab[neffects-1].olen);
	fflush(outformat->fp);

	/* Also done. */
	if (efftab[f].olen != ST_BUFSIZ)
	  break;
      }
    }


  /* Very Important:
   * Effect stop is called BEFORE files close.
   * Effect may write out more data after.
   */

  for (e = 1; e < neffects; e++) {
    (* efftab[e].h->stop)(&efftab[e]);
    if (efftabR[e].name)
      (* efftabR[e].h->stop)(&efftabR[e]);
  }

  /* If problems closing input file, just warn user since
   * we are exiting anyways.
   */
  if ((* informat->h->stopread)(informat) == ST_EOF)
    st_warn(informat->st_errstr);
  fclose(informat->fp);

  /* problem closing output file, just warn user since we
   * are exiting anyways.
   */
  if ((* outformat->h->stopwrite)(outformat) == ST_EOF)
    st_warn(outformat->st_errstr);
  
  fclose(outformat->fp);

  if (r->do_volume && r->clipped > 0)
    st_report("Volume change clipped %d samples", r->clipped);
  return 0;
}

static int flow_effect_out(sox_request *r)
{
  int e, havedata, flowstatus = 0;
  ft_t outformat = r->outformat;
  struct st_effect *efftab = r->efftab;
  int neffects = r->neffects;
  
  do {
    /* run entire chain BACKWARDS: pull, don't push.*/
    /* this is because buffering system isn't a nice queueing system */
    for(e = neffects - 1; e > 0; e--)
      {
	flowstatus = flow_effect(e, r);
	if (flowstatus)
	  break;
      }

    /* If outputing and output data was generated then write it */
    if ((efftab[neffects-1].olen>efftab[neffects-1].odone))
      {
	(* outformat->h->write)(outformat,
				efftab[neffects-1].obuf,
				(st_ssize_t)efftab[neffects-1].olen);
	efftab[neffects-1].odone = efftab[neffects-1].olen;
	fflush(outformat->fp);
      }    

    if (outformat->st_errno)
      {
	st_warn("Error writing: %s",outformat->st_errstr);
	break;
      }

    /* If any effect will never again produce data, give up.  This
     * works because of the pull status: the effect won't be able to
     * shut us down until all downstream buffers have been emptied.
     */
    if (flowstatus < 0)
      break;

    /* if stuff still in pipeline, set up to flow effects again */
    havedata = 0;
    for(e = 0; e < neffects - 1; e++)
      if (efftab[e].odone < efftab[e].olen) {
	havedata = 1;
	break;
      }
  } while (havedata);

  return flowstatus;
}

static int flow_effect(int e, sox_request *r)
{
  st_ssize_t i, done, idone, odone, idonel, odonel, idoner, odoner;
  st_sample_t *ibuf, *obuf;
  int effstatus;
  struct st_effect *efftab = r->efftab;
  struct st_effect *efftabR = r->efftabR;
  st_sample_t *ibufl = r->ibufl;
  st_sample_t *ibufr = r->ibufr;
  st_sample_t *obufl = r->obufl;
  st_sample_t *obufr = r->obufr;


  /* I have no input data ? */
  if (efftab[e-1].odone == efftab[e-1].olen)
    return 0;

  if (! efftabR[e].name) {
    /* No stereo data, or effect can handle stereo data so
     * run effect over entire buffer.
     */
    idone = efftab[e-1].olen - efftab[e-1].odone;
    odone = ST_BUFSIZ;
    effstatus = (* efftab[e].h->flow)(&efftab[e],
				      &efftab[e-1].obuf[efftab[e-1].odone],
				      efftab[e].obuf, &idone, &odone);
    efftab[e-1].odone += idone;
    efftab[e].odone = 0;
    efftab[e].olen = odone;
    done = idone + odone;
  } else {

    /* Put stereo data in two seperate buffers and run effect
     * on each of them.
     */
    idone = efftab[e-1].olen - efftab[e-1].odone;
    odone = ST_BUFSIZ;
    ibuf = &efftab[e-1].obuf[efftab[e-1].odone];
    for(i = 0; i < idone; i += 2) {
      ibufl[i/2] = *ibuf++;
      ibufr[i/2] = *ibuf++;
    }

    /* left */
    idonel = (idone + 1)/2;         /* odd-length logic */
    odonel = odone/2;
    effstatus = (* efftab[e].h->flow)(&efftab[e],
				      ibufl, obufl, &idonel, &odonel);

    /* right */
    idoner = idone/2;               /* odd-length logic */
    odoner = odone/2;
    effstatus = (* efftabR[e].h->flow)(&efftabR[e],
				       ibufr, obufr, &idoner, &odoner);

    obuf = efftab[e].obuf;
    /* This loop implies left and right effect will always output
     * the same amount of data.
     */
    for(i = 0; i < odoner; i++) {
      *obuf++ = obufl[i];
      *obuf++ = obufr[i];
    }
    efftab[e-1].odone += idonel + idoner;
    efftab[e].odone = 0;
    efftab[e].olen = odonel + odoner;
    done = idonel + idoner + odonel + odoner;
  }
  if (done == 0) {    
    st_fail_errno(r->outformat, ST_EINVAL, "Effect took & gave no samples!");
    return -1;
  }
  if (effstatus == ST_EOF)
    return -1;
  return 1;
}

static int drain_effect(int e, sox_request *r)
{
  st_ssize_t i, olen, olenl, olenr;
  st_sample_t *obuf;
  struct st_effect *efftab = r->efftab;
  struct st_effect *efftabR = r->efftabR;
  st_sample_t *obufl = r->obufl;
  st_sample_t *obufr = r->obufr;

  if (! efftabR[e].name) {
    efftab[e].olen = ST_BUFSIZ;
    (* efftab[e].h->drain)(&efftab[e],efftab[e].obuf,
			   &efftab[e].olen);
  }
  else {
    olen = ST_BUFSIZ;

    /* left */
    olenl = olen/2;
    (* efftab[e].h->drain)(&efftab[e], obufl, &olenl);

    /* right */
    olenr = olen/2;
    (* efftab[e].h->drain)(&efftabR[e], obufr, &olenr);

    obuf = efftab[e].obuf;
    /* This loop implies left and right effect will always output
     * the same amount of data.
     */
    for(i = 0; i < olenr; i++) {
      *obuf++ = obufl[i];
      *obuf++ = obufr[i];
    }
    efftab[e].olen = olenl + olenr;
  }
  return(efftab[e].olen);
}

/*
 * If no effect given, decide what it should be.
 * Smart ruleset for multiple effects in sequence.
 *      Puts user-specified effect in right place.
 */
static int
checkeffect(sox_request *r, char **errstr)
{
        int i;
        int needchan = 0, needrate = 0, haschan = 0, hasrate = 0;
        int effects_mask = 0;
	ft_t informat = r->informat;
	ft_t outformat = r->outformat;
	eff_t effp, effpR, effpU;
	  
        needrate = (informat->info.rate != outformat->info.rate);
        needchan = (informat->info.channels != outformat->info.channels);

        for (i = 0; i < r->nuser_effects; i++)
        {
            if (r->user_efftab[i].h->flags & ST_EFF_CHAN)
            {
                haschan++;
            }
            if (r->user_efftab[i].h->flags & ST_EFF_RATE)
            {
                hasrate++;
            }
        }

        if (haschan > 1) {
	  *errstr = "Can not specify multiple effects that modify channel #";
	  return -1;
	}
	
        if (hasrate > 1)
            st_report("Can not specify multiple effects that change sample rate");

        /* --------- add the effects ------------------------ */

        /* efftab[0] is always the input stream and always exists */
        r->neffects = 1;

        /* If reducing channels then its faster to run all effects
         * after the avg effect.
         */
        if (needchan && !(haschan) &&
            (informat->info.channels > outformat->info.channels))
        {
	  effp = &(r->efftab[r->neffects]);
            /* Find effect and update initial pointers */
            st_geteffect(effp, "avg");

            /* give default opts for added effects */
            (* effp->h->getopts)(effp,(int)0,(char **)0);

            /* Copy format info to effect table */
            effects_mask = st_updateeffect(effp, informat,
                                           outformat, effects_mask);

            r->neffects++;
        }

        /* If reducing the number of samples, its faster to run all effects
         * after the resample effect.
         */
        if (needrate && !(hasrate) &&
            (informat->info.rate > outformat->info.rate))
        {
	  effp = &(r->efftab[r->neffects]);
	  effpR = &(r->efftabR[r->neffects]);
	  
	  st_geteffect(effp, "resample");

            /* set up & give default opts for added effects */
            (* effp->h->getopts)(effp,(int)0,(char **)0);

            /* Copy format info to effect table */
            effects_mask = st_updateeffect(effp, informat,
                                           outformat, effects_mask);

            /* Rate can't handle multiple channels so be sure and
             * account for that.
             */
            if (effp->ininfo.channels > 1)
            {
                memcpy(effpR, effp, sizeof(struct st_effect));
            }

            r->neffects++;
        }

        /* Copy over user specified effects into real efftab */
        for(i = 0; i < r->nuser_effects; i++)
        {
	  effp = &(r->efftab[r->neffects]);
	  effpR = &(r->efftabR[r->neffects]);
	  effpU = &(r->user_efftab[i]);
	  
            memcpy(effp, effpU, sizeof(struct st_effect));

            /* Copy format info to effect table */
            effects_mask = st_updateeffect(effp, informat,
                                           outformat, effects_mask);

            /* If this effect can't handle multiple channels then
             * account for this.
             */
            if ((effp->ininfo.channels > 1) &&
                !(effp->h->flags & ST_EFF_MCHAN))
            {
                memcpy(effpR, effp, sizeof(struct st_effect));
            }

            r->neffects++;
        }

        /* If rate effect hasn't been added by now then add it here.
         * Check adding rate before avg because its faster to run
         * rate on less channels then more.
         */
        if (needrate && !(effects_mask & ST_EFF_RATE))
	  {
	    effp = &(r->efftab[r->neffects]);
	    effpR = &(r->efftabR[r->neffects]);
	    
	    st_geteffect(effp, "resample");

            /* set up & give default opts for added effects */
            (* effp->h->getopts)(effp,(int)0, (char **)0);

            /* Copy format info to effect table */
            effects_mask = st_updateeffect(effp, informat,
                                           outformat, effects_mask);

            /* Rate can't handle multiple channels so be sure and
             * account for that.
             */
            if (effp->ininfo.channels > 1)
            {
                memcpy(effpR, effp, sizeof(struct st_effect));
            }

            r->neffects++;
        }

        /* If code up until now still hasn't added avg effect then
         * do it now.
         */
        if (needchan && !(effects_mask & ST_EFF_CHAN))
        {
	  effp = &(r->efftab[r->neffects]);
            st_geteffect(effp, "avg");

            /* set up & give default opts for added effects */
            (* effp->h->getopts)(effp,(int)0,(char **)0);

            /* Copy format info to effect table */
            effects_mask = st_updateeffect(effp, informat,
                                           outformat, effects_mask);

            r->neffects++;
        }
	return 0;
}

static st_sample_t volumechange(st_sample_t *buf, st_ssize_t ct, 
	                        double vol)
{
        double y;
        st_sample_t *p,*top;
        st_ssize_t clips=0;

        p = buf;
        top = buf+ct;
        while (p < top) {
            y = vol * *p;
            if (y < -2147483647.0) {
                y = -2147483647.0;
                clips++;
            }
            else if (y > 2147483647.0) {
                y = 2147483647.0;
                clips++;
            }
            *p++ = y + 0.5;
        }
        return clips;
}

static int filetype(fd)
int fd;
{
        struct stat st;

        fstat(fd, &st);

        return st.st_mode & S_IFMT;
}

#if 0
/* called from util.c:fail */
void cleanup(void) {
        /* Close the input file and outputfile before exiting*/
        if (informat && informat->fp)
                fclose(informat->fp);
        if (outformat && outformat->fp) {
                fclose(outformat->fp);
                /* remove the output file because we failed, if it's ours. */
                /* Don't if its not a regular file. */
                if (filetype(fileno(outformat->fp)) == S_IFREG)
                    unlink(outformat->filename);
        }
}
#endif
void cleanup(void)
{
}
