/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/* 
   db_server.c
   -----------

   USAGE:
   db_server -dblogin <username/password>[@host] -maxrows <rows>
 
   -dblogin defaults to "flights/flights@muddy"
   -maxrows defaults to 1000

   DESCRIPTION:
   db_server provides connectivity to an oracle database.  db_server expects a frame like this:

   {c foobar
      :query "select airline, flight_id, depart_time, arrive_time, stop_nonstop, meal, equip_type from airline_flight where depart_airport = 'BOS' and arrive_airport = 'LAX'"
      :maxrows 10 }

   :query (required) contains the sql statement to be sent to the database
   :maxrows (optional) contains the maximum number of rows to be returned.  Values exceeding the maxrows limit set at startup are ignored

   db_server returns a frame like this:

   {c foobar
       :query_result {c result
                    :column_names ( "airline"
                                    "flight_id"
                                    "depart_time"
                                    "arrive_time"
                                    "stop_nonstop"
                                    "meal"
                                    "equip_type" ) 
                    :nfound 2
                    :values ( ( "AA"
                                "11"
                                "815"
                                "1148"
                                "0"
                                "B"
                                "757" ) 
                              ( "UA"
                                "171"
                                "715"
                                "1040"
                                "0"
                                "BS"
                                "762" )  )  } }
    
    :column_names is a list of column names (strings)  for the reply tuples
    :nfound is an integer indicating the number of rows returned
    :values is a list of lists containing the tuples from the database

    BUGS:
    --  reply tuples are limited to 128 columns. 

    UPDATE LOG:
    12/24/98 SAM handles multiple connections.
    12/28/98 HUGO new reply frame format
*/


#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include "galaxy/galaxy_all.h"
#define SERVER_FUNCTIONS_INCLUDE "db_server_server.h"
#define USE_SERVER_DATA
#include "galaxy/server_functions.h"
#include "sql.h"

#ifdef __cplusplus
}
#endif

#define MAX_COLS 128
#define TYPE_NULL 0
#define TYPE_INT 1
#define TYPE_STRING 2

#define ROWBUF_SIZE 2048

char rowbuf[ROWBUF_SIZE];
char *username_passwd;
int server_max_rows;
int request_max_rows;
Gal_Object *tlist;
static int ncols = 0;

Gal_Object build_column_list(char *str);
Gal_Object list_from_line(char *str);
Gal_Frame process_query(char *sql_query);
Gal_Frame sql_query(Gal_Frame frame, void *server_data);
void init_sql(void);
void exit_sql();
void sig_handler(int);

static char *oas[] = {
  "-dblogin <un/pw>", "<username>/<password>[@host] for database login", "flights/flights@muddy",
  "-maxrows <mr>", "maximum number of rows to be returned", "1000",
  NULL
};

/* required operation for a server */
Gal_Frame reinitialize(Gal_Frame frame, void *server_data)
{
  return frame;
}

/* etablishes connection to the database */
void init_sql() {
  sql_connect(username_passwd);
  fflush(stdout);
}

/* disconnects from the database */
void exit_sql() {
  sql_disconnect();
  fflush(stdout);
}

/* pulls key/value pairs from input frame, calls process_query, and sends reply frame*/
Gal_Frame sql_query(Gal_Frame frame, void *server_data)
{
  char *sql_string;
  char *key = ":query";
  Gal_Frame result_frame;

  printf("\nreceived request frame:\n");
  Gal_PPFrame(frame); 
  printf("\n");

  sql_string = Gal_GetString(frame, key); 
  request_max_rows = Gal_GetInt(frame, ":maxrows");
  if((request_max_rows <= 0)||(request_max_rows > server_max_rows))
    request_max_rows = server_max_rows;

  result_frame = process_query(sql_string);
  
  Gal_SetProp(frame, ":query_result", Gal_FrameObject(result_frame));
  Gal_DelProp(frame, key);
  Gal_DelProp(frame, ":maxrows");
  
  printf("\nsending reply frame:\n");
  Gal_PPFrame(frame);
  request_max_rows = server_max_rows;
  return frame;
}

/* for debugging with purify */
static void exit_for_purify(void)
{
  exit_sql();
  free(tlist);
  exit(0);
}

/* sends query to database and processes resulting tuples */
Gal_Frame process_query(char *sql_query) {
  int nlines;
  Gal_Frame nf;
  Gal_Object no;
  char str[ROWBUF_SIZE];
  Gal_Object column_list;

  ncols = 0;
  nlines = 0;

  start_query(sql_query);
  fetch_header(rowbuf);
  column_list = build_column_list(rowbuf);

  while(fetch_row(rowbuf) > 0)  {
    strncpy(str, rowbuf, ROWBUF_SIZE);
    no = list_from_line(str);
    if(no && (nlines < request_max_rows))
      tlist[nlines++] = no;
  }

  nf = Gal_MakeFrame("result", GAL_CLAUSE);
  Gal_SetProp(nf,":column_names", column_list);
  Gal_SetProp(nf,":nfound",Gal_IntObject(nlines));
  Gal_SetProp(nf,":values",Gal_ListObject(tlist, nlines));

  return nf;
}

/* SAM 8/30/99: Copied string_downcase from sls_util since
   I didn't want to have to link against it just because of
   this one function. */

static void string_downcase (char *string)
{ 
  unsigned int i;

  for (i = 0; i < strlen(string); i++)
    string[i] = tolower ((int)string[i]);
}

/* pulls column names from header
   and puts them in a list object.
   header looks like ( xxx yyy ) */
Gal_Object build_column_list(char *str)
{
  char *cp;
  char token[128];
  char buf[ROWBUF_SIZE];
  Gal_Object colnames[MAX_COLS];
  Gal_Object collist;

  strncpy(buf, rowbuf, ROWBUF_SIZE);
  cp = buf;
  ncols = 0;

  while(((cp=Gal_NextToken(cp,token,128)) != NULL) && (ncols < MAX_COLS)) {
    if(token[0] == '(') continue;
    if(token[0] == ')') continue;

    string_downcase(token);
    colnames[ncols++] = Gal_StringObject(token);
  }
  collist = Gal_ListObject(colnames, ncols);
  return collist;
}

/* grabs the next token from the result set buffer.  differentiates
   between strings and ints */
char *get_next_value(char *str,int *type,int *ival,char *cval)
{
  char *cp;
  char temp[1028];
  int i=0;
  for(cp=str;*cp;cp++) if(!isspace((int)*cp)) break;
  if(*cp == 0) return(NULL);
  if(*cp == '"'){
    cp++;
    while(*cp && *cp != '"') temp[i++] = *cp++;
    if(*cp == 0) return(NULL);
    cp++;
    temp[i] = 0;
    *type = TYPE_STRING;
    for(i = strlen(temp)-1;i>0;i--){
      if(isspace((int)temp[i])) temp[i] = 0;
      else break;
    }
    strcpy(cval,temp);
    return(cp);
  }
  else{
    cp = Gal_NextToken(cp,temp, 1028);
    i = atoi(temp);
    *type = TYPE_INT;
    *ival = i;
    return(cp);
  }
  return(NULL);
}

/* generates a list object from a result row */
Gal_Object list_from_line(char *str)
{
  char *cp = str;
  int i;
  char cval[128];
  int ival;
  int type;
  Gal_Object nf;
  Gal_Object no[MAX_COLS];
  
  for(i=0;i<ncols;i++){
    cp = get_next_value(cp,&type,&ival,cval);
    if(cp == NULL){
      printf("Bad column %s\n",str);
      return(NULL);
    }
    if(type == TYPE_INT){
      no[i] = Gal_IntObject(ival);
    }
    else if(type == TYPE_STRING){
      no[i] = Gal_StringObject(cval);
    }

  }
  nf = Gal_ListObject(no, ncols);
  return(nf);
}

/* signal handler that disconnects server from database and exits */
void sig_handler(int signum) {
  exit_sql();
  free(tlist);
  exit(0);
}

/* main() required to implement multiple connections. */
/* SAM 9/22/99: No longer. */

void *_GalSS_init_server(GalIO_ServerStruct *s, int argc, char **argv)
{
  
  /* contents of _GalSS_init_server */
  Gal_AddSignalHandler(SIGINT, sig_handler);
  
  if (GalUtil_OACheckUsage(argc, argv, oas, NULL) == 0)
    exit(-1);

  GalUtil_OAExtract(argc, argv, oas, "-dblogin", GAL_OA_STRING, &username_passwd); 
  GalUtil_OAExtract(argc, argv, oas, "-maxrows", GAL_OA_INT, &server_max_rows);
  request_max_rows = server_max_rows;

  tlist  = (Gal_Object*)malloc(server_max_rows*sizeof(Gal_Object));
  if (tlist == NULL) {
    fflush(stdout);
    fprintf(stderr, "%s:  malloc() failed.\n", argv[0]);
    exit(1);
  }

  init_sql();
  return (void *) NULL;
}

