/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

#include "galaxy/sysdep.h"
#include <stdio.h>
#include "pg_interface.h"

char *pgport = NULL;
char *pgoptions = NULL;
char *pgtty = NULL;
    
PGconn     *conn;
PGresult   *res;

char *cursor = "DECLARE mycursor CURSOR FOR "; 
char *queryBuff[2048];

void exit_nicely(PGconn *conn)
{
    PQfinish(conn);
    exit(1);
}


int begin_query(char *query)
{
  char tok[32];
  int transaction_type;
  
  /* First, we figure out if we've got a select,
     an insert, etc. */
  printf("Sscanf returned %d\n", sscanf(query, " %32s", tok));
  printf("Query type is `%s'\n", tok);
  fflush(stdout);
  
  if (!_gal_strcasecmp(tok, "select")) {
    transaction_type = PG_SELECT_QUERY;
  } else if (!_gal_strcasecmp(tok, "insert")) {
    transaction_type = PG_INSERT_QUERY;
  } else {
    GalUtil_Warn("Command was neither an insert nor a select");
    transaction_type = PG_OTHER_QUERY;
  }

  /* start a transaction block */
  res = PQexec(conn, "BEGIN");
  if (PQresultStatus(res) != PGRES_COMMAND_OK)
    {
      fprintf(stderr, "BEGIN command failed\n");
      PQclear(res);
      exit_nicely(conn);
    }

  /*
   * should PQclear PGresult whenever it is no longer needed to avoid
   * memory leaks
   */
  PQclear(res);

  /*
   * fetch instances from the pg_database, the system catalog of
   * databases
   */

  switch (transaction_type) {
  case PG_SELECT_QUERY:
    /* Do this for a select. */

    strcpy((char *)queryBuff, cursor);
    strcat((char *)queryBuff, query);

    res = PQexec(conn, (char*)queryBuff); 
    if (PQresultStatus(res) != PGRES_COMMAND_OK) {
      fprintf(stderr, "DECLARE CURSOR command failed\n");
      PQclear(res);
      exit_nicely(conn);
    }
    PQclear(res);
    res = PQexec(conn, "FETCH ALL in mycursor");
    if (PQresultStatus(res) != PGRES_TUPLES_OK) {
      fprintf(stderr, "FETCH ALL command didn't return tuples properly\n");
      PQclear(res);
      exit_nicely(conn);
    }
    return PG_SELECT_QUERY;
    
  case PG_INSERT_QUERY:
  case PG_OTHER_QUERY:
    res = PQexec(conn, query);
    if (PQresultStatus(res) != PGRES_COMMAND_OK) {
      fprintf(stderr, "Other command failed: %s: %s\n", PQresStatus(PQresultStatus(res)), PQresultErrorMessage(res));
      PQclear(res);
      exit_nicely(conn);
    } else {
      printf("Query succeeded\n");
    }
    return transaction_type;
  default:
    return PG_OTHER_QUERY;
  }
}


int end_query(int query_type) {

  if (query_type == PG_SELECT_QUERY) {
    /* close the cursor */
    res = PQexec(conn, "CLOSE mycursor");
    PQclear(res);
  }

  /* commit the transaction */
  /* res = PQexec(conn, "COMMIT");
     PQclear(res); */
  
  return 0;
}

void sql_connect(char *username_passwd)
{
  /* SAM 3/30/01: username_passwd needs to be parsed in order to
     find <username>/<password>/<db>[@host]. */

  char *pghost = (char *) NULL;
  char *pguser = (char *) NULL;
  char *pgpasswd = (char *) NULL;
  char *pgdb = (char *) NULL;
  char *cp_root = (char *) NULL;

  if (username_passwd) {
    char *tmp_conninfo = strdup(username_passwd);
    char *cp;

    cp_root = tmp_conninfo;

    cp = strrchr(tmp_conninfo, '@');
    /* Insert null to delimit host. */
    if (cp) {
      cp[0] = '\0';
      pghost = cp + 1;
    }

    /* Now, get the user. */
    cp = strchr(tmp_conninfo, '/');
    if (!cp) {
      fprintf(stderr, "Ill-formed user/pass/db string\n");
      exit(1);
    }
    if (cp != tmp_conninfo) {
      /* Leave it NULL if the slash is the first thing. */
      pguser = tmp_conninfo;
    }
    cp[0] = '\0';
    tmp_conninfo = cp + 1;

    /* Now, get the password. */
    cp = strchr(tmp_conninfo, '/');
    if (!cp) {
      fprintf(stderr, "Ill-formed user/pass/db string\n");
      exit(1);
    }
    if (cp != tmp_conninfo) {
      /* Leave it NULL if the slash is the first thing. */
      pgpasswd = tmp_conninfo;
    }
    cp[0] = '\0';

    /* Set the db if it's a nonzero string. */
    if (strlen(cp + 1) != 0) {
      pgdb = cp + 1;
    }
  }
  
  conn = PQsetdbLogin(pghost, pgport, pgoptions, pgtty,
		      pgdb, pguser, pgpasswd);
  /*
   * check to see that the backend connection was successfully made
   */
  if (PQstatus(conn) == CONNECTION_BAD) {
    char *dbName = pgdb;

    if (!dbName) dbName = "<default>";
    
    fprintf(stderr, "Connection to database '%s' failed.\n", dbName);
    fprintf(stderr, "%s", PQerrorMessage(conn));
    exit_nicely(conn);
  }

  if (cp_root)
    free(cp_root);

  /* debug = fopen("/tmp/trace.out","w"); */
  /* PQtrace(conn, debug);  */
}


void sql_disconnect(void)
{
    /* close the connection to the database */
    PQfinish(conn);
}


void print_tuples(void)
{
  int i, j, nFields;

    /* first, print out the attribute names */
    nFields = PQnfields(res);
    for (i = 0; i < nFields; i++)
        printf("%-10s", PQfname(res, i));
    printf("\n\n");

    /* next, print out the instances */
    for (i = 0; i < PQntuples(res); i++)
    {
        for (j = 0; j < nFields; j++)
            printf("%-10s", PQgetvalue(res, i, j));
        printf("\n");
    }
    printf("\nNumber of tuples:  %d\n",PQntuples(res));
}

static int pg_num_attributes = 0;
static Gal_Object *pg_attributes = (Gal_Object *) NULL;
#define GAL_PG_ATTRIBUTE_INCREMENT 64

void insure_num_attributes(int num_attributes)
{
  while (num_attributes >= pg_num_attributes) {
    if (!pg_attributes) {
      pg_attributes = (Gal_Object *) calloc(GAL_PG_ATTRIBUTE_INCREMENT,
					    sizeof(Gal_Object));
      pg_num_attributes = GAL_PG_ATTRIBUTE_INCREMENT;
    } else {
      pg_attributes = (Gal_Object *) realloc(pg_attributes, ((pg_num_attributes + GAL_PG_ATTRIBUTE_INCREMENT) * sizeof(Gal_Object)));
      pg_num_attributes += GAL_PG_ATTRIBUTE_INCREMENT;
    }
  }
}

Gal_Object build_attribute_list()
{
  Gal_Object attribute_list;
  int i,  nFields;

  nFields = PQnfields(res);
  insure_num_attributes(nFields);
  for (i = 0; i < nFields; i++) {
    pg_attributes[i] = Gal_StringObject(PQfname(res, i));
  }
  attribute_list = Gal_ListObject(pg_attributes, nFields);
  return attribute_list;
}

Gal_Object build_tuple_list(int index)
{
  Gal_Object tuple_list;
  int nFields,j;

  nFields = PQnfields(res);
  insure_num_attributes(nFields);
  for (j = 0; j < nFields; j++) {  
    pg_attributes[j] = Gal_StringObject(PQgetvalue(res, index, j));
  }

  tuple_list = Gal_ListObject(pg_attributes, nFields);
  return tuple_list;
}


int get_rowcount()
{
  return PQntuples(res);
}
