/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

/* 
   pgdb_sever.c
   -----------

   USAGE:
   pgdb_sever -dblogin <username/password>[@host] -maxrows <rows>
 
   -dblogin defaults to "flights/flights@muddy"
   -maxrows defaults to 1000

   DESCRIPTION:
   pgdb_sever provides connectivity to an oracle database.  pgdb_sever expects a frame like this:

   {c foobar
      :query "select airline, flight_id, depart_time, arrive_time, stop_nonstop, meal, equip_type from airline_flight where depart_airport = 'BOS' and arrive_airport = 'LAX'"
      :maxrows 10 }

   :query (required) contains the sql statement to be sent to the database
   :maxrows (optional) contains the maximum number of rows to be returned.  Values exceeding the maxrows limit set at startup are ignored

   pgdb_sever returns a frame like this:

   {c foobar
       :query_result {c result
                    :column_names ( "airline"
                                    "flight_id"
                                    "depart_time"
                                    "arrive_time"
                                    "stop_nonstop"
                                    "meal"
                                    "equip_type" ) 
                    :nfound 2
                    :values ( ( "AA"
                                "11"
                                "815"
                                "1148"
                                "0"
                                "B"
                                "757" ) 
                              ( "UA"
                                "171"
                                "715"
                                "1040"
                                "0"
                                "BS"
                                "762" )  )  } }
    
    :column_names is a list of column names (strings)  for the reply tuples
    :nfound is an integer indicating the number of rows returned
    :values is a list of lists containing the tuples from the database

    BUGS:
    --  reply tuples are limited to 128 columns. 

    UPDATE LOG:
    12/24/98 SAM handles multiple connections.
    12/28/98 HUGO new reply frame format
*/


#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include "galaxy/galaxy_all.h"
#define SERVER_FUNCTIONS_INCLUDE "pgdb_server_server.h"
#define USE_SERVER_DATA
#include "galaxy/server_functions.h"
#include <postgres.h>
#include "pg_interface.h"

#ifdef __cplusplus
}
#endif

/* #define MAX_COLS 128 */
#define TYPE_NULL 0
#define TYPE_INT 1
#define TYPE_STRING 2
#define ROWBUF_SIZE 2048

char rowbuf[ROWBUF_SIZE];
char *username_passwd;
int server_max_rows;
int request_max_rows;
Gal_Object *tlist;
static int ncols = 0;

Gal_Frame process_query(char *sql_query);
Gal_Frame sql_query(Gal_Frame frame, void *server_data);
void init_sql(void);
void exit_sql();
void sig_handler(int);

static char *oas[] = {
  "-dblogin <un/pw/db@host>", "<username>/<password>/<db>[@host] for database login", NULL,
  "-maxrows <mr>", "maximum number of rows to be returned", "1000",
  NULL
};

/* required operation for a server */
Gal_Frame reinitialize(Gal_Frame frame, void *server_data)
{
  return frame;
}

/* etablishes connection to the database */
void init_sql() {
  sql_connect(username_passwd);
  fflush(stdout);
}

/* disconnects from the database */
void exit_sql() {
  sql_disconnect();
  fflush(stdout);
}

/* pulls key/value pairs from input frame, calls process_query, and sends reply frame*/
Gal_Frame sql_query(Gal_Frame frame, void *server_data)
{
  char *sql_string;
  char *key = ":query";
  Gal_Frame result_frame;

  printf("\nreceived request frame:\n");
  Gal_PPFrame(frame); 
  printf("\n");

  sql_string = Gal_GetString(frame, key); 
  request_max_rows = Gal_GetInt(frame, ":maxrows");
  if((request_max_rows <= 0)||(request_max_rows > server_max_rows))
    request_max_rows = server_max_rows;

  result_frame = process_query(sql_string);
  
  Gal_SetProp(frame, ":query_result", Gal_FrameObject(result_frame));
  Gal_DelProp(frame, key);
  Gal_DelProp(frame, ":maxrows");
  
  printf("\nsending reply frame:\n");
  Gal_PPFrame(frame);
  request_max_rows = server_max_rows;
  return frame;
}

/* for debugging with purify */
static void exit_for_purify(void)
{
  exit_sql();
  free(tlist);
  exit(0);
}

/* sends query to database and processes resulting tuples */
Gal_Frame process_query(char *sql_query) {
  int nlines, i;
  Gal_Frame nf;
  /* Gal_Object no;
     char str[1024]; */
  Gal_Object column_list;
  int query_type;
  
  ncols = 0;
  nlines = 0;

  query_type = begin_query(sql_query);

  switch (query_type) {
  case PG_SELECT_QUERY:
    column_list = build_attribute_list();

    nlines = get_rowcount();
    if (nlines > request_max_rows)
      nlines = request_max_rows;

    for(i=0; i< nlines; i++) {
      tlist[i] = build_tuple_list(i); 
    }

    end_query(query_type);

    nf = Gal_MakeFrame("result", GAL_CLAUSE);
    Gal_SetProp(nf,":column_names", column_list);
    Gal_SetProp(nf,":nfound",Gal_IntObject(nlines));
    Gal_SetProp(nf,":values",Gal_ListObject(tlist,nlines));
    break;
  case PG_INSERT_QUERY:
    end_query(query_type);
  case PG_OTHER_QUERY:
  default:
    nf = Gal_MakeFrame("result", GAL_CLAUSE);
    Gal_SetProp(nf,":nfound",Gal_IntObject(0));
    Gal_SetProp(nf,":values",Gal_ListObject((Gal_Object *) NULL, 0));
    break;    
  }
  return nf;
}

/* signal handler that disconnects server from database and exits */
void sig_handler(int signum) {
  exit_sql();
  free(tlist);
  exit(0);
}

/* main() required to implement multiple connections. */
/* SAM 9/22/99: no longer. */

void *_GalSS_init_server(GalIO_ServerStruct *s, int argc, char **argv)
{
  
  /* contents of _GalSS_init_server */
  Gal_AddSignalHandler(SIGINT, sig_handler);
  
  if (GalUtil_OACheckUsage(argc, argv, oas, NULL) == 0)
    exit(-1);

  GalUtil_OAExtract(argc, argv, oas, "-dblogin", GAL_OA_STRING, &username_passwd); 
  GalUtil_OAExtract(argc, argv, oas, "-maxrows", GAL_OA_INT, &server_max_rows);
  request_max_rows = server_max_rows;

  tlist  = (Gal_Object*)malloc(server_max_rows*sizeof(Gal_Object));
  if (tlist == NULL) {
    fflush(stdout);
    fprintf(stderr, "%s:  malloc() failed.\n", argv[0]);
    exit(1);
  }

  init_sql();
  return (void *) NULL;
}











