# This file (c) Copyright 1998 - 2002 The MITRE Corporation
# 
# This file is part of the Galaxy Communicator system. It is licensed
# under the conditions described in the file LICENSE in the root 
# directory of the Galaxy Communicator system.

# Now, we try to write the mapping support functions.
# The code in this file will be evaluated in a context
# which has Target, Source, Debug and Altered set.

from transducer_core import CanonicalizeValue, \
     ExtractCanonicalizedValue, OperationError, SetCanonicalizedValue, \
     _I_ExtractCanonicalizedValue, _I_SetCanonicalizedValue

from Galaxy import ObjectEqual, GAL_LIST, GetObjectType, GAL_FRAME, \
     OPr, PP_TYPE, Frame

# Simple predicates.

def In(a, b):
    return a in b

def Exists(a):
    return a is not None

def Eq(a, b):
    return ObjectEqual(a, b)

def Neq(a, b):
    return not ObjectEqual(a, b)

def Geq(a, b):
    try:
	return a > b
    except:
	return 0

def Leq(a, b):
    try:
	return a < b
    except:
	return 0

def Not(a):
    return not a

# Basic testing stuff.

# I think Some and Every need to be like Test.

def Some(a, subpath, testfn, *args, **kw):
    return _Test(a, subpath, _SomeCore, "Some", testfn, args, kw)

def Every(a, subpath, testfn, *args, **kw):
    return _Test(a, subpath, _EveryCore, "Every", testfn, args, kw)

def Test(a, testfn, *args, **kw):
    return _Test(a, None, _TestCore, "Test", testfn, args, kw)

def _Test(path, subpath, core_fn, core_name, testfn, args, kw):
    global Source
    if Debug:
	print "== Matching", core_name, path, subpath, testfn, args
    path = CanonicalizePath(Source, path)
    try:
	v = ExtractCanonicalizedValue(path)
    except OperationError:
	if Debug: print "== failed (error)"
	return 0
    if path.object is not Source:
	old_source = Source
	source_to_target = 1
	Source = path.object
    else:
	source_to_target = 0
    r = core_fn(subpath, testfn, v, args)
    if source_to_target:
	Source = old_source
    return r

def _TestCore(subpath, testfn, v, args):
    r = apply(testfn, (v,) + args)
    if Debug:
	if r: print "== succeeded"
	else: print "== failed"
    return r

def _SomeCore(subpath, testfn, v, args):
    if GetObjectType(v) is not GAL_LIST:
	if Debug:
	    print "== failed (error)"
	return 0
    r = 0
    p = CanonicalizePath(None, subpath)
    for elt in v:
	try:
	    elt = _I_ExtractCanonicalizedValue(elt, p.path_elements)
	except OperationError:
	    continue
	r = apply(testfn, (elt,) + args)
	if r: break
    if Debug:
	if r: print "== succeeded"
	else: print "== failed"
    return r

def _EveryCore(subpath, testfn, v, args):
    if GetObjectType(v) is not GAL_LIST:
	if Debug:
	    print "== failed (error)"
	return 0
    r = 0
    p = CanonicalizePath(None, subpath)
    for elt in v:
	try:
	    elt = _I_ExtractCanonicalizedValue(elt, p.path_elements)
	except OperationError:
	    continue
	r = apply(testfn, (elt,) + args)
	if not r: break
    if Debug:
	if r: print "== succeeded"
	else: print "== failed"
    return r

class Path:
    def __init__(self, *args):
	if not args:
	    self.object = None
	    self.path_elements = ["self"]
	elif type(args[0]) is type(""):
	    self.object = None
	    self.path_elements = args
	else:
	    self.object = args[0]
	    if not args[1:]:
		self.path_elements = ["self"]
	    else:
		self.path_elements = args[1:]

def CanonicalizePath(prefix, element):
    if not isinstance(element, Path):
	if type(element) is type([]):
	    element = tuple(element)
	elif type(element) is not type(()):
	    element = (element,)
	return apply(Path, (prefix,) + element)
    elif element.object is None:
	element.object = prefix
	return element
    else:
	return element

# List accessor.

def List(path):
    p = CanonicalizePath(Source, path)
    try:
	v = ExtractCanonicalizedValue(p)
    except OperationError:
	return []
    if GetObjectType(v) is not GAL_LIST:
	return []
    else: return v

def Value(path):
    p = CanonicalizePath(Source, path)
    try:
	v = ExtractCanonicalizedValue(p)
    except OperationError:
	return None
    return v

# Classes for the basic operations.

class _FrameOperation:
    def __init__(self, source_path, target_path):
	self.op = "NOP"
	if source_path is not None:
	    self.source_path_obj = CanonicalizePath(Source, source_path)
	    self.source_frame = self.source_path_obj.object
	    self.source_path = self.source_path_obj.path_elements	    
	else:
	    self.source_path = self.source_path_obj = self.source_frame = None
	if target_path is not None:
	    self.target_path_obj = CanonicalizePath(Target, target_path)
	    self.target_frame = self.target_path_obj.object
	    self.target_path = self.target_path_obj.path_elements	
	else:
	    self.target_path = self.target_path_obj = self.target_frame = None
    def Print(self):
	print self.op, self.ArgsForPrinting()
    def ArgsForPrinting(self):
	return self.source_path, self.target_path
    def PrintPreamble(self):
	source_frame = self.source_frame
	target_frame = self.target_frame
	print
	print "== Apply"
	self.Print()
	print
	print "Source frame:"
	if source_frame is not None:
	    print OPr(source_frame, PP_TYPE)
	else:
	    print "None"
	print "Target frame:"
	if target_frame is not None:
	    print OPr(target_frame, PP_TYPE)
	else:
	    print "None"
    def PrintPostamble(self, res):
	print "Result:"
	if res is not None:
	    print OPr(res, PP_TYPE)
	else:
	    print "(no result)"
	print "== (end Apply)"
	print
    def Apply(self):
	source_frame = self.source_frame
	target_frame = self.target_frame
	# If there's a source key and the source frame
	# doesn't have that key, just return.
	if Debug:
	    self.PrintPreamble()
	if self.source_path:
	    # Get the source key value. Postprocess it.
	    # If you can't get the value, give up.
	    try:
		source_val = _I_ExtractCanonicalizedValue(source_frame,
							  self.source_path)
	    except OperationError:
		if Debug:
		    print "Not modifying."
		    print "== (end Apply)"
		    print
		return target_frame
	else:
	    # No source key, use the source_frame as the source_val.
	    source_val = source_frame
	res = self.InternalApply(source_val, target_frame)
	if Debug:
	    self.PrintPostamble(res)
	return res
    def InternalApply(self, source_val, target_frame):
	return target_frame

# Add is to add literal values.

class _AddOperation(_FrameOperation):
    def __init__(self, target_path, target_val, kw):
	if kw.has_key("how"):
	    if kw["how"] != "append":
		raise SpecificationError, "value of how is not append"
	    self.do_append = 1
	else:
	    self.do_append = 0
	_FrameOperation.__init__(self, None, target_path)
	self.op = "ADD"
	self.given_value = target_val
    def ArgsForPrinting(self):
	return self.target_path, self.given_value
    def InternalApply(self, source_val, target_frame):
	global Altered
	# Can't apply to instances.
	if GetObjectType(target_frame) is not GAL_FRAME:
	    return target_frame
	v, altered = _I_SetCanonicalizedValue(target_frame, self.target_path, \
					      self.given_value, self.do_append)
	if altered: Altered = 1
	return v

# Map does the following:
# (1) Uses the source value extract via the source path.
# (2) Maps the value. If there is a map name, use the map to
# process the value; otherwise, don't change it.
# (3) Sets the result in the target path. Remember, with
# Map there must always be a path, even implicit, and the default
# path is "self".
# Implications: Map() copies all source frames to the target.

class _MapOperation(_FrameOperation):
    def __init__(self, source_path, target_path, map_to_use, kw):
	self.map_to_use = map_to_use
	self.do_append = 0
	_FrameOperation.__init__(self, source_path, target_path)
	if kw.has_key("how"):
	    if kw["how"] != "append":
		raise SpecificationError, "value of how is not append"
	    self.do_append = 1
	self.op = "MAP"
    def ArgsForPrinting(self):
	return self.source_path, self.target_path, self.map_to_use
    def TranslateValue(self, source_val):
	# Should we do conversions if there is no map name? No.
	if Debug:
	    print
	    print "== Transfer"
	    print "Input:"
	    if source_val is not None:
		print OPr(source_val, PP_TYPE)
	    else:
		print "None"
	if self.map_to_use:
	    v = self._Recurse(source_val)
	else:
	    v = source_val
	if Debug:
	    print "Output:"
	    if v is not None:
		print OPr(v, PP_TYPE)
	    else:
		print "None"
	    print "== (end Transfer)"
	    print
	return v
    def TranslateValueInternal(self):
	global Target
	if callable(self.map_to_use):
	    self.map_to_use()
	elif type(self.map_to_use) is type({}):
	    Target = Source
	    if GetObjectType(Source) is not GAL_FRAME:
		# Just do a key lookup.
		if self.map_to_use.has_key(Source):
		    Target = self.map_to_use[Source]		    
    def InternalApply(self, source_val, target_frame):
	global Altered
	# Can't apply to instances.	
	if GetObjectType(target_frame) is not GAL_FRAME:
	    return target_frame
	new_val = self.TranslateValue(source_val)
	# Under some circumstances, new_val will be None.
	# It's a verrrry bad thing to let this be added
	# to a frame, so we only do something if new_val is not None.
	if new_val is not None:
	    v, altered = _I_SetCanonicalizedValue(target_frame,
						  self.target_path, \
						  new_val, self.do_append)
	    if altered: Altered = 1
	else:
	    v = target_frame
	return v
    def _Recurse(self, source_val):
	global Source, Target
	old_source = Source
	old_target = Target
	Source = source_val
	if GetObjectType(source_val) is GAL_FRAME:
	    Target = Frame(name = "__NOT_SET__")
	    Target.parent = None
	else:
	    Target = None
	self.TranslateValueInternal()
	v = Target
	Source = old_source
	Target = old_target
	return v

class _DeleteOperation(_FrameOperation):
    def __init__(self, target_path):
	# Always apply to the target frame.
	_FrameOperation.__init__(self, None, target_path)
	self.op = "DELETE"
    def ArgsForPrinting(self):
	return (self.target_path,)
    def InternalApply(self, source_val, result_frame):
	global Altered
	# Getting started. This is like the guts of
	# SetCanonicalizedValue.
	if not self.target_path:
	    sub_path = []
	    final_element = "self"
	else:	
	    sub_path = self.target_path[:-1]
	    final_element = self.target_path[-1]

	if final_element in ["self", "parent", "frame_name"]:
	    # Can't delete self, parent, frame_name.
	    return result_frame
		
	# Now, normal processing. First, extract the value
	# using the subpath.
	try:
	    target_frame = _I_ExtractCanonicalizedValue(result_frame,
							sub_path)
	except OperationError:
	    return result_frame

	# We can be sure that pred_hack is not set.
	
	# Can't apply to instances.
	if GetObjectType(target_frame) is not GAL_FRAME:
	    return result_frame
	elif final_element in [":pred", ":predicate",
			       ":preds", ":predicates"]:
	    target_frame.preds = []
	    Altered = 1
	elif target_frame.has_key(final_element):
	    del target_frame[final_element]
	    Altered = 1
	return result_frame

def Delete(path):
    global Target, Altered
    i = _DeleteOperation(path)
    v = i.Apply()
    if v is not Target:
	Altered = 1
    Target = v

# Operations.

def Map(element, new_path, map_to_use = None, **kw):
    global Target, Altered
    i = _MapOperation(element, new_path, map_to_use, kw)
    v = i.Apply()
    if v is not Target:
	Altered = 1
    Target = v

# Toplevel adder.

def Add(path, value, **kw):
    global Target, Altered
    i = _AddOperation(path, value, kw)
    v = i.Apply()
    if v is not Target:
	Altered = 1
    Target = v

# Toplevel flow of control.

def DoAll(*rules):
    for rule in rules: rule()

# This is a little trickier, since we need to identify if a frame
# has been altered.

def DoOne(*rules):
    global Altered
    old_altered = Altered
    Altered = 0
    for rule in rules:
	rule()
	if Altered: break
    Altered = old_altered

