/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

package org.mitre.galaxycommunicator.visualizer;

import java.awt.*;
import java.awt.font.*;

import galaxy.util.Logger;

/**
 * Encapsulates a Communicator enity (e.g, service provider, Hub) graphic.
 *
 * @author Steve Wohlever, MITRE Corporation, wohlever@mitre.org
 */
public class CommunicatorEntityGraphic
{    
    // Radius and center of this entity's circle graphic.
    private int radius;
    private int centerX, centerY;
    
    // Coordinates of upper left corner of square that encompasses this 
    // entity's circle graphic.
    private int upperLeftX, upperLeftY;
    
    // Angle relative to the Hub (0 = straight above the Hub, increases 
    // clockwise). -1 indicates that this entity represents the Hub.
    private double angleRadians = -1;
    private double angleDegrees = -1;

    private Color circleColor;

    private String label;
    private Font font;
    private Color labelColor;
    private int labelLowerLeftX, labelLowerLeftY;
    private int labelWidth, labelHeight;
    private int labelBackgroundUpperLeftX, labelBackgroundUpperLeftY;
    private int labelBackgroundWidth, labelBackgroundHeight;
    private boolean drawLabelBackground;
    private Color labelBackgroundColor;

    private Image icon;
    private int iconWidth, iconHeight;
    private int iconUpperLeftX, iconUpperLeftY;
    
    private VisualizerUI ui;
    private VisualizerDrawingArea drawingArea;

    // Time that a thread periodically waits after loading an icon before it
    // tries to determine the icon's dimensions. Loading icons is apparently 
    // an asynchronous event in Java, so the dimension info is not immediately
    // available.
    private final int DEFAULT_ICON_LOAD_WAIT_MSEC = 500;

    public CommunicatorEntityGraphic(int centerX, int centerY, int radius, 
				     Color color, String label, Font font,
				     Color labelColor, 
				     boolean drawLabelBackground,
				     Color labelBackgroundColor, Image icon,
				     VisualizerUI ui)
    {
	this.radius = radius;
	this.label = label;
	this.font = font;
	this.labelColor = labelColor;
	this.drawLabelBackground = drawLabelBackground;
	this.labelBackgroundColor = labelBackgroundColor;
	this.icon = icon;
	this.ui = ui;
	drawingArea = ui.getDrawingArea();

	setLocation(centerX, centerY);
	setColor(color);
	
	if(icon != null) {
	    iconWidth = iconHeight = -1;
	    while(iconWidth == -1 || iconHeight == -1) {
		iconWidth = icon.getWidth(ui);
		iconHeight = icon.getHeight(ui);
		
		try {
		    Thread.sleep(DEFAULT_ICON_LOAD_WAIT_MSEC);
		} catch(Exception ex) {
		    Logger.getLogger().logErrorMessage("Caught exception in CommunicatorEntityGraphic while waiting for image dimensions: " + ex.toString(), ex);
		}
	    }
	}
    }


    //========================================================================
    // Settor and gettor methods.
    //========================================================================
    
    public void setAngleRadians(double angleRadians)
    {
	this.angleRadians = angleRadians;
	angleDegrees = Math.toDegrees(angleRadians);
    }
    
    public void setLocation(int centerX, int centerY)
    { 
	Graphics graphics = null;
	FontMetrics fontMetrics = null;
	LineMetrics lineMetrics = null;

	this.centerX = centerX;
	this.centerY = centerY;
	upperLeftX = centerX - radius;
	upperLeftY = centerY - radius;

	if(icon != null && angleDegrees == 0) {

	    // Set the icon's Y coordinate for the case in which the
	    // graphic is above the Hub. This coordinate needs to be computed
	    // before the label can be placed.
	  
	    iconUpperLeftY = upperLeftY - 2 - iconHeight;
	}

	if(label != null) {
	    graphics = drawingArea.getGraphics();

	    fontMetrics = graphics.getFontMetrics(font);
	    lineMetrics = fontMetrics.getLineMetrics(label, graphics);
	    
	    labelWidth = fontMetrics.stringWidth(label);
	    labelHeight = (int) lineMetrics.getHeight();
	    
	    if(angleDegrees == -1) {
		
		// The Hub case.		    
		labelLowerLeftX = centerX - labelWidth/2;
		labelLowerLeftY = centerY + labelHeight/4;
		
	    } else {
		
		// The service providers.
		if(angleDegrees == 0) { 
		    labelLowerLeftX = upperLeftX - labelWidth/2 + radius;
		    if(icon != null && angleDegrees == 0)
			labelLowerLeftY = iconUpperLeftY - ((int) lineMetrics.getDescent()) - 2;
		    else
			labelLowerLeftY = upperLeftY - ((int) lineMetrics.getDescent()) - 2;
		} else if(angleDegrees == 180) {
		    labelLowerLeftX = upperLeftX - labelWidth/2 + radius;
		    labelLowerLeftY = upperLeftY + 2*radius + labelHeight + 2;
		} else {
		    
		    if(angleDegrees < 180) {
			labelLowerLeftX = upperLeftX + 2*radius + 2;
		    } else {
			labelLowerLeftX = upperLeftX - labelWidth - 2;
		    }
		    
		    labelLowerLeftY = upperLeftY + radius + labelHeight/4;
		}
	    }

	    // Subtract 2 and add 4 to the upper left X and width to
	    // prevent the rectangle's vertical edges from ending on
	    // the text edges. Add 2 to the upper left Y since 
	    // since without this, the rectangle extends too far above
	    // the text for some reason. Modify the height to take
	    // into account letters that extend below the baseline of
	    // the text (e.g., the letter "p").
	    
	    labelBackgroundUpperLeftX = labelLowerLeftX-2;
	    labelBackgroundUpperLeftY = labelLowerLeftY - labelHeight + 2;
	    labelBackgroundWidth = labelWidth+4;
	    labelBackgroundHeight = labelHeight + ((int) lineMetrics.getDescent());		
	}
	
	if(icon != null) { 

	    // Set the icon's Y coordinate.
	    if(angleDegrees != 0) {
	       
		if(label != null) {
		    iconUpperLeftY = labelLowerLeftY + ((int) lineMetrics.getDescent()) + 4;
		} else if(angleDegrees == 180) {
		    iconUpperLeftY = upperLeftY + 2*radius + 2;
		} else {
		    iconUpperLeftY = (upperLeftY + radius) - iconHeight/2;
		}
	    }

	    // Set the icon's X coordinate	  
	    if(angleDegrees == 0 || angleDegrees == 180) {
		iconUpperLeftX = centerX - iconWidth/2; 
	    } else if(angleDegrees < 180) {
		if(label != null)
		    iconUpperLeftX = labelLowerLeftX;
		else
		    iconUpperLeftX = upperLeftX + 2*radius + 2;
	    } else {
		if(label != null)
		    iconUpperLeftX = labelLowerLeftX + (labelWidth - iconWidth);
		else
		    iconUpperLeftX  = upperLeftX - iconWidth - 2;
	    }
	}

    }
    
    public void setColor(Color color)
    {
	circleColor = color;
    }
    
    public void setLabelColor(Color labelColor)
    {
	this.labelColor = labelColor;
    }
    
    public void setDrawLabelBackground(boolean drawLabelBackground)
    {
	this.drawLabelBackground = drawLabelBackground;
    }

    public void setLabelBackgroundColor(Color labelBackgroundColor)
    {
	this.labelBackgroundColor = labelBackgroundColor;
    }
    
    public int getRadius()
    {
	return radius;
    }
       
    //========================================================================
    // UI methods.
    //========================================================================
    
    /**
     * Tests if a mouse click falls within the bounds of this graphic.
     */
    public boolean isClicked(int x, int y)
    {
	// First check for a click in the circle.
	int deltaX = centerX - x;
	int deltaY = centerY - y;
	double distance = Math.sqrt(deltaX*deltaX + deltaY*deltaY);
	if(distance <= radius)
	    return true;
	
	// Check for a click in the label.
	if(label != null) {
	    if(x >= labelLowerLeftX && x <= (labelLowerLeftX + labelWidth) &&
	       y <= labelLowerLeftY && y >= (labelLowerLeftY - labelHeight)) {
		return true;
	    }
	}
	
	// Check for a click in the icon.
	if(icon != null) {
	    if(x >= iconUpperLeftX && x <= (iconUpperLeftX + iconWidth) &&
	       y >= iconUpperLeftY && y <= (iconUpperLeftY + iconHeight)) {
		return true;
	    }
	}
	
	return false;
    }
    
    public void render(Graphics graphics)
    {
	// Draw the filled circle.
	graphics.setColor(circleColor);
	graphics.fillOval(upperLeftX, upperLeftY, radius*2, radius*2);
	
	// Draw the icon.
	if(icon != null) {
	    graphics.drawImage(icon, iconUpperLeftX, iconUpperLeftY, ui);
	}
	
	// Draw the label.
	if(label != null) {

	    // Draw the label background.
	    if(drawLabelBackground) {		
	
		graphics.setColor(labelBackgroundColor);
		graphics.fillRect(labelBackgroundUpperLeftX,
				  labelBackgroundUpperLeftY,
				  labelBackgroundWidth,
				  labelBackgroundHeight);
	    }
	    
	    graphics.setFont(font);
	    graphics.setColor(labelColor);
	    graphics.drawString(label, labelLowerLeftX, labelLowerLeftY);
	}
    }
}
