/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

package org.mitre.galaxycommunicator.visualizer;

import java.awt.*;
import java.awt.geom.*;

/**
 * Encapsulates a line graphic.
 *
 * @author Steve Wohlever, MITRE Corporation, wohlever@mitre.org
 */

public class Line
{
    // Hub end points
    private int x1, y1; 
    
    // Provider end points
    private int x2, y2; 
    
    private Color color;

    private float lineWidth = 1;
    private BasicStroke stroke;

    // Angle (0 = pointing straight up, increases clockwise).
    private double angleRadians = 0;
    
    private int arrowheadXPoints[];
    private int arrowheadYPoints[];
    private Shape arrowhead;
    private boolean hasArrowhead = false;
    private boolean arrowHeadPointsAtProvider = false;
    private int arrowheadWidth = 0;
    private int arrowheadHeight = 0;

    public Line(int x1, int y1, int x2, int y2, int lineWidth, 
		Color color, int arrowheadWidth, int arrowheadHeight)
    {
	this.x1 = x1;
	this.y1 = y1;
	this.x2 = x2;
	this.y2 = y2;

	if(arrowheadWidth > 0 && arrowheadHeight > 0) {
	    hasArrowhead = true;
	    this.arrowheadWidth = arrowheadWidth;
	    this.arrowheadHeight = arrowheadHeight;
	    arrowheadXPoints = new int[3];
	    arrowheadYPoints = new int[3];
	}

	setLineWidth(lineWidth);
	setColor(color);
    }

    public Line(int x1, int y1, int x2, int y2, int lineWidth, 
		Color color)
    {
	this.x1 = x1;
	this.y1 = y1;
	this.x2 = x2;
	this.y2 = y2;

	hasArrowhead = false;
       
	setLineWidth(lineWidth);
	setColor(color);
    }
    
    /**
     * Sets this line's arrowhead to point at the service provider or Hub.
     *
     * @param arrowHeadPointsAtProvider if true, points at the provider,
     *                                  otherwise points at the Hub
     */
    public void setArrowheadPointsAtProvider(boolean arrowHeadPointsAtProvider)
    {
	this.arrowHeadPointsAtProvider = arrowHeadPointsAtProvider;
    }
    
    public void setLocation(int x1, int y1, int x2, int y2)
    {
	this.x1 = x1;
	this.y1 = y1;
	this.x2 = x2;
	this.y2 = y2;

	// If the line has an arrowhead...
	if(hasArrowhead) {
	    int anchorX = 0;
	    int anchorY = 0;
	    double rotateAngleRadians = 0;
	    
	    // If pointing at the provider, rotate the arrowhead 
	    // relative to that end of the line. Else, rotate relative
	    // to the Hub end.
	    if(arrowHeadPointsAtProvider) {
		anchorX = x2;
		anchorY = y2;
		rotateAngleRadians = angleRadians;
	    } else {
		anchorX = x1;
		anchorY = y1;
		rotateAngleRadians = angleRadians + Math.toRadians(180);
	    }
	    
	    // Set the points of the arrowhead polygon.
	    arrowheadXPoints[0] = anchorX - arrowheadWidth/2;
	    arrowheadXPoints[1] = anchorX;
	    arrowheadXPoints[2] = anchorX + arrowheadWidth/2;

	    arrowheadYPoints[0] = anchorY;
	    arrowheadYPoints[1] = anchorY - arrowheadHeight;
	    arrowheadYPoints[2] = anchorY;
	   
	    // Create the arrowhead pointing straight up.
	    Polygon arrowHeadPolygon = new Polygon(arrowheadXPoints, 
						   arrowheadYPoints, 3);
	    
	    // Create the rotation transform object.
	    AffineTransform rotationTransform = 
		AffineTransform.getRotateInstance(rotateAngleRadians, 
						  anchorX, anchorY);
	    
	    // Rotate the arrowhead.
	    arrowhead = rotationTransform.createTransformedShape(arrowHeadPolygon);
	}
    }
    
    public void setColor(Color color)
    {
	this.color = color;
    }
    
    public void setLineWidth(int lineWidth)
    {
	this.lineWidth = lineWidth;
	stroke = new BasicStroke(lineWidth);
    }
    
    public void setAngleRadians(double angleRadians)
    {
	this.angleRadians = angleRadians;
    }
    
    public void render(Graphics graphics)
    {
	if(graphics instanceof Graphics2D) {
	    Graphics2D graphics2D = (Graphics2D) graphics;
	    
	    // Draw the line.
	    graphics2D.setColor(color);
	    graphics2D.setStroke(stroke);
	    graphics2D.drawLine(x1, y1, x2, y2);
	    
	    // If the line has an arrowhead...
	    if(hasArrowhead) {
		graphics2D.draw(arrowhead);
		graphics2D.fill(arrowhead);
	    }
	}
    }
}
