/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

package org.mitre.galaxycommunicator.visualizer;

import java.awt.*;
import java.util.*;
import java.io.File;

import galaxy.util.Logger;

/**
 * Encapsulates a service provider graphic.
 *
 * @author Steve Wohlever, MITRE Corporation, wohlever@mitre.org
 */

public class ProviderGraphic
{  
    // Id of the provider associated with this graphic.
    private String id;
    
    // The provider graphic.
    private CommunicatorEntityGraphic graphicObject;
    
    // The link between the provider and the Hub.
    private Line linkLine;
    
    // Center of this provider graphic.
    private int x, y;
    
    // Flag that indicates if this provider is "available" (i.e., can be
    // connected to the Hub).
    private boolean available;
    
    // Flag that indicates if this provider is connected to the Hub.
    private boolean connected;
    
    // Message traffic lines (one for each direction: Hub to provider and
    // provider to Hub).
    private Line messageTrafficToProviderLine;
    private Line messageTrafficToHubLine;
    
    // Flags that indicate if the message traffic lines are active.
    private boolean messageTrafficToProviderOn = false;
    private boolean messageTrafficToHubOn = false;
    
    // Angle relative to the Hub (0 = straight above the Hub, increases 
    // clockwise).
    private double angleRadians;
    
    private ProviderInfo provider;
    
    private volatile boolean flashing = false;
        
    private int linkLineLength; 

    private Color color;

    // Maps provider ids to provider graphic objects. This contains all
    // provider graphics of providers that are "available" (i.e., can be
    // connected to Hub).
    static Map availableProviderMap = new HashMap();

    private VisualizerProperties props;

    public ProviderGraphic(ProviderInfo provider, int providerRadius,
			   int linkLineLength, Font font, Color color, 
			   VisualizerUI ui, boolean available, boolean connected)
    {
	this.props =  ui.getProperties();
	this.provider = provider;
	this.linkLineLength = linkLineLength;
	this.color = color;

	id = provider.getId();
	
	linkLine = new Line(0, 0, 0, 0, props.getLinkLineWidth(), 
			    props.getLinkLineColor());
	
	messageTrafficToProviderLine = 
	    new Line(0, 0, 0, 0, props.getMessageTrafficLineWidth(),
		     props.getMessageTrafficToProviderLineColor(), props.getArrowheadWidth(), props.getArrowheadHeight());
	messageTrafficToProviderLine.setArrowheadPointsAtProvider(true);
	
	messageTrafficToHubLine = 
	    new Line(0, 0, 0, 0, props.getMessageTrafficLineWidth(),
		     props.getMessageTrafficToHubLineColor(), props.getArrowheadWidth(), props.getArrowheadHeight());
	messageTrafficToHubLine.setArrowheadPointsAtProvider(false);
	
	int circleRadius = providerRadius;
	
	// Try to build icon.
	String iconFileName = provider.getIconFileName();
	Image icon = null;
	if(iconFileName != null) {
	    
	    // Check if the icon file exists.
	    File file = new File(iconFileName);
	    if(file.exists()) {
		icon = Toolkit.getDefaultToolkit().getImage(iconFileName);
	    } else {
		Logger.getLogger().logWarningMessage("Icon file " + iconFileName + " does not exist");
	    }
	}
	
	graphicObject = new CommunicatorEntityGraphic(0, 0, circleRadius, props.getUnavailableColor(), provider.getLabel(), font, props.getUnavailableColor(), false, props.getLabelBackgroundColor(), icon, ui);

	setAvailabilityStatus(available);
	setConnectionStatus(connected);
    }
    
    public static void clearAvailableProviderMap()
    {
	availableProviderMap.clear();
    }

    public void setFlashing(boolean flashing)
    {
	this.flashing = flashing;
    }
    
    public boolean isFlashing()
    {
	return flashing;
    }
    
    public void setColor(Color color)
    {
	this.color = color;
    }

    public void setLabelBackgroundColor(Color labelBackgroundColor)
    {
	graphicObject.setLabelBackgroundColor(labelBackgroundColor);
    }

    public Color getColor()
    {
	return color;
    }

    public ProviderInfo getProvider()
    {
	return provider;
    }

    public CommunicatorEntityGraphic getGraphicObject()
    {
	return graphicObject;
    }
   
    public void setAvailabilityStatus(boolean available)
    {
	this.available = available;
	if(available) {
	    availableProviderMap.put(id, this);
	    graphicObject.setLabelColor(props.getLabelForegroundColor());
	    linkLine.setColor(props.getLinkLineColor());
	} else {
	    availableProviderMap.remove(id);
	    graphicObject.setLabelColor(props.getUnavailableColor());
	    linkLine.setColor(props.getUnavailableColor());
	}
	
	graphicObject.setDrawLabelBackground(available);
    }
    
    public boolean isAvailable()
    {
	return available;
    }
    
    public void setConnectionStatus(boolean connected)
    {
	this.connected = connected;
    }
    
    public boolean isConnected()
    {
	return connected;
    }
    
    public boolean isClicked(int x, int y)
    {
	return graphicObject.isClicked(x, y);
    }
    
    public void setLocation(int x, int y, Point hubCenter, int hubRadius)
    {
	this.x = x;
	this.y = y;
	graphicObject.setLocation(x, y);
	linkLine.setLocation(hubCenter.x, hubCenter.y, x, y);
	
	// Traffic lines should fit between provider and Hub with room to
	// spare.
	int spareRoom = 20;
	int lineLength = linkLineLength - hubRadius - graphicObject.getRadius() - props.getArrowheadHeight() - spareRoom;
	
	// Distance of Hub-side end of message flow line from center of Hub
	int distanceFromHub = hubRadius + spareRoom/2;
	
	// Perpendicular distance of message flow lines from link line.
	int distanceFromLink = props.getArrowheadWidth()/2 + 10;
	
	// Hub-side coordinates of the message flow line. Calculate the
	// values for the "to provider" message flow line first.
	Point hubPoint = VisualizerDrawingArea.generateRelativePoint(hubCenter, angleRadians, distanceFromHub);
	
	// Provider-side coordinates of the message flow line.
	Point providerPoint = VisualizerDrawingArea.generateRelativePoint(hubPoint, angleRadians, lineLength);
	
	// X/Y translations that move line "distanceFromLink" distance
	// away from link line.
	int deltaX = (int) (distanceFromLink*Math.cos(angleRadians));
	int deltaY = (int) (distanceFromLink*Math.sin(angleRadians));
	
	// Set the "to provider" message flow line.
	messageTrafficToProviderLine.setLocation(hubPoint.x-deltaX, hubPoint.y-deltaY, providerPoint.x-deltaX, providerPoint.y-deltaY);
	
	distanceFromHub += props.getArrowheadHeight();
	
	// Calculate the coordinates for the "to Hub" message flow line.
	hubPoint = VisualizerDrawingArea.generateRelativePoint(hubCenter, angleRadians, distanceFromHub);
	providerPoint = VisualizerDrawingArea.generateRelativePoint(hubPoint, angleRadians, lineLength);
	// Set the "to Hub" message flow line.
	messageTrafficToHubLine.setLocation(hubPoint.x+deltaX, hubPoint.y+deltaY, providerPoint.x+deltaX, providerPoint.y+deltaY);
    }
    
    public void setAngleRadians(double angleRadians)
    {
	this.angleRadians = angleRadians;
	messageTrafficToProviderLine.setAngleRadians(angleRadians);
	messageTrafficToHubLine.setAngleRadians(angleRadians);
	graphicObject.setAngleRadians(angleRadians);
    }
    
    public double getAngleRadians()
    {
	return angleRadians;
    }
    
    public Line getMessageTrafficToProviderLine()
    {
	return messageTrafficToProviderLine;
    }
    
    public Line getMessageTrafficToHubLine()
    {
	return messageTrafficToHubLine;
    }
    
    public String getProviderId()
    {
	return id;
    }
    
    public void setToProviderGraphic(boolean on)
    {
	messageTrafficToProviderOn = on;
    }
    
    public void setToHubGraphic(boolean on)
    {
	messageTrafficToHubOn = on;
    }
    
    public boolean hasMessageTrafficToProvider()
    {
	return messageTrafficToProviderOn;
    }
    
    public boolean hasMessageTrafficToHub()
    {
	return messageTrafficToHubOn;
    }
    
    public void render(Graphics graphics)
    {
	if(available) {
	    if(flashing)
		graphicObject.setColor(props.getProviderHighlightColor());
	    else
		graphicObject.setColor(color);
	} else {
	    graphicObject.setColor(props.getUnavailableColor());
	}
	// Render the line first so that it is under the provider graphic.
	if(connected) {
	    linkLine.setLineWidth(props.getLinkLineWidth());
	    linkLine.render(graphics);
	}
	graphicObject.render(graphics);
    }
}
