/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

package org.mitre.galaxycommunicator.visualizer;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

import galaxy.server.MainServer;
import galaxy.util.Logger;

/**
 * Encapsulates the drawing area that displays the Hub and service provider
 * graphics and all associate state info.
 *
 * @author Steve Wohlever, MITRE Corporation, wohlever@mitre.org
 */
public class VisualizerDrawingArea extends JPanel 
    implements MouseListener
{    
    // Center of Hub.
    private Point hubCenter;

    // The Hub graphic.
    private CommunicatorEntityGraphic hub = null;

    // List of provider graphics. This is used to repeatedly access the 
    // graphics in a specific order.
    private java.util.List providerGraphics;
    
    // Maps provider ids to provider graphic objects. This contains all
    // current provider graphics.
    private Map providerMap;

    // The UI that contains this drawing area.
    private VisualizerUI ui;

    // Flag that indicates if provider graphics can be selected by clicking
    // on them with the mouse.
    private boolean providerSelectionEnabled = true;

    // Font used to display Hub and provider labels.
    private Font labelFont;

    // The common diagnostic logger.
    private Logger logger = null;

    // The visualizer properties.
    private VisualizerProperties props = null;

    public VisualizerDrawingArea(VisualizerUI ui) 
    {
	this.ui = ui;
	props = ui.getProperties();

	Dimension size = new Dimension(props.getDrawingAreaWidth(),
				       props.getDrawingAreaHeight());
	logger = Logger.getLogger();
	
	labelFont = new Font("SansSerif", Font.BOLD, props.getLabelFontSize());

	hubCenter = new Point(size.width/2, size.height/2);

	// Initialize data structures.
	providerGraphics = new ArrayList();
	providerMap = new HashMap();

	setBackground(Color.white);
	setPreferredSize(size);

	// Add this panel as a mouse listener.
	addMouseListener(this);
    }

    public void setLabelBackgroundColor(Color color)
    {
	ProviderGraphic providerGraphic;
	for(int i = 0; i < providerGraphics.size(); i++) {
	    providerGraphic = (ProviderGraphic) providerGraphics.get(i);
	    providerGraphic.setLabelBackgroundColor(color);
	}
	repaint();
    }

    public void setHubRunningColor(Color color)
    {
	if(ui.isHubRunning()) {
	    hub.setColor(color);
	    repaint();
	}
    }

    public void setHubStoppedColor(Color color)
    {
	if(!ui.isHubRunning()) {
	    hub.setColor(color);
	    repaint();
	}
    }

    public void setProviderColor(Color color)
    {
	ProviderGraphic providerGraphic;
	for(int i = 0; i < providerGraphics.size(); i++) {
	    providerGraphic = (ProviderGraphic) providerGraphics.get(i);
	    providerGraphic.setColor(color);
	}

	repaint();
    }

    public void setToProviderArrowColor(Color color)
    {
	repaint();
    } 

    public void setToHubArrowColor(Color color)
    {
	repaint();
    }

    //========================================================================
    // Various graphical methods.
    //========================================================================

    /**
     * This method redraws the drawing area contents.
     */
    protected void paintComponent(Graphics g) 
    {
	super.paintComponent(g);

	if(hub == null)
	    addHubGraphic();

	// Render the graphics.
	ProviderGraphic providerGraphic;
	for(int i = 0; i < providerGraphics.size(); i++) {
	    providerGraphic = (ProviderGraphic) providerGraphics.get(i);
	    providerGraphic.render(g);
	}
	
	// Render the Hub such that it is on top of all the link and
	// connection lines.
	if(hub != null)
	    hub.render(g);

	// Render the message flow lines last.
	for(int i = 0; i < providerGraphics.size(); i++) {
	    providerGraphic = (ProviderGraphic) providerGraphics.get(i);
	    if(providerGraphic.hasMessageTrafficToProvider()) {
		providerGraphic.getMessageTrafficToProviderLine().setColor(props.getMessageTrafficToProviderLineColor());
		providerGraphic.getMessageTrafficToProviderLine().render(g);
	    }

	    if(providerGraphic.hasMessageTrafficToHub()) {
		providerGraphic.getMessageTrafficToHubLine().setColor(props.getMessageTrafficToHubLineColor());
		providerGraphic.getMessageTrafficToHubLine().render(g);
	    }
	}
    }

    /**
     * Calculates the coordinates of a point at a given angle to and distance
     * from a target point.
     *
     * @param point the target point
     * @param angleRadians angle to the target point (0 = straight above the
     *                     target point, increases clockwise)
     * @param distance distance in pixels from the target point
     * @return the generated point
     */
    static Point generateRelativePoint(Point point, double angleRadians,
				       int distance)
    {
	int x, y;
	    x = point.x + (int) (distance*Math.sin(angleRadians));
	    y = point.y - (int) (distance*Math.cos(angleRadians));
	    return new Point(x, y);
    }

    /**
     * Resets the drawing area. This includes removing all graphics from
     * the drawing area.
     */
    public void reset()
    {
	providerGraphics.clear();
	providerMap.clear();
	ProviderGraphic.clearAvailableProviderMap();

	repaint();
    }


    //========================================================================
    // Methods for modifying the drawing area contents.
    //========================================================================

    private void addHubGraphic()
    {
	if(hub != null) {
	    logger.logWarningMessage("VisualizerDrawingArea.addHubGraphic: Hub has already been added");
	    return;
	}

	hub = new CommunicatorEntityGraphic(hubCenter.x, hubCenter.y, 
					    props.getHubRadius(), 
					    props.getHubStoppedColor(),
					    "Hub", labelFont,
					    props.getLabelForegroundColor(),
					    false, null, null, ui);

    }  

    private void setProviderLocations() 
    {
	// Set new provider locations.
	int numProviders = providerGraphics.size();
	double angleIncrementDegrees = (double) (360/numProviders);
	Point point;
	double angleRadians;
	ProviderGraphic providerGraphic;

	for(int idx=0; idx<numProviders; ++idx) {
	    angleRadians = Math.toRadians(idx * angleIncrementDegrees);
	    point = VisualizerDrawingArea.generateRelativePoint(hubCenter, angleRadians, props.getLinkLineLength());
	    providerGraphic = (ProviderGraphic) providerGraphics.get(idx);
	    providerGraphic.setAngleRadians(angleRadians);
	    providerGraphic.setLocation(point.x, point.y, hubCenter, props.getHubRadius());
	}
    }

    public void addProviderGraphic(ProviderInfo provider)
    {	
	String providerId = provider.getId();
	ProviderGraphic providerGraphic = (ProviderGraphic) providerMap.get(providerId);
	if(providerGraphic != null) {
	    logger.logWarningMessage("VisualizerDrawingArea.addProviderGraphic: Provider \"" + providerId + "\" already exists");
	    return;
	}

	// Set up initial graphic.
	providerGraphic = new ProviderGraphic(provider, 
					      props.getProviderRadius(),
					      props.getLinkLineLength(),
					      labelFont,
					      props.getProviderColor(),
					      ui, provider.isAvailable(), 
					      provider.isConnected());
	
	// Add the provider graphic to the list.
	providerGraphics.add(providerGraphic);
	providerMap.put(providerId, providerGraphic);
	
	setProviderLocations();
	repaint();
    } 

    public void removeProviderGraphic(String providerId)
    {	
	ProviderGraphic providerGraphic = (ProviderGraphic) providerMap.get(providerId);
	if(providerGraphic == null) {
	    logger.logWarningMessage("VisualizerDrawingArea.removeProviderGraphic: Provider \"" + providerId + "\" does not exists");
	} else {
	    if(providerGraphics.remove(providerGraphic)) {
		setProviderLocations();
		repaint();
	    } else {
		logger.logWarningMessage("VisualizerDrawingArea.removeProviderGraphic: Could not fine graphic for provider \"" + providerId + "\"");
	    }
	}
    }

    public void disconnectProviders()
    {
	int numProviders = providerGraphics.size();
	ProviderGraphic providerGraphic = null;
	for(int idx=0; idx<numProviders; ++idx) {
	    providerGraphic = (ProviderGraphic) providerGraphics.get(idx);
	    providerGraphic.setConnectionStatus(false);
	}	
	repaint();
    }

    public void setAllProvidersUnavailable()
    {
	int numProviders = providerGraphics.size();
	ProviderGraphic providerGraphic = null;
	for(int idx=0; idx<numProviders; ++idx) {
	    providerGraphic = (ProviderGraphic) providerGraphics.get(idx);
	    providerGraphic.setAvailabilityStatus(false);
	}	
	repaint();
    }

    public void setProviderAvailable(String providerId)
    {
	ProviderGraphic providerGraphic = findProviderGraphic(providerId);
	if(providerGraphic != null) {
	    providerGraphic.setAvailabilityStatus(true);
	    repaint();
	    
	} else {
	    logger.logWarningMessage("VisualizerDrawingArea.setProviderAvailable: can't find provider \"" + providerId + "\"");
	}
    }

    private ProviderGraphic findProviderGraphic(String providerId)
    {
	return (ProviderGraphic) providerMap.get(providerId);
    }


    // Create a class for this to reduce the calls to "new Runnable".
    class RepaintAreaRunnable implements Runnable
    {
	public void run()
	{
	    repaint();
	}
    }
    
    private final RepaintAreaRunnable repaintAreaRunnable 
	= new RepaintAreaRunnable();

    // Need to do this in a non-Swing thread. Otherwise, the call to 
    // Thread.sleep would interfere with the repaint.
    public void flashProvider(String providerId)
    {
	ProviderGraphic providerGraphic = findProviderGraphic(providerId);
	if(providerGraphic == null) {
	    logger.logWarningMessage("VisualizerDrawingArea.flashProvider: can't find provider \"" + providerId + "\"");
	    return;
	}

	if(!providerGraphic.isConnected())
	    return;

	providerGraphic.getGraphicObject().setColor(props.getProviderHighlightColor());
	providerGraphic.setFlashing(true);
	
	SwingUtilities.invokeLater(repaintAreaRunnable);
	
	try {
	    Thread.sleep(props.getProviderFlashMsec());
	} catch(Exception ex) {
	}

	providerGraphic.getGraphicObject().setColor(props.getProviderColor());
	providerGraphic.setFlashing(false);

	SwingUtilities.invokeLater(repaintAreaRunnable);
    }

   

    // Need to do this in a non-Swing thread. Otherwise, the call to 
    // Thread.sleep would interfere with the repaint.
    public void sendMessage(String providerId, boolean toProvider)
    {
	ProviderGraphic providerGraphic = findProviderGraphic(providerId);
	if(providerGraphic == null) {
	    logger.logWarningMessage("VisualizerDrawingArea.sendMessage: sendMessage can't find provider \"" + providerId + "\"");
	    return;
	}

	if(!providerGraphic.isConnected())
	    return;

	if(toProvider)
	    providerGraphic.setToProviderGraphic(true);
	else
	    providerGraphic.setToHubGraphic(true);

	// If the provider is flashing, wait until it is done.
	while(providerGraphic.isFlashing()) {
	    try {
		Thread.sleep(props.getProviderFlashMsec());
	    } catch(Exception ex) {
	    }
	}

	SwingUtilities.invokeLater(repaintAreaRunnable);
	
	try {
	    Thread.sleep(props.getArrowFlashMsec());
	} catch(Exception ex) {
	}

	if(toProvider)
	    providerGraphic.setToProviderGraphic(false);
	else
	    providerGraphic.setToHubGraphic(false);

	SwingUtilities.invokeLater(repaintAreaRunnable);
    }

    public void highlightProvider(String providerId, boolean highlight)
    {
	ProviderGraphic providerGraphic = findProviderGraphic(providerId);
	if(providerGraphic == null) {
	    logger.logWarningMessage("VisualizerDrawingArea.highlightProvider: can't find provider \"" + providerId + "\"");
	    return;
	}

	if(highlight)
	    providerGraphic.getGraphicObject().setColor(props.getProviderHighlightColor());
	else
	    providerGraphic.getGraphicObject().setColor(props.getProviderColor());
	
	repaint();
    }

    public void setProviderConnectionStatus(String providerId, 
					    boolean connected)
    {
	ProviderGraphic providerGraphic = findProviderGraphic(providerId);
	if(providerGraphic != null) {
	    providerGraphic.setConnectionStatus(connected);
	    repaint();
	} else {
	    logger.logWarningMessage("VisualizerDrawingArea.setProviderConnectionStatus: can't find provider \"" + providerId + "\"");
	}
    }

    public void setProviderAvailabilityStatus(String providerId, 
					      boolean available)
    {
	ProviderGraphic providerGraphic = findProviderGraphic(providerId);
	if(providerGraphic != null) {
	    providerGraphic.setAvailabilityStatus(available);
	    repaint();
	} else {
	    logger.logWarningMessage("VisualizerDrawingArea.setProviderAvailabilityStatus: can't find provider \"" + providerId + "\"");
	}
    }

    /**
     * Tests if the specified provider is connected to the Hub.
     *
     * @param providerId the id of the provider to test
     * @return true if the provider is connected to the Hub, false otherwise
     */
    public boolean isConnectedToHub(String providerId)
    {
	ProviderGraphic providerGraphic = findProviderGraphic(providerId);
	return providerGraphic.isConnected();
    }

    /**
     * Enables/disables provider selection with the mouse.
     *
     * @param enabled enabled if true, disabled if false
     */
    public void setProviderSelectionEnabled(boolean enabled)
    {
	providerSelectionEnabled = enabled;
    }


    void setHubIsRunning(boolean hubIsRunning)
    {
	if(hubIsRunning)
	    hub.setColor(props.getHubRunningColor());
	else
	    hub.setColor(props.getHubStoppedColor());

	repaint();
    }


    //=========================================================================
    // MouseListener interface
    //=========================================================================

    public void mouseClicked(MouseEvent event)
    {
	int modifiers = event.getModifiers();

        switch(modifiers) {
	case InputEvent.BUTTON3_MASK: // right mouse button
	    showProviderPopupMenu(event);
	    break;
	default:
	}
	    
    } 
    
    public void mouseEntered(MouseEvent event)
    {}
    
    public void mouseExited(MouseEvent event) 
    {}
    
    public void mousePressed(MouseEvent event)
    {}
    
    public void mouseReleased(MouseEvent event) 
    {}

    private void showProviderPopupMenu(MouseEvent event)
    {
	int x = event.getX();
	int y = event.getY();

	int numProviders = providerGraphics.size();
	ProviderGraphic providerGraphic = null;
	boolean doRepaint = false;
	for(int idx=0; idx<numProviders; ++idx) {
	    providerGraphic = (ProviderGraphic) providerGraphics.get(idx);
	    if(providerGraphic.isClicked(x, y)) {
		ui.showPopupMenu(providerGraphic.getProviderId(), event);
	    }
	}
    }
}
