/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

package org.mitre.galaxycommunicator.visualizer;

import java.io.*;
import java.net.Socket;
import java.util.*;
import javax.swing.SwingUtilities;

import galaxy.lang.*;
import galaxy.server.*;

/**
 * The Communicator visualizer connection object.
 *
 * @author Steve Wohlever, MITRE Corporation, wohlever@mitre.org
 */
public class VisualizerServer extends galaxy.server.Server 
{
    private VisualizerUI ui;
    private List providerIds;
    private List providers; 
    private Map providerMap;
    private Map providerHighlightCounters;
    
    public static final int INITIAL_NUM_PROVIDERS = 10;

    private final int GAL_HUB_SESSION_ALARM_STATUS = 0;
    private final int GAL_HUB_SESSION_ALARM_ENABLED = 1; 
    private final int GAL_HUB_SESSION_ALARM_DISABLED = 2;
    private final int GAL_HUB_SESSION_ALARM_RESET = 3;
    private final int GAL_HUB_SESSION_ALARM_EXPIRED = 4;
    private final int GAL_HUB_SESSION_LOCK_STATUS = 0;
    private final int GAL_HUB_SESSION_LOCK_SET = 1;
    private final int GAL_HUB_SESSION_LOCK_RELEASED = 2;
    private final int GAL_HUB_TOKEN_STATUS = 0;
    private final int GAL_HUB_TOKEN_CREATED = 1;
    private final int GAL_HUB_TOKEN_DESTROYED = 2;
    private final int GAL_HUB_SESSION_SETTINGS = 0;
    private final int GAL_HUB_SESSION_CREATED = 1;
    private final int GAL_HUB_SESSION_DESTROYED = 2;
    private final int GAL_HUB_SESSION_HISTORY = 3;
    private final int GAL_HUB_LISTENER_INITIALIZED = 0;
    private final int GAL_HUB_LISTENER_INITIALIZATION_ERROR = 1;
    private final int GAL_HUB_MESSAGE_QUEUE_CONTENTS = 0;

    private final int DEFAULT_GRAPHIC_UPDATE_WAIT_MSEC = 50;

    public VisualizerServer(VisualizerMainServer mainServer, Socket socket) 
	throws IOException
    {
	super(mainServer, socket);
	ui = mainServer.getUI();
	ui.setServer(this);

	providerIds = new ArrayList(INITIAL_NUM_PROVIDERS);
	providers = new ArrayList(INITIAL_NUM_PROVIDERS);
	providerMap = new HashMap(INITIAL_NUM_PROVIDERS);
	providerHighlightCounters = new HashMap(INITIAL_NUM_PROVIDERS);
    }
   
    public void serverOpReinitialize(GFrame frame)
    {
	ui.reset();
    } 

    /** Cleanup routine that is run when this server stops. */
    protected void cleanup()
    {
	final VisualizerUI _ui = ui;
	Runnable runnable = new Runnable() {
		public void run() {
		    _ui.processHubStop();
		}
	    };
	SwingUtilities.invokeLater(runnable);
    }
    
    //========================================================================
    // GUI-related dispatch methods
    //========================================================================
    
    public GFrame serverOpProviderStatus(GFrame frame)
    {
	GVector providerList = frame.getList(":provider_list");
	if(providerList == null) {
	    logWarningMessage("VisualizerServer.serverOpProviderStatus received a frame with a null `:provider_list` value.");
	    return null;
	}

	int numProviders = providerList.size();

	String id;
	String screenName;
	String iconFileName;
	GVector menuItemList;
	int numMenuItems = 0;
	List menuItems = null;
	ProviderInfo provider;
	GFrame providerFrame;
	GVector menuItem;
	boolean isConnected, isAvailable;

	// Process each provider in the incoming list.
	for(int providerIdx=0; providerIdx<numProviders; ++providerIdx) {

	    providerFrame = (GFrame) providerList.elementAt(providerIdx);

	    /*
	    if(!providerFrame.containsProperty(":connected")) {
		logWarningMessage("VisualizerServer.serverOpProviderStatus received a list item without a `:connected` value. Skipping.");
		continue;
	    }
	    */

	    id = providerFrame.getString(":provider_id");
	    if(id == null) {
		logWarningMessage("VisualizerServer.serverOpProviderStatus received a list item with a null `:provider_id` value. Skipping.");
		continue;
	    }

	    provider = (ProviderInfo) providerMap.get(id);

	    // Set defaults. 
	    isConnected = false;
	    isAvailable = false;

	    if(providerFrame.containsProperty(":connected"))
		isConnected = providerFrame.getIntAsBoolean(":connected").booleanValue();

	    if(providerFrame.containsProperty(":available"))
		isAvailable = providerFrame.getIntAsBoolean(":available").booleanValue();
	    
	    if(provider == null) {
		
		// Create a new provider.
		if(providerFrame.containsProperty(":init")) {
	 
		    screenName = providerFrame.getString(":screen_name");
		    if(screenName == null) {
			logWarningMessage("VisualizerServer.serverOpProviderStatus received a list item with a null `:screen_name` value. Skipping.");
			continue;
		    }
		    
		    // Icons are optional.
		    iconFileName = providerFrame.getString(":icon");
		    
		    // Menu items are optional.
		    menuItemList = providerFrame.getList(":menu_items");
		    if(menuItemList != null) {
			numMenuItems = menuItemList.size();
			menuItems = new ArrayList(numMenuItems);
			for(int menuItemIdx=0; menuItemIdx<numMenuItems; ++menuItemIdx) {
			    menuItem = (GVector) menuItemList.elementAt(menuItemIdx);
			    menuItems.add(new FrameMenuItem((String) menuItem.elementAt(0),
							    (GFrame) menuItem.elementAt(1)));
			}
		    }
		    
		    provider = new ProviderInfo(id, screenName, iconFileName, menuItems, 
						providerFrame.containsProperty(":hub_is_client"),
						isAvailable, isConnected);
		    providerIds.add(id);
		    providers.add(provider);		    
		    providerMap.put(id, provider);
		    providerHighlightCounters.put(id, new Integer(0));
		    
		    // Reset for next pass.
		    menuItems = null;
		} else {
		    logWarningMessage("VisualizerServer.serverOpProviderStatus could not find service provider with id \"" + id + "\". Skipping.");
		    continue;
		}
	    } else if(!provider.isHubTheClient() && !isConnected) {
		
		// If a Hub client disconnects, remove it from the GUI.
		providerIds.remove(id);
		providers.remove(provider);
		providerMap.remove(id);
		providerHighlightCounters.remove(id);
		ui.removeProvider(id);
		continue;
	    } else {
		// Update connection and availability status if appropriate.
		if(providerFrame.containsProperty(":connected"))
		    provider.setConnectionStatus(isConnected);
		
		if(providerFrame.containsProperty(":available"))
		    provider.setAvailabilityStatus(isAvailable);
	    }
	}   
	
	ui.update(providers, providerIds);

	return null;
    }

    public GFrame serverOpAddMenuItems(GFrame frame)
    {
	GVector menuItemList = frame.getList(":menu_items");
	if(menuItemList == null) {
	    logWarningMessage("VisualizerServer.serverOpAddMenuItems received a frame with a null `:menu_items` value.");
	    return null;
	}

	int numMenuItems = menuItemList.size();
	List menuItems = new ArrayList(numMenuItems);
	GVector menuItem;

	for(int menuItemIdx=0; menuItemIdx<numMenuItems; ++menuItemIdx) {
	    menuItem = (GVector) menuItemList.elementAt(menuItemIdx);
	    menuItems.add(new FrameMenuItem((String) menuItem.elementAt(0),
					    (GFrame) menuItem.elementAt(1)));
	}
 
	ui.addMenuItems(menuItems);

	return null;
    }

    public GFrame serverOpMessageEvent(GFrame frame)
    {
	String id = frame.getString(":provider_id");
	if(id == null) {
	    logWarningMessage("VisualizerServer.serverOpMessageEvent received a frame with a null `:provider_id` value.");
	    return null;
	}

	if(!providerIds.contains(id)) {
	    logWarningMessage("VisualizerServer.serverOpMessageEvent could not find service provider with id \"" + id + "\"");
	    return null;
	}

	GFrame msg = frame.getFrame(":msg");
	if(msg == null) {
	    logWarningMessage("VisualizerServer.serverOpMessageEvent received a frame with a null `:msg` value.");
	    return null;
	}

	Integer msgType = frame.getInt(":msg_type");
	if(msgType == null) {
	    logWarningMessage("VisualizerServer.serverOpMessageEvent received a frame with a null `:msg_type` value.");
	    return null;
	}

	String destination = frame.getString(":destination");
	if(destination == null) {
	    logWarningMessage("VisualizerServer.serverOpMessageEvent received a frame with a null `:destination` value.");
	    return null;
	}

	boolean toProvider = destination.equalsIgnoreCase("server");

	// Get the "reply requested" flag.
	int replyRequested = 0;
	if(msg.containsProperty(GFrame.GAL_ROUND_TRIP_FRAME_KEY))
	    replyRequested = msg.getInt(GFrame.GAL_ROUND_TRIP_FRAME_KEY).intValue();

	switch(msgType.intValue()) {
	case GalaxyMessage.GAL_MESSAGE_MSG_TYPE:
	    
	    // Display the message traffic graphic.
	    ui.sendMessage(id, toProvider);
	    
	    if(toProvider) {
		if(replyRequested == 1) {
		    Integer highlightCount = (Integer) providerHighlightCounters.get(id);
		    
		    // Highlight the provider if a) a reply is requested and b)
		    // it is not already highlighted.
		    if(highlightCount.intValue() == 0) {
			ui.highlightProvider(id, true);
		    }
		    
		    // Increment the highlight count.
		    providerHighlightCounters.put(id, new Integer(highlightCount.intValue() + 1));

		} else {
		    
		    // If no reply is requested, simply flash the provider.
		    ui.flashProvider(id);
		}
	    } 
	    break;
	case GalaxyMessage.GAL_REPLY_MSG_TYPE:
	case GalaxyMessage.GAL_DESTROY_MSG_TYPE:
	case GalaxyMessage.GAL_ERROR_MSG_TYPE:

	    // Display the message traffic graphic.
	    ui.sendMessage(id, toProvider);
	    Integer highlightCount = (Integer) providerHighlightCounters.get(id);

	    if(toProvider) {
		if(highlightCount.intValue() == 0)
		    ui.flashProvider(id);
	    } else {
		if(highlightCount.intValue() == 1) {
		    providerHighlightCounters.put(id, new Integer(0));
		    
		    // Turn off the provider highlight since all current reply
		    // requests have been accounted for.
		    ui.highlightProvider(id, false);
		} else if(highlightCount.intValue() > 1) {
		    providerHighlightCounters.put(id, new Integer(highlightCount.intValue() - 1));
		}
	    }
	    break;
	case GalaxyMessage.GAL_POSTPONE_MSG_TYPE:

	    // Display the message traffic graphic.
	    ui.sendMessage(id, toProvider);
	    break;
	default:
	    logWarningMessage("VisualizerServer.serverOpMessageEvent does not process messages of type " + msgType);
	}

	// Wait a little to allow the UI to update its graphics.
	try {
	    Thread.sleep(DEFAULT_GRAPHIC_UPDATE_WAIT_MSEC);
	} catch(Exception ex) {
	}

	return null;
    }

    //========================================================================
    // Hub output dispatch methods (debug and non-debug modes)
    //========================================================================

    public GFrame serverOpSessionStatus(GFrame frame)
    {
	StringBuffer textBuffer = new StringBuffer("Session");
	
	if(frame.containsProperty(":action")) {
	    int action = frame.getInt(":action").intValue();
	    switch(action) {
	    case GAL_HUB_SESSION_SETTINGS:
		textBuffer.append(" Settings");
		break;
	    case GAL_HUB_SESSION_CREATED:
		textBuffer.append(" Created");
		break;
	    case GAL_HUB_SESSION_DESTROYED:
		textBuffer.append(" Destroyed");
		break;
	    case GAL_HUB_SESSION_HISTORY:
		textBuffer.append(" History");
		break;
	    default:
		textBuffer.append(": Unknown Action");
	    }
	}
	
	if(frame.containsProperty(":session_id"))
	    textBuffer.append("\nsession id: " + frame.getString(":session_id"));
	
	if(frame.containsProperty(":settings"))
	    textBuffer.append("\nsettings:\n" + frame.getFrame(":settings").toFormattedString());
	
	if(frame.containsProperty(":history"))
	    textBuffer.append("\nhistory:\n" + frame.getFrame(":history").toFormattedString());
	
	if(frame.containsProperty(":debug"))
	    ui.setDebugText(textBuffer.toString());
	else {
	    ui.hubStatus(textBuffer.toString() + "\n\n");
	}
	return null;
    }

    public GFrame serverOpSessionLockStatus(GFrame frame)
    {
	StringBuffer textBuffer = new StringBuffer("Session Lock");
	
	if(frame.containsProperty(":action")) {
	    int action = frame.getInt(":action").intValue();
	    switch(action) {
	    case GAL_HUB_SESSION_LOCK_STATUS:
		textBuffer.append(" Status");
		break;
	    case GAL_HUB_SESSION_LOCK_SET:
		textBuffer.append(" Set");
		break;
	    case GAL_HUB_SESSION_LOCK_RELEASED:
		textBuffer.append(" Released");
		break;
	    default:
		textBuffer.append(": Unknown Action");
	    }
	}

	GVector list, listElement;
	list = frame.getList(":providers_sessions");
	if(list != null) {
	    for(int idx=0; idx<list.size(); ++idx) {
		listElement = (GVector) list.elementAt(idx);
		textBuffer.append("\nsession " + (String) listElement.elementAt(1) + " can write to provider " + (String) listElement.elementAt(0));
	    }
	}

	list = frame.getList(":providers_sessions_readers");
	if(list != null) {
	    for(int idx=0; idx<list.size(); ++idx) {
		listElement = (GVector) list.elementAt(idx);
		textBuffer.append("\nprovider " + (String) listElement.elementAt(0) + " only reading from session " + (String) listElement.elementAt(1));
	    }
	}

	list = frame.getList(":providers_sessions_writers");
	if(list != null) {
	    for(int idx=0; idx<list.size(); ++idx) {
		listElement = (GVector) list.elementAt(idx);
		textBuffer.append("\nprovider " + (String) listElement.elementAt(0) + " only writing to session " + (String) listElement.elementAt(1));
	    }
	}
	 
	if(frame.containsProperty(":lock_type"))
	    textBuffer.append("\nlock type: " + frame.getInt(":lock_type"));
	
	if(frame.containsProperty(":session_id"))
	    textBuffer.append("\nsession id: " + frame.getString(":session_id"));

	if(frame.containsProperty(":debug"))
	  ui.setDebugText(textBuffer.toString());
	else {
	    ui.hubStatus(textBuffer.toString() + "\n\n");
	}
	return null;
    }

    public GFrame serverOpTokenStatus(GFrame frame)
    {
	StringBuffer textBuffer = new StringBuffer("Token");
	
	if(frame.containsProperty(":action")) {
	    int action = frame.getInt(":action").intValue();
	    switch(action) {
	    case GAL_HUB_TOKEN_STATUS:
		textBuffer.append(" Status");
		break;
	    case GAL_HUB_TOKEN_CREATED:
		textBuffer.append(" Created");
		break;
	    case GAL_HUB_TOKEN_DESTROYED:
		textBuffer.append(" Destroyed");
		break;
	    default:
		textBuffer.append(": Unknown Action");
	    }
	}
	
	if(frame.containsProperty(":tidx"))
	    textBuffer.append("\ntidx: " + frame.getInt(":tidx"));
	
	if(frame.containsProperty(":state"))
	    textBuffer.append("\nstate:\n" + frame.getFrame(":state").toFormattedString());
	if(frame.containsProperty(":debug"))
	  ui.setDebugText(textBuffer.toString());
	else {
	    ui.hubStatus(textBuffer.toString() + "\n\n");
	}
	return null;
    }

    public GFrame serverOpHubDebugInfo(GFrame frame)
    {
	StringBuffer textBuffer = new StringBuffer("Hub Debug Info");

	if(frame.containsProperty(":type"))
	    textBuffer.append("\ntype: " + frame.getString(":type"));

	if(frame.containsProperty(":msg"))
	    textBuffer.append("\nmsg:\n" + frame.getString(":msg"));

	if(frame.containsProperty(":frame"))
	    textBuffer.append("\nframe:\n" + frame.getFrame(":frame").toFormattedString());

	ui.setDebugText(textBuffer.toString());
	return null;
    }

    public GFrame serverOpHubDebugPrompt(GFrame frame)
    {
	Environment env = getCurrentEnvironment();
	try {
	    env.postponeReply();
	} catch(IOException ioex) {
	    logErrorMessage("Caught exception: " + ioex.toString(), ioex);
	}
	ui.setEnv(env);
	return null;
    }

    protected void sendDebugCommand(Environment env, String cmd)
    {
	GFrame replyFrame = new Clause("hub_debug_cmd");
	replyFrame.setProperty(":debug_cmd", cmd);
	try {
	    env.reply(replyFrame);
	} catch(IOException ioex) {
	    logErrorMessage("Caught exception: " + ioex.toString(), ioex);
	}
    }

    //========================================================================
    // Strictly non-debug Hub output dispatch methods
    //========================================================================

    public GFrame serverOpListenerStatus(GFrame frame)
    {
	StringBuffer textBuffer = new StringBuffer("Listener");
	
	if(frame.containsProperty(":action")) {
	    int action = frame.getInt(":action").intValue();
	    switch(action) {
	    case GAL_HUB_LISTENER_INITIALIZED:
		textBuffer.append(" Initialized");
		break;
	    case GAL_HUB_LISTENER_INITIALIZATION_ERROR:
		textBuffer.append(" Initialization Error");
		break;
	    default:
		textBuffer.append(": Unknown Action");
	    }
	}

	if(frame.containsProperty(":port"))
	    textBuffer.append("\nport: " + frame.getInt(":port"));
	
	if(frame.containsProperty(":reused"))
	    textBuffer.append("\nreused: " + frame.getInt(":reused"));

	textBuffer.append("\n\n");
	System.out.print(textBuffer.toString());

	return null;
    }

    public GFrame serverOpSessionAlarmStatus(GFrame frame)
    {
	StringBuffer textBuffer = new StringBuffer("Session Alarm");
	
	if(frame.containsProperty(":action")) {
	    int action = frame.getInt(":action").intValue();
	    switch(action) {
	    case GAL_HUB_SESSION_ALARM_STATUS:
		textBuffer.append(" Status");
		break;
	    case GAL_HUB_SESSION_ALARM_ENABLED:
		textBuffer.append(" Enabled");
		break;
	    case GAL_HUB_SESSION_ALARM_DISABLED:
		textBuffer.append(" Disabled");
		break;
	    case GAL_HUB_SESSION_ALARM_RESET:
		textBuffer.append(" Reset");
		break;
	    case GAL_HUB_SESSION_ALARM_EXPIRED:
		textBuffer.append(" Expired");
		break;
	    default:
		textBuffer.append(": Unknown Action");
	    }
	}

	if(frame.containsProperty(":alarm_key"))
	    textBuffer.append("\nalarm key: " + frame.getString(":alarm_key"));

	if(frame.containsProperty(":session_id"))
	    textBuffer.append("\nsession id: " + frame.getString(":session_id"));

	if(frame.containsProperty(":expiration_secs"))
	    textBuffer.append("\nexpiration secs: " + frame.getInt(":expiration_secs"));

	if(frame.containsProperty(":remaining_secs"))
	    textBuffer.append("\nremaining secs: " + frame.getInt(":remaining_secs"));
	
	textBuffer.append("\n\n");
	System.out.print(textBuffer.toString());

	return null;
    }

    public GFrame serverOpNewServiceType(GFrame frame)
    {
	StringBuffer textBuffer = new StringBuffer("New Service Type");

	if(frame.containsProperty(":name"))
	    textBuffer.append("\nname: " + frame.getString(":name"));

	if(frame.containsProperty(":port"))
	    textBuffer.append("\nport: " + frame.getInt(":port"));

	textBuffer.append("\n\n");
	System.out.print(textBuffer.toString());

	return null;
    }

    public GFrame serverOpMessageQueueStatus(GFrame frame)
    {
	StringBuffer textBuffer = new StringBuffer("Message Queue");

	if(frame.containsProperty(":action")) {
	    int action = frame.getInt(":action").intValue();
	    switch(action) {
	    case GAL_HUB_MESSAGE_QUEUE_CONTENTS:
		textBuffer.append(" Contents");
		if(frame.containsProperty(":queue")) {
		    Int32 int32 = frame.getInt32(":queue");
		    if(int32 != null) {
			int[] array = int32.getIntArray();
			textBuffer.append("\nqueue: "); 
			for(int idx=0; idx<array.length; ++idx)
			    textBuffer.append(array[idx] + " ");
		    }
		}
		break;
	    default:
		textBuffer.append(": Unknown Action");
	    }
	}
	
	textBuffer.append("\n\n");
	System.out.print(textBuffer.toString());

	return null;
    }
}
