/*
  This file (c) Copyright 1998 - 2002 The MITRE Corporation
  
  This file is part of the Galaxy Communicator system. It is licensed
  under the conditions described in the file LICENSE in the root 
  directory of the Galaxy Communicator system.
*/

package org.mitre.galaxycommunicator.visualizer;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.Border;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.InvocationTargetException;

import galaxy.server.*;
import galaxy.lang.*;
import galaxy.util.Logger;

/**
 * @author Steve Wohlever, MITRE Corporation, wohlever@mitre.org
 */
public class VisualizerUI extends JFrame 
    implements ActionListener, WindowListener
{
    // The main drawing area components.
    private VisualizerDrawingArea drawingArea;

    // Control menu components that need to be accessible.
    private JMenu controlMenu;
    private JMenuItem hubControlMenuItem;
    private JMenuItem hubOutputToggleMenuItem;
    private JMenuItem hubStatusToggleMenuItem;
    private JMenuItem hubConfigMenuItem;

    private JMenuItem setLabelBackgroundColorMenuItem;
    private JMenuItem setHubRunningColorMenuItem;
    private JMenuItem setHubStoppedColorMenuItem;
    private JMenuItem setProviderColorMenuItem;
    private JMenuItem setToProviderArrowColorMenuItem;
    private JMenuItem setToHubArrowColorMenuItem;

    // The Hub output pane components.
    private JTextArea hubOutputArea;
    private int hubOutputAreaRows = 10;
    private int hubOutputAreaColumns = 50;
    private JFrame hubOutputFrame;
    private JButton pauseHubOutputButton;
    private JButton clearHubOutputButton;
    private StringBuffer hubOutputAreaBuffer;

    // The Hub status pane components.
    private JTextArea hubStatusArea;
    private int hubStatusAreaRows = 10;
    private int hubStatusAreaColumns = 50;
    private JFrame hubStatusFrame;
    private JButton pauseHubStatusButton;
    private JButton clearHubStatusButton;
    private StringBuffer hubStatusAreaBuffer;

    private JPanel hubOutputPanel;
    private JPanel hubStatusPanel;

    // Label that displays the current program file.
    private JLabel programFileLabel;

    // The menu bar;
    private JMenuBar menuBar; 

    private JFileChooser fileChooser; 

    // Service provider-independent menu.
    private JMenu frameMenu = null;

    // The Hub process.
    private Process hubProcess;

    // Run status flag of Hub process.
    private volatile boolean hubRunning;
    
    // The Hub "standard out" and "standard error" streams and 
    // processing threads.
    private InputStream hubStdOut;
    private InputStream hubStdErr;
    private OutputThread hubProcessStdOutThread;
    private OutputThread hubProcessStdErrThread;

    // Waits for the Hub process to exit.
    private WaitingThread waiterThread;

    // List of service providers (in the order that the Hub specified).
    private java.util.List providers = null;

    // List of the service provider ids.
    private java.util.List providerIds = null;

    // Maps JMenuItems to service provider FrameMenuItems.
    private Map providerMenuItemMap = null;

    // Maps service provider ids to popup menus.
    private Map popupMenuMap = null;

    // Maps JMenuItems to service provider ids.
    private Map menuItemToProviderIdMap = null;

    // Maps JMenuItems to provider-independent menu item frames.
    private Map menuItemToFrameMap = null;

    // The server command line argument parser.
    private ServerArgParser argParser;

    private JColorChooser colorChooser;

    // The Galaxy Communicator visualizer server associated with this UI.
    private VisualizerServer server;

    private Logger logger = null;

    private DebugUI debugUI = null;
    private HubFrame hubFrame = null;
    private int port = -1;

    private final int MAX_HUB_ARGS = 100;

    private VisualizerProperties props = null;

    public VisualizerUI(String title, ServerArgParser argParser, int port) throws Exception
    {
	super(title);
	this.argParser = argParser;
	this.port = port;

	// Process properties and command line arguments.
	props = new VisualizerProperties(argParser.getStringArg("-props", null));

	logger = Logger.getLogger();

	providerMenuItemMap = new HashMap();
	popupMenuMap = new HashMap();
	menuItemToProviderIdMap = new HashMap();
	providerIds = new ArrayList();
	menuItemToFrameMap = new HashMap();
	hubFrame = new HubFrame(props);

	// Get the frame's content pane.
	Container contentPane = getContentPane();

	if(argParser.isArg("-use_hub_output_dialog"))
	    props.setUseHubOutputDialog(true);

	if(argParser.isArg("-use_hub_status_dialog"))
	    props.setUseHubStatusDialog(true);

	props.setServerLocationFile(argParser.getStringArg("-server_location_file", null));

	debugUI = new DebugUI();

	if(argParser.isArg("-debug")) {
	    props.setDebug(true);
	}

	if(argParser.isArg("-validate")) {
	    props.setValidate(true);
	}
	
	hubFrame.setValidateEnabled(props.getDebug());
	hubFrame.setValidateEnabled(props.getValidate());

	props.setProgramFile(argParser.getStringArg("-pgm_file", props.getProgramFile()));
	props.setHubPath(argParser.getStringArg("-hub", props.getHubPath()));

	props.setInit(argParser.getStringArg("-init", props.getInit()));
	props.setLocations(argParser.getStringArg("-locations", props.getLocations()));
	props.setHubVerbosity(argParser.getIntegerArg("-hub_verbosity", props.getHubVerbosity()));

	hubFrame.setInitValues(props.getInit());
	hubFrame.setLocations(props.getLocations());
	hubFrame.setVerbosity(props.getHubVerbosity());

	// Build a nice border.
	Border raisedBevelBorder = BorderFactory.createRaisedBevelBorder();
	Border loweredBevelBorder = BorderFactory.createLoweredBevelBorder();
	Border border = BorderFactory.createCompoundBorder(raisedBevelBorder, 
							   loweredBevelBorder);

	// Create a drawing area.
	drawingArea = new VisualizerDrawingArea(this);
	drawingArea.setBorder(border);

	// Add this UI as a window listener.
	addWindowListener(this);

	// Process the program file if it uses the DEMO_HOME prefix
	String programFile = props.getProgramFile();
	String demoHomePath = props.getDemoHomePath();
	String demoHomePrefix = props.getDemoHomePrefix();
	if(programFile != null && demoHomePath != null &&
	   programFile.startsWith(demoHomePrefix)) {
	    programFile = demoHomePath + programFile.substring(demoHomePrefix.length(), programFile.length());
	}

	fileChooser = new JFileChooser();

	// Set up the main menu.
        menuBar = new JMenuBar();
	setJMenuBar(menuBar);


	// Set up the file menu.
	JMenu fileMenu = new JMenu("File");
	menuBar.add(fileMenu);

	JMenuItem openMenuItem = new JMenuItem("Select Program File...");
	openMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		openMenuItem_actionPerformed(e);
	    }
	});
	fileMenu.add(openMenuItem);

	JMenuItem selectHubMenuItem = new JMenuItem("Select Hub Executable...");
	selectHubMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		selectHubMenuItem_actionPerformed(e);
	    }
	});
	fileMenu.add(selectHubMenuItem);

	JMenuItem selectLocationFileMenuItem = new JMenuItem("Select Server Location File...");
	selectLocationFileMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		selectLocationFileMenuItem_actionPerformed(e);
	    }
	});
	fileMenu.add(selectLocationFileMenuItem);

	fileMenu.addSeparator();

	JMenuItem exitMenuItem = new JMenuItem("Exit");
	exitMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		exitMenuItem_actionPerformed(e);
	    }
	});
	fileMenu.add(exitMenuItem);


	// Set up the control menu
	controlMenu = new JMenu("Controls");
	menuBar.add(controlMenu);

	hubControlMenuItem = new JMenuItem("Start Hub");
	hubControlMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		hubControlMenuItem_actionPerformed(e);
	    }
	});
	controlMenu.add(hubControlMenuItem);

	hubOutputToggleMenuItem = new JMenuItem("Hide Hub Output");
	hubOutputToggleMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		hubOutputToggleMenuItem_actionPerformed(e);
	    }
	});
	controlMenu.add(hubOutputToggleMenuItem);

	hubStatusToggleMenuItem = new JMenuItem("Hide Hub Status");
	hubStatusToggleMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		hubStatusToggleMenuItem_actionPerformed(e);
	    }
	});
	controlMenu.add(hubStatusToggleMenuItem);

	hubConfigMenuItem = new JMenuItem("Configure Hub...");
	hubConfigMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		hubConfigMenuItem_actionPerformed(e);
	    }
	});
	controlMenu.add(hubConfigMenuItem);

	JMenu graphicsMenu = new JMenu("Graphics");
	menuBar.add(graphicsMenu);    

	JMenu hubGraphicsMenu = new JMenu("Hub Graphics...");
	JMenu providerGraphicsMenu = new JMenu("Service Provider Graphics...");

	graphicsMenu.add(hubGraphicsMenu);
	graphicsMenu.add(providerGraphicsMenu);

	setHubRunningColorMenuItem = new JMenuItem("Set Hub \"Running\" Color...");
	setHubRunningColorMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		setHubRunningColorMenuItem_actionPerformed(e);
	    }
	});
	hubGraphicsMenu.add(setHubRunningColorMenuItem);

	setHubStoppedColorMenuItem = new JMenuItem("Set Hub \"Stopped\" Color...");
	setHubStoppedColorMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		setHubStoppedColorMenuItem_actionPerformed(e);
	    }
	});
	hubGraphicsMenu.add(setHubStoppedColorMenuItem);

	setToHubArrowColorMenuItem = new JMenuItem("Set \"To Hub\" Message Traffic Color...");
	setToHubArrowColorMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		setToHubArrowColorMenuItem_actionPerformed(e);
	    }
	});
	hubGraphicsMenu.add(setToHubArrowColorMenuItem);

	setProviderColorMenuItem = new JMenuItem("Set Service Provider Color...");
	setProviderColorMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		setProviderColorMenuItem_actionPerformed(e);
	    }
	});
	providerGraphicsMenu.add(setProviderColorMenuItem);

	setLabelBackgroundColorMenuItem = new JMenuItem("Set Label Background Color...");
	setLabelBackgroundColorMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		setLabelBackgroundColorMenuItem_actionPerformed(e);
	    }
	});
	providerGraphicsMenu.add(setLabelBackgroundColorMenuItem);

	setToProviderArrowColorMenuItem = new JMenuItem("Set \"To Service Provider\" Message Traffic Color...");
	setToProviderArrowColorMenuItem.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		setToProviderArrowColorMenuItem_actionPerformed(e);
	    }
	});
	providerGraphicsMenu.add(setToProviderArrowColorMenuItem);

	// Initialize the program file label.
	programFileLabel = new JLabel();
	if(programFile == null) {
	    programFileLabel.setText("  Current Program File: none");
	} else {
	    File file = new File(programFile);
	    programFileLabel.setText("  Current Program File: " + file.getName());
	}

	// Build the Hub control panel.
	JPanel topPanel = new JPanel();
	topPanel.setLayout(new BoxLayout(topPanel, BoxLayout.Y_AXIS));
	topPanel.add(drawingArea);

	// Set up the Hub output pane.
	hubOutputArea = new JTextArea(hubOutputAreaRows, hubOutputAreaColumns);
	hubOutputArea.setEditable(false);
	hubOutputArea.setLineWrap(true);
	hubOutputArea.setBorder(border);
	
	hubStatusArea = new JTextArea(hubStatusAreaRows, hubStatusAreaColumns);
	hubStatusArea.setEditable(false);
	hubStatusArea.setLineWrap(true);
	hubStatusArea.setBorder(border);

	clearHubOutputButton = new JButton("Clear Output");
	clearHubOutputButton.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		clearHubOutputButton_actionPerformed(e);
	    }
	});

	clearHubStatusButton = new JButton("Clear Status");
	clearHubStatusButton.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		clearHubStatusButton_actionPerformed(e);
	    }
	});

	pauseHubOutputButton = new JButton("Pause Output");
	pauseHubOutputButton.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		pauseHubOutputButton_actionPerformed(e);
	    }
	});

	pauseHubStatusButton = new JButton("Pause Status");
	pauseHubStatusButton.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		pauseHubStatusButton_actionPerformed(e);
	    }
	});

	JButton closeHubOutputDialogButton = new JButton("Close");
	closeHubOutputDialogButton.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		closeHubOutputDialogButton_actionPerformed(e);
	    }
	}); 

	JButton closeHubStatusDialogButton = new JButton("Close");
	closeHubStatusDialogButton.addActionListener(new ActionListener()  {
	    public void actionPerformed(ActionEvent e) {
		closeHubStatusDialogButton_actionPerformed(e);
	    }
	}); 

	// Set up the Hub output pane button panel.
	JPanel hubOutputButtonPanel = new JPanel();
	hubOutputButtonPanel.add(clearHubOutputButton);
	hubOutputButtonPanel.add(pauseHubOutputButton);
	hubOutputButtonPanel.add(closeHubOutputDialogButton);

	// Set up the Hub status pane button panel.
	JPanel hubStatusButtonPanel = new JPanel();
	hubStatusButtonPanel.add(clearHubStatusButton);
	hubStatusButtonPanel.add(pauseHubStatusButton);
	hubStatusButtonPanel.add(closeHubStatusDialogButton);

	// Build the Hub output panel.
	JScrollPane hubOutputScrollPane = new JScrollPane(hubOutputArea);
	hubOutputPanel = new JPanel(new BorderLayout());
	hubOutputPanel.add(new JLabel("  Hub Output"), BorderLayout.NORTH);
	hubOutputPanel.add(hubOutputScrollPane, BorderLayout.CENTER);
	hubOutputPanel.add(hubOutputButtonPanel, BorderLayout.SOUTH);
	hubOutputAreaBuffer = new StringBuffer();

	// Build the Hub status panel.
	JScrollPane hubStatusScrollPane = new JScrollPane(hubStatusArea);
	hubStatusPanel = new JPanel(new BorderLayout());
	hubStatusPanel.add(new JLabel("  Hub Status"), BorderLayout.NORTH);
	hubStatusPanel.add(hubStatusScrollPane, BorderLayout.CENTER);
	hubStatusPanel.add(hubStatusButtonPanel, BorderLayout.SOUTH);
	hubStatusAreaBuffer = new StringBuffer();
	    
	JPanel textPanel = new JPanel(new BorderLayout());

	// Create the main panel.
	JPanel mainPanel = new JPanel(new BorderLayout());

	mainPanel.add(topPanel, BorderLayout.NORTH);
	mainPanel.add(textPanel, BorderLayout.CENTER);

	if(props.getUseHubOutputDialog()) {
	    hubOutputFrame = new JFrame("Hub Output");
	    hubOutputFrame.getContentPane().add(hubOutputScrollPane, BorderLayout.CENTER);
	    hubOutputFrame.getContentPane().add(hubOutputButtonPanel, BorderLayout.SOUTH);
	    hubOutputFrame.addWindowListener(this);
	    hubOutputFrame.pack();
	    closeHubOutputDialogButton.setVisible(true);
	    hubOutputToggleMenuItem.setText("Show Hub Output");
	} else {
	    textPanel.add(hubOutputPanel, BorderLayout.NORTH);
	    closeHubOutputDialogButton.setVisible(false);
	}

	if(props.getUseHubStatusDialog()) {
	    hubStatusFrame = new JFrame("Hub Status");
	    hubStatusFrame.getContentPane().add(hubStatusScrollPane, BorderLayout.CENTER);
	    hubStatusFrame.getContentPane().add(hubStatusButtonPanel, BorderLayout.SOUTH);
	    hubStatusFrame.addWindowListener(this);
	    hubStatusFrame.pack();
	    closeHubStatusDialogButton.setVisible(true);
	    hubStatusToggleMenuItem.setText("Show Hub Status");
	} else {
	    textPanel.add(hubStatusPanel, BorderLayout.SOUTH);
	    closeHubStatusDialogButton.setVisible(false);
	}

	mainPanel.add(programFileLabel, BorderLayout.SOUTH);

	// Add the main panel to the frame.
	contentPane.add(mainPanel);
	
	colorChooser = new JColorChooser();

	// Pack the frame.
	pack();

	// Make the UI visible.
	setVisible(true);
    }

    VisualizerDrawingArea getDrawingArea()
    {
	return drawingArea;
    }

    VisualizerProperties getProperties()
    {
	return props;
    }

    //=========================================================================
    // Update methods
    //=========================================================================

    public void removeProvider(String providerId) 
    { 
      final String _providerId = providerId;
      Runnable runnable = new Runnable() {
	  public void run() {drawingArea.removeProviderGraphic(_providerId);}
	};

      try {
	SwingUtilities.invokeAndWait(runnable);
      } catch(InterruptedException iex) {
	logger.logErrorMessage("Caught exception during call to SwingUtilities.invokeAndWait: " + iex.toString(), iex);
      } catch(InvocationTargetException itex) {
	logger.logErrorMessage("Caught exception during call to SwingUtilities.invokeAndWait: " + itex.toString() + " -> " + itex.getTargetException().toString(), itex);
      }
    }

    public void update(java.util.List providers, java.util.List providerIds)
    {
	final java.util.List _providers = providers;
	final java.util.List _providerIds = providerIds;
	Runnable runnable = new Runnable() {
		public void run() {updateProviders(_providers, _providerIds);}
	    };

	// To ensure that providers and providerIds are not changed before they are used.
	try {
	  SwingUtilities.invokeAndWait(runnable);
	} catch(InterruptedException iex) {
	  logger.logErrorMessage("Caught exception during call to SwingUtilities.invokeAndWait: " + iex.toString(), iex);
	} catch(InvocationTargetException itex) {
	  logger.logErrorMessage("Caught exception during call to SwingUtilities.invokeAndWait: " + itex.toString() + " -> " + itex.getTargetException().toString(), itex);
	}
    }

    private void updateProviders(java.util.List providers, java.util.List providerIds)
    {
	FrameMenuItem providerMenuItem;
	java.util.List providerMenuItems;
	int numMenuItems;
	JMenuItem menuItem;
	JPopupMenu popupMenu;
	String providerId;
	ProviderInfo provider;
	String iconFileName;
	int numProviders = providers.size();

	this.providers = providers;
	this.providerIds = providerIds;

	if(!hubRunning) {
	    // "Turn on" the Hub icon in the drawing area. This is here
            // to support the use of Hubs that are not started by this UI.
	    // Do not set the hubIsRunning flag since it is used to manage
	    // Hubs started by this UI.
	    drawingArea.setHubIsRunning(true);

	    // Since an external Hub is apparently being used, disable 
            // the local menu for starting the Hub.
            controlMenu.setEnabled(false);
        }

	for(int providerIdx=0; providerIdx<numProviders; ++providerIdx) {
	    provider = (ProviderInfo) providers.get(providerIdx);
	    providerId = (String) provider.getId();

	    // Initialize new providers.
	    if(!provider.isInitialized()) {
		
		// Process the icon file name
		iconFileName = provider.getIconFileName();
		String demoHomePath = props.getDemoHomePath();
		String demoHomePrefix = props.getDemoHomePrefix();
		if(iconFileName != null && demoHomePath != null &&
		   iconFileName.startsWith(demoHomePrefix)) {
		    iconFileName = demoHomePath + iconFileName.substring(demoHomePrefix.length(), iconFileName.length());
		    provider.setIconFileName(iconFileName);
		}
		
		providerMenuItems = provider.getMenuItems();
		if(providerMenuItems != null) {
		    numMenuItems = providerMenuItems.size();
		    if(numMenuItems > 0) {
			popupMenu = new JPopupMenu();
			popupMenuMap.put(providerId, popupMenu);
			for(int menuItemIdx=0; menuItemIdx< numMenuItems; ++menuItemIdx) {
			    providerMenuItem = (FrameMenuItem) providerMenuItems.get(menuItemIdx);
			    menuItem = new JMenuItem(providerMenuItem.getName());
			    menuItem.addActionListener(this);
			    
			    providerMenuItemMap.put(menuItem, providerMenuItem);
			    menuItemToProviderIdMap.put(menuItem, providerId);
			    
			    popupMenu.add(menuItem);
			}
		    }
		}
		initializeProviderGraphic(provider);
		provider.setInitialized(true);
	    } else { 
	      drawingArea.setProviderConnectionStatus(providerId, provider.isConnected());
	      drawingArea.setProviderAvailabilityStatus(providerId, provider.isAvailable());
	    }
	}
    }

    public void addMenuItems(java.util.List menuItems)
    {
	final java.util.List _menuItems = menuItems;
	Runnable runnable = new Runnable() {
	    public void run() {addMenuItemsToMenuBar(_menuItems);}
	};
	SwingUtilities.invokeLater(runnable);
    }

    private void addMenuItemsToMenuBar(java.util.List menuItems)
    {
        if(!hubRunning) {
	    // "Turn on" the Hub icon in the drawing area. This is here
            // to support the use of Hubs that are not started by this UI.
	    // Do not set the hubIsRunning flag since it is used to manage
	    // Hubs started by this UI.
	    drawingArea.setHubIsRunning(true);

            // Since an external Hub is apparently being used, disable 
            // the local menu for starting the Hub.
            controlMenu.setEnabled(false);
        }

	int numMenuItems = menuItems.size();
	if(numMenuItems > 0) {
	    if(frameMenu == null) {
		frameMenu = new JMenu("Send Message");
		menuBar.add(frameMenu);
	    }
	    
	    FrameMenuItem frameMenuItem;
	    JMenuItem menuItem;
	    for(int idx=0; idx<numMenuItems; ++idx) {
		frameMenuItem = (FrameMenuItem) menuItems.get(idx);
		menuItem = new JMenuItem(frameMenuItem.getName());
		menuItem.addActionListener(this);
		menuItemToFrameMap.put(menuItem, frameMenuItem.getFrame());
		frameMenu.add(menuItem);
	    }
	    validate();
	}
    }

    private void initializeProviderGraphic(ProviderInfo provider)
    {
	drawingArea.addProviderGraphic(provider);
    }

    /** Set this UI's server. */
    public void setServer(VisualizerServer server)
    {
	this.server = server;
	debugUI.setServer(server);
    }
    
    public void setEnv(Environment env)
    {
	debugUI.setEnv(env);
	debugUI.setVisible(true);
    }

    public void setDebugText(String text)
    {
	final String _text = text;
	Runnable runnable = new Runnable() {
		public void run() {debugUI.setText(_text); }
	    };
	SwingUtilities.invokeLater(runnable);
    }

    //=========================================================================
    // Various event handling methods
    //=========================================================================

    /** 
     * Processes menu item selections. 
     *
     * @param event the menu selection event
     */
    public void actionPerformed(ActionEvent event)
    {
	JMenuItem source = (JMenuItem) event.getSource();
	
	FrameMenuItem providerMenuItem = (FrameMenuItem) providerMenuItemMap.get(source);
	GFrame frame = (GFrame) menuItemToFrameMap.get(source);
	if(providerMenuItem != null) {
	    String providerId = (String) menuItemToProviderIdMap.get(source);
	    if(!drawingArea.isConnectedToHub(providerId)) {
		JOptionPane.showMessageDialog(this, "Cannot process menu option for provider \"" + providerId + "\" since it is not connected to the Hub.", "Error", JOptionPane.ERROR_MESSAGE);
		return;
	    }

	    frame = providerMenuItem.getFrame();
	    try {
		if(server != null) {
		    Environment env = server.getCurrentEnvironment();
		    env.writeFrame(frame);
		}
	    } catch(IOException ioex) {
		logger.logErrorMessage("VisualizerUI caught exception while sending frame: " + ioex.toString(), ioex);
	    }
	} else if(frame != null) {
	    try {
		if(server != null) {
		    Environment env = server.getCurrentEnvironment();
		    env.writeFrame(frame);
		}
	    } catch(IOException ioex) {
		logger.logErrorMessage("VisualizerUI caught exception while sending frame: " + ioex.toString(), ioex);
	    }
	}
    }

    
    /** Clears the Hub output pane. */
    public void clearHubOutputButton_actionPerformed(ActionEvent e)
    {
	hubOutputArea.setText(null);
    } 

    public void pauseHubOutputButton_actionPerformed(ActionEvent e)
    {
	if(pauseHubOutputButton.getText().equals("Pause Output")) {
	    pauseHubOutputButton.setText("Resume Output");
	} else {
	    pauseHubOutputButton.setText("Pause Output");
	    appendHubOutput(hubOutputAreaBuffer.toString());
	    hubOutputAreaBuffer.delete(0, hubOutputAreaBuffer.length());
	}
    } 

    public void clearHubStatusButton_actionPerformed(ActionEvent e)
    {
	hubStatusArea.setText(null);
    } 

    public void pauseHubStatusButton_actionPerformed(ActionEvent e)
    {
	if(pauseHubStatusButton.getText().equals("Pause Status")) {
	    pauseHubStatusButton.setText("Resume Status");
	} else {
	    pauseHubStatusButton.setText("Pause Status");
	    appendHubOutput(hubStatusAreaBuffer.toString());
	    hubStatusAreaBuffer.delete(0, hubStatusAreaBuffer.length());
	}
    } 

    public void closeHubOutputDialogButton_actionPerformed(ActionEvent e)
    {
	hubOutputFrame.setVisible(false);
	hubOutputToggleMenuItem.setText("Show Hub Output");
    }

    public void closeHubStatusDialogButton_actionPerformed(ActionEvent e)
    {
	hubStatusFrame.setVisible(false);
	hubStatusToggleMenuItem.setText("Show Hub Status");
    }

    public void openMenuItem_actionPerformed(ActionEvent e)
    {	 
	File currentFile = new File(props.getProgramFile());
	fileChooser.setCurrentDirectory(currentFile);
	int returnVal = fileChooser.showOpenDialog(this);
	if(returnVal == JFileChooser.APPROVE_OPTION) {
	    File file = fileChooser.getSelectedFile();
	    props.setProgramFile(file.getAbsolutePath());
	    programFileLabel.setText("  Current Program File: " + file.getName());
	}
    }

    public void selectHubMenuItem_actionPerformed(ActionEvent e)
    {
	fileChooser.setCurrentDirectory(new File(props.getHubPath()));
	int returnVal = fileChooser.showOpenDialog(this);
	if(returnVal == JFileChooser.APPROVE_OPTION) {
	    File file = fileChooser.getSelectedFile();
	    props.setHubPath(file.getAbsolutePath());
	}
    }
    
    public void selectLocationFileMenuItem_actionPerformed(ActionEvent e)
    {
	String serverLocationFile = props.getServerLocationFile();
	if(serverLocationFile != null)
	    fileChooser.setCurrentDirectory(new File(serverLocationFile));
	int returnVal = fileChooser.showOpenDialog(this);
	if(returnVal == JFileChooser.APPROVE_OPTION) {
	    File file = fileChooser.getSelectedFile();
	    props.setServerLocationFile(file.getAbsolutePath());
	}
    }

    /** Exits the application. */
    public void exitMenuItem_actionPerformed(ActionEvent e)
    {
	processHubStop();
	System.exit(0);
    }  

    public void hubControlMenuItem_actionPerformed(ActionEvent e)
    {
	if(hubControlMenuItem.getText().equals("Start Hub")) {
	    String programFile = props.getProgramFile();
	    if(programFile == null) {
		JOptionPane.showMessageDialog(this, "You must select a program file before starting the Hub.", "Hub Error", JOptionPane.ERROR_MESSAGE);
		return; 
	    }
	    
	    try{
		// Determine the Hub command line arguments.
		String[] tempCmdArray = new String[MAX_HUB_ARGS];

		int idx = 0;

		// Set the Hub executable.
		tempCmdArray[idx] = props.getHubPath(); ++idx;

		// Set the -gui parameter.
		tempCmdArray[idx] = "-gui"; ++idx;
		tempCmdArray[idx] = "localhost:" + port; ++idx;

		// Set the program file.
		tempCmdArray[idx] = "-pgm_file"; ++idx;
		tempCmdArray[idx] = programFile; ++idx;

		// Set -debug;
		if(hubFrame.debugEnabled()) {
		    tempCmdArray[idx] = "-debug"; ++idx;
		}
		
		// Set -validate.
		if(hubFrame.validateEnabled()) {
		    tempCmdArray[idx] = "-validate"; ++idx;
		}
		
		// Set -verbosity.
		tempCmdArray[idx] = "-verbosity"; ++idx;
		tempCmdArray[idx] = Integer.toString(hubFrame.getVerbosity()); ++idx;
 
		// Set -server_location_file.
		String serverLocationFile = props.getServerLocationFile();
		if(serverLocationFile != null) {
		    tempCmdArray[idx] = "-server_location_file"; ++idx;
		    tempCmdArray[idx] = serverLocationFile; ++idx;
		}

		// Set -init.
		String initValues = hubFrame.getInitValues();
		if(initValues != null && initValues.length() > 0) {
		    tempCmdArray[idx] = "-init"; ++idx;
		    tempCmdArray[idx] = initValues; ++idx;
		}

		// Set -locations.
		String locations = hubFrame.getLocations();
		if(locations != null && locations.length() > 0) {
		    tempCmdArray[idx] = "-locations"; ++idx;
		    tempCmdArray[idx] = locations; ++idx;
		}

		// Build the final command line array.
		int cmdLength = idx;
		String[] cmdArray = new String[cmdLength];
		for(idx = 0; idx<cmdLength; ++idx) {
		    cmdArray[idx] = tempCmdArray[idx];
		}

		hubFrame.setEnabled(false);
		hubProcess = Runtime.getRuntime().exec(cmdArray);
		hubRunning = true;
		hubControlMenuItem.setText("Stop Hub");
		drawingArea.setHubIsRunning(hubRunning);
		waiterThread = new WaitingThread(hubProcess);
		waiterThread.start();
		hubStdOut = hubProcess.getInputStream();
		hubStdErr = hubProcess.getErrorStream();
		hubProcessStdOutThread = new OutputThread("HubStdOut", hubStdOut);
		hubProcessStdOutThread.start();
		hubProcessStdErrThread = new OutputThread("HubStdErr", hubStdErr);
		hubProcessStdErrThread.start();
	    } catch(IOException ioex) {
		JOptionPane.showMessageDialog(this, "Exception caught: " + ioex.getMessage(), "Hub Error", JOptionPane.ERROR_MESSAGE);
		processHubStop();
	    }
	    
	} else {
	    processHubStop();
	}
    }
    
    public void hubOutputToggleMenuItem_actionPerformed(ActionEvent e)
    {
	Component hubOutputComponent;
	if(hubOutputFrame != null)
	    hubOutputComponent = hubOutputFrame;
	else
	    hubOutputComponent = hubOutputPanel;

	if(hubOutputComponent.isVisible()) {
	    hubOutputComponent.setVisible(false);
	    hubOutputToggleMenuItem.setText("Show Hub Output");
	} else {
	    hubOutputComponent.setVisible(true);
	    hubOutputToggleMenuItem.setText("Hide Hub Output");
	}

	pack();
	repaint();
    }

    public void hubStatusToggleMenuItem_actionPerformed(ActionEvent e)
    {
	Component hubStatusComponent = hubStatusPanel;
	if(hubStatusFrame != null)
	    hubStatusComponent = hubStatusFrame;
	else
	    hubStatusComponent = hubStatusPanel;

	if(hubStatusComponent.isVisible()) {
	    hubStatusComponent.setVisible(false);
	    hubStatusToggleMenuItem.setText("Show Hub Status");
	} else {
	    hubStatusComponent.setVisible(true);
	    hubStatusToggleMenuItem.setText("Hide Hub Status");
	}

	pack();
	repaint();
    }

    public void hubConfigMenuItem_actionPerformed(ActionEvent e)
    {
	if(!hubFrame.isVisible()) {
	    Point point = getLocationOnScreen();
	    hubFrame.setVisible(true);
	}
    }

    public void setHubRunningColorMenuItem_actionPerformed(ActionEvent e)
    {
	Color currentColor = props.getHubRunningColor();
	Color newColor = colorChooser.showDialog(this, "Set Hub \"Running\" Color", 
						 currentColor);
	if(newColor != null) {
	    currentColor = newColor;	 
	    props.setHubRunningColor(newColor);
	    drawingArea.setHubRunningColor(newColor);
	    logger.logMessage("Hub \"running\" color = " + newColor.toString(),
			      MainServer.INITIALIZATION_DETAILS_VERBOSITY_LEVEL);
	}
    }
    
    public void setHubStoppedColorMenuItem_actionPerformed(ActionEvent e)
    {
	Color currentColor = props.getHubStoppedColor();
	Color newColor = colorChooser.showDialog(this, "Set Hub \"Stopped\" Color", 
						 currentColor);
	if(newColor != null) {
	    currentColor = newColor;	
	    props.setHubStoppedColor(newColor);
	    drawingArea.setHubStoppedColor(newColor);
	    logger.logMessage("Hub \"stopped\" color = " + newColor.toString(),
			      MainServer.INITIALIZATION_DETAILS_VERBOSITY_LEVEL);
	}
    } 

    public void setToHubArrowColorMenuItem_actionPerformed(ActionEvent e)
    {
	Color currentColor = props.getMessageTrafficToHubLineColor();
	Color newColor = colorChooser.showDialog(this, "Set \"To Hub\" Message Traffic Color", 
						 currentColor);
	if(newColor != null) {
	    currentColor = newColor;
	    props.setMessageTrafficToHubLineColor(newColor);
	    drawingArea.setToHubArrowColor(newColor);
	    logger.logMessage("\"To Hub\" message traffic color = " + newColor.toString(),
			      MainServer.INITIALIZATION_DETAILS_VERBOSITY_LEVEL);
	}
    }

    public void setProviderColorMenuItem_actionPerformed(ActionEvent e)
    {
	Color currentColor = props.getProviderColor();
	Color newColor = colorChooser.showDialog(this, "Set Service Provider Color", 
						 currentColor);
	if(newColor != null) {
	    currentColor = newColor;
	    props.setProviderColor(newColor);
	    drawingArea.setProviderColor(newColor);
	    logger.logMessage("Service provider color = " + newColor.toString(),
			      MainServer.INITIALIZATION_DETAILS_VERBOSITY_LEVEL);
	}
    }

    public void setLabelBackgroundColorMenuItem_actionPerformed(ActionEvent e)
    {
	Color currentColor = props.getLabelBackgroundColor();
	Color newColor = colorChooser.showDialog(this, "Set Label Background Color", 
						 currentColor);
	if(newColor != null) {
	    currentColor = newColor;
	    props.setLabelBackgroundColor(newColor);
	    drawingArea.setLabelBackgroundColor(newColor);
	    logger.logMessage("Label background color = " + newColor.toString(),
			      MainServer.INITIALIZATION_DETAILS_VERBOSITY_LEVEL);
	}
    }

    public void setToProviderArrowColorMenuItem_actionPerformed(ActionEvent e)
    {
	Color currentColor = props.getMessageTrafficToProviderLineColor();
	Color newColor = colorChooser.showDialog(this, "Set \"To Service Provider\" Message Traffic Color", 
						 currentColor);
	if(newColor != null) {
	    currentColor = newColor;
	    props.setMessageTrafficToProviderLineColor(newColor);
	    drawingArea.setToProviderArrowColor(newColor);
	    logger.logMessage("\"To Provider\" message traffic color = " + newColor.toString(),
			      MainServer.INITIALIZATION_DETAILS_VERBOSITY_LEVEL);
	}
    }
        
    //=========================================================================
    // Hub control methods
    //=========================================================================

    /** 
     * Returns the run status of the Hub process. 
     *
     * @return true if the Hub started by this UI is running, false otherwise
     */
    public boolean isHubRunning()
    {
	return hubRunning;
    }

    /**
     * Marks the Hub as being stopped and sets the "enabled" status of various
     * buttons accordingly on the UI.
     */
    void processHubStop()
    {	
	// "Turn off" the Hub icon in the drawing area.
	drawingArea.setHubIsRunning(false);

	// Reset the UI buttons appropriately.
	hubControlMenuItem.setText("Start Hub");

	// Reset the UI.
	resetUI();

	debugUI.setVisible(false);

	hubFrame.setEnabled(true);

	if(!isHubRunning()) {
            
            // In case this menu item was disabled due to the use of an
            // external Hub.
	    controlMenu.setEnabled(true);
	    return;
        }

	// If the UI still has a running "child" Hub process, stop it.
	if(hubProcess != null) {
	    hubProcess.destroy();
	}

	// Mark the Hub as being stopped.
	hubRunning = false;

	// Close the Hub output streams.
	try {
	    if(hubStdOut != null)
		hubStdOut.close();
	    
	    if(hubStdErr != null)
		hubStdErr.close();
	} catch(IOException ioe) {
	}    


	// Interrupt the Hub process watcher thread.
	if(waiterThread != null)
	    waiterThread.interrupt();
    }

    //=========================================================================
    // Methods for updating the UI.
    //=========================================================================

    public void reset()
    {
	Runnable runnable = new Runnable() {
	    public void run() {resetUI();}
	};
	SwingUtilities.invokeLater(runnable);
    }
    
    private void resetUI()
    {
	// Clear local data structures.
	
	if(providers != null)
	    providers.clear();
	
	if(providerIds != null)
	    providerIds.clear();
	
	if(providerMenuItemMap != null)
	    providerMenuItemMap.clear();
	
	if(popupMenuMap != null)
	    popupMenuMap.clear();
	
	if(menuItemToProviderIdMap != null)
	    menuItemToProviderIdMap.clear();
	
	if(menuItemToFrameMap != null)
	    menuItemToFrameMap.clear();
	
	// Remove the provider-independent menu.
	if(frameMenu != null) {
	    menuBar.remove(frameMenu);
	    frameMenu.removeAll();
	    frameMenu = null;
	}
	
	// Reset the drawing area.
	drawingArea.reset();
	
	validate();
	repaint();
    }
    
    /** 
     * Displays a popup menu for a specified service provider (if it has one).
     *
     * @param providerId id of the service provider
     * @param event the mouse event that triggered this method call
     */
    void showPopupMenu(String providerId, MouseEvent event)
    {
	JPopupMenu menu = (JPopupMenu) popupMenuMap.get(providerId);
	if(menu != null)
	    menu.show(event.getComponent(), event.getX(), event.getY());
    }

    /** Appends text to the Hub output pane. */
    private void appendHubOutput(String text)
    {
	if(pauseHubOutputButton.getText().equals("Pause Output")) {
	    hubOutputArea.append(text);
	} else {
	    hubOutputAreaBuffer.append(text);
	}
    }

    /** Appends text to the Hub status pane. */
    private void appendHubStatus(String text)
    {
	if(pauseHubStatusButton.getText().equals("Pause Status")) {
	    hubStatusArea.append(text);
	} else {
	    hubStatusAreaBuffer.append(text);
	}
    }

    // Creating these shared runnable objects is really, really ugly,
    // but I am trying to get rid of a memory leak (the JVM garbage collector
    // does not do a great job reclaiming memory apparently). So, I am
    // trying to reduce the number of objects created.
    class SetProviderConnectionStatusRunnable implements Runnable
    {
	public String providerId;
	public boolean isConnected;
	public void run() 
	{
	    drawingArea.setProviderConnectionStatus(providerId, isConnected);
	}
    }
    
    SetProviderConnectionStatusRunnable setProviderConnectionStatusRunnable =
	new SetProviderConnectionStatusRunnable();

    public void setProviderConnectionStatus(String providerId, 
					    boolean isConnected)
    {
	setProviderConnectionStatusRunnable.providerId = providerId;
	setProviderConnectionStatusRunnable.isConnected = isConnected;

	// Need to do invokeAndWait since the argument is a shared object
	// (i.e., don't want to modify it until this method completes).
	try {
	  SwingUtilities.invokeAndWait(setProviderConnectionStatusRunnable);
	} catch(InterruptedException iex) {
	  logger.logErrorMessage("Caught exception during call to SwingUtilities.invokeAndWait: " + iex.toString(), iex);
	} catch(InvocationTargetException itex) {
	  logger.logErrorMessage("Caught exception during call to SwingUtilities.invokeAndWait: " + itex.toString() + " -> " + itex.getTargetException().toString(), itex);
	}
    }

    class SetProviderAvailabilityStatusRunnable implements Runnable
    {
	public String providerId;
	public boolean isAvailable;
	public void run() 
	{
	    drawingArea.setProviderAvailabilityStatus(providerId, isAvailable);
	}
    }

    SetProviderAvailabilityStatusRunnable setProviderAvailabilityStatusRunnable =
	new SetProviderAvailabilityStatusRunnable();
    
    public void setProviderAvailabilityStatus(String providerId, 
					      boolean isAvailable)
    {
	setProviderAvailabilityStatusRunnable.providerId = providerId;
	setProviderAvailabilityStatusRunnable.isAvailable = isAvailable;

	// Need to do invokeAndWait since the argument is a shared object
	// (i.e., don't want to modify it until this method completes).
	try {
	  SwingUtilities.invokeAndWait(setProviderAvailabilityStatusRunnable);
	} catch(InterruptedException iex) {
	  logger.logErrorMessage("Caught exception during call to SwingUtilities.invokeAndWait: " + iex.toString(), iex);
	} catch(InvocationTargetException itex) {
	  logger.logErrorMessage("Caught exception during call to SwingUtilities.invokeAndWait: " + itex.toString() + " -> " + itex.getTargetException().toString(), itex);
	}
    }
    
    class HighlightProviderRunnable implements Runnable
    {
	public String providerId;
	public boolean highlight;
	public void run() 
	{
	    drawingArea.highlightProvider(providerId, highlight);
	}
    }

    HighlightProviderRunnable highlightProviderRunnable =
	new HighlightProviderRunnable();
    

    /** Turns a provider's highligh on or off. */
    public void highlightProvider(String providerId, boolean highlight)
    {
	highlightProviderRunnable.providerId = providerId;
	highlightProviderRunnable.highlight = highlight;

	// Need to do invokeAndWait since the argument is a shared object
	// (i.e., don't want to modify it until this method completes).
	try {
	  SwingUtilities.invokeAndWait(highlightProviderRunnable);
	} catch(InterruptedException iex) {
	  logger.logErrorMessage("Caught exception during call to SwingUtilities.invokeAndWait: " + iex.toString(), iex);
	} catch(InvocationTargetException itex) {
	  logger.logErrorMessage("Caught exception during call to SwingUtilities.invokeAndWait: " + itex.toString() + " -> " + itex.getTargetException().toString(), itex);
	}
    }

    /** Flashes a provider in the drawing area. */
    public void flashProvider(String providerId) 
    {
	drawingArea.flashProvider(providerId);
    } 

    /** Illustrates message traffic in the drawing area. */
    public void sendMessage(String providerId, boolean toProvider) 
    {
	drawingArea.sendMessage(providerId, toProvider);
    } 

    public void hubStatus(String text)
    {
	final String _text = text;
	Runnable runnable = new Runnable() {
		public void run() {
		    appendHubStatus(_text);
		}
	    };
	SwingUtilities.invokeLater(runnable);
    }

    //=========================================================================
    // WindowListener interface
    //=========================================================================


    public void windowActivated(WindowEvent event)
    {}

    public void windowClosed(WindowEvent event)
    {}

    public void windowClosing(WindowEvent event)
    {
	Object source = event.getSource();
	if(source.equals(hubOutputFrame)) {
	    hubOutputToggleMenuItem_actionPerformed(null);
	} else if(source.equals(this)) {
	    processHubStop();
	    System.exit(0);
	}
    }

    public void windowDeactivated(WindowEvent event)
    {}

    public void windowDeiconified(WindowEvent event)
    {}

    public void windowIconified(WindowEvent event)
    {}

    public void windowOpened(WindowEvent event)
    {}


    //=========================================================================
    // Locally-defined utility classes
    //=========================================================================
    
    /**
     * This thread waits for a specified process to end.
     */
    class WaitingThread extends Thread
    {
	private Process process;
	
	public WaitingThread(Process process)
	{
	    super("WaitingThread");
	    this.process = process;
	}
	
	public void run()
	{
	    try {
		process.waitFor();
		
		Runnable runnable = new Runnable() {
		    public void run() {
			processHubStop();
		    }
		};
		SwingUtilities.invokeLater(runnable);
			
	    } catch(InterruptedException iex) {
		logger.logWarningMessage("WaitingThread.run was interrupted");
	    }
	}
    }

    /**
     * This thread sends data from a text stream to a UI text area.
     */	
    class OutputThread extends Thread 
    {
	private BufferedReader reader;
	
	public OutputThread(String name, InputStream inStream) 
	{
	    super(name);
	    reader = new BufferedReader(new InputStreamReader(inStream));
	}
	
	public void run() 
	{
	    String line;
	    while(true) {
		try {
		    line = reader.readLine();
		    if(line != null) {

			final String text = "  " + line + "\n";	
			Runnable runnable = new Runnable() {
			    public void run() {appendHubOutput(text);}
			};
			SwingUtilities.invokeLater(runnable);
		    }
		} catch(IOException ioex) {
		    // Stream has most likely been closed.
		    return;
		} catch(Exception ex) {
		    logger.logErrorMessage("Exception in OutputThread.run: " + ex.toString(), ex);
		    return;
		}
	    }
	}
    }
}
